/* cdffilter -- Do low and high pass filtering on fields */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

static char *hbuff[HBMAX][LINE];       /* header buffer */
static char cline[LINE];               /* character line buffer */
static long nsbuff,nvbin,nvbout;       /* slice buffer sizes */
static float *sbuff,*vbuff;            /* slice buffer pointers */
static float *ifield;                  /* pointer to specified index field */
static float *sfield;                  /* pointer to specified field */
static float *dfield;                  /* pointer to filtered field */
static float *tfield;                  /* pointer to temporary field */
static long instance;                  /* looping var over instances */
static long loop;                      /* the other looping variable */
static struct field *fp;               /* field structure temporary pointer */
static struct field *fps;              /* field structure pointer for
					   specified field */
static float bad,badlim;               /* bad data values */
static long start,incr,size;           /* loop variables */
static int highpass;                   /* 1 = high pass, 0 = low pass */
static long index;                     /* field looping variable */
static double sum;                     /* integration temporary */
static float lambda;                   /* filter smoothing length */
static float dx;                       /* step in index field */
static float alpha;                    /* smoothing parameter */


main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if (argc != 6) {
    fprintf(stderr,
         "Usage: cdffilter -[lh] smooth_len filtered_fld field dim\n");
    exit(1);
  }

/* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdffilter: float format expected\n");
    exit(1);
  }

/* get the flag */
  if (strcmp(argv[1],"-l") != 0 || strcmp(argv[1],"-h") != 0)
    highpass = (strcmp(argv[1],"-h") == 0) ? 1 : 0;
  else {
    fprintf(stderr,"cdffilter: flag %s unknown\n",argv[1]);
    exit(1);
  }

/* get lambda */
  lambda = atof(argv[2]);
  if (lambda <= 0.) {
    fprintf(stderr,"cdffilter: inappropriate smoothing length: %g\n",lambda);
    exit(1);
  }

/* get bad data values */
  bad = (seekpar(hbuff,HBMAX,"bad",cline) == OK) ? atof(cline) : BAD;
  badlim = (seekpar(hbuff,HBMAX,"badlim",cline) == OK) ? atof(cline) : BADLIM;

/* compute input buffer lengths */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');

/* add stuff to header, comment first */
  sprintf(cline,"cdffilter: %s\n",argv[1]);
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s\n",argv[2],argv[3],argv[4],argv[5]);
  addcline(hbuff,HBMAX,cline);

/* unfiltered field next -- same dimensions for filtered field */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[4])) == NULL) {
    fprintf(stderr,"cdffilter: %s field not found in variable slice\n",argv[4]);
    exit(1);
  }
  addfld(hbuff,HBMAX,'v',argv[3],fp->smul,fp->sadd,fp->bytes,fp->dim,
       fp->dname1,fp->dsize1,fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,fp->dname4,fp->dsize4);

/* allocate buffers */
  nvbout = elemcnt(hbuff,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbout);

/* assign field and structure pointers */
  sfield = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[4],&fps);
  dfield = getptr(hbuff,HBMAX,vbuff,'v','d',argv[3]);
  if ((ifield = getptr2(hbuff,HBMAX,sbuff,'s','r',argv[5],&fp)) == NULL) {
    fprintf(stderr,"cdffilter: %s field doesn't exist\n",argv[5]);
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->fname,fp->dname1) != 0)) {
    fprintf(stderr,"cdffilter: %s field is not an index field\n",argv[5]);
    exit(1);
  }

/* write header */
  puthdr(stdout,hbuff,HBMAX);

/* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* loop on variable slices */
  while (getslice(stdin,nvbin,vbuff) != EOF) {

/* loop on subspace instances */
    instance = 0;
    while (subspace1(fps,argv[5],instance,&start,&incr,&size) != FAIL) {
      tfield = getbuff(size);

/* do a low pass filtering job -- forward filtering first */
      for (loop = 0; loop < size; loop++) {
        index = start + incr*loop;
        if (loop == 0 || fabs(sfield[index]) > badlim ||
             fabs(sfield[index - incr]) > badlim) dfield[index] = sfield[index];
        else {
          dx = ifield[loop] - ifield[loop - 1];
          alpha = dx/lambda;
          if (alpha <= 0. || alpha >= 1.) {
            fprintf(stderr,"cdffilter: smoothing length < grid size\n");
            exit(1);
          }
          dfield[index] = alpha*sfield[index] +
               (1. - alpha)*dfield[index - incr];
        }
        tfield[loop] = dfield[index];
      }

/* backward filtering for symmetry */
      for (loop = size - 1; loop >= 0; loop--) {
        index = start + incr*loop;
        if (loop == size - 1 || fabs(sfield[index]) > badlim ||
             fabs(sfield[index + incr]) > badlim) dfield[index] = sfield[index];
        else {
          dx = ifield[loop + 1] - ifield[loop];
          alpha = dx/lambda;
          if (alpha <= 0. || alpha >= 1.) {
            fprintf(stderr,"cdffilter: smoothing length < grid size\n");
            exit(1);
          }
          dfield[index] = alpha*tfield[loop] +
               (1. - alpha)*dfield[index + incr];
        }
      }

/* if high pass filter is desired, take the complement */
      if (highpass) {
        for (loop = 0; loop < size; loop++) {
          index = start + incr*loop;
          if (fabs(sfield[index]) < badlim) {
            dfield[index] = sfield[index] - dfield[index];
          }
        }
      }

/* increment instance */
      instance++;
      free(tfield);
    }
/* write variable slice */
    putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}
