 /*  cdfmatlab.c - This program is designed to read in a file from the 
 *  cdf data and convert it to Matlab format so that it 
 *  may be used in conjunction with other data. This program requires 
 *  the variables to be 0 or 1 dimensional only.
 *  This program also creates a file called "cdfmatlab.log" that
 *  contains the names of all the variables found in the cdf file.
 */
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "cdfhdr.h"

#define   CSIZE 1 /* char size in bytes */
#define   SSIZE 2 /* short size in bytes */
#define   LSIZE 4 /* long size in bytes */
#define   FSIZE 4 /* float size in bytes */
#define   DSIZE 8 /* double size in bytes */

static struct myfield
/*  my version of cdf's field structure so that I can store away
 *  the data I need for each variable.
 */
{
  char  name[LINE];
  int dm;
  char  dname[LINE];
  int dsize;
  long  pos;
};


static double  t4[4],t8[8],t16[16];  /* static time intervals */
static double  *ddata[HBMAX];
static double  *stddata[HBMAX];
static double  *time4, *time8, *time16;/* new time variables for Matlab */

static float   *stat;      /* Static slice */
static float *var;     /* Variable slice read in */
static float dtime;      /* time between slices */

static int tcount;     /* # of slices * # per slice */
static int count;      /*  */
static int scount=0;     /* keep track of the number of slices */
static int echeck;
static int stcheck;    /* make sure there are some static elements */
static int vcount;     /* keep track of the number of variables */
static int stcount;    /* keep track of static variables */
static int i,j;
static int poshour, posmin, possec;/* indicate where the variable is in the
           order of things */
static int fghour = 0,   /* indicates whether the correct time */
     fgmin = 0,          /* elements are located in the data file */
     fgsec = 0;

static long  nhead[5];   /* array for header of each variable 
           to be written in matlab format */
static long  snum;     /* static number count */
static long  vnum;     /* variable number count */

static char  hdbuffer[HBMAX][LINE];  /* Header for cdf file */
static char  tstring[LINE];    /* temporary string variable */
static char  pname[LINE];    /* string to obtain info from parameters */
static char  logfilename[LINE];  /* log file name */

static struct myfield  *vwhere[LINE];  /* pointer to cdf structure for
           each of the possible variables */
static struct myfield  *stwhere[LINE]; /* static */
static struct field  *where;   /* variable */

static FILE  *fin,*fout;

main(argc,argv)
int argc;
char *argv[];
{
  if (argc != 2) {
    fprintf(stderr,"Usage: cdfmatlab infile\n");
    exit(1);
  }
  if (strlen(argv[1]) > LINE - 5) {
    fprintf(stderr,"cdfmatlab: infile name too long\n");
    exit(1);
  }
  sprintf(logfilename,"%s.log",argv[1]);
  if ((fout=fopen(logfilename,"w")) == NULL) {
    fprintf(stderr,"cdfmatlab: can't open %s.log for output\n",argv[1]);
    exit(1);
  }
  if ((fin=fopen(argv[1],"r")) == NULL) {
    fprintf(stderr,"cdfmatlab: can't open %s for input\n",argv[1]);
    exit(1);
  }

  /* Read in the header from the cdf file. */
  gethdr(fin,hdbuffer,HBMAX);

  snum = elemcnt(hdbuffer,HBMAX,'s');
  vnum = elemcnt(hdbuffer,HBMAX,'v');


  /* Obtain all of the information from the header. */
  hdread();

  /* Count the number of slices in the file. */
  slicecount();

  /* Read in the data and store it away.  */
  varread();

  /* Set up the header for each variable and write it out.*/
  dtset();

  /* Close the files. */
  fclose(fout);
  fclose(fin);
  exit(0);
}


hdread()
/*  Read in the info from the header. */
{

  /*  Obtain all of the variables in the static slice. */
  stcount = 0;
  fprintf(fout,"static slice:\n");
  stwhere[stcount] = (struct myfield *) malloc(sizeof(struct myfield));

  if(stwhere[stcount] == NULL){
    fprintf(stderr,"cdfmatlab: unable to allocate memory\n");
    exit(1); 
  }

  where = getfld(hdbuffer,HBMAX,'s',stcount);
  if(where != NULL)
  {
    if (where->dim > 1) {
      fprintf(stderr,"cdfmatlab: field %s greater than 1 dimension\n",
           where->fname);
      exit(1);
    }
    /* Store away the necessary information. */
    strcpy(stwhere[stcount]->name,where->fname);
    stwhere[stcount]->dm = where->dim;
    strcpy(stwhere[stcount]->dname,where->dname1);
    stwhere[stcount]->dsize = where->dsize1;
    stwhere[stcount]->pos = where->ptr;
    stcheck = 1;
  }
  while(1)
  {

    fprintf(fout," %s\n",stwhere[stcount]->name);

    stcount++;
    stwhere[stcount] = (struct myfield *) malloc(sizeof(struct myfield));
    if(stwhere[stcount] == NULL){
      fprintf(stderr,"cdfmatlab: unable to allocate memory\n");
      exit(1); 
    }
    where = getfld(hdbuffer,HBMAX,'s',stcount);
    if(where == NULL)break;
    /* Store away the necessary information. */
    strcpy(stwhere[stcount]->name,where->fname);
    stwhere[stcount]->dm = where->dim;
    strcpy(stwhere[stcount]->dname,where->dname1);
    stwhere[stcount]->dsize = where->dsize1;
    stwhere[stcount]->pos = where->ptr;
  }


  /*  Now obtain all of the variables in the variable slice 
 *  and store them in myfield structure so as to preserve the data. 
 */
  vcount = 0;

  fprintf(fout,"variable slice:\n");
  vwhere[vcount] = (struct myfield *) malloc(sizeof(struct myfield));

  if(vwhere[vcount] == NULL){
    fprintf(stderr,"cdfmatlab: unable to allocate memory\n");
    exit(1); 
  }

  where = getfld(hdbuffer,HBMAX,'v',vcount);
  /* Store away the necessary information. */
  strcpy(vwhere[vcount]->name,where->fname);
  vwhere[vcount]->dm = where->dim;
  strcpy(vwhere[vcount]->dname,where->dname1);
  vwhere[vcount]->dsize = where->dsize1;
  vwhere[vcount]->pos = where->ptr;
  while(1)
  {

    fprintf(fout," %s\n",vwhere[vcount]->name);

    vcount++;
    vwhere[vcount] = (struct myfield *) malloc(sizeof(struct myfield));
    if(vwhere[vcount] == NULL){
      fprintf(stderr,"cdfmatlab: unable to allocate memory\n");
      exit(1); 
    }
    where = getfld(hdbuffer,HBMAX,'v',vcount);
    if(where == NULL)break;
    /* Store away the necessary information. */
    if (where->dim > 1) {
      fprintf(stderr,"cdfmatlab: field %s greater than 1 dimension\n",
           where->fname);
      exit(1);
    }
    strcpy(vwhere[vcount]->name,where->fname);
    vwhere[vcount]->dm = where->dim;
    strcpy(vwhere[vcount]->dname,where->dname1);
    vwhere[vcount]->dsize = where->dsize1;
    vwhere[vcount]->pos = where->ptr;
  }

}





slicecount()
/*  Count the number of slices in the file in order to more efficiently
 *  handle the large amounts of data in memory. After rewinding, the
 *  header will be reread in and the so that it will seem as if nothing
 *  happened.
 */
{
  /*  Read in the static slice. */
  stat = getbuff(snum);
  getslice(fin,snum,stat);
  free(stat);

  /*  Read in the variable slices. */
  scount = 0;
  /* malloc some space for the variable field pointer */
  var = getbuff(vnum);
  while(getslice(fin,vnum,var) != EOF)
  {
    scount++;
    free(var);
    var = getbuff(vnum);
  }
  free(var);


  rewind(fin);
  /*  Reread in the header from the cdf file. */
  gethdr(fin,hdbuffer,HBMAX);

}



varread()
/*  Read in the static and variable slices to store away the necessary
 *  information.
 */
{


  /*  Allocate space for the static variables */
  for(i=0; i<stcount; i++)
  {
    if(stwhere[i]->dm == 0)
    {   
      tcount = 1;
      stddata[i] = (double *)malloc(sizeof(double)*tcount);
    }
    else {  
      tcount = (stwhere[i]->dsize);
      stddata[i] = (double *)malloc(sizeof(double)*tcount);
    }

    if(stddata[i] == NULL)
    { 
      fprintf(stderr,"cdfmatlab: cannot malloc enough memory\n");
      exit(1);
    }
  }

  /*  Read in the static slice. */
  stat = getbuff(snum);
  getslice(fin,snum,stat);

  /*  Now store away the static data  */
  for(i=0; i<stcount; i++)
  {
    if(stwhere[i]->dm == 0) stddata[i][0]=stat[stwhere[i]->pos];
    else {
      for(j=0; j< stwhere[i]->dsize; j++)
        stddata[i][j]=stat[(stwhere[i]->pos)+j];
    }
  }


  /*  Allocate the space to store the data from the variable slices.  */
  for(i=0; i<vcount; i++)
  {
    if(vwhere[i]->dm == 0)
    {   
      tcount = scount;
      ddata[i] = (double *)malloc(sizeof(double)*tcount);
    }
    else {  
      tcount = (vwhere[i]->dsize) * scount;
      ddata[i] = (double *)malloc(sizeof(double)*tcount);
    }

    if(ddata[i] == NULL)
    { 
      fprintf(stderr,"cdfmatlab: cannot malloc enough memory\n");
      exit(1);
    }
  }

  count=0;
  /*  Read in the variable slices. */
  /*  Malloc some space for the variable field pointer. */
  var = getbuff(vnum);
  while(getslice(fin,vnum,var) != EOF)
  {
    for(i=0; i<vcount; i++)
    {
      if(vwhere[i]->dm == 0) ddata[i][count]=var[vwhere[i]->pos];
      else {
        for(j=0; j< vwhere[i]->dsize; j++)
          ddata[i][count*(vwhere[i]->dsize)+j]=var[(vwhere[i]->pos)+j];
      }
    }
    count++;

    free(var);
    var = getbuff(vnum);
  }
  free(var);
}





dtset()
/*  This routine will set up the headers and write out each variable. */
{

  nhead[0] = 1000L; /* see page 3-76 of the Matlab manual */
  nhead[2] = 1L;
  nhead[3] = 0L;

  /*  Write out the static slice. */
  if(stcheck == 1)
  {
    for(i=0;i<stcount;i++)
    {
      if(stwhere[i]->dm == 0)nhead[1] = 1L ;
      else { 
        nhead[1] = (long) stwhere[i]->dsize; 
      }
      strcpy(pname,stwhere[i]->name);
      nhead[4] = (long)strlen(pname)+1;
      if(fwrite(nhead,LSIZE,5,stdout) != 5)echeck = 1;
      if(fwrite(pname,CSIZE,strlen(pname)+1,stdout) != strlen(pname)+1)echeck =1;
      if(fwrite(stddata[i],DSIZE,nhead[1],stdout)!=nhead[1])echeck = 1;
      if(echeck == 1){
        fprintf(stderr,"cdfmatlab: write failure\n"); 
        exit(1);
      }
    }
  }


  /*  Write out the variable slice. */
  for(i=0;i<vcount;i++)
  {
    if(vwhere[i]->dm == 0)nhead[1] = (long) scount;
    else { 
      nhead[1] = (long) vwhere[i]->dsize*scount; 
    }
    strcpy(pname,vwhere[i]->name);
    nhead[4] = (long)strlen(pname)+1;
    if(fwrite(nhead,LSIZE,5,stdout) != 5)echeck = 1;
    if(fwrite(pname,CSIZE,strlen(pname)+1,stdout) != strlen(pname)+1)echeck =1;
    if(fwrite(ddata[i],DSIZE,nhead[1],stdout)!=nhead[1])echeck = 1;
    if(echeck == 1){
      fprintf(stderr,"cdfmatlab: write failure\n"); 
      exit(1);
    }
  }


}





