/* cdfplot.c -- General plotting and graphing routine that accepts a
 * common data format file on standard input and plots fields in each
 * variable slice according to instructions in the command line
 * argument list.  Annotated by comment in header, command list,
 * and zero dimensional fields in slice being treated.
 */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#include "pgraf.h"

#define MIN(a,b) ( (a)<(b) ? (a):(b))
#define MAX(a,b) ((a)>(b) ? (a):(b))

#define SPTRMAX 15         /* max length of element stack */
#define LNEMAX  7          /* mx number of points for line command */
#define XLOC 0.05          /* x location of start of annotation lines */
#define YMAX 0.95          /* upper limit on annotation lines */
#define YDEC 0.027         /* vertical spacing of annotation lines */
#define YMIN 0.05          /* lower limit on annotation lines */
#define ELSEP ','          /* element separator character */
#define CSEP '/'           /* command separator character */
#define ESCCHAR '\\'       /* escape character */
#define NCONT 10           /* default number of contours */
#define QUANT 0.05         /* default quantization of axis labels */
#define MAXLBL 3           /* maximum number of station labels */
#define VLEN (-0.3)        /* station vector arrow length */
#define VWID 0.1           /* station vector arrow width */

static char hbuff[HBMAX][LINE],com[HBMAX][LINE],stack[SPTRMAX][LINE];
                    /* header buffer, comment buffer, sub- command stack */
static int iarg,sptr;             /* loop temporary and stack pointer */
static int axdone,wintype;        /* axes set flag and window type flag*/
static long ns,nv;                /* lengths of slice buffers */
static float *sbuff,*vbuff;       /* slice buffer pointers */
static float badlim;              /* bad data limit */
static float xlow,xhigh,ylow,yhigh; /* window size variables*/
static char parval[LINE];         /* buffer to hold parameter values */
static void windax(),lenprob(),setlims(),syntax(),comex(),push();
static float *varspace(),varname(),*getvar1(),midcomp();
static char *pop(),getel();

main(argc,argv)
int argc;
char *argv[];
{
  /* check for arguments */
  if (argc == 1) {
    fprintf(stderr,"Usage: cdfplot \"command_list\" ... < cdf_file\n");
    fprintf(stderr,"     command_list = command/command...\n");
    fprintf(stderr,"     commands: ntx,nty,t  (set tic marks)\n");
    fprintf(stderr,"             x1,x2,x");
    fprintf(stderr,"  (scale x-axis)\n");
    fprintf(stderr,"             y1,y2,y");
    fprintf(stderr,"  (scale y-axis)\n");
    fprintf(stderr,"             x1,y1,x2,y2,... [ltype,]d (draw line)\n");
    fprintf(stderr,"             xvar,yvar,[ltype,]p  (line graph)\n");
    fprintf(stderr,"             xvar,yvar,[mtype,]m  (mark graph)\n");
    fprintf(stderr,"             var,[cint,]ltype,c  (contour plot)\n");
    fprintf(stderr,"             var,xi,yi,[cint,]ltype,uc (uneven contour)\n");
    fprintf(stderr,"             var,[nf,pf,],ltype,f  (fill)\n");
    fprintf(stderr,"             var,xi,yi,[nf,pf,]ltype,uf (uneven fill)\n");
    fprintf(stderr,"             var,[skipx,skipy,]g  (number grid)\n");
    fprintf(stderr,"             var,xi,yi,[skipx,skipy,]ug (uneven grid)\n");
    fprintf(stderr,"             xvar,yvar,[sx,sy,]v  (vector plot)\n");
    fprintf(stderr,"             xvar,yvar,xi,yi,[sx,sy,]uv");
    fprintf(stderr," (uneven vector plot)\n");
    fprintf(stderr,"             mtype,x_loc,y_loc,[xvar,yvar,sx,sy,");
    fprintf(stderr,"[var,...]],skip,s (station plot)");
    fprintf(stderr,"             text,x,y,l (label)\n");
    fprintf(stderr,"             text,lx (x-axis label)\n");
    fprintf(stderr,"             text,ly (y-axis label)\n");
    fprintf(stderr,"             [[xlow,ylow,]xhigh,yhigh][wintype,]a");
    fprintf(stderr," (omit annotation)\n");
    fprintf(stderr,"     var,xvar,yvar = var_name[*scale_factor]\n");
    exit(1);
  }
  /* set axes flag to not done */
  axdone = 0;
  /* set window size to default */
  xlow = ylow = 0.;
  yhigh = xhigh = 10.;
  wintype = 0;
  /* get header from standard input */
  gethdr(stdin,hbuff,HBMAX);
  /* check format and die if not float */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfplot: format not float\n");
    exit(1);
  }
  /* get static and variable slice lengths */
  ns = elemcnt(hbuff,HBMAX,'s');
  nv = elemcnt(hbuff,HBMAX,'v');
  /* get static slice */
  if ((sbuff = (float *)malloc(ns*sizeof(float))) == NULL) {
    fprintf(stderr,"cdfplot: can't allocate enough memory\n");
    exit(1);
  }
  getslice(stdin,ns,sbuff);
  /* allocate space for variable slices */
  if ((vbuff = (float *)malloc(nv*sizeof(float))) == NULL) {
    fprintf(stderr,"cdfplot: can't allocate enough memory\n");
    exit(1);
  }
  /* open graphics */
  gopen();
  /* check parameters for badlim and set if exists -- otherwise default */
  if (seekpar(hbuff,HBMAX,"badlim",parval) == OK) badlim = atof(parval);
  else badlim = BADLIM;
  badset(badlim);
  /* loop through variable slices */
  while (getslice(stdin,nv,vbuff) != EOF) {
    /* loop through command lists */
    for (iarg = 1; iarg < argc; iarg++) {
      /* do plot for this slice and command list */
      comex(argv[iarg]);
    }
  }
  /* close graphics */
  gclose();
  exit(0);
}

static char getel(arg,el)
/* Gets an element from a command list (arg) and returns terminating
 * character.  Maintains an internal memory of where it is in
 * the command list.
 */
char *arg,*el;
{
  static int pointer = 0;
  int i,escape;
  char c;
  /* loop through command list */
  escape = 0;
  i = 0;
  while (((c = arg[pointer++]) != '\0') &&
      (((c != ELSEP) && (c != CSEP)) || escape)) {
    /* do we have an escape character? */
    if ((escape == 0) && (c == ESCCHAR)) {
      escape = 1;
    }
    /* nope, add character to buffer */
    else {
      el[i++] = c;
      escape = 0;
    }
  }
  /* null terminate string */
  el[i] = '\0';
  if (c == '\0') pointer = 0;
  return(c);
}

static void push(el)
/* This routine pushes string onto stack */
char *el;
{
  if (sptr < 0) sptr = 0;
  if (strlen(el) >= LINE - 1) {
    fprintf(stderr,"cdfplot: command element too long\n");
    exit(1);
  }
  strcpy(stack[sptr],el);
  sptr++;
  if (sptr > SPTRMAX - 1) sptr = SPTRMAX - 1;
}

static char *pop()
/* This routine pops stack pointer and returns pointer to string */
{
  sptr--;
  if (sptr < 0) sptr = 0;
  return(stack[sptr]);
}

static int nlabels,gotvel;  /* number of labels for station data
                              and flag to indicate station velocities */
static int skip;            /* number of points to skip between shown points*/
static char beta[LINE],labels[MAXLBL][LINE],x_loc[LINE],y_loc[LINE];
                            /* temporaries for station data */
static float mfactl[MAXLBL],velxt[5],velyt[5];
                            /* more temporaries for station data */
static float *labptr[MAXLBL],*locxptr,*locyptr,*velxptr,*velyptr;
                            /* station data -- slice buffer pointers */
static long lablen[MAXLBL],locxlen,locylen,velxlen,velylen;
                            /* more temporaries for station data */


static void comex(clist)
/* This routine makes plots according to the command list using
 * pgraf.  It assumes that graphics has already been opened.
 */
char *clist;
{
  /* temporaries for comex */
  char alpha[LINE],xname[LINE],xdname[LINE],yname[LINE],ydname[LINE];
  char lxname[LINE],lyname[LINE],locx[LINE],locy[LINE];
  char endchar,graph;
  int nclines,loop,ntx,nty,sxdone,sydone,ltype,xsize,ysize,nx,ny;
  int cfgot,skipx,skipy,xnew,ynew,xlnew,ylnew,aomit,xoccur,yoccur,loop2;
  int xmsize,ymsize,unv,xsflg,ysflg;
  int xlabel,ylabel,i,chkx,chky,scldne;
  float xloc,yloc,sx1,sx2,sy1,sy2,cint,cfn,cfp,sx,sy,mxfact,myfact,xlab,ylab;
  float middle,rx,ry,*xvar,*yvar,*var,x[LNEMAX],y[LNEMAX],*lcx,*lcy;
  struct field *fb;
  /* function declarations */
  char *pop();
  float *getvar1(),varname(),*varspace();
  /* inititialization */
  xnew = ynew = xlnew = ylnew = scldne = 0;
  ntx = 6;
  nty = 6;
  sx1 = sy1 = 0.;
  unv = 0;
  sx2 = sy2 = 1.;
  ltype = 1;
  aomit = 0;
  xlabel = ylabel = 0;
  sxdone = sydone = 0;
  xsflg = ysflg = 0;
  /* get comment from header buffer */
  nclines = getcmt(hbuff,HBMAX,com);
  /* clear screen */
  gclear();
  /* make plot */
  /* loop on commands */
  for (;;) {
    /* push command elements on stack, except last one */
    sptr = 0;
    while (((endchar = getel(clist,alpha)) != CSEP) && (endchar != '\0')) {
      push(alpha);
    }
    /* if-else-if through commands */
    /* set number of tics on each axis */
    if (strcmp(alpha,"t") == 0) {
      if (sptr != 2) syntax(clist);
      nty = atoi(pop());
      ntx = atoi(pop());
    }
    /* scale x axis */
    else if (strcmp(alpha,"x") == 0) {
      if (sptr != 2) syntax(clist);
      sx2 = atof(pop());
      sx1 = atof(pop());
      sxdone = 1;
      if(axdone == 1 ) {
        fprintf(stderr," cdfplot: axes have already been scaled \n");
        exit(1);
      }
    }
    /* scale y axis */
    else if (strcmp(alpha,"y") == 0) {
      if (sptr != 2) syntax(clist);
      sy2 = atof(pop());
      sy1 = atof(pop());
      sydone = 1;
      if (axdone == 1) {
        fprintf(stderr," cdfplot: axes have already been scaled \n");
        exit(1);
      }
    }
    /* draw a line */
    else if(strcmp(alpha,"d") == 0) {
      i = 0;
      if(sptr%2 == 1)ltype = atoi(pop());
      if(sptr == 0) syntax(clist);
      /*  get the variables */
      while (sptr != 0) {
        y[i] = atof(pop());
        x[i] = atof(pop());
        i++;
      }
      if(axdone == 0) {
        if(sxdone == 0) scale(i,QUANT,x,&sx2,&sx1);
        if(sydone == 0) scale(i,QUANT,y,&sy2,&sy1);
        if(xlabel == 0) {
          lxname[0] = 'X';
          lxname[1] = ' ';
          lxname[2] = ' ';
        }
        if(ylabel == 0) {
          lyname[0] = 'Y';
          lyname[1] = ' ';
          lyname[2] = ' ';
          lyname[3] = ' ';
        }
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      line(1,ltype,i,x,y);
    }
    /* graph one variable against another */
    else if ((strcmp(alpha,"p") == 0) || (strcmp(alpha,"m") == 0)) {
      if (sptr == 3) ltype = atoi(pop());
      if (sptr != 2) syntax(clist);
      /* get pointers to the variables */
      yvar = getvar1(yname,ydname,&ysize,&myfact,&ynew,&yoccur);
      xvar = getvar1(xname,xdname,&xsize,&mxfact,&xnew,&xoccur);
      if ((xvar == NULL) || (yvar == NULL)) {
        fprintf(stderr,"cdfplot: %s or %s unknown field\n",xname,yname);
        exit(1);
      }
      /* check for consistency */
      if (((xoccur > 1) && (yoccur > 1) && (xoccur != yoccur))
          || (xsize != ysize) || (strcmp(xdname,ydname) != 0)) {
        fprintf(stderr,
		"cdfplot: %s, %s, inconsistent variables for graphing\n",
		xname,yname);
        exit(1);
      }
      /* draw axes if needed */
      if (axdone == 0) {
        if (sxdone == 0) scale(xsize*xoccur,QUANT,xvar,&sx2,&sx1);
        if (sydone == 0) scale(ysize*yoccur,QUANT,yvar,&sy2,&sy1);
        if (xlabel == 0) strcpy(lxname,xname);
        if (ylabel == 0) strcpy(lyname,yname);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* draw graph, considering multi-dimensional arrays as multiple graphs */
      xmsize = xoccur > 1 ? xsize: 0;
      ymsize = yoccur > 1 ? ysize: 0;
      for (loop2 = 0; loop2 < (xoccur > yoccur ? xoccur: yoccur); loop2++) {
        graph = alpha[0];
        if (graph == 'p') {
          line(1,ltype,xsize,xvar + xmsize*loop2,yvar + ymsize*loop2);
        }
        else {
          for (loop = 0; loop < xsize; loop++) {
            mark(1,ltype,*(xvar + xmsize*loop2 + loop),
                *(yvar + ymsize*loop2 + loop),0.5);
          }
        }
      }
      /* add comment */
      if (graph == 'p')
        sprintf(alpha,"graph(%s*%g,%s*%g,%d)\n",
            xname,mxfact,yname,myfact,ltype);
      else
        sprintf(alpha,"mark(%s*%g,%s*%g,%d)\n",
            xname,mxfact,yname,myfact,ltype);
      strcpy(com[nclines++],alpha);
      /* reset the line type to default */
      ltype = 1;
      /* release space if necessary */
      if (xnew) free(xvar);
      if (ynew) free(yvar);
    }
    /* contour plot */
    else if ((strcmp(alpha,"c") == 0) || strcmp(alpha,"uc") == 0) {
      ltype = atoi(pop());
      cint = 0.;
      if ((sptr == 2) || (sptr == 4)) cint = atof(pop());
      /* find the variable */
      if(strcmp(alpha,"uc") == 0) {
	if (sptr != 3) syntax(clist);
        myfact = varname(locy);
        if((fb = seekfld(hbuff,HBMAX,'v',locy)) == NULL) {
          if((fb = seekfld(hbuff,HBMAX,'s',locy)) == NULL) {
            fprintf(stderr,"cdfplot: variable field %s nonexistent\n",locy);
            exit(1);
          }
          else ysflg = 1;
        }
        if(fb->dim != 1) {
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locy);
          exit(1);
        }
        chky = fb->dsize1;
        if(ysflg == 0) lcy = varspace(&vbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        else lcy = varspace(&sbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        mxfact = varname(locx);
        if((fb = seekfld(hbuff,HBMAX,'v',locx)) == NULL) {
          if((fb = seekfld(hbuff,HBMAX,'s',locx)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locx);
            exit(1);
          }
          else xsflg = 1;
        }
        if(fb->dim != 1){
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locx);
          exit(1);
        }
        chkx = fb->dsize1;
        if(xsflg == 0) lcx = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
        else lcx = varspace(&sbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
        unv = 1;
      }
      if (sptr != 1) syntax(clist);
      mxfact = varname(xname);
      if ((fb = seekfld(hbuff,HBMAX,'v',xname)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",xname);
        exit(1);
      }
      if (fb->dim != 2) {
        fprintf(stderr,
             "cdfplot: contour requires 2 dimensions for %s\n",xname);
        exit(1);
      }
      nx = fb->dsize1;
      ny = fb->dsize2;
      var = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xnew);
      if( unv == 1 && (nx != chkx || ny != chky)) {
        fprintf(stderr,"cdfplot: number of points and number of point ");
        fprintf(stderr,"locations must match \n");
        exit(1);
      }
      /* compute axes scales for uneven contours if needed */
      if(unv == 1 && axdone == 0) {
        if(sxdone == 0) scale(nx,QUANT,lcx,&sx2,&sx1);
        if(sydone == 0) scale(ny,QUANT,lcy,&sy2,&sy1);
        if(xlabel == 0) strcpy(lxname,locx);
        if(ylabel == 0) strcpy(lyname,locy);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* compute default contour interval if needed */
      if (cint <= 0.) cint = dcon(nx,ny,NCONT,var);
      /* draw axes if needed */
      if (axdone == 0) {
        if (sxdone == 0) setlims(&sx1,&sx2,fb->dname1,fb->dsize1);
        if (sydone == 0) setlims(&sy1,&sy2,fb->dname2,fb->dsize2);
        if (xlabel == 0) strcpy(lxname,fb->dname1);
        if (ylabel == 0) strcpy(lyname,fb->dname2);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* add comment */
      sprintf(alpha,"contour(%s*%g,%.3g,%d)\n",xname,mxfact,cint,ltype);
      strcpy(com[nclines++],alpha);
      /* draw contours */
      if(unv == 1) conto2(1,nx,ny,cint,ltype,var,lcx,lcy);
      else conto(1,nx,ny,cint,ltype,var);
      /* reset the line type to default */
      ltype = 1;
      /* release space if necessary */
      if (xnew) free(var);
      if (xlnew) free(lcx);
      if (ylnew) free(lcy);
    }
    /* fill routine */
    else if ((strcmp(alpha,"f") == 0) || (strcmp(alpha,"uf") == 0)) {
      ltype = atoi(pop());
      cfgot = 0;
      if (((sptr == 3) && (strcmp(alpha,"f") == 0)) || (sptr == 5)) {
        cfp = atof(pop());
        cfn = atof(pop());
        cfgot = 1;
      }
      /* find the point locations if necessary */
      if(strcmp(alpha,"uf") == 0) {
	if (sptr != 3) syntax(clist);
        unv = 1;
        myfact =  varname(locy);
        if((fb = seekfld(hbuff,HBMAX,'v',locy)) == NULL) {
          if((fb = seekfld(hbuff,HBMAX,'s',locy)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locy);
            exit(1);
          }
          else ysflg = 1;
        }
        if(fb->dim != 1) {
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locy);
          exit(1);
        }
        chky = fb->dsize1;
        if(ysflg == 0) lcy = varspace(&vbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        else lcy = varspace(&sbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        mxfact = varname(locx);
        if((fb = seekfld(hbuff,HBMAX,'v',locx)) == NULL ) {
          if((fb = seekfld(hbuff,HBMAX,'s',locx)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locx);
            exit(1);
          }
          else xsflg =  1;
        }

        if(fb->dim != 1){
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locx);
          exit(1);
        }
        chkx = fb->dsize1;
        if(xsflg == 0) lcx = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
        else lcx = varspace(&sbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
      }
      if (sptr != 1) syntax(clist);
      /* find the variable */
      mxfact = varname(xname);
      if ((fb = seekfld(hbuff,HBMAX,'v',xname)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",xname);
        exit(1);
      }
      if (fb->dim != 2) {
        fprintf(stderr,"cdfplot: fill requires 2 dimensions for %s\n",xname);
        exit(1);
      }
      nx = fb->dsize1;
      ny = fb->dsize2;
      var = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xnew);
      if(( unv == 1) && (chkx != nx || chky != ny)) {
        fprintf(stderr,"cdfplot: number of points and number of point");
        fprintf(stderr," locations must match\n");
        exit(1);
      }

      /* compute default fill boundaries if needed */
      if (cfgot == 0) {
        cint = dcon(nx,ny,NCONT,var);
	middle = midcomp(nx,ny,cint,var);
        cfn = middle - cint;
        cfp = middle + cint;
      }
      /* draw axes for uneven fill */
      if(( axdone == 0) && (unv == 1)) {
        if(sxdone == 0) scale(nx,QUANT,lcx,&sx2,&sx1);
        if(sydone == 0) scale(ny,QUANT,lcy,&sy2,&sy1);
        if(xlabel == 0) strcpy(lxname,locx);
        if(ylabel == 0) strcpy(lyname,locy);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* draw axes if needed */
      if (axdone == 0) {
        if (sxdone == 0) setlims(&sx1,&sx2,fb->dname1,fb->dsize1);
        if (sydone == 0) setlims(&sy1,&sy2,fb->dname2,fb->dsize2);
        if (xlabel == 0) strcpy(lxname,fb->dname1);
        if (ylabel == 0) strcpy(lyname,fb->dname2);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* print uneven fill if needed */
      if(unv == 1){
        fill2(1,nx,ny,cfn,cfp,ltype,var,lcx,lcy);
      }
      /* add comment */
      sprintf(alpha,"fill(%s*%g,%.5g,%.5g,%d)\n",xname,mxfact,cfn,cfp,ltype);
      strcpy(com[nclines++],alpha);
      /* do the fill */
      if(unv != 1) fill(1,nx,ny,cfn,cfp,ltype,var);
      /* reset the line type to default */
      ltype = 1;
      /* reset unv */
      unv = 0;
      /* release space if necessary */
      if (xnew) free(var);
      if (xlnew) free(lcx);
      if (ylnew) free(lcy);
    }
    /* grid of numbers */
    else if ((strcmp(alpha,"g") == 0) || (strcmp(alpha,"ug") == 0)){
      cfgot = 0;
      if (((sptr == 3) && (strcmp(alpha,"g") == 0)) || (sptr == 5)) {
        skipy = atof(pop());
        skipx = atof(pop());
        cfgot = 1;
      }
      if(strcmp(alpha,"ug") == 0) unv = 1;
      else unv = 0;
      /* find the variable */
      if(unv == 1) {
	if (sptr != 3) syntax(clist);
        myfact = varname(locy);
        if ((fb = seekfld(hbuff,HBMAX,'v',locy)) == NULL) {
          if((fb = seekfld(hbuff,HBMAX,'s',locy)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locy);
            exit(1);
          }
          else ysflg = 1;
        }

        if(fb->dim != 1){
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locy);
          exit(1);
        }
        chky = fb->dsize1;
        if(ysflg == 0) lcy = varspace(&vbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        else lcy = varspace(&sbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        mxfact = varname(locx);
        if((fb = seekfld(hbuff,HBMAX,'v',locx)) == NULL) {
          if((fb = seekfld(hbuff,HBMAX,'s',locx)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locx);
            exit(1);
          }
          else xsflg = 1;
        }
        if( fb->dim !=1){
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locx);
          exit(1);
        }
        chkx = fb->dsize1;
        if(xsflg == 0) lcx = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
        else lcx = varspace(&sbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
      }
      if (sptr != 1) syntax(clist);
      mxfact = varname(xname);
      if ((fb = seekfld(hbuff,HBMAX,'v',xname)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",xname);
        exit(1);
      }
      if (fb->dim != 2) {
        fprintf(stderr,"cdfplot: grid requires 2 dimensions for %s\n",xname);
        exit(1);
      }
      nx = fb->dsize1;
      ny = fb->dsize2;
      var = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xnew);
      if((unv == 1) && (chkx != nx || chky != ny)) {
        fprintf(stderr,"cdfplot: number of points and number of point");
        fprintf(stderr," locations must match\n");
        exit(1);
      }
      /* compute default grid skips if needed */
      if (cfgot == 0) {
        skipx = nx/7 + 1;
        skipy = ny/7 + 1;
      }
      /* draw axes for uneven grid if needed */
      if((unv == 1) && (axdone == 0)) {
        if(sxdone == 0) scale(nx,QUANT,lcx,&sx2,&sx1);
        if(sydone == 0) scale(ny,QUANT,lcy,&sy2,&sy1);
        if(xlabel == 0) strcpy(lxname,locx);
        if(ylabel == 0) strcpy(lyname,locy);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* draw axes if needed */
      if (axdone == 0) {
        if (sxdone == 0) setlims(&sx1,&sx2,fb->dname1,fb->dsize1);
        if (sydone == 0) setlims(&sy1,&sy2,fb->dname2,fb->dsize2);
        if (xlabel == 0) strcpy(lxname,fb->dname1);
        if (ylabel == 0) strcpy(lyname,fb->dname2);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* add comment */
      sprintf(alpha,"grid(%s*%g)\n",xname,mxfact);
      strcpy(com[nclines++],alpha);
      /* do grid operation */
      /* make grid uneven if needed */
      if(unv == 1) grid2(1,nx,ny,skipx,skipy,var,lcx,lcy);
      else grid(1,nx,ny,skipx,skipy,var);
      /* reset uneven flag */
      unv = 0;
      /* release space if necessary */
      if (xnew) free(var);
      if (xlnew) free(lcx);
      if (ylnew) free(lcy);
    }
    /* vector plot */
    else if ((strcmp(alpha,"v") == 0) || (strcmp(alpha,"uv") == 0)) {
      if ((sptr != 4) && (sptr != 6) && (sptr != 2)) syntax(clist);
      if(strcmp(alpha,"uv") == 0) unv = 1;
      /* get the scaling factors */
      if( sptr == 6 || (unv == 0 && sptr == 4)) {
        sy = atof(pop());
        sx = atof(pop());
        scldne = 1;
      }
      /* find the variables */
      if(unv == 1) {
	if (sptr != 4) syntax(clist);
        myfact = varname(locy);
        if ((fb = seekfld(hbuff,HBMAX,'v',locy)) == NULL) {
          if (( fb = seekfld(hbuff,HBMAX,'s',locy)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locy);
            exit(1);
          }
          else ysflg = 1;
        }
        if (fb->dim != 1) {
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locy);
          exit(1);
        }
        chky = fb->dsize1;
        if(ysflg == 0) lcy = varspace(&vbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        else lcy = varspace(&sbuff[fb->ptr],fb->tsize,myfact,&ylnew);
        mxfact = varname(locx);
        if((fb = seekfld(hbuff,HBMAX,'v',locx)) == NULL) {
          if((fb = seekfld(hbuff,HBMAX,'s',locx)) == NULL) {
            fprintf(stderr,"cdfplot: field %s nonexistent\n",locx);
            exit(1);
          }
          else xsflg = 1;
        }
        if (fb->dim != 1) {
          fprintf(stderr,"cdfplot: %s must have only one dimension\n",locx);
          exit(1);
        }
        chkx = fb->dsize1;
        if(xsflg == 0) lcx = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
        else lcx = varspace(&sbuff[fb->ptr],fb->tsize,mxfact,&xlnew);
      }
      if (sptr != 2) syntax(clist);
      myfact = varname(yname);
      if ((fb = seekfld(hbuff,HBMAX,'v',yname)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",yname);
        exit(1);
      }
      if (fb->dim != 2) {
        fprintf(stderr,"cdfplot: vector requires 2 dimensions for %s\n",yname);
        exit(1);
      }
      nx = fb->dsize1;
      ny = fb->dsize2;
      strcpy(xdname,fb->dname1);
      strcpy(ydname,fb->dname2);
      yvar = varspace(&vbuff[fb->ptr],fb->tsize,myfact,&ynew);
      if ((unv == 1) && ( chkx != nx || chky != ny )) {
        fprintf(stderr,"cdfplot: number of points and number of point ");
        fprintf(stderr,"locations must match\n");
        exit(1);
      }
      mxfact = varname(xname);
      if ((fb = seekfld(hbuff,HBMAX,'v',xname)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",xname);
        exit(1);
      }
      if (fb->dim != 2) {
        fprintf(stderr,"cdfplot: vector requires 2 dimensions for %s\n",xname);
        exit(1);
      }
      if ((nx != fb->dsize1) || (ny != fb->dsize2) ||
          (strcmp(xdname,fb->dname1) != 0) ||
          (strcmp(ydname,fb->dname2) != 0)) {
        fprintf(stderr,"cdfplot: domains incompatible in vector plot\n");
        exit(1);
      }
      xvar = varspace(&vbuff[fb->ptr],fb->tsize,mxfact,&xnew);
      /* draw axes if needed */
      /* draw uneven axes if needed */
      if ((unv == 1) && (axdone == 0)) {
        if(sxdone == 0) scale(nx,QUANT,lcx,&sx2,&sx1);
        if(sydone == 0) scale(ny,QUANT,lcy,&sy2,&sy1);
        if(xlabel == 0) strcpy(lxname,locx);
        if(ylabel == 0) strcpy(lyname,locy);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      else if (axdone == 0) {
        if (sxdone == 0) setlims(&sx1,&sx2,fb->dname1,fb->dsize1);
        if (sydone == 0) setlims(&sy1,&sy2,fb->dname2,fb->dsize2);
        if (xlabel == 0) strcpy(lxname,fb->dname1);
        if (ylabel == 0) strcpy(lyname,fb->dname2);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* figure out the vector sizing if needed */
      if (scldne == 0)
        vscale(1,nx,ny,xvar,yvar,&sx,&sy);
      /* add comment */
      sprintf(alpha,"vector(%s*%g,%s*%g,%.3g,%.3g)\n",
          xname,mxfact,yname,myfact,sx,sy);
      strcpy(com[nclines++],alpha);
      /* draw vectors */
      if (unv == 1) vector2(1,nx,ny,sx,sy,xvar,yvar,lcx,lcy);
      else vector(1,nx,ny,sx,sy,xvar,yvar);
      unv = 0;
      /* release space if necessary */
      if (xnew) free(xvar);
      if (ynew) free(yvar);
      if (xlnew) free(lcx);
      if (ylnew) free(lcy);
    }
    /* station plot */
    else if (strcmp(alpha,"s") == 0) {
      if((sptr != 4) && ((sptr < 8) || (sptr > 8 + MAXLBL))) syntax(clist);
      skip = atof(pop());
      /* get label variable names */
      nlabels = 0;
      while (sptr > 7) {
        mfactl[nlabels] = varname(labels[nlabels]);
        if ((fb = seekfld(hbuff,HBMAX,'v',labels[nlabels])) == NULL) {
          fprintf(stderr,"cdfplot: variable field %s nonexistent\n",
              labels[nlabels]);
          exit(1);
        }
        labptr[nlabels] = vbuff + fb->ptr;
        lablen[nlabels++] = fb->tsize;
      }
      /* get scales and velocity variable names */
      gotvel = 0;
      if (sptr == 7) {
        sy = atof(pop());
        sx = atof(pop());
        myfact = varname(yname);
        if ((fb = seekfld(hbuff,HBMAX,'v',yname)) == NULL) {
          fprintf(stderr,"cdfplot: variable field %s nonexistent\n",yname);
          exit(1);
        }
        velyptr = vbuff + fb->ptr;
        velylen = fb->tsize;
        mxfact = varname(xname);
        if ((fb = seekfld(hbuff,HBMAX,'v',xname)) == NULL) {
          fprintf(stderr,"cdfplot: variable field %s nonexistent\n",xname);
          exit(1);
        }
        velxptr = vbuff + fb->ptr;
        velxlen = fb->tsize;
        gotvel = 1;
      }
      /* get location and station plot type */
      strcpy(y_loc,pop());
      if ((fb = seekfld(hbuff,HBMAX,'v',y_loc)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",y_loc);
        exit(1);
      }
      locyptr = vbuff + fb->ptr;
      locylen = fb->tsize;
      strcpy(x_loc,pop());
      if ((fb = seekfld(hbuff,HBMAX,'v',x_loc)) == NULL) {
        fprintf(stderr,"cdfplot: variable field %s nonexistent\n",x_loc);
        exit(1);
      }
      locxptr = vbuff + fb->ptr;
      locxlen = fb->tsize;
      ltype = atoi(pop());
      if (sptr != 0) syntax(clist);
      /* check for length consistency */
      if (locylen != locxlen) lenprob();
      if (gotvel && ((velxlen != locxlen) || (velylen != locxlen))) lenprob();
      for (loop = 0; loop < nlabels; loop++) {
        if (lablen[loop] != locxlen) lenprob();
      }
      /* draw axes if needed */
      if (axdone == 0) {
        if (xlabel == 0) strcpy(lxname,x_loc);
        if (ylabel == 0) strcpy(lyname,y_loc);
        windax(1,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
        axdone = 1;
      }
      /* add comment */
      sprintf(alpha,"station(%d,%s,%s",ltype,x_loc,y_loc);
      if (gotvel) {
        sprintf(beta,"%s,%s*%g,%s*%g",alpha,xname,mxfact,yname,myfact);
        strcpy(com[nclines++],beta);
        sprintf(beta,",%.3g,%.3g",sx,sy);
        if (nlabels == 0) {
          strcat(beta,")\n");
          strcpy(com[nclines++],beta);
        }
        else {
          strcat(beta,"\n");
          strcpy(com[nclines++],beta);
          strcpy(alpha,"");
          loop = nlabels;
          while (--loop >= 0) {
            sprintf(beta,",%s*%g",labels[loop],mfactl[loop]);
            strcat(alpha,beta);
          }
          strcat(alpha,")\n");
          strcpy(com[nclines++],alpha);
        }
      }
      else {
        strcat(alpha,")\n");
        strcpy(com[nclines++],alpha);
      }
      /* do station plot for all stations */
      for (loop = 0; loop < locxlen; loop+= (skip + 1)) {
        /* skip if out of range */
        if ((locxptr[loop]<MIN(sx1,sx2))||(locxptr[loop]>MAX(sx1,sx2)))
             continue;
        if ((locyptr[loop]<MIN(sy1,sy2))||(locyptr[loop]>MAX(sy1,sy2)))
             continue;
        /* labels */
        loop2 = nlabels;
        strcpy(beta,"");
        while (--loop2 >= 0) {
          if (fabs(labptr[loop2][loop]) > badlim) {
            if (loop2 == 0) strcat(beta,"*");
            else strcat(beta,"*;");
          }
          else {
            if (loop2 == 0)
              sprintf(alpha,"%.0f",labptr[loop2][loop]*mfactl[loop2]);
            else
              sprintf(alpha,"%.0f;",labptr[loop2][loop]*mfactl[loop2]);
            strcat(beta,alpha);
          }
        }
        if (nlabels > 0) label(1,locxptr[loop],locyptr[loop],beta);
        /* identifier mark */
        mark(1,ltype,locxptr[loop],locyptr[loop],0.5);
        /* skip if no velocities */
        if (!gotvel) continue;
        /* velocity vector */
        velxt[0] = locxptr[loop];
        velyt[0] = locyptr[loop];
        rx = sx*velxptr[loop]*mxfact;
        ry = sy*velyptr[loop]*myfact;
        velxt[4] = velxt[1] = velxt[0] + rx;
        velyt[4] = velyt[1] = velyt[0] + ry;
        velxt[2] = velxt[1] + VLEN*rx + VWID*ry;
        velyt[2] = velyt[1] - VWID*rx + VLEN*ry;
        velxt[3] = velxt[1] + VLEN*rx - VWID*ry;
        velyt[3] = velyt[1] + VWID*rx + VLEN*ry;
        if ((fabs(velxptr[loop]) < badlim) &&
            (fabs(velyptr[loop]) < badlim)) line(1,1,5,velxt,velyt);
      }
      /* reset line type to 1 */
      ltype = 1;
    }
    /* omit annotation */
    else if (strcmp(alpha,"a") == 0) {
      aomit = 1;
      if (sptr == 5 || sptr == 1 || sptr == 3) wintype = atoi(pop());
      if ((sptr == 4) || (sptr == 2)) {
        yhigh = atof(pop());
        xhigh = atof(pop());
      }
      if (sptr == 2) {
        ylow = atof(pop());
        xlow = atof(pop());
      }
      if (sptr != 0) syntax(clist);
      /* reset the window flags for the new window */
      axdone = 0;
      sxdone = 0;
      sydone = 0;
      /* make number of tics proportional to window size */
      ntx = (xhigh - xlow)*6/10;
      nty = (yhigh - ylow)*6/10;
    }
    /* label */
    else if (strcmp(alpha,"l") == 0) {
      if (sptr != 3) syntax(clist);
      if (axdone != 1) {
        fprintf(stderr,"cdfplot: do plotting command before making label\n");
        exit(1);
      }
      /* pop comment location and text */
      ylab = atof(pop());
      xlab = atof(pop());
      strcpy(xname,pop());
      /* convert underscores to spaces */
      loop = 0;
      while (xname[loop] != '\0') {
        if (xname[loop] == '_') xname[loop] = ' ';
        loop++;
      }
      /* write label */
      label(1,xlab,ylab,xname);
    }
    /* x-axis label */
    else if (strcmp(alpha,"lx") == 0) {
      if (sptr != 1) syntax(clist);
      if (axdone == 1) {
        fprintf(stderr,"cdfplot: do axis label before plotting\n");
        exit(1);
      }
      /* store the label in the appropriate place and set the x label flag */
      strcpy(lxname,pop());
      xlabel = 1;
      /* convert underscores to spaces */
      loop = 0;
      while (lxname[loop] != '\0') {
        if (lxname[loop] == '_') lxname[loop] = ' ';
        loop++;
      }
    }
    /* y-axis label */
    else if (strcmp(alpha,"ly") == 0) {
      if (sptr != 1) syntax(clist);
      if (axdone == 1) {
        fprintf(stderr,"cdfplot: do axis label before plotting\n");
        exit(1);
      }
      /* store the label in the appropriate place and set the y label flag */
      strcpy(lyname,pop());
      ylabel = 1;
      /* convert underscores to spaces */
      loop = 0;
      while (lyname[loop] != '\0') {
        if (lyname[loop] == '_') lyname[loop] = ' ';
        loop++;
      }
    }
    /* unknown command */
    else {
      fprintf(stderr,"cdfplot: %s unknown command\n",alpha);
      exit(1);
    }
    /* get out if end of command list */
    if (endchar == '\0') break;
  }
  /* write annotation */
  if (aomit == 1) {
    gpause();
    return;
  }
  window(2,10.,15.,0.,10.);
  /* first, header comment and locally added stuff */
  xloc = XLOC;
  yloc = YMAX;
  for (loop = 0; loop < nclines; loop++) {
    label(2,xloc,yloc,com[loop]);
    yloc -= YDEC;
    if (yloc < YMIN) yloc = YMIN;
  }
  /* next, scalar fields in static slice */
  loop = 0;
  while ((fb = getfld(hbuff,HBMAX,'s',loop++)) != NULL) {
    if (fb->dim == 0) {
      sprintf(alpha,"%s = %g",fb->fname,sbuff[fb->ptr]);
      label(2,xloc,yloc,alpha);
      yloc -= YDEC;
      if (yloc < YMIN) yloc = YMIN;
    }
  }
  /* finally, scalar fields in variable slice */
  loop = 0;
  while ((fb = getfld(hbuff,HBMAX,'v',loop++)) != NULL) {
    if (fb->dim == 0) {
      sprintf(alpha,"%s = %g",fb->fname,vbuff[fb->ptr]);
      label(2,xloc,yloc,alpha);
      yloc -= YDEC;
      if (yloc < YMIN) yloc = YMIN;
    }
  }
  /* graphics pause */
  axdone = 0;
  gpause();
}

static float *getvar1(fname,dname,size,mfact,new,occurs)
/* Return a pointer to a 1-D variable and fill in appropriate stuff
 * in argument list.  Return NULL if not found.
 */
char *fname,*dname;
int *size,*new,*occurs;
float *mfact;
{
  struct field *f;
  char type;
  float *var;
  float varname(),*varspace();
  char *pop();
  *mfact = varname(fname);
  type = 'v';
  f = seekfld(hbuff,HBMAX,type,fname);
  if (f == NULL) {
    type = 's';
    f = seekfld(hbuff,HBMAX,type,fname);
  }
  if (f == NULL) {
    strcpy(dname,"");
    *size = 0;
    return(NULL);
  }
  if (f->dim == 1) {
    strcpy(dname,f->dname1);
    *size = f->dsize1;
  }
  else if (f->dim == 2) {
    strcpy(dname,f->dname2);
    *size = f->dsize2;
  }
  else if (f->dim == 3) {
    strcpy(dname,f->dname3);
    *size = f->dsize3;
  }
  else if (f->dim == 4) {
    strcpy(dname,f->dname4);
    *size = f->dsize4;
  }
  else {
    fprintf(stderr,"cdfplot: trying to graph zero-dimensional field\n");
    exit(1);
  }
  *occurs = f->tsize/(*size);
  if (type == 'v') var = &vbuff[f->ptr];
  else var = &sbuff[f->ptr];
  var = varspace(var,f->tsize,*mfact,new);
  return(var);
}

static void syntax(command)
/* report syntax error and die */
char *command;
{
  fprintf(stderr,"cdfplot: syntax error in %s\n",command);
  exit(1);
}

static void setlims(s1,s2,dname,size)
/* This routine adjusts window space limits based on assumptions about
 * what certain parameters mean.  A parameter with name dname0 is assumed
 * to be a new starting point, while one named ddname is assumed to be
 * a grid increment.
 */
float *s1,*s2;
char *dname;
int size;
{
  char temp[LINE],value[LINE];
  /* look for new lower limit in parameter list */
  strcpy(temp,dname);
  strcat(temp,"0");
  if (seekpar(hbuff,HBMAX,temp,value) == OK) {
    *s2 -= *s1;
    *s1 = atof(value);
    *s2 += *s1;
  }
  /* now look for increment */
  strcpy(temp,"d");
  strcat(temp,dname);
  if (seekpar(hbuff,HBMAX,temp,value) == OK) {
    *s2 = *s1 + (size - 1)* atof(value);
  }
}

static float varname(name)
/* This routine pops the variable name off the stack and parses it for
 * a multiplier factor.  It puts the name in "name", and returns, the
 * multiplier factor, if found.  Otherwise, it returns 1.0.
 */
char *name;
{
  char temp[LINE],mult[LINE],c;
  char *pop();
  int sw,loop,poop,goop;
  /* pop stuff into temporary storage */
  strcpy(temp,pop());
  /* extract name and multiplier */
  sw = loop = goop = poop = 0;
  while ((c = temp[loop++]) != '\0') {
    if (c == '*') {
      sw = 1;
      continue;
    }
    if (sw == 0) {
      name[goop++] = c;
    }
    else {
      mult[poop++] = c;
    }
  }
  /* terminate strings */
  mult[poop] = '\0';
  name[goop] = '\0';
  /* return proper multiplier */
  if (mult[0] == '\0') return(1.0);
  else return(atof(mult));
}

static float *varspace(optr,size,mfact,new)
/* This routine just returns optr and sets *new = 0 if mfact = 1.0.
 * Otherwise it allocates new space and scales the variable with
 * mfact.  *new = 1 is returned.
 */
float *optr,mfact;
long size;
int *new;
{
  long loop;
  float *nptr;
  if (mfact == 1.0) {
    *new = 0;
    return(optr);
  }
  else {
    *new = 1;
    if ((nptr = (float *)malloc(sizeof(float)*size)) == NULL) {
      fprintf(stderr,"cdfplot: can't get enough memory\n");
      exit(1);
    }
    for (loop = 0; loop < size; loop++) {
      nptr[loop] = optr[loop]*mfact;
    }
    return(nptr);
  }
}

/* error exit if station plot variables don't match in length */
static void lenprob()
{
  fprintf(stderr,"cdfplot: inconsistent field lengths in station plot\n");
  exit(1);
}

/* draw out the windows and the axes using wintype to decide what to show */
static void windax(nw,xlow,xhigh,ylow,yhigh,sx1,sx2,ntx,lxname,
     sy1,sy2,nty,lyname)
float xlow,xhigh,ylow,yhigh ;         /* location of the window */
float sx1,sx2,sy1,sy2;                /* values for axes */
int nw,ntx,nty;
char *lxname,*lyname;
{
  if (xhigh > 15.) xhigh = 15.;
  if (yhigh > 10.) yhigh = 10.;
  window(nw,xlow,xhigh,ylow,yhigh);
  if( wintype == 0) axes(nw,sx1,sx2,ntx,lxname,sy1,sy2,nty,lyname);
  else if(wintype == 1) axesnox(nw,sx1,sx2,ntx,sy1,sy2,nty,lyname);
  else if(wintype == 2) axesnoy(nw,sx1,sx2,ntx,lxname,sy1,sy2,nty);
  else axesnoxy(nw,sx1,sx2,ntx,sy1,sy2,nty);
}

/* compute a middle value in function */
static float midcomp(nx,ny,cint,var)
int nx,ny;
float cint;
float *var;
{
  long i,ival,count;
  float meanval,ratio;
  meanval = 0.;
  count = 0;
  for (i = 0; i < nx*ny; i++) {
    if (fabs(var[i]) < badlim) {
      meanval += var[i];
      count++;
    }
  }
  if (count > 0) meanval /= (float)count;
  else meanval = 0.;
  if (cint < 1.e-6) cint = 1.e-6;

/* quantize meanval by cint if possible with 32 bit longs */
  ratio = meanval/cint + .5;
  if (fabs(ratio) < 2.e9) {
    ival = meanval/cint + .5;
    meanval = ival*cint;
  }
  return(meanval);
}
