/* this program generates a cdf file of sounding data from an
 * ascii file of the form:
		comment line
		pres  temp  mixrat_or_dewpt [wind_speed wind_direction]
  		 "     "           "
			etc.
 * with the output having a pres, temp and dewpt or mixrat fields depending
 * on the flags used (-q means that mixing ratio is to be assumed, -d means
 * dewpoint)
 * if -c is specified, expect x-wind and y-wind instead of windspeed and
 * direction
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define MAXEL 200
#define DTOR 57.3

static int numt;                            /* number of levels */
static int loop;                            /* looping variables */
static int mixrat,stat;                  /* mixing ratio flag and no dp flag */
static int components;                 /* wind as components rather than s/d */
static char pline[82];                      /* line buffer */
static char buffer[HBMAX][LINE];            /* cdf output header */
static char cline[LINE],dtype[LINE];        /* cdf stuff */
static float *sbuff,*vbuff,*cdfpres;
static float *cdftemp,*cdfmr;
static float *cdfu,*cdfv;
static long nsbuff,nvbuff;
static float pt[MAXEL],t[MAXEL],dp[MAXEL];  /* field buffers */
static float ws[MAXEL],wd[MAXEL];

main(argc,argv)
int argc;
char *argv[];
{
  int c;
  extern int optind;

/* check arguments */
  if (argc == 1 || argc > 3) {
    fprintf(stderr,"Usage: cdfua -[dq] [-c]\n");
    exit(1);
  }
  mixrat = components = 0;
  while ((c = getopt(argc,argv,"qdc")) != -1) {
    switch (c) {
      case 'q':
        mixrat = 1;
        break;
      case 'd':
        mixrat = 0;
        break;
      case 'c':
        components = 1;
        break;
      case '?':
        exit(1);
    }
  }
  
/* read in the comment line */
  if (fgets(pline,80,stdin) == NULL) {
    fprintf(stderr,"%s: empty input file\n",argv[0]);
    exit(1);
  }

/* read in the sounding info */
  numt = 0;
  while (fgets(cline,80,stdin) != NULL) {
    stat = sscanf(cline,"%f %f %f %f %f",
         &pt[numt],&t[numt],&dp[numt],&ws[numt],&wd[numt]);
    if (stat == 3) {
      ws[numt] = wd[numt] = 0.;
    }
    else if (stat != 5) {
      fprintf(stderr,"%s: bad input format\n",argv[0]);
      exit(1);
    }
    numt++;
  }

/* create null header */
  nullhdr(buffer,HBMAX,"float");

/* add stuff to header */
  sprintf(cline,"%s: %s\n",argv[0],argv[1]);
  addcline(buffer,HBMAX,cline);
  addcline(buffer,HBMAX,pline);

/* add static fields */
  addfld(buffer,HBMAX,'s',"pres",100.0,0.0,'s',1,"pres",numt);

/* add var fields */
  addfld(buffer,HBMAX,'v',"temp",100.0,0.0,'s',1,"pres",numt);
  if (mixrat == 1) sprintf(dtype,"mixrat");
  else sprintf(dtype,"dewpt");
  addfld(buffer,HBMAX,'v',dtype,100.0,0.0,'s',1,"pres",numt);
  addfld(buffer,HBMAX,'v',"u",100.,0.,'s',1,"pres",numt);
  addfld(buffer,HBMAX,'v',"v",100.,0.,'s',1,"pres",numt);

/* make static and variable slice buffers */
  nsbuff = elemcnt(buffer,HBMAX,'s');
  nvbuff = elemcnt(buffer,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

/* get pointers to data */
  cdfpres = getptr(buffer,HBMAX,sbuff,'s','d',"pres");
  cdftemp = getptr(buffer,HBMAX,vbuff,'v','d',"temp");
  cdfmr = getptr(buffer,HBMAX,vbuff,'v','d',dtype);
  cdfu = getptr(buffer,HBMAX,vbuff,'v','d',"u");
  cdfv = getptr(buffer,HBMAX,vbuff,'v','d',"v");

/* put data into buffers */
  for (loop = 0 ;loop < numt; loop++) {
    cdfpres[loop] = pt[loop];
    cdftemp[loop] = t[loop];
    cdfmr[loop] = dp[loop];
    if (components) {
      cdfu[loop] = ws[loop];
      cdfv[loop] = wd[loop];
    }
    else {
      cdfu[loop] = -ws[loop]*sin(wd[loop]/DTOR);
      cdfv[loop] = -ws[loop]*cos(wd[loop]/DTOR);
    }
  }

/* output cdf info */
  puthdr(stdout,buffer,HBMAX);
  putslice(stdout,nsbuff,sbuff);
  putslice(stdout,nvbuff,vbuff);
  exit(0);
}
