/* cdfuaplot -- This program plots upper air sounding data, with
 * temperature and dewpoint on a skew t log p chart and the winds
 * in u-v component form.  An optional flag makes program accept
 * mixing ratio instead of dewpoint.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define ZHI 16.
#define ZLO 0.
#define ZTICS 9
#define THI 40.
#define TLO -40.
#define TTICS 5
#define UMAX 30.
#define UTICS 5
#define DU 3.
#define ZS 7.63                      /* based on temp = 260 K */
#define SKEW 4.
#define DH 0.6
#define NX 21
#define NY 17
#define L 2.45e6
#define CP 1.005e3
#define KAPA 0.286
#define EPS 0.623
#define A 17.269
#define B 35.86
#define C 6.1078
#define D 273.16
#define ITMAX 5

static char hbuff[HBMAX][LINE];             /* header buffer */
static char cline[LINE];                    /* line buffer */
static float *sbuff,*vbuff;                 /* slice buffer */
static long nsbuff,nvbuff;                  /* slice buffer length */
static float *pt,*t,*dp,*pw,*u,*v;          /* field pointers */
static char *comment;                       /* comment pointer */
static int numt,numw;                    /* number of thermo and wind levels */
static float *station,*hour,*date,*month,*year;  /* more field pointers */
static float *latgrid,*longgrid;            /* for gridded output */
static struct field *fp;
static char pval[LINE];                     /* parameter value */
static float es();
static float *zt,*zw;                       /* pointers to height fields */
static int loop,ix,iy;                      /* looping variables */
static int mixrat;                          /* mixing ratio flag */
static int parcelline;                      /* draw a parcel track flag */
static float thetes0;                /* specified value of thetes for parcel */
static float badlim,bad;                    /* bad data limit */
static float x[2],y[2];                     /* temps for drawing lines */
static float umax,qs,esat,temp,pres,height; /* more temporaries */
static float theta[NX][NY];                 /* potential temperature */
static float thetes[NX][NY];                /* wet bulb potential temp */
static float logq[NX][NY];                  /* log of mixing ratio */
static float dewpoint(),es();

main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if (argc > 7 && strcmp(argv[1],"-q") == 0) {
/* mixing ratio instead of dewpt */
    for (loop = 2; loop < argc; loop++) {
      argv[loop - 1] = argv[loop];
    }
    argc--;
    mixrat = 1;
  }
  else mixrat = 0;
  parcelline = 0;
  if (argc == 7) comment = NULL;
  else if (argc == 8) comment = argv[7];
  else if (argc == 9) {
    comment = argv[7];
    parcelline = 1;
  }
  else {
    fprintf(stderr,
         "Usage: cdfuaplot [-q] pt t dp(or q) pw u v ['comment' [thes0]]\n");
    exit(1);
  }
/* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"cdfuaplot: need float format\n");
    exit(1);
  }
/* get static slice */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  getslice(stdin,nsbuff,sbuff);
/* set up for variable slices */
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbuff);
/* set up pointers */
/* pressure at thermodynamic levels */
  if ((pt = getptr2(hbuff,HBMAX,vbuff,'v','r',argv[1],&fp)) == NULL) {
    pt = getptr2(hbuff,HBMAX,sbuff,'s','d',argv[1],&fp);
  }
  numt = fp->tsize;
/* temperature */
  t = getptr(hbuff,HBMAX,vbuff,'v','d',argv[2]);
/* dewpoint (or mixing ratio) */
  dp = getptr(hbuff,HBMAX,vbuff,'v','d',argv[3]);
/* pressure at wind levels */
  if ((pw = getptr2(hbuff,HBMAX,vbuff,'v','r',argv[4],&fp)) == NULL) {
    pw = getptr2(hbuff,HBMAX,sbuff,'s','d',argv[4],&fp);
  }
  numw = fp->tsize;
/* u (from west) wind component */
  u = getptr(hbuff,HBMAX,vbuff,'v','d',argv[5]);
/* v (from south) wind component */
  v = getptr(hbuff,HBMAX,vbuff,'v','d',argv[6]);
/* station */
  station = getptr(hbuff,HBMAX,vbuff,'v','r',"station");
/* hour */
  if ((hour = getptr(hbuff,HBMAX,vbuff,'v','r',"hour")) == NULL) {
    hour = getptr(hbuff,HBMAX,vbuff,'v','r',"hourav");
  }
/* date */
  date = getptr(hbuff,HBMAX,vbuff,'v','r',"date");
/* month */
  month = getptr(hbuff,HBMAX,vbuff,'v','r',"month");
/* year */
  year = getptr(hbuff,HBMAX,vbuff,'v','r',"year");
/* gridded latitude and longitude */
  latgrid = getptr(hbuff,HBMAX,sbuff,'s','r',"latgrid");
  longgrid = getptr(hbuff,HBMAX,sbuff,'s','r',"longgrid");
/* open graphics */
  gopen();
/* get bad data limit */
  if (seekpar(hbuff,HBMAX,"badlim",pval) == OK) {
    badset(badlim = atof(pval));
  }
  else {
    badset(badlim = BADLIM);
  }
  if (seekpar(hbuff,HBMAX,"bad",pval) == OK) bad = atof(pval);
  else bad = BAD;
/* allocate nominal height fields */
  zt = getbuff(numt);
  zw = getbuff(numw);
/* compute background fields */
  for (ix = 0; ix < NX; ix++) {
    for (iy = 0; iy < NY; iy++) {
/* potential temperature */
      height = ZLO + (ZHI - ZLO)*((float)iy)/((float)(NY - 1));
      temp = TLO + (THI - TLO)*((float)ix)/((float)(NX - 1)) - SKEW*height;
      pres = 1000.*exp(-height/ZS);
      theta[ix][iy] = (temp + D)*pow(1000./pres,KAPA) - D;
/* log saturation mixing ratio */
      esat = es(temp);
      logq[ix][iy] = log10(qs = EPS*esat/(pres - esat));
      if (logq[ix][iy] < -4.2) logq[ix][iy] = -4.2;
/* saturated equivalent potential temperature */
      thetes[ix][iy] = (theta[ix][iy] + D)*exp(L*qs/(CP*(temp + D))) - D;
      if (thetes[ix][iy] > 90.) thetes[ix][iy] = 90.;
    }
  }
/* for parcel line only */
  if (parcelline) {
    thetes0 = atof(argv[8]);
  }
/* loop on variable slices */
  while (getslice(stdin,nvbuff,vbuff) != EOF) {
/* compute the nominal height fields (const scale height assumption) */
    for (loop = 0; loop < numt; loop++) {
      if (fabs(pt[loop]) < badlim) {
        zt[loop] = ZS*log(1000./pt[loop]);
      }
      else zt[loop] = 1.001*badlim;
    }
    for (loop = 0; loop < numw; loop++) {
      if (fabs(pw[loop]) < badlim) {
        zw[loop] = ZS*log(1000./pw[loop]);
      }
      else zw[loop] = 1.001*badlim;
    }
/* compute the maximum wind */
    umax = 0.;
    for (loop = 0; loop < numw; loop++) {
      if((fabs(u[loop]) < badlim) && (fabs(u[loop]) > umax))
           umax = fabs(u[loop]);
      if((fabs(v[loop]) < badlim) && (fabs(v[loop]) > umax))
           umax = fabs(v[loop]);
    }
    umax = ((int)(umax/UMAX) + 1)*UMAX;
/* skew the temperatures */
    for (loop = 0; loop < numt; loop++) {
/* convert mixing ratio to dewpoint if necessary */
      if (mixrat) dp[loop] = dewpoint(dp[loop],pt[loop],t[loop]);
      t[loop] += zt[loop]*SKEW;
      dp[loop] += zt[loop]*SKEW;
    }
/* make the plot */
    gclear();
    window(1,0.,7.5,0.,10.);
    window(2,7.5,15.,0.,10.);
    axes(1,TLO,THI,TTICS,"temp (C)",ZLO,ZHI,ZTICS,"height (km)");
    axes(2,-umax,umax,UTICS,"u-v wind (m/s)",
         ZLO,ZHI,ZTICS,"height (km)");
/* constant pressure lines */
    for (pres = 900.; (height = ZS*log(1000./pres)) < ZHI; pres -= 100.) {
      x[0] = TLO;
      x[1] = THI;
      y[0] = y[1] = height;
      line(1,1,2,x,y);
      x[0] = -umax;
      x[1] = umax;
      line(2,1,2,x,y);
      sprintf(pval,"%.0f",pres);
      label(1,TLO + 2.,height - 0.3,pval);
    }
/* background contours */
    conto(1,NX,NY,20.,3,theta);
    conto(1,NX,NY,0.5,3,logq);
    conto(1,NX,NY,20.,3,thetes);
/* for parcel line only */
    if (parcelline) {
      for (ix = 0; ix < NX; ix++) {
        for (iy = 0; iy < NY; iy++) {
          thetes[ix][iy] -= thetes0;
        }
      }
      conto(1,NX,NY,1000.,1,thetes);
    }
/* zero C line */
    x[0] = 0.;
    y[0] = 0.;
    x[1] = THI;
    y[1] = THI/SKEW;
/* draw the data */
    line(1,1,2,x,y);
    line(1,1,numt,t,zt);
    line(1,1,numt,dp,zt);
    line(2,1,numw,u,zw);
    line(2,3,numw,v,zw);
/* comments */
    height = ZHI - 2.*DH;
    if (comment != NULL) {
      label(2,-umax + DU,height,comment);
      height -= DH;
    }
    if (parcelline) {
      sprintf(cline,"parcel theta-es: %.1f C",thetes0);
      label(2,-umax + DU,height,cline);
      height -= DH;
    }
    if (latgrid != NULL) {
      sprintf(pval,"latitude: %.1f",*latgrid);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    if (longgrid != NULL) {
      sprintf(pval,"longitude: %.1f",*longgrid);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    if (station != NULL) {
      sprintf(pval,"station: %.0f",*station);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    if (hour != NULL) {
      sprintf(pval,"hour: %.0f",*hour);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    if (date != NULL) {
      sprintf(pval,"date: %.0f",*date);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    if (month != NULL) {
      sprintf(pval,"month: %.0f",*month);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    if (year != NULL) {
      sprintf(pval,"year: %.0f",*year);
      label(2,-umax + DU,height,pval);
      height -= DH;
    }
    gpause();
  }
/* close graphics */
  gclose();
  exit(0);
}

/* teten's formula for water saturation vapor pressure (mb)
 * temperature in C
 */
static float es(t)
float t;
{
  return(C*exp(A*t/(t + D - B)));
}

/* convert mixing ratio (g/kg) and pressure (mb) to dewpoint (C) */
static float dewpoint(mr,p,t)
float mr;                   /* mixing ratio */
float p;                    /* pressure */
float t;                    /* temperature (starting value for iteration) */
{
  float e,dp;
  int poop;
  if (fabs(mr) > badlim || fabs(p) > badlim || fabs(t) > badlim) return(bad);
  if (t < -40.) return(bad);
  dp = t;
  e = mr*p/(1000.*EPS + mr);
  for (poop = 0; poop < ITMAX; poop++) {
    dp = dp + (e - es(dp))/(es(dp + 1.) - es(dp));
  }
  return(dp);
}
