/* hrdair -- Translate hrd in situ aircraft tape data to Candis
 *      format.  Assumes that tape has been read by the rtape utility.
 *      This is the second version of this program.  David Leon
 *      pioneered the first version.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define HRDBUF 5000               /* max number of shorts */
#define MAXNAME 10                /* maximum size of a P-3 field name */
#define LOGRECS 106               /* maximum number of P-3 fields */
#define DIVOFFSET 11              /* convert position in divisor array
                                     to position in record 5 */
#define NLOGREC 10                /* number of logical records in phys rec 5 */

/* hrd stuff */
short hrdin[HRDBUF];              /* input buffer */
int recsize;                      /* record size */
int rectype;                      /* type of record */
int nhrdflds;                     /* number of hrd fields */
int nrecflds;                     /* number of recorded fields */
int rec3flag;                     /* set if a record 3 has been processed */
int rec1flag;                     /* set if a record 1 has been processed */
char rtheader[9];                 /* rtape header string */
float *divisor;                   /* divisors for calibration */
struct {                          /* positions and names of P-3 fields */
  int addr;
  char name[MAXNAME];
} hrdlist[] = {
  16,"ra",
  17,"ps",
  18,"ta",
  19,"tw1",
  20,"rd",
  21,"rs",
  22,"gs",
  23,"ts",
  24,"wgs",
  25,"tk",
  26,"hd",
  27,"pc",
  28,"rl",
  29,"aa",
  30,"sa",
  31,"lw",
  32,"pq",
  33,"td",
  34,"ru",
  35,"sw3",
  36,"utail",
  37,"vtail",
  38,"wtail",
  40,"ga",
  41,"pa",
  42,"dv",
  43,"ht",
  44,"sp",
  45,"rh",
  46,"tv",
  47,"was",
  48,"gm",
  49,"ama",
  50,"da",
  51,"gsx",
  52,"gsy",
  53,"tx",
  54,"ty",
  55,"wx",
  56,"wy",
  57,"wz",
  58,"ws",
  59,"wd",
  61,"ee",
  62,"mr",
  63,"pt",
  64,"et",
  65,"wxb",
  66,"wyb",
  67,"wsb",
  68,"wdb",
  69,"av1",
  70,"av2",
  71,"wac",
  72,"bt1",
  73,"bt2",
  74,"bt3",
  75,"nav",
  76,"itmp",
   0,""                         /* this terminates the list */
};
int aircraft,year,mon,day,starthour,startmin,startsec,endhour,endmin,endsec;
long start,endit,seconds;
short getshort();                 /* routine to extract shorts */

/* candis stuff */
char hbuf[HBMAX][LINE];
char cline[LINE];
float *sbuff,*vbuff;
long nsbuff,nvbuff;
float *tptime,*lon,*lat,*ptr[LOGRECS];
float *sw1,*sw2,*fl1,*fl2;

main()
{
  int i,j,lrec;
  short *hrdx;
  float dayoffset,oldtptime;

/* loop on hrd record access */
  rtheader[8] = '\0';
  rec1flag = rec3flag = 0;
  while (fread(rtheader,sizeof(char),8,stdin) == 8) {

/* check the record size */
    recsize = atoi(rtheader)/2;
    if (recsize < 0 || recsize > HRDBUF*2) {
      fprintf(stderr,"hrdair: input record = %d bytes is too big\n",recsize);
      exit(1);
    }

/* read the record */
    if (fread(hrdin,sizeof(short),recsize,stdin) != recsize) {
      fprintf(stderr,"hrdair: terminating with bad record\n");
      exit(0);
    }

/* get the record type */
    rectype = getshort(hrdin,0);

/* do things for various record types */

/* tape header record */
    if (rectype == 1) {
      if (rec1flag == 1) {
        fprintf(stderr,"hrdair: multiple type 1 records!\n");
        exit(1);
      }
      rec1flag = 1;

/* collect header data for later use */
      aircraft = getshort(hrdin,2);
      year = getshort(hrdin,3);
      mon = getshort(hrdin,4);
      day = getshort(hrdin,5);
      starthour = getshort(hrdin,6);
      startmin = getshort(hrdin,7);
      startsec = getshort(hrdin,8);
      endhour = getshort(hrdin,9);
      endmin = getshort(hrdin,10);
      endsec = getshort(hrdin,11);
    }

/* calibration coefficients */
    else if (rectype == 3) {
      if (!rec1flag) {
        fprintf(stderr,"hrdair: found rec 3 without preceding rec 1\n");
        exit(1);
      }
      if (rec3flag == 1) {
        fprintf(stderr,"hrdair: multiple type 3 records!\n");
        exit(1);
      }
      rec3flag = 1;

/* save the calibration coefficients */
      nhrdflds = (recsize - 2)/2;
      divisor = getbuff(nhrdflds);
      for (i = 0; i < nhrdflds; i++) {
        divisor[i] = (float)getshort(hrdin,3 + 2*i);
      }

/* check for zero divisors on recorded fields */
      nrecflds = 0;
      while (hrdlist[nrecflds].addr > 0) {
        i = hrdlist[nrecflds].addr - 1;
        if (i > nhrdflds) {
          fprintf(stderr,"hrdair: address in hrdlist out of range\n");
          exit(1);
        }
        if (divisor[i - DIVOFFSET] == 0.) {
          fprintf(stderr,"hrdair: divisor at %d is zero\n",i + 1);
          exit(1);
        }
        nrecflds++;
      }

/* set up output Candis file */
/* comments first */
      nullhdr(hbuf,HBMAX,"float");
      addcline(hbuf,HBMAX,"hrdair:\n");
      sprintf(cline,"  aircraft = %d\n",aircraft);
      addcline(hbuf,HBMAX,cline);
      sprintf(cline,"  date = %d-%d-%d\n",year,mon,day);
      addcline(hbuf,HBMAX,cline);
      sprintf(cline,"  start time = %d:%d:%d\n",starthour,startmin,startsec);
      addcline(hbuf,HBMAX,cline);
      sprintf(cline,"  end time = %d:%d:%d\n",endhour,endmin,endsec);
      addcline(hbuf,HBMAX,cline);
      start = starthour*3600 + startmin*60 + startsec;
      endit = endhour*3600 + endmin*60 + endsec;
      seconds = endit - start + 1;
      if (seconds < 86400) seconds += 86400;
      sprintf(cline,"  total seconds = %d\n",seconds);
      addcline(hbuf,HBMAX,cline);

/* one proforma field in static slice */
      addfld(hbuf,HBMAX,'s',"aircraft",1.,0.,'s',0);

/* variable slice */
      addfld(hbuf,HBMAX,'v',"tptime",1.,0.,'l',0);
      addfld(hbuf,HBMAX,'v',"lon",100000.,0.,'l',0);
      addfld(hbuf,HBMAX,'v',"lat",100000.,0.,'l',0);
      addfld(hbuf,HBMAX,'v',"sw1",1.,0.,'s',0);
      addfld(hbuf,HBMAX,'v',"sw2",1.,0.,'s',0);
      addfld(hbuf,HBMAX,'v',"fl1",1.,0.,'s',0);
      addfld(hbuf,HBMAX,'v',"fl2",1.,0.,'s',0);
      for (i = 0; i < nrecflds; i++) {
        addfld(hbuf,HBMAX,'v',hrdlist[i].name,divisor[hrdlist[i].addr - 1],
             0.,'s',0);
      }

/* allocate space for slice buffers */
      nsbuff = elemcnt(hbuf,HBMAX,'s');
      nvbuff = elemcnt(hbuf,HBMAX,'v');
      sbuff = getbuff(nsbuff);
      vbuff = getbuff(nvbuff);

/* fill static slice */
      sbuff[0] = (float)aircraft;

/* get variable slice pointers */
      tptime = getptr(hbuf,HBMAX,vbuff,'v','d',"tptime");
      lon = getptr(hbuf,HBMAX,vbuff,'v','d',"lon");
      lat = getptr(hbuf,HBMAX,vbuff,'v','d',"lat");
      sw1 = getptr(hbuf,HBMAX,vbuff,'v','d',"sw1");
      sw2 = getptr(hbuf,HBMAX,vbuff,'v','d',"sw2");
      fl1 = getptr(hbuf,HBMAX,vbuff,'v','d',"fl1");
      fl2 = getptr(hbuf,HBMAX,vbuff,'v','d',"fl2");
      for (i = 0; i < nrecflds; i++) {
        ptr[i] = getptr(hbuf,HBMAX,vbuff,'v','d',hrdlist[i].name);
      }

/* write header */
      puthdr(stdout,hbuf,HBMAX);

/* write static slice */
      putslice(stdout,nsbuff,sbuff);

/* initialize tptime */
      *tptime = start - 1;
      dayoffset = 0.;
    }

/* calibrated data record */
    else if (rectype == 5) {
      if (!rec1flag || !rec3flag) {
        fprintf(stderr,"hrdair: found rec 5 without preceding rec 1 or 3\n");
        exit(1);
      }

/* loop over logical records */
      hrdx = hrdin;
      for (lrec = 0; lrec < NLOGREC; lrec++) {

/* compute tptime */
        oldtptime = *tptime;
        *tptime = (float)(getshort(hrdx,2)*3600 + getshort(hrdx,3)*60 +
             getshort(hrdx,4)) + dayoffset;;
        if (*tptime < oldtptime) {
          dayoffset += 86400.;
          *tptime += 86400.;
        }
        if (*tptime != oldtptime + 1) {
          fprintf(stderr,"hrdair: warning: data gap between %g and %g\n",
               oldtptime,*tptime);
        }

/* compute lon and lat */
        *lon = (float)getshort(hrdx,13)/divisor[13 - DIVOFFSET] +
             (float)getshort(hrdx,14)/(60.*divisor[14 - DIVOFFSET]);
        *lat = (float)getshort(hrdx,11)/divisor[11 - DIVOFFSET] +
             (float)getshort(hrdx,12)/(60.*divisor[12 - DIVOFFSET]);

/* compute the stuff not included in the calibration scheme */
        *sw1 = (float)getshort(hrdx,7);
        *sw2 = (float)getshort(hrdx,8);
        *fl1 = (float)getshort(hrdx,9);
        *fl2 = (float)getshort(hrdx,10);

/* compute the other stuff */
        for (i = 0; i < nrecflds; i++) {
          j = hrdlist[i].addr - 1;
          *ptr[i] = (float)getshort(hrdx,j)/divisor[j - DIVOFFSET];
        }

/* write a variable slice */
        putslice(stdout,nvbuff,vbuff);

/* increment logical record pointer */
        hrdx += LOGRECS;
      }
    }
  }
}

/* getshort -- get a short from the buffer */
short getshort(buff,position)
short *buff;
int position;
{
  short w;
  char *cp,*wp;

/* byte swapping if VAX-type ordering of bytes in shorts -- tested on
   a decstation 5000 */
#if VAX
  cp = (char*)(buff + position);
  wp = (char*)(&w);
  *wp = *(cp + 1);
  *(wp + 1) = *cp;
  return(w);

/* otherwise, no byte swapping */
#else
  return(buff[position]);
#endif
}
