/* dispimage.c -- X11 version. */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "mcidas.h"

#define BORDER 5
#define DEF_FONT "9x15"
#define NCOLORS 21

static unsigned char *imagelist[MAXIMAGES]; /* list of images */
static Pixmap pixlist[MAXIMAGES];  /* list of pixmaps */
static Display *mydisplay;         /* X11 stuff */
static Window mywindow,rootwin;
static XImage *myimage;
static GC mygc;
static XEvent myevent;
static XSizeHints myhint;
static XWMHints xwmh = {(InputHint|StateHint),False,NormalState,0,0,0,0,0,0,};
static int myscreen;
static unsigned long mybackground,myforeground;
static int xret,yret,xsize,ysize,border,depth;
static Colormap cmap;
static Visual *vis;
static XColor exact,xcvals[NCOLORS];
static XFontStruct *fontstruct;     /* Font descriptor */
static XGCValues gcv;               /* structure for making graphics context */
static int color;
static char cnames[NCOLORS][10] = {"grey0","grey5","grey10","grey15",
     "grey20","grey25","grey30","grey35","grey40","grey45","grey50",
     "grey55","grey60","grey65","grey70","grey75","grey80","grey85",
     "grey90","grey95","grey100"};

extern int gargc;
extern char **gargv;

/* setupdisp -- this cranks up the X11 display and returns
 * the pixel depth of the display.
 */
setupdisp(nx,ny)
int nx,ny;                         /* size of image */
{
  int loop;

/* initialize storage pointers */
  for (loop = 0; loop < MAXIMAGES; loop++) imagelist[loop] = 0;

/* set up the X11 connection -- take default server and screen */
  if ((mydisplay = XOpenDisplay("")) == NULL) {
    fprintf(stderr,"setupdisp: can't open default X11 display\n");
    exit(1);
  }
  myscreen = DefaultScreen(mydisplay);

/* find out things about the workstation display */
  depth = DefaultDepth(mydisplay,myscreen);
  color = depth > 1;
  vis = DefaultVisual(mydisplay,myscreen);

/* set up hints for the window manager */
  myhint.x = 1;
  myhint.y = 1;
  myhint.width = nx;
  myhint.height = ny;
  myhint.flags = PPosition | PSize;

/* make background black, foreground white */
  mybackground = BlackPixel(mydisplay,myscreen);
  myforeground = WhitePixel(mydisplay,myscreen);

/* create a window */
  mywindow = XCreateSimpleWindow(mydisplay,DefaultRootWindow(mydisplay),
       myhint.x,myhint.y,myhint.width,myhint.height,BORDER,
       myforeground,mybackground);

/* set standard properties */
  XSetStandardProperties(mydisplay,mywindow,"mcdisp","mcdisp",None,
       gargv,gargc,&myhint);

/* get font */
  if ((fontstruct = XLoadQueryFont(mydisplay,DEF_FONT)) == NULL) {
    fprintf(stderr,"dispimage: can't find font %s\n",DEF_FONT);
    exit(1);
  }

/* set up GCs */
  gcv.foreground = myforeground;
  gcv.background = mybackground;
  gcv.font = fontstruct->fid;
  mygc = XCreateGC(mydisplay,mywindow,(GCForeground|GCBackground|GCFont),&gcv);

/* set events of interest */
  XSelectInput(mydisplay,mywindow,ExposureMask);

/* map the window */
  XMapWindow(mydisplay,mywindow);

/* wait for an exposure event */
  while (1) {
    XNextEvent(mydisplay,&myevent);
    if ((myevent.type == Expose) && (myevent.xexpose.count == 0)) break;
  }

/* for color screen, get default color map and colors */
  if (color) {
    cmap = XDefaultColormap(mydisplay,myscreen);
    for (loop = 0; loop < NCOLORS; loop++) {
      if (XAllocNamedColor(mydisplay,cmap,cnames[loop],&exact,xcvals + loop)
           == 0) colorfail(cnames[loop]);
    }
  }

/* return pixel depth */
  return(depth);
}

/* colorfail -- fatal error */
colorfail(color)
char *color;
{
  fprintf(stderr,"mcdisp: can't find color %s in default table\n",color);
  exit(1);
}

/* getcolors -- this returns a list of pixel values for the color list */
getcolors(cvals)
int cvals[];
{
  int i;
  for (i = 0; i < NCOLORS; i++) {
    cvals[i] = xcvals[i].pixel;
  }
  return(NCOLORS);
}

/* sendimage -- in the X11 version this creates initial image on
 * a pixmap.
 */
sendimage(index,nx,ny,image_depth,image)
int index;                         /* instance of image -- position in list */
int nx,ny;                         /* dimensions of image */
int image_depth;                   /* depth of image */
unsigned char *image;              /* pointer to the image */
{
  int loop;
  Window troot;
  int tx,ty;
  unsigned int tw,th,tbw,td;

/* set up pixmap */
  pixlist[index] = XCreatePixmap(mydisplay,mywindow,nx,ny,image_depth);

/* test that creation succeeded */
  tw = th = 0;
  XGetGeometry(mydisplay,pixlist[index],&troot,&tx,&ty,&tw,&th,&tbw,&td);
  if (tw != nx || th != ny) {
    fprintf(stderr,"sendimage: can't allocate pixmap %d\n",index);
    exit(1);
  }

/* clean it */
  XSetFunction(mydisplay,mygc,GXclear);
  XFillRectangle(mydisplay,pixlist[index],mygc,0,0,nx,ny);

/* only monochrome or 8 bit color works */
  if (image_depth != 1 && image_depth != 8) {
    fprintf(stderr,"dispimage: %d bit color not supported\n",image_depth);
    exit(1);
  }

/* create the image structure */
  if ((myimage = XCreateImage(mydisplay,vis,image_depth,ZPixmap,0,
       image,nx,ny,32,0)) == NULL) {
    fprintf(stderr,"sendimage: can't allocate ximage structure\n");
    exit(1);
  }

/* blap the image onto the pixmap */
  XSetFunction(mydisplay,mygc,GXcopy);
  XPutImage(mydisplay,pixlist[index],mygc,myimage,0,0,0,0,nx,ny);

/* free the image structure */
  XFree(myimage);
}

/* dispimage -- display a particular image */
dispimage(index,infoptr)
int index;                       /* instance of image to be displayed */
struct goes *infoptr;            /* information on this image */
{

/* first, check for an exposure event, and process if it is there */
  while (XCheckWindowEvent(mydisplay,mywindow,ExposureMask,&myevent) != 0);

/* display image */
  XSetFunction(mydisplay,mygc,GXcopy);
  XCopyArea(mydisplay,pixlist[index],mywindow,mygc,0,0,infoptr->xsize,
       infoptr->ysize,0,0);

/* synchronize */
  XSync(mydisplay,0);
}

/* imageclear -- Clear the image from the image
 * window.  (This is a no-op for X11 version.)
 */
imageclear()
{
}

/* finishimage -- No-op for the X11 version */
finishimage(index,nx,ny)
int index;
int nx,ny;
{
}

/* cleanup -- This routine clears screen
 * and then exits.
 */
cleanup(nimages)
int nimages;
{
  exit(0);
}

/* drawline -- Draw a line on image in inverting mode.
 * Drawing space goes from 0 to 639 in x and 0 to 349 in y.
 * To the right is more positive in x, and down is more positive in y.
 */
drawline(index,xstart,ystart,xend,yend)
int index;
int xstart,ystart,xend,yend;
{
  XSetFunction(mydisplay,mygc,GXequiv);
  XDrawLine(mydisplay,pixlist[index],mygc,xstart,ystart,xend,yend);
}

/* drawtext -- Draw text on image in inverting mode.
 */
drawtext(index,x,y,text)
int index;
int x,y;
char *text;
{
  XSetFunction(mydisplay,mygc,GXcopy);
  XDrawImageString(mydisplay,pixlist[index],mygc,x,y,text,strlen(text));
}
