/* dumbnav -- This routine assumes that the GOES satellite is in a nearly
 * circular geostationary orbit nearly
 * over the equator and that the center of the bitmap points nearly
 * toward the center of the earth.  It also assumes that
 * satellite scan lines are nearly east-west.  The navheader argument
 * contains information about the satellite position and orientation.
 * Lon and lat are the desired longitude and latitude.  X and y are
 * pointers to integers ranging from 0 to 639 in x and 0 to 349 in y,
 * which give the x-y location on the bitmap.  If the desired point is
 * invisible, dumbnav returns -2.  If it is outside the above range,
 * dumbnav returns -1.  Otherwise it returns 0.
 * We assume for the GVAR satellite that the earth is a sphere.
 */

#include <stdio.h>
#include <math.h>
#include "mcidas.h"
#define EARTH_RADIUS 6370.
#define DEG_PER_RAD 57.2958
#define SATRAD 42164.365

double gst();

static struct vec {             /* vector structure */
  float x,y,z;
};
static struct mat {             /* matrix structure */
  float xx,xy,xz,yx,yy,yz,zx,zy,zz;
};
static float satlat;            /* satellite latitude */
static float satlon;            /* satellite longitude */
static struct vec s;            /* satellite to earth center */
static float slen;              /* length of s vector */
static struct vec l;            /* earth center to desired point */
static struct vec v;            /* satellite to desired point */
static float vlen;              /* length of v vector */
static struct vec u;            /* up in image plane */
static struct vec r;            /* right in image plane */
static float axisth,axisaz;     /* angles giving orientation of satellite */
static float rlen;              /* length of r vector */
static float rad_sat;           /* distance of satellite from earth center */
static float line_angle,elem_angle;   /* angles of desired point */
static float angle_from_ep;     /* angle between position and epoch */
static float angle_from_peri;   /* angle between position and perigee */
static float angle_from_asn;    /* angle between position and asnode */
static int line,element;        /* GOES image position */
static int epdate,epmonth,epyear,epday,ephour,epmin,epsec;  /* epoch info */
static double gmtangle;         /* angle of gmt to V. E. at epoch time */
static float pitch,yaw,roll;    /* camera misalignment */
static struct vec mv_mul();     /* matrix-vector multiplier */
static struct mat mm_mul();     /* matrix-matrix multiplier */
static struct mat trans();      /* transpose taker */
static struct mat rot;          /* rotate in horiz. plane from x to sat */
static struct mat fix;          /* fix alignment problems */

/* variables for GVAR */
static float phi,psi,sinphi,sinpsi,oisq,sinoi,cosoi,uu,sinuu,cosuu,
  asc,sinasc,cosasc,sublat,sublon,radius,sinlat,coslat,sinlon,coslon,
  scan,elev,rma,pma,sinelev,coselev,sinscan,cosscan;
static struct mat b,rpy,bt;
static struct vec xs,ug,f,ft;

int dumbnav(navheader,lon,lat,x,y,fast)
struct goes *navheader;         /* contains nav information */
float lon,lat;                  /* inputs desired latitude and longitude */
int *x,*y;                      /* returned image coordinates */
int *fast;                      /* flag requesting fast computation */
{

  /* GOES or GVAR? */
  if (strcmp(navheader->navtype,"GOES") == 0) {
    
    /* get satellite latitude and longitude */
    /* first, get epoch date and time */
    if (!(*fast)) {
      epdate = navheader->epochd;
      epmonth = epdate/100;
      epdate -= 100*epmonth;
      epyear = epmonth/100;
      epmonth -= epyear*100;
      epyear += 1900;
      epday = julian_date(epyear,epmonth,epdate) - julian_date(epyear,1,0);
      epsec = navheader->epocht;
      epmin = epsec/100;
      epsec -= epmin*100;
      ephour = epmin/100;
      epmin -= ephour*100;
      
      /* next, get zeroth order angle from epoch time */
      angle_from_ep = 361.*(navheader->day - epday +
			    (navheader->hour - ephour)/24. +
			    (navheader->minute - epmin)/1440. +
			    (navheader->second - epsec)/86400.);
      
      /* get gst for epoch time and date */
      gmtangle = gst(epyear,epday,ephour,epmin,epsec);
      
      /* get angle of satellite relative to V. E. at epoch time
	 assuming circ orb */
      satlon = navheader->asnode + navheader->perigee + navheader->meana;
      
      /* get angle of satellite from perigee at picture time */
      angle_from_peri = angle_from_ep + navheader->meana;
      
      /* get angle of satellite from ascending node at picture time */
      angle_from_asn = angle_from_ep + navheader->meana + navheader->perigee;
      
      /* compute satellite latitude at picture time */
      satlat = navheader->orbinc*sin(angle_from_asn/DEG_PER_RAD);
      
      /* get angle of satellite relative to gmt meridian at epoch time */
      satlon -= gmtangle;
      
      /* correct for non-circular orbit -- thus getting sat lon at
	 picture time */
      satlon += 2.*DEG_PER_RAD*navheader->eccen*
	sin(angle_from_peri/DEG_PER_RAD);
      
      /* confine satlon to -180 -- +180 */
      while (satlon > 180.) satlon -= 360.;
      while (satlon < -180.) satlon += 360.;
      
      /* find vector from satellite to center of earth */
      rad_sat = navheader->semima*(1. - navheader->eccen*
				   cos(angle_from_peri/DEG_PER_RAD));
      s.x = - rad_sat*cos(satlon/DEG_PER_RAD);
      s.y = - rad_sat*sin(satlon/DEG_PER_RAD);
      s.z = - rad_sat*satlat/DEG_PER_RAD;
      slen = sqrt(s.x*s.x + s.y*s.y + s.z*s.z);
      
      /* find up and right vectors in image plane */
      axisaz = (navheader->rascen - gmtangle - angle_from_ep)/DEG_PER_RAD;
      axisth = (90. - navheader->declin)/DEG_PER_RAD;
      u.x = sin(axisth)*cos(axisaz);
      u.y = sin(axisth)*sin(axisaz);
      u.z = cos(axisth);
      r.x = s.y*u.z - s.z*u.y;
      r.y = s.z*u.x - s.x*u.z;
      r.z = s.x*u.y - s.y*u.x;
      rlen = sqrt(r.x*r.x + r.y*r.y + r.z*r.z);
      r.x /= rlen;
      r.y /= rlen;
      r.z /= rlen;
      
      /* small angle corrections for pitch, yaw, and roll */
      pitch = navheader->pitch/DEG_PER_RAD;
      yaw = navheader->yaw/DEG_PER_RAD;
      roll = navheader->roll/DEG_PER_RAD;
      
      /* find rotator from x-axis to vert plane containing center of earth */
      rot.xx = s.x/slen;
      rot.yx = s.y/slen;
      rot.zx = 0.;
      rot.xy = -s.y/slen;
      rot.yy = s.x/slen;
      rot.zy = 0.;
      rot.xz = 0.;
      rot.yz = 0.;
      rot.zz = 1.;
      
      /* construct approximate pitch-roll-yaw corrector */
      fix.xx = cos(roll)*cos(pitch);
      fix.xy = -sin(roll);
      fix.xz = sin(pitch);
      fix.yx = sin(roll);
      fix.yy = cos(roll)*cos(yaw);
      fix.yz = sin(yaw);
      fix.zx = -sin(pitch);
      fix.zy = -sin(yaw);
      fix.zz = cos(yaw)*cos(pitch);
      
      /* rotate to actual coordinate system */
      fix = mm_mul(rot,mm_mul(fix,trans(rot)));
      
      /* fix up and right vectors */
      u = mv_mul(fix,u);
      r = mv_mul(fix,r);
      
      /* end of stuff we don't have to repeat */
    }
    *fast = 1;
    
    /* find vector from center of earth to desired point (spherical earth) */
    l.x = EARTH_RADIUS*cos(lon/DEG_PER_RAD)*cos(lat/DEG_PER_RAD);
    l.y = EARTH_RADIUS*sin(lon/DEG_PER_RAD)*cos(lat/DEG_PER_RAD);
    l.z = EARTH_RADIUS*sin(lat/DEG_PER_RAD);
    
    /* find vector from satellite to desired point */
    v.x = s.x + l.x;
    v.y = s.y + l.y;
    v.z = s.z + l.z;
    
    /* if v and l have positive dot product, the point is invisible */
    if (v.x*l.x + v.y*l.y + v.z*l.z > 0.) {
      *x = *y = 1000000;
      return(-2);
    }
    vlen = sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
    
    /* get the line angle */
    line_angle = DEG_PER_RAD*asin((v.x*u.x + v.y*u.y + v.z*u.z)/vlen);
    
    /* get the element angle */
    elem_angle = DEG_PER_RAD*asin((v.x*r.x + v.y*r.y + v.z*r.z)/vlen);
    
    /* get the line */
    line = navheader->lintot*(.5*navheader->deglin - line_angle)/
      navheader->deglin + 1.5;
    
    /* get the element */
    element = navheader->eletot*(elem_angle + .5*navheader->degele)/
      navheader->degele + 1.5;
  }

  /* check for GVAR */
  else if (strcmp(navheader->navtype,"GVAR") == 0) {

    /* these calculations are derived from CSU code of Smith and Phillips
       modified by D. Farmer at Unidata and further modified by
       Dan Vietor at Purdue -- they are simplified by assuming a
       spherical earth */
    if (!(*fast)) {

      /* compute earth-centric coordinate axes at satellite */
      phi = navheader->phi;
      psi = navheader->psi;
      sinphi = sin(phi);
      sinpsi = sin(psi);
      oisq = sinphi*sinphi + sinpsi*sinpsi;
      sinoi = pow(oisq,0.5);
      cosoi = pow(1. - oisq,0.5);
      uu = (sinphi == 0. && sinpsi == 0.) ? 0. : atan2(sinphi,sinpsi);
      sinuu = sin(uu);
      cosuu = cos(uu);
      asc = navheader->lam - uu;
      sinasc = sin(asc);
      cosasc = cos(asc);
      sublat = phi;
      sublon = asc + atan2(cosoi*sinuu,cosuu);
      b.xx = -cosasc*sinuu - sinasc*cosuu*cosoi;
      b.yx = -sinasc*sinuu + cosasc*cosuu*cosoi;
      b.zx = cosuu*sinoi;
      b.xy = -sinasc*sinoi;
      b.yy = cosasc*sinoi;
      b.zy = -cosoi;
      b.xz = -cosasc*cosuu + sinasc*sinuu*cosoi;
      b.yz = -sinasc*cosuu - cosasc*sinuu*cosoi;
      b.zz = -sinphi;

      /* compute the vector which points to satellite from earth center */
      radius = (SATRAD + navheader->dr)/EARTH_RADIUS;
      xs.x = -b.xz*radius;
      xs.y = -b.yz*radius;
      xs.z = -b.zz*radius;

      /* compute the rotation matrix to fix satellite misalignment */
      roll = navheader->roll;
      yaw = navheader->yaw;
      pitch = navheader->pitch;
      rpy.xx = 1. - .5*(pitch*pitch + yaw*yaw);
      rpy.xy = -yaw;
      rpy.xz = pitch;
      rpy.yx = yaw + pitch*roll;
      rpy.yy = 1. - .5*(yaw*yaw + roll*roll);
      rpy.yz = -roll;
      rpy.zx = -pitch + roll*yaw;
      rpy.zy = roll + pitch*yaw;
      rpy.zz = 1. - .5*(pitch*pitch + roll*roll);
      /* this is wrong!
      bt = mm_mul(b,rpy);
      */

      /* this is probably right */
      bt = mm_mul(rpy,mm_mul(b,trans(rpy)));
    }
    *fast = 1;

    /* compute the vector which points from satellite to earth point */
    sinlat = sin(lat/DEG_PER_RAD);
    coslat = cos(lat/DEG_PER_RAD);
    sinlon = sin(lon/DEG_PER_RAD);
    coslon = cos(lon/DEG_PER_RAD);
    ug.x = coslat*coslon;
    ug.y = coslat*sinlon;
    ug.z = sinlat;
    f.x = ug.x - xs.x;
    f.y = ug.y - xs.y;
    f.z = ug.z - xs.z;
    if (ug.x*f.x + ug.y*f.y + ug.z*f.z > 0.) {
      *x = *y = 1000000;
      return(-2);
    }

    /* compute the angles of the earth point rel to satellite axis */
    ft = mv_mul(trans(bt),f);
    scan = atan(ft.x/pow(ft.y*ft.y + ft.z*ft.z,.5));
    elev = -atan(ft.y/ft.z);

    /* do some corrections */
    rma = navheader->rma;
    pma = navheader->pma;
    sinelev = sin(elev);
    coselev = cos(elev);
    sinscan = sin(scan);
    cosscan = cos(scan);
    elev = elev + rma*(1. - coselev/cosscan) +
      pma*sinelev*((1. + sinscan)/cosscan);
    scan = scan - rma*sinelev;

    /* compute scan line and element */
    element = (navheader->scanbnds + scan)/navheader->scanpix +
      navheader->scanoffset;
    line = (navheader->elevbnds - elev)/navheader->elevdln +
      navheader->lineoffset;
  }

  /* oops -- unknown type! */
    else {
    fprintf(stderr,"dumbnav: navtype %s unknown\n",navheader->navtype);
    exit(1);
  }
  
  /* scale to the window */
  *x = (element + navheader->xres/2 - navheader->xcoor)/navheader->xres;
  *y = (line + navheader->yres/2 - navheader->ycoor)/navheader->yres;

  /* check for being in window */
  if (*x < 0 || *x >= navheader->xsize ||
      *y < 0 || *y >= navheader->ysize) return(-1);
  else return(0);
}

/* gst -- Computes Greenwich Sidereal Time in degrees.
 * Data obtained from "Practical Astronomy with your Calculator"
 * by Peter Duffett-Smith, Cambridge University Press.
 */
double gst(year,day,hour,minute,second)
int year,day,hour,minute,second;
{
  static double bt[] = {17.397221, 17.413525, 17.363611, 17.379644,
    17.395558, 17.411473, 17.361678, 17.377592, 17.393506, 17.409421,
    17.359625, 17.375540, 17.391454, 17.407368, 17.357573, 17.373487,
    17.389402, 17.405316, 17.355521, 17.371435, 17.387349, 17.403264,
    17.353468, 17.369383, 17.385297, 17.401211};
  double a = 0.065709, b, c = 1.002743, d = 0.997257;
  double gmt,result;

  if (year < 1975 || year > 2000) {
    fprintf(stderr,"navpac: year %d out of range for gst calc\n",year);
    exit(1);
  }
  else b = bt[year - 1975];
  result = day*a;
  result -= b;
  gmt = (double)hour + (double)minute/60. + (double)second/3600.;
  gmt *= c;
  result += gmt;
  if (result > 24.) result -= 24.;
  if (result < 0.) result += 24.;
  return(360.*result/24.);
}

/* matrix-vector multiplier */
static struct vec mv_mul(m,v)
struct mat m;
struct vec v;
{
  struct vec r;
  r.x = m.xx*v.x + m.xy*v.y + m.xz*v.z;
  r.y = m.yx*v.x + m.yy*v.y + m.yz*v.z;
  r.z = m.zx*v.x + m.zy*v.y + m.zz*v.z;
  return(r);
}

/* matrix-matrix multiplier */
static struct mat mm_mul(m,n)
struct mat m;
struct mat n;
{
  struct mat r;
  r.xx = m.xx*n.xx + m.xy*n.yx + m.xz*n.zx;
  r.yx = m.yx*n.xx + m.yy*n.yx + m.yz*n.zx;
  r.zx = m.zx*n.xx + m.zy*n.yx + m.zz*n.zx;
  r.xy = m.xx*n.xy + m.xy*n.yy + m.xz*n.zy;
  r.yy = m.yx*n.xy + m.yy*n.yy + m.yz*n.zy;
  r.zy = m.zx*n.xy + m.zy*n.yy + m.zz*n.zy;
  r.xz = m.xx*n.xz + m.xy*n.yz + m.xz*n.zz;
  r.yz = m.yx*n.xz + m.yy*n.yz + m.yz*n.zz;
  r.zz = m.zx*n.xz + m.zy*n.yz + m.zz*n.zz;
  return(r);
}

/* transpose taker */
static struct mat trans(m)
struct mat m;
{
  struct mat r;
  r.xx = m.xx;
  r.xy = m.yx;
  r.xz = m.zx;
  r.yx = m.xy;
  r.yy = m.yy;
  r.yz = m.zy;
  r.zx = m.xz;
  r.zy = m.yz;
  r.zz = m.zz;
  return(r);
}
