/* getlwtoa3.c -- This subroutine extracts the "LWTOA3" product from the 
   pc-mcidas data stream using calls to the getpacket routine.  The string
   argument is the data from the header packet.  This product contains
   satellite images. */

#include <stdio.h>
#include <math.h>
#include "mcidas.h"

static struct goes header;	/* this is where we store header info */
static unsigned char *imagebuffer = NULL;	/* this is where we store
						   image */
static unsigned char *linebuffer = NULL;	/* temp storage for packed 

						   line */
static int trycount;		/* number of failed sync tries */
static int imagesize = 0, linesize = 0;		/* current buffer sizes */
static int maximagesize = 0, maxlinesize = 0;	/* remember max sizes */
int getsync(), getskip(), getstream();
int getline(), getnib();
long getword();
short getshort();
float hhmmss();
float dddmmss();

getlwtoa3(textline)
unsigned char *textline;
{
    long temp, temp2;
    int actline, line, loop;
    long idntfr, imcact, etime1, etime2, imgday, imgtm, imgsnd, iofnc,
     iofec, iofni, iofei;
    float reflon, refdis, reflat, refyaw, ratrol, ratptc, ratyaw,
      edtime, imcrol, imcptc, imcyaw, ldr[13], rddr[11], dgl[9],
      doy[9], solinc, exptim, raawds[55], yaawds[55], spc1[12], paawds[55],
      spc2[18], rmawds[55], pmawds[55];
    long year,day,hour,min;
    float sec;

/* announce to the world */
    fprintf(stderr, "getlwtoa3: found a satellite image\n");

/* initialize stream reader */
    resetstream();

/* check for a proper sync */
    if (trycount = getsync()) {
	fprintf(stderr, "getlwtoa3: sync error %d at beginning of area dir\n",
		trycount);
	if (trycount == TRYMAX) {
	    fprintf(stderr, "getlwtoa3: sync error count too high\n");
	    return;
	}
	if (trycount == -1) {
	    fprintf(stderr, "getlwtoa3: unexpected eof at area beginning\n");
	}
    }
/* throw away two words */
    getskip(8);

/* get satellite id */
    header.satid = getword();

/* get year and day of year */
    temp = getword();
    temp2 = temp / 1000;
    temp -= 1000 * temp2;
    if (temp2 < 57)
	temp2 += 2000;
    else
	temp2 += 1900;
    header.year = temp2;
    header.day = temp;

/* get hour minute and second */
    temp = getword();
    temp2 = temp / 10000;
    header.hour = temp2;
    temp -= 10000 * temp2;
    temp2 = temp / 100;
    header.minute = temp2;
    temp -= 100 * temp2;
    header.second = temp;

/* get coordinates, sizes, resolutions, and bandmap */
    header.ycoor = getword();
    header.xcoor = getword();
    header.zcoor = getword();
    header.ysize = getword();
    header.xsize = getword();
    header.nbyte = getword();
    header.yres = getword();
    header.xres = getword();
    header.zres = getword();
    header.dprelen = getword();
    getskip(12);
    header.bandmap = getword();

    /* get byte offsets and validity code */
    getskip(56);
    header.dataoset = getword();
    header.navoset = getword();
    header.validity = getword();

/* get source and calibration types */
    getskip(60);
    getstream(header.source, 4);
    header.source[4] = '\0';
    getstream(header.calib, 4);
    header.calib[4] = '\0';

/* finish off area directory */
    getskip(44);

/* print out area directory information */
#if VERBOSE
    printf("satid = %d; time = %d:%d:%d:%d:%d\n", header.satid, header.year,
	   header.day, header.hour, header.minute, header.second);
    printf("prefix len = %d; data, nav offsets = %d %d; val flag = %d;\n",
       header.dprelen, header.dataoset, header.navoset, header.validity);
    printf("xcoor = %d; ycoor = %d; zcoor = %d; xsize = %d; ysize = %d;\n",
    header.xcoor, header.ycoor, header.zcoor, header.xsize, header.ysize);
    printf("nbyte = %d; xres = %d; yres = %d; zres = %d;\n", header.nbyte,
	   header.xres, header.yres, header.zres);
    printf("bandmap = %8x\n", header.bandmap);
    printf("source type = %s; calib type = %s\n", header.source, header.calib);
#endif

/* should find a sync */
    if (trycount = getsync()) {
	fprintf(stderr, "getlwtoa3: sync error %d at head of nav codicil\n",
		trycount);
	if (trycount == TRYMAX) {
	    fprintf(stderr, "getlwtoa3: sync error count too high\n");
	    return;
	}
    }
/* get navigation type */
    getstream(header.navtype, 4);
    header.navtype[4] = '\0';

/* print out navigation type */
#if VERBOSE
    printf("navigation type = %s\n", header.navtype);
#endif

/* select nav type -- for now, only GOES or GVAR */

/* GOES navigation type */
    if (strcmp(header.navtype, "GOES") == 0) {

/* get orbit parameters */
	getskip(12);
	header.epochd = getword();
	header.epocht = getword();
	header.semima = (float) (getword()) / 100.;
	header.eccen = (float) (getword()) / 1.e6;
	header.orbinc = (float) (getword()) / 1000.;
	header.meana = (float) (getword()) / 1000.;
	header.perigee = (float) (getword()) / 1000.;
	header.asnode = (float) (getword()) / 1000.;

/* attitude parameters */
	header.declin = dddmmss(getword());
	header.rascen = dddmmss(getword());
	header.piclin = getword();

/* spin */
	header.spinp = (float) getword();

/* frame geometry */
	header.deglin = dddmmss(getword());
	temp = getword();
	temp2 = temp / 100000;
	temp -= 100000 * temp2;
	temp *= temp2;
	header.lintot = temp;
	header.scanners = temp2;
	header.degele = dddmmss(getword());
	header.eletot = getword();

/* camera geometry */
	header.pitch = dddmmss(getword());
	header.yaw = dddmmss(getword());
	header.roll = dddmmss(getword());

/* other parameters */
	getskip(4);
	header.iajust = getword();
	header.iajtim = hhmmss(getword());
	header.iajtim += (double) (getword()) / 100.;
	getskip(4);
	header.iseang = dddmmss(getword());
	getskip(8);

/* betas for this image */
	header.scanline1 = getword();
	header.timesl1 = hhmmss(getword());
	header.timesl1 += (double) (getword()) / 100.;
	header.betacount1 = getword();
	header.scanline2 = getword();
	header.timesl2 = hhmmss(getword());
	header.timesl2 += (double) (getword()) / 100.;
	header.betacount2 = getword();

/* gammas for this image */
	header.gamma = (float) (getword()) / 100.;
	header.gammadot = (float) (getword()) / 100.;

/* print parameters */
#if VERBOSE
	printf("epochd = %d; epocht = %d\n", header.epochd, header.epocht);
	printf("semima = %g; eccen = %g; orbinc = %g;\n", header.semima,
	       header.eccen, header.orbinc);
	printf("meana = %g; perigee = %g; asnode = %g;\n", header.meana,
	       header.perigee, header.asnode);
	printf("declin = %g; rascen = %g; piclen = %d\n", header.declin,
	       header.rascen, header.piclin);
	printf("spinp = %g\n", header.spinp);
	printf("deglin = %g; lintot = %d; degele = %g; eletot = %d\n",
	     header.deglin, header.lintot, header.degele, header.eletot);
	printf("pitch = %g; yaw = %g; roll = %g\n", header.pitch, header.yaw,
	       header.roll);
	printf("iajust = %d; iajtim = %g; iseang = %g\n", header.iajust,
	       header.iajtim, header.iseang);
	printf("scanline1 = %d; timesl1 = %g; betacount1 = %d\n",
	       header.scanline1, header.timesl1, header.betacount1);
	printf("scanline2 = %d; timesl2 = %g; betacount2 = %d\n",
	       header.scanline2, header.timesl2, header.betacount2);
	printf("gamma = %g; gammadot = %g\n", header.gamma, header.gammadot);
#endif

/* skip rest of GOES codicil for now */
	getskip(348);
    }
    /* GVAR navigation type */
    if (strcmp(header.navtype, "GVAR") == 0) {
	idntfr = getword();
	imcact = getword();
	getskip(8);
	reflon = getword() / 1.e7;
	refdis = getword() / 1.e7;
	reflat = getword() / 1.e7;
	refyaw = getword() / 1.e7;
	ratrol = getword() / 1.e7;
	ratptc = getword() / 1.e7;
	ratyaw = getword() / 1.e7;
	etime1 = getword();
	etime2 = getword();
	edtime = getword() / 100.;
	imcrol = getword() / 1.e7;
	imcptc = getword() / 1.e7;
	imcyaw = getword() / 1.e7;
	for (loop = 0; loop < 13; loop++)
	    ldr[loop] = getword() / 1.e7;
	for (loop = 0; loop < 11; loop++)
	    rddr[loop] = getword() / 1.e7;
	for (loop = 0; loop < 9; loop++)
	    dgl[loop] = getword() / 1.e7;
	for (loop = 0; loop < 9; loop++)
	    doy[loop] = getword() / 1.e7;
	solinc = getword() / 1.e7;
	exptim = getword() / 100.;
	for (loop = 0; loop < 55; loop++)
	    raawds[loop] = getword();
	for (loop = 0; loop < 12; loop++)
	    spc1[loop] = getword() / 1.e7;
	for (loop = 0; loop < 55; loop++)
	    paawds[loop] = getword();
	for (loop = 0; loop < 55; loop++)
	    yaawds[loop] = getword();
	for (loop = 0; loop < 18; loop++)
	    spc2[loop] = getword() / 1.e7;
	for (loop = 0; loop < 55; loop++)
	    rmawds[loop] = getword();
	for (loop = 0; loop < 55; loop++)
	    pmawds[loop] = getword();
	imgday = getword();
	imgtm = getword();
	imgsnd = getword();
	getskip(36);
	iofnc = getword();
	iofec = getword();
	iofni = getword();
	iofei = getword();
	getskip(1028);

	/* unlike the GOES case, we do preliminary calculations here */
	/* do restricted case for now */
	if (imgsnd != 1) {
	  fprintf(stderr,"getlwtoa3: imgsnd %d != 1 (no soundings)\n",imgsnd);
	  exit(1);
	}
	if ((imcact & 128) != 128) {
	  fprintf(stderr,"getlwtoa3: imc off not yet implemented\n");
	  exit(1);
	}
	else {
	  
	  /* assign reference navigation values to header */
	  header.lam = reflon;
	  header.dr = refdis;
	  header.phi = reflat;
	  header.psi = refyaw;
	  header.roll = ratrol;
	  header.pitch = ratptc;
	  header.yaw = ratyaw;
	  header.rma = 0.;
	  header.pma = 0.;
	}

	/* constants for converting from angles to line/element */
	header.elevbnds = 0.220896;
	header.elevdln = 28.e-6;
	header.scanbnds = 0.24544;
	header.scanpix = 16.e-6;
	header.lineoffset = 4.5;
	header.scanoffset = 1.;

	/* print out if desired */
#if VERBOSE
	printf("lam,dr,phi,psi: %g %g %g %g\n",header.lam,header.dr,
	       header.phi,header.psi);
	printf("roll,pitch,yaw: %g %g %g\n",header.roll,header.pitch,
	       header.yaw);
	printf("rma,pma: %g %g\n",header.rma,header.pma);
#endif
    } 
    else {

/* unimplemented nav type -- quit */
	fprintf(stderr, "getlwtoa3: unimplemented nav type %s\n",
		header.navtype);
	return;
    }

/* set up array for image -- deal only with zres = nbyte = 1 for now */
    if (header.zres != 1 || header.nbyte != 1) {
	fprintf(stderr, "getlwtoa3: zres (%d) != 1 or nbyte (%d) != 1\n",
		header.zres, header.nbyte);
	return;
    }
/* allocate space for image if new image is bigger */
    imagesize = header.xsize * header.ysize;
    if (imagesize > maximagesize) {
	maximagesize = imagesize;
	if ((imagebuffer = (unsigned char *) malloc(sizeof(char) * imagesize))
	    == NULL) {
	    fprintf(stderr, "getlwtoa3: can't get enough memory\n");
	    exit(1);
	}
    }
/* allocate space for line buffer if new line is bigger */
    linesize = header.xsize;
    if (linesize > maxlinesize) {
	maxlinesize = linesize;
	if ((linebuffer = (unsigned char *) malloc(3 * sizeof(char) * linesize))
	    == NULL) {
	    fprintf(stderr, "getlwtoa3: can't get enough memory\n");
	    exit(1);
	}
    }
/* loop on lines */
    for (line = 0; line < header.ysize; line++) {
	if ((actline = getline(header.xsize, header.ysize, line)) < 0) {
	    fprintf(stderr,
		    "getlwtoa3: serious error on line %d decode -- truncate\n", line);
	    break;
	}
/* update actual line to expected line */
	line = actline;
    }

/* got the image -- save it somewhere */
    putimage(&header, imagebuffer);
}

/* getline -- This routine gets a line from the input stream and puts * it 

   in the image buffer.  The image buffer is arranged to contain * a
   sequence  of lines, so that it looks like a 2-D C array [ysize][xsize]. 
 */

int packptr, nibflag;		/* stuff for unpacking nibbles */

getline(nx, ny, exline)
long nx, ny;			/* dimensions of image */
int exline;			/* expected line number */
{
    long mode, line, bytes;
    int elem;
    unsigned char hold;

/* get sync */
    if (trycount = getsync()) {
	fprintf(stderr, "getline: sync error %d at head of expected line %d\n",
		trycount, exline);
	if (trycount == -1) {
	    fprintf(stderr, "getline: unexpected eof\n");
	    return (-1);
	}
	if (trycount == TRYMAX || trycount == -1) {
	    fprintf(stderr, "getline: sync error count %d too high\n", trycount);
	    return (-1);
	}
    }
/* get mode, line number, and number of bytes in packed line */
    mode = getshort();
    line = getshort();
    getskip(2);
    bytes = getshort();

/* 
   printf("expected line = %d; actual = %d; mode = %d; bytes = %d;\n",
   exline,line,mode,bytes); */

    if (line < 0 || line >= ny) {
	fprintf(stderr, "getline: indicated line number = %d out of range\n", line);
	return (-1);
    }
    if (line != exline) {
	fprintf(stderr, "getline: line mismatch, expected %d, got %d\n",
		exline, line);
    }
    if (bytes < 0 || bytes > 3 * nx) {
	fprintf(stderr, "getline: number of packed bytes = %d out of range\n",
		bytes);
	return (-1);
    }
/* get a packed line */
    if (getstream(linebuffer, bytes) == -1) {
	fprintf(stderr, "getline: unexpected eof\n");
	return (-1);
    }
/* unpack the line -- start with initializing pointers and flags */
    packptr = 0;
    nibflag = 0;
    hold = 0;

/* loop on expected elements */
    for (elem = 0; elem < nx; elem++) {
	imagebuffer[line * nx + elem] = 0;

/* grab a nibble and test it for various cases */
	switch (getnib()) {
	case 0x0:		/* escape case */
	    hold = 16 * getnib();
	    hold += getnib();
	    break;
	case 0x1:		/* a bunch of simple increments */
	    hold -= 6;
	    break;
	case 0x2:
	    hold += 6;
	    break;
	case 0x3:
	    hold -= 5;
	    break;
	case 0x4:
	    hold += 5;
	    break;
	case 0x5:
	    hold -= 4;
	    break;
	case 0x6:
	    hold += 4;
	    break;
	case 0x7:
	    hold -= 3;
	    break;
	case 0x8:
	    hold += 3;
	    break;
	case 0x9:
	    hold -= 2;
	    break;
	case 0xa:
	    hold += 2;
	    break;
	case 0xb:
	    hold -= 1;
	    break;
	case 0xc:
	    hold += 1;
	    break;
	case 0xd:
	    break;
	case 0xe:		/* this and next same as previous */
	    imagebuffer[line * nx + elem++] = hold;
	    if (elem >= nx) {
		fprintf(stderr, "getline: decoding error #1 on line %d\n", line);
		elem--;
	    }
	    break;
	case 0xf:		/* this and next two same as previous */
	    imagebuffer[line * nx + elem++] = hold;
	    if (elem >= nx) {
		fprintf(stderr, "getline: decoding error #2 on line %d\n", line);
		elem--;
	    }
	    imagebuffer[line * nx + elem++] = hold;
	    if (elem >= nx) {
		fprintf(stderr, "getline: decoding error #3 on line %d\n", line);
		elem--;
	    }
	    break;
	default:
	    fprintf(stderr, "getline: a nibble is not a nibble!\n");
	    return (-1);
	}

/* hold is the new pixel value */
	imagebuffer[line * nx + elem] = hold;
    }

/* check if decoding ended up in correct place */
    if (bytes != packptr + nibflag) {
	fprintf(stderr, "getline: phase error in decoding on line %d\n", line);
    }
    return (line);
}

/* get a nibble from the line buffer */

int getnib()
{
    int result;
    result = linebuffer[packptr];
    if (nibflag) {
	result = 0xf & result;
	packptr++;
    } else {
	result = 0xf & (result >> 4);
    }
    nibflag = !nibflag;
    return (result);
}

/* convert hhmmss format to float seconds */
float hhmmss(word)
long word;
{
    long hours, minutes;
    hours = word / 10000;
    word -= hours * 10000;
    minutes = word / 100;
    word -= minutes * 100;
    return ((float) (word + 60 * minutes + 3600 * hours));
}

/* convert dddmmss format to degrees */
float dddmmss(word)
long word;
{
    long degrees, minutes;
    degrees = word / 10000;
    word -= degrees * 10000;
    minutes = word / 100;
    word -= minutes * 100;
    return ((float) (degrees * 3600 + minutes * 60 + word) / 3600.);
}
