%{
/* mccmd -- command module for mcdisp */
#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "cdfhdr.h"
#include "mcidas.h"

/* number of usleep ticks per second */
#define MILLION 1000000

/* default sleep interval */
#define SLEEPINT 200000

/* min and max sleep intervals */
#define SLMIN 50000
#define SLMAX 2000000

/* default monochrome cut values */
#define TCUT 273.
#define WCUT 245.
#define BCUT .3
#define TDITHER 20.
#define WDITHER 10.
#define BDITHER .15

/* default color cut values */
#define CTCUT 263.
#define CWCUT 245.
#define CBCUT .3
#define CTDITHER 40.
#define CWDITHER 20.
#define CBDITHER .3

/* local defines */
#define SMAX 10
#define LSIZE 80

/* global definitions */
int visible,watervapor,infrared;
int sleepint;
float wcut,tcut,bcut;
float tdither,bdither,wdither;
float range1,range2;
int gargc;
char **gargv;
int option;
extern char *optarg;
extern int optind,opterr;

/* local definitions */
static float dithval;               /* global dither value for monochrome */
static float cutval;                /* global cut value for monochrome */
static float fsleepint;             /* sleep interval in seconds */
static int help = 0;                /* flag a help request */
static int cutflag = 0;             /* flag a change in cut value */
static int dith = 0;                /* flag a new dither value */
static int rangeflag = 0;           /* flag a new set of ranges */
static int sleepflag = 0;           /* flag a change in sleep interval */
static int errorflag = 0;           /* flag errors */
static int listflag = 0;            /* flag a list request */
static int goflag = 0;              /* flag ready to go */
static int reproc;                  /* reprocess flag */
static int nstack = 0;              /* the stack pointer */
static char stack[SMAX][LSIZE];     /* the stack */
static void push();                 /* push a token to command stack */
static char *pop();                 /* pop a token from stack */
static void docmd();                /* routine that does commands */
static void syntax();               /* report a syntax error */
static void doproc();               /* image processing routine */
int intrhand();                     /* interrupt handler */

/* more local stuff */
static char *image[MAXIMAGES];      /* pointers to images */
static int nimages;                 /* number of images read */
static int loop;                    /* looping variable */
int nx,ny;                          /* size of image */
int image_depth;                    /* depth display hardware likes */
static struct goes list[MAXIMAGES]; /* data on images */
char *getimage();                   /* routine that returns image */
int cleanup();                      /* routine that cleans up stuff */
%}
%%
			fprintf(stderr,"\nmcdisp> ");

"\n"			{docmd(); fprintf(stderr,"mcdisp> ");}
go			goflag = 1;
l			listflag = 1;
q			{cleanup(nimages); exit(0);}
"?"			help = 1;
c			cutflag = 1;
d			dith = 1;
r			rangeflag = 1;
si			sleepflag = 1;
[-0-9.+e]+		push(yytext);
" "|"\t"		;
.			errorflag = 1;
%%

/* push a string onto the stack */
static void push(txt)
char *txt;
{
  if (nstack >= SMAX) {
    fprintf(stderr,"mccmd -- too many numbers in a row\n");
    exit(1);
  }
  strcpy(stack[nstack++],txt);
}

/* pop a string from the stack */
static char *pop()
{
  if (nstack <= 0) {
    fprintf(stderr,"mccmd: popping empty stack\n");
    exit(1);
  }
  return(stack[--nstack]);
}

/* docmd -- do the commands */
static void docmd()
{

/* error */
  if (errorflag) {
    errorflag = 0;
    nstack = 0;
    fprintf(stderr,"unknown command\n");
    return;
  }

/* help routine */
  if (help) {
    help = 0;
    nstack = 0;
    fprintf(stderr,"cut_value c       (set center of grey scale)\n");
    fprintf(stderr,"dith_value d      (set half-width of grey scale)\n");
    fprintf(stderr,
         "low_val hi_val r  (set lower and upper limits of grey scale)\n");
    fprintf(stderr,"sleepint_value si (set looping interval)\n");
    fprintf(stderr,"go                (display the images)\n");
    fprintf(stderr,"?                 (help command)\n");
    fprintf(stderr,"q                 (exit the program)\n");
    return;
  }

/* list parameters */
  if (listflag) {
    listflag = 0;
    fprintf(stderr,"sleepint = %g\n",((float)sleepint)/MILLION);
  }

/* set cut */
  if (cutflag) {
    cutflag = 0;
    if (nstack == 1) {
      cutval = wcut = bcut = tcut = atof(pop());
      reproc++;
    }
    else {
      syntax();
      return;
    }
  }

/* set dith */
  if (dith) {
    dith = 0;
    if (nstack == 1) {
      dithval = tdither = wdither = bdither = atof(pop());
      reproc++;
    }
    else {
      syntax();
      return;
    }
  }

/* set range */
  if (rangeflag) {
    rangeflag = 0;
    if (nstack == 2) {
      range2 = atof(pop());
      range1 = atof(pop());
      dithval = tdither = wdither = bdither = (range2 - range1)/2.;
      cutval = wcut = bcut = tcut = (range2 + range1)/2.;
      reproc++;
    }
    else {
      syntax();
      return;
    }
  }

/* set sleepint */
  if (sleepflag) {
    sleepflag = 0;
    if (nstack == 1) {
      fsleepint = atof(pop());
      sleepint = MILLION*fsleepint;
      if (sleepint < SLMIN) sleepint = SLMIN;
      if (sleepint > SLMAX) sleepint = SLMAX;
    }
    else {
      syntax();
      return;
    }
  }

/* go do it */

/* but reprocess images if necessary */
  if (goflag && reproc) {
    reproc = 0;
    doproc();
  }

/* upon loop start, set up signal processing */
  signal(SIGINT,intrhand);

/* loop on display */
  while (goflag) {
    for (loop = 0; loop < nimages; loop++) {
      if (goflag) {
        dispimage(loop,list + loop);
        usleep(sleepint);
      }
      else {
        break;
      }
    }
    usleep(2*sleepint);
  }
}

/* doproc -- process images */
static void doproc()
{
  for (loop = 0; loop < nimages; loop++) {
    fprintf(stderr,"processing image %d\n",loop);
    procimage(loop,image[loop],list + loop,gargv[loop + optind],image_depth);
  }
}

/* syntax -- report a syntax error and zero stack */
static void syntax()
{
  fprintf(stderr,"syntax error\n");
  nstack = 0;
}

/* interrupt handler */
int intrhand(sig,code,scp,addr)
int sig,code;
struct sigcontext *scp;
char *addr;
{
  goflag = 0;
}

/* main program */
main(argc,argv)
int argc;
char **argv;
{
  int nxsize,nysize;
  float nlon,nlat;

/* check command line arguments */
  opterr = 0;
  nxsize = nysize = 0;
  while ((option = getopt(argc,argv,"s:")) != -1) {
    switch (option) {
    case 's':
      if (sscanf(optarg,"%d,%d,%f,%f",&nxsize,&nysize,&nlon,&nlat) != 4) {
        fprintf(stderr,"Usage: mcdisp [-s nx,ny,nlon,nlat] ifile1 ...\n");
        exit(1);
      }
      break;
    }
  }
  if (argc - optind < 1) {
    fprintf(stderr,"Usage: mcdisp [-s nx,ny,nlon,nlat] ifile1 ...\n");
    exit(1);
  }
  if (argc - optind >= MAXIMAGES) {
    fprintf(stderr,
         "mcdisp: %d is too many images -- dropping some early ones\n",
         argc - optind);
    optind = argc - MAXIMAGES + 1;
  }

/* globalize arguments */
  gargc = argc;
  gargv = argv;

/* set default sleep interval */
  sleepint = SLEEPINT;

/* loop on image retrieval */
  fprintf(stderr,"mcdisp: getting images...\n");
  nimages = argc - optind;
  nx = ny = 0;
  for (loop = 0; loop < nimages; loop++) {

/* getimage reads the file, allocates an image buffer, and fills it */
    image[loop] = getimage(argv[loop + optind],loop,list + loop,
         nxsize,nysize,nlon,nlat);
    fprintf(stderr,"     ...got %s...\n",argv[loop + optind]);

/* get maximum image size */
    if (list[loop].xsize > nx) nx = list[loop].xsize;
    if (list[loop].ysize > ny) ny = list[loop].ysize;
  }

/* set up display */
  image_depth = setupdisp(nx,ny);

/* set up default globals */
  if (image_depth == 1) {
    tcut = TCUT;
    bcut = BCUT;
    wcut = WCUT;
    tdither = TDITHER;
    bdither = BDITHER;
    wdither = WDITHER;
  }
  else {
    tcut = CTCUT;
    bcut = CBCUT;
    wcut = CWCUT;
    tdither = CTDITHER;
    bdither = CBDITHER;
    wdither = CWDITHER;
  }

/* clear the window */
  imageclear();

/* process the images */
  doproc();

/* hand control to command processor */
  yylex();
}
