/* nwsgrid -- This program interpolates station data to a two-dimensional
 * grid.  Multi-level station data, such as produced by radiosondes,
 * results in a three-dimensional grid.  However, the grid at each level
 * results only from data at that level.  The interpolation is done by
 * a weighted averaging technique inside a specified radius of influence ri.
 * The weighting function is (ri*ri - r*r)/(ri*ri + r*r), where r is
 * the distance between the data and grid point.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
/* maximum number of fields that can be gridded */
#define MAXFIELDS 50

char hbuff[HBMAX][LINE];           /* header buffer */
char cline[LINE];                  /* character line buffer */
char pval[LINE];                   /* another character line buffer */
char xngrid[LINE],yngrid[LINE];    /* index field names */
char vname[MAXFIELDS][LINE];       /* field names */
char vgname[MAXFIELDS][LINE];      /* gridded field names */
long nsbin,nsbout,nvbin,nvbout;    /* slice buffer lengths */
float *sbuff,*vbuff;               /* slice buffers */
float bad,badlim;                  /* bad data values */
float *x,*y;                       /* index field pointers */
float *xloc,*yloc;                 /* station location pointers */
float *p[MAXFIELDS],*gp[MAXFIELDS];  /* pointers to ungridded and gridded */
float *hour,*hourav;               /* pointers to hour data */
int nlev[MAXFIELDS];               /* number of levels */
float *fin,*fout;                  /* contiguous buffers for gridding */
int loop,poop,goop;                /* looping variables */
int nfields;                       /* number of fields to be gridded */
int nstns;                         /* number of stations */
struct field *fp;                  /* field pointer temporary */

main(argc,argv)
int argc;
char *argv[];
{
  char *sname,*xname,*yname;         /* input strings */
  float x0,dx,y0,dy,radius;          /* input floats */
  int nx,ny;                         /* input ints */
/* check arguments */
  if (argc != 11) {
    fprintf(stderr,
         "Usage: nwsgrid sname xname x0 dx nx yname y0 dy ny radius\n");
    exit(1);
  }
  sname = argv[1];
  xname = argv[2];
  yname = argv[6];
  x0 = atof(argv[3]);
  dx = atof(argv[4]);
  y0 = atof(argv[7]);
  dy = atof(argv[8]);
  radius = atof(argv[10]);
  nx = atoi(argv[5]);
  ny = atoi(argv[9]);
/* get header */
  gethdr(stdin,hbuff,HBMAX);
/* check for float format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"nwsgrid: expecting float format\n");
    exit(1);
  }
/* get input buffer lengths */
  nsbin = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');
/* see if xname and yname fields exist */
  if ((fp = seekfld(hbuff,HBMAX,'v',xname)) == NULL) {
    fprintf(stderr,"nwsgrid: field %s doesn't exist\n",xname);
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,sname) != 0)) {
    fprintf(stderr,"nwsgrid: %s dimension size or name incorrect\n",xname);
    exit(1);
  }
  if ((fp = seekfld(hbuff,HBMAX,'v',yname)) == NULL) {
    fprintf(stderr,"nwsgrid: field %s doesn't exist\n",yname);
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,sname) != 0)) {
    fprintf(stderr,"nwsgrid: %s dimension size or name incorrect\n",yname);
    exit(1);
  }
/* pick up number of stations */
  nstns = fp->tsize;
/* get bad data values */
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;
/* add new things to header, starting with index fields */
  sprintf(xngrid,"%sgrid",xname);
  sprintf(yngrid,"%sgrid",yname);
  addfld(hbuff,HBMAX,'s',xngrid,100.,0.,'l',1,xngrid,nx);
  addfld(hbuff,HBMAX,'s',yngrid,100.,0.,'l',1,yngrid,ny);
/* add index parameters */
  sprintf(cline,"%s0",xngrid);
  sprintf(pval,"%f",x0);
  addpar(hbuff,HBMAX,cline,pval);
  sprintf(cline,"d%s",xngrid);
  sprintf(pval,"%f",dx);
  addpar(hbuff,HBMAX,cline,pval);
  sprintf(cline,"%s0",yngrid);
  sprintf(pval,"%f",y0);
  addpar(hbuff,HBMAX,cline,pval);
  sprintf(cline,"d%s",yngrid);
  sprintf(pval,"%f",dy);
  addpar(hbuff,HBMAX,cline,pval);
/* add comment */
  sprintf(cline,"%s: %s\n",argv[0],argv[1]);
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s\n",argv[2],argv[3],argv[4],argv[5]);
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s %s\n",argv[6],argv[7],argv[8],argv[9],
       argv[10]);
  addcline(hbuff,HBMAX,cline);
/* find all fields starting with sname as dimension - dim = 1 or 2 is ok */
  nfields = loop = 0;
  while ((fp = getfld(hbuff,HBMAX,'v',loop++)) != NULL) {
    if ((fp->dim >= 1) && (fp->dim <= 2) && (strcmp(fp->dname1,sname) == 0)) {
/* don't grid xname and yname */
      if (strcmp(fp->fname,xname) == 0) continue;
      if (strcmp(fp->fname,yname) == 0) continue;
      if (nfields >= MAXFIELDS) {
        fprintf(stderr,"nwsgrid: too many fields to be gridded\n");
        exit(1);
      }
/* store gridded and ungridded names for later ref and add the field */
      strcpy(vname[nfields],fp->fname);
      sprintf(vgname[nfields],"%sgrid",fp->fname);
      addfld(hbuff,HBMAX,'v',vgname[nfields++],fp->smul,fp->sadd,fp->bytes,
           fp->dim + 1,xngrid,nx,yngrid,ny,fp->dname2,fp->dsize2);
    }
  }
/* add variable slice field hourav */
  addfld(hbuff,HBMAX,'v',"hourav",10.,0.,'s',0);
/* write modified header */
  puthdr(stdout,hbuff,HBMAX);
/* allocate static slice space */
  nsbout = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbout);
/* get static slice */
  getslice(stdin,nsbin,sbuff);
/* set the index fields */
  x = getptr(hbuff,HBMAX,sbuff,'s','d',xngrid);
  for (loop = 0; loop < nx; loop++) x[loop] = x0 + loop*dx;
  y = getptr(hbuff,HBMAX,sbuff,'s','d',yngrid);
  for (loop = 0; loop < ny; loop++) y[loop] = y0 + loop*dy;
/* put the static slice */
  putslice(stdout,nsbout,sbuff);
/* allocate variable slice buffer */
  nvbout = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbout);
/* assign pointers for non-gridded and gridded fields */
  for (loop = 0; loop < nfields; loop++) {
    gp[loop] = getptr(hbuff,HBMAX,vbuff,'v','d',vgname[loop]);
    p[loop] = getptr2(hbuff,HBMAX,vbuff,'v','d',vname[loop],&fp);
/* collect info on levels */
    if (fp->dim == 1) nlev[loop] = 1;
    else nlev[loop] = fp->dsize2;
  }
/* assign pointers for station locations */
  xloc = getptr(hbuff,HBMAX,vbuff,'v','d',xname);
  yloc = getptr(hbuff,HBMAX,vbuff,'v','d',yname);
/* assign hourav pointer */
  hourav = getptr(hbuff,HBMAX,vbuff,'v','d',"hourav");
/* allocate space for contiguous station and grid buffers */
  fin = getbuff(nstns);
  fout = getbuff(nx*ny);
/* loop on variable slice read */
  while (getslice(stdin,nvbin,vbuff) != EOF) {
/* loop on gridable fields */
    for (loop = 0; loop < nfields; loop++) {
/* loop on levels */
      for (poop = 0; poop < nlev[loop]; poop++) {
/* transfer station data to contiguous array */
        for (goop = 0; goop < nstns; goop++) {
          fin[goop] = p[loop][goop*nlev[loop] + poop];
        }
/* do gridding */
        grid(fin,xloc,yloc,nstns,x0,dx,nx,y0,dy,ny,radius,fout);
/* transfer gridded data from contiguous array */
        for (goop = 0; goop < nx*ny; goop++) {
          gp[loop][goop*nlev[loop] + poop] = fout[goop];
        }
      }
    }
/* compute hourav as average of all hours */
    if ((hour = getptr2(hbuff,HBMAX,vbuff,'v','r',"hour",&fp)) != NULL) {
      *hourav = 0.;
      for (loop = 0; loop < fp->tsize; loop++) *hourav += hour[loop];
      *hourav /= fp->tsize;
    }
    else *hourav = bad;
/* put variable slice */
    putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}

/* this routine does the gridding */

#define I(ix,iy) ix*ny + iy
#define NSTATMIN 3

grid(svals,xloc,yloc,nstns,x0,dx,nx,y0,dy,ny,radius,gvals)
float *svals;      /* station values */
float *xloc,*yloc; /* station locations */
int nstns;         /* number of stations */
float x0,dx,y0,dy; /* index parameters */
int nx,ny;         /* sizes of index fields */
float *gvals;      /* gridded values */
float radius;      /* weighting factors */
{
  int ix,iy,istns;             /* looping variables */
  int x1,x2,y1,y2;             /* limit values on grid */
  static float *denom = NULL;  /* work space pointer */
  static short *nstats = NULL; /* more work space */
  float xgrid,ygrid;           /* difference values of x and y */
  float weight;                /* weighting factor */
  float rsq,radsq;             /* squared radii */
/* allocate work space only once */
  if (denom == NULL) {
    denom = getbuff(nx*ny);
    if ((nstats = (short*)malloc(nx*ny*sizeof(short))) == NULL) {
      fprintf(stderr,"nwsgrid: can't get enough memory\n");
      exit(1);
    }
    radsq = radius*radius;
  }
/* zero work space */
  for (ix = 0; ix < nx*ny; ix++) {
    gvals[ix] = denom[ix] = 0.;
    nstats[ix] = 0;
  }
/* loop on stations */
  for (istns = 0; istns < nstns; istns++) {
/* skip stations with bad locations or data */
    if (fabs(xloc[istns]) > badlim) continue;
    if (fabs(yloc[istns]) > badlim) continue;
    if (fabs(svals[istns]) > badlim) continue;
/* compute grid points within approximate range of influence */
    x1 = (xloc[istns] - x0 - radius)/dx - .5;
    x2 = (xloc[istns] - x0 + radius)/dx + 1.5;
    y1 = (yloc[istns] - y0 - radius)/dy - .5;
    y2 = (yloc[istns] - y0 + radius)/dy + 1.5;
/* limit to actual grid */
    if (x1 < 0) x1 = 0;
    if (x1 > nx) x1 = nx;
    if (x2 < 0) x2 = 0;
    if (x2 > nx) x2 = nx;
    if (y1 < 0) y1 = 0;
    if (y1 > ny) y1 = ny;
    if (y2 < 0) y2 = 0;
    if (y2 > ny) y2 = ny;
/* loop on grid portion near station */
    for (ix = x1; ix < x2; ix++) {
      for (iy = y1; iy < y2; iy++) {
/* compute weight */
        xgrid = x0 + dx*ix - xloc[istns];
        ygrid = y0 + dy*iy - yloc[istns];
        rsq = xgrid*xgrid + ygrid*ygrid;
/*
if (ix == 0 && iy == 4) fprintf(stderr,"rsq = %f; radsq = %f\n",rsq,radsq);
*/
        if (rsq > radsq) {
          weight = 0.;
        }
        else {
          rsq = sqrt(rsq);
          weight = (radius - rsq)/(radius + rsq);
/* increment grid fields */
          gvals[I(ix,iy)] += svals[istns]*weight;
          denom[I(ix,iy)] += weight;
          nstats[I(ix,iy)] += 1;
        }
/*
if (ix == 0 && iy == 4) fprintf(stderr,"  xgrid = %f; ygrid = %f; rsq = %f; istns = %d; svals = %f\n",xgrid,ygrid,rsq,istns,svals[istns]);
*/
      }
    }
  }
/* normalize gridded field */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      if ((denom[I(ix,iy)] == 0.) || (nstats[I(ix,iy)] < NSTATMIN)) {
        gvals[I(ix,iy)] = bad;
      }
      else gvals[I(ix,iy)] /= denom[I(ix,iy)];
    }
  }
}
