/* nwspv -- Calculate the potential vorticity from the cdf */
/*          gridded surface data */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#define TINY 1.e-30
#define DTOR 0.017453292		/* degrees to radian conversion */
#define I(x,y,z) x*latl*presl+y*presl+z

float ckdenom(),diff();     		/* function declarations */
char setflag();			        /* function declaration */
char *hbuff[HBMAX][LINE];		/* header buffer */
char cline[LINE];			/* character line buffer */
long nsbuff,nvbin,nvbout;		/* slice buffer sizes */
float *sbuff,*vbuff;		        /* slice buffer pointers */
float *longfield,*latfield,*pfield;     /* pointer to index fields */
float *ptfield,*ufield,*vfield;		/* pointers to data fields */
float *pvfield;				/* pointer to output pv field */
float bad,badlim;			/* bad data values */
float dlonggrid,dlatgrid,dpres;		/* index field increments */
float dptdp,dudp,dvdp,dptdy,dptdx,dudy,dvdx;/* intermidiate results */
float denom;				/* denominator */
float f;				/* coriolis parameter */
long il,il2,ih,ih2,im;			/* temp index field points */
char longname[LINE],latname[LINE],presname[LINE];/* index field names */
char flag;				/* diffence type name */
long longl,latl,presl;			/* loop lengths */
long x,y,z;				/* loop indexes */
struct field *fp,*fppt,*fpu,*fpv;	/* data field struct pointers */
struct field *fplong,*fplat,*fppres;    /* index field struct pointers*/


main(argc,argv)
int argc;
char *argv[];

{
/* ERROR TRAPS AND INITIALIZATIONS */
/* ******************************* */

/* check arguments */
  if (argc != 5) {
    fprintf(stderr, "Usage: nwspv u_field v_field vtheta_field pv_field\n");
    exit(1);
  }

/* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"nwspv: float format expected\n");
    exit(1);
  }

/* get bad data values */
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;

/* compute buffer lengths */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');

/* add stuff to header, comment first */
  sprintf(cline,"nwspv: %s %s\n",argv[1],argv[2]);
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s\n",argv[3],argv[4]);
  addcline(hbuff,HBMAX,cline);

/* get information on field dimensions */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[1])) == NULL) {
    fprintf(stderr,"nwspv: can't find %s field\n",argv[1]);
    exit(1);
  }
  if (fp->dim != 3) {
    fprintf(stderr,"nwspv: expecting 3-D velocity and pot temp fields\n");
    exit(1);
  }
/* get longitude, latitude and pressure names and sizes */
  strcpy(longname,fp->dname1);
  strcpy(latname,fp->dname2);
  strcpy(presname,fp->dname3);
  longl = fp->dsize1;
  latl = fp->dsize2;
  presl = fp->dsize3;
  addfld(hbuff,HBMAX,'v',argv[4],100.,0.,'s',3,
      longname,longl,latname,latl,presname,presl);

/* compute output buffer length */
  nvbout = elemcnt(hbuff,HBMAX,'v');

/* allocate buffers */
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbout);

/* check for existence of data fields and assign pointers */
  if ((ptfield = getptr2(hbuff,HBMAX,vbuff,'v','r',argv[3],&fppt)) == NULL) {
    fprintf(stderr,"nwspv:potential temperature field not found\n");
    exit(1);
  }
  if (((ufield = getptr2(hbuff,HBMAX,vbuff,'v','r',argv[1],&fpu))
      == NULL) || ((vfield = getptr2(hbuff,HBMAX,vbuff,'v','r',argv[2],
      &fpv)) == NULL)) {
    fprintf(stderr,"nwspv: wind velocity fields not found\n");
    exit(1);
  }

/* check for existance of index fields and assign pointers */
  if (((longfield = getptr2(hbuff,HBMAX,sbuff,'s','r',longname,
      &fplong)) == NULL) || ((latfield = getptr2(hbuff,HBMAX,sbuff,
      's','r',latname,&fplat)) == NULL) || ((pfield = getptr2(hbuff,
      HBMAX,sbuff,'s','r',presname,&fppres)) == NULL)) {
    fprintf(stderr,"nwspv: index fields not found\n");
    exit(1);
  }

/* assign pointer for output field */
  pvfield = getptr(hbuff,HBMAX,vbuff,'v','d',argv[4]);

/* get index field increment values */
  if(seekpar(hbuff,HBMAX,"dpres",cline) == OK) dpres = atof(cline);
  else {
    fprintf(stderr,"nwspv: pressure increment not found\n");
    exit(1);
  }
  if(seekpar(hbuff,HBMAX,"dlatgrid",cline) == OK) dlatgrid = atof(cline);
  else {
    fprintf(stderr,"nwspv: lattitude increment not found\n");
    exit(1);
  }
  if(seekpar(hbuff,HBMAX,"dlonggrid",cline) == OK) dlonggrid = atof(cline);
  else {
    fprintf(stderr,"nwspv: longitude increment not found\n");
    exit(1);
  }

/* MAIN PROGRAM */
/* ************ */

/* write header */
  puthdr(stdout,hbuff,HBMAX);

/* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* loop while there is a variable slice */
  while (getslice(stdin,nvbin,vbuff) != EOF) {

/* loop over all interior data points */
    for(x = 0; x < longl; x++)  /* loop over longitude */ {
      for(y = 0; y < latl; y++) /* loop over lattitude */ {
        for(z = 0; z < presl; z++)  /* loop over pressure */ {
          pvfield[I(x,y,z)] = bad;
/* differences over pressure */
          il2 = I(x,y,(z-2));
          il = I(x,y,(z-1));
          im = I(x,y,z);
          ih = I(x,y,(z+1));
          ih2 = I(x,y,(z+2));
	  denom = ckdenom(2.0*dpres);
          flag = setflag(ptfield,presl,z);
          dptdp = diff(ptfield,flag);
          flag = setflag(ufield,presl,z);
          dudp = diff(ufield,flag);
          flag = setflag(vfield,presl,z);
          dvdp = diff(vfield,flag);
          if((fabs(dptdp) <= badlim) && (fabs(dudp) <= badlim)
              && (fabs(dvdp) <= badlim)) {
/* differences over latitude */
	    il2 = I(x,(y-2),z);
            il = I(x,(y-1),z);
            ih = I(x,(y+1),z);
	    ih2 = I(x,(y+2),z);
	    denom = ckdenom(2.0*dlatgrid)*111.195;
	    flag = setflag(ptfield,latl,y);
	    dptdy = diff(ptfield,flag);
	    flag = setflag(ufield,latl,y);
	    dudy = diff(ufield,flag)
		-tan(latfield[y]*DTOR)*ufield[im]/6400.0;
            if((fabs(dptdy) <= badlim) && (fabs(dudy) <= badlim)) {
/* differences over longitude */
	      il2 = I((x-2),y,z);
              il = I((x-1),y,z);
              ih = I((x+1),y,z);
	      ih2 = I((x+2),y,z);
	      denom = ckdenom(2.0*dlonggrid)*111.195*cos(latfield[y]*DTOR);
	      flag = setflag(ptfield,longl,x);
	      dptdx = diff(ptfield,flag);
	      flag = setflag(vfield,longl,x);
	      dvdx = diff(vfield,flag);
	      if((fabs(dptdx) <= badlim) && (fabs(dvdx) <= badlim)) {
/* calculate coriolis parameter and pv */
		f = 0.1454*sin(latfield[y]*DTOR);
		pvfield[I(x,y,z)] = (-(dptdy*dudp)+(dptdx*dvdp)
        	    -(dptdp*(dvdx-dudy+f)))*98.0;
	      }
            }
          }
        }
      }
    }
/* write variable slice */
  putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}

/* this function determines which differencing method is to be used */
/* for each derivative. */
/* flag='c': centered difference */
/* flag='b': backwards difference */
/* flag='f': forward difference */
/* flag=' ': result set to 'bad' */

char setflag(ifpt,length,mid)
float *ifpt;
long length,mid;
{

  if((mid != 0) && (mid != length-1) && (fabs(ifpt[il]) <= badlim)
      && (fabs(ifpt[ih]) <= badlim)) return('c');
  else if(fabs(ifpt[im]) <= badlim) {
    if((mid == 0) || (fabs(ifpt[il]) > badlim)) {
      if((mid < length-2) && (fabs(ifpt[ih]) <= badlim)
	  && (fabs(ifpt[ih2]) <= badlim))
        return('f');
    }
    else if((mid == (length-1)) || (fabs(ifpt[ih]) > badlim)) {
      if((mid > 1) && (fabs(ifpt[il]) <= badlim) && (fabs(ifpt[il2]) <= badlim))
        return('b');
      else return(' ');
    }
  }
}


/* this function calculates the derivatives according to the flag */
/* set in setflag */
float diff(ifpt,flg)
float *ifpt;
char flg;
{

  switch(flg) {
    case 'c':
      return((ifpt[ih]-ifpt[il])/denom);
    case 'b' :
      return((3.0*ifpt[im]-4.0*ifpt[il]+ifpt[il2])/denom);
    case 'f':
      return((-ifpt[ih2]+4.0*ifpt[ih]-3.0*ifpt[im])/denom);
    case ' ':
    default:
      return(bad);
  }
}


float ckdenom(d)      /* checks for division by zero */
float d;
{
  if(fabs(d) < TINY)
  {
    fprintf(stderr,"nwspv: division by zero\n");
    exit(1);
  }
  return(d);
}
