/* nwsts -- Do time and space rationalization for National Weather
 * service synoptic weather (surface and upper air) gridded data.
 * Compute a new time variable called "jday", which is the Julian
 * day minus some reference day.  Also compute augrid and avgrid
 * which are gridded east and north velocity components in terms of
 * degrees per day.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define I(ix,iy,iz) (iz + iy*nz + ix*ny*nz)
#define AU "augrid"
#define AV "avgrid"
#define JD "julday"
#define SECPERDAY 86400.
#define RADEARTH 6.4e6
#define DEGPERRAD 57.2959

char hbuff[HBMAX][LINE];               /* header buffer */
char cline[LINE];                      /* line buffer */
float *sbuff,*vbuff;                   /* slice buffer pointers */
long nsbuff,nvbin,nvbout;              /* slice lengths */
float *u,*v,*au,*av,*jday,*year,*month,*day,*hour,*lon,*lat;
                                       /* field pointers */
int nx,ny,nz;                          /* array sizes */
long totalsize;                        /* size of velocity grids */
int dimens;                            /* number of field dimensions */
int loop,ix,iy,iz;                     /* looping variables */
float bad,badlim;                      /* bad data stuff */
struct field *fp;
char lonname[LINE],latname[LINE];      /* latitude and longitude field names */
char zname[LINE];                      /* vertical dimension name */
long refday;                           /* reference Julian day */
float latcon,loncon;                   /* computed conversion factors */
long julian_date();

main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if (argc != 8) {
    fprintf(stderr,"Usage: nwsts refday u v year month day hour\n");
    exit(1);
  }
  refday = atoi(argv[1]);
/* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"nwsts: float format needed\n");
    exit(1);
  }
/* establish bad data stuff */
  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;
/* get input sizes */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');
/* seek information on grid */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[2])) == NULL) {
    fprintf(stderr,"nwsts: field %s not found\n",argv[2]);
    exit(1);
  }
  if (fp->dim == 2) {
    nz = 1;
    strcpy(zname,"");
  }
  else if (fp->dim == 3) {
    nz = fp->dsize3;
    strcpy(zname,fp->dname3);
  }
  else {
    fprintf(stderr,"nwsts: dimensionality %d of %s not 2 or 3\n",
         fp->dim,argv[2]);
  }
  nx = fp->dsize1;
  ny = fp->dsize2;
  totalsize = nx*ny*nz;
  strcpy(lonname,fp->dname1);
  strcpy(latname,fp->dname2);
  dimens = fp->dim;
/* add appropriate fields to header buffer */
  addfld(hbuff,HBMAX,'v',JD,24.,0.,'l',0);
  addfld(hbuff,HBMAX,'v',AU,100.,0.,'s',dimens,lonname,nx,latname,ny,zname,nz);
  addfld(hbuff,HBMAX,'v',AV,100.,0.,'s',dimens,lonname,nx,latname,ny,zname,nz);
/* add comment */
  sprintf(cline,"nwsts: %s %s %s\n",argv[1],argv[2],argv[3]);
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s\n",argv[4],argv[5],argv[6],argv[7]);
  addcline(hbuff,HBMAX,cline);
/* allocate slice buffers */
  sbuff = getbuff(nsbuff);
  nvbout = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbout);
/* allocate field pointers */
  au = getptr2(hbuff,HBMAX,vbuff,'v','d',AU,&fp);
  if (fp->tsize != totalsize) {
    fprintf(stderr,"nwsts: size inconsistency in %s\n",AU);
    exit(1);
  }
  av = getptr2(hbuff,HBMAX,vbuff,'v','d',AV,&fp);
  if (fp->tsize != totalsize) {
    fprintf(stderr,"nwsts: size inconsistency in %s\n",AV);
    exit(1);
  }
  u = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[2],&fp);
  if (fp->tsize != totalsize) {
    fprintf(stderr,"nwsts: size inconsistency in %s\n",argv[2]);
    exit(1);
  }
  v = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[3],&fp);
  if (fp->tsize != totalsize) {
    fprintf(stderr,"nwsts: size inconsistency in %s\n",argv[3]);
    exit(1);
  }
  year = getptr(hbuff,HBMAX,vbuff,'v','d',argv[4]);
  month = getptr(hbuff,HBMAX,vbuff,'v','d',argv[5]);
  day = getptr(hbuff,HBMAX,vbuff,'v','d',argv[6]);
  hour = getptr(hbuff,HBMAX,vbuff,'v','d',argv[7]);
  jday = getptr(hbuff,HBMAX,vbuff,'v','d',JD);
  lon = getptr(hbuff,HBMAX,sbuff,'s','d',lonname);
  lat = getptr(hbuff,HBMAX,sbuff,'s','d',latname);
/* write header */
  puthdr(stdout,hbuff,HBMAX);
/* transfer static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);
/* loop on variable slices */
  while (getslice(stdin,nvbin,vbuff) != EOF) {
/* do time computation */
    *jday = julian_date((long)(*year),(long)(*month),(long)(*day)) - refday;
    *jday += *hour/24.;
/* do velocity computation */
    for (iy = 0; iy < ny; iy++) {
/* compute conversion factors */
      if (fabs(lat[iy]) > 89.) {
        fprintf(stderr,"nwsts: cannot do polar regions\n");
        exit(1);
      }
      latcon = SECPERDAY*DEGPERRAD/RADEARTH;
      loncon = latcon/cos(lat[iy]/DEGPERRAD);
      for (ix = 0; ix < nx; ix++) {
        for (iz = 0; iz < nz; iz++) {
/* compute common index */
          loop = I(ix,iy,iz);
/* make conversions */
          if (fabs(u[loop]) < badlim) au[loop] = u[loop]*loncon;
          else au[loop] = bad;
          if (fabs(v[loop]) < badlim) av[loop] = v[loop]*latcon;
          else av[loop] = bad;
        }
      }
    }
/* write variable slice */
    putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}
