/* dispimage.c -- X11 version. */

#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#define BORDER 5
#define DEF_FONT "9x15" 
#define NCOLORS 34
#define NGRAY 23
#define NMONO 2
#define MAXPOLY 20
#define MAXIMAGES 50
#define MAXBUTTONS 200
#define MAXBUTTONLABELSIZE 20
#define DIALOGUEBOXMAXINPUT 200

static unsigned char *imagelist[MAXIMAGES]; /* list of images */
static nx,ny;                      /* size of pixmaps */
static depth;                      /* depth of pixmaps */
static Pixmap pixlist[MAXIMAGES];  /* list of pixmaps */
static Display *mydisplay;         /* X11 stuff */
static Window mywindow;
static XImage *myimage;
static GC mygc;
static XEvent myevent;
static XSizeHints myhint;
static int myscreen;
static unsigned long mybackground;
static unsigned long myforeground;
static Colormap cmap;
static Visual *vis;
static XColor exact,xcvals[NCOLORS];
static XFontStruct *fontstruct;     /* Font descriptor */
static XGCValues gcv;               /* structure for making graphics context */
static int ncolors;                 /* actual number of colors */
static char cnames[NCOLORS][20] = {"black","white","gray0","gray5",
     "gray10","gray15","gray20","gray25","gray30","gray35","gray40",
     "gray45","gray50","gray55","gray60","gray65","gray70","gray75",
     "gray80","gray85","gray90","gray95","gray100",
     "brown","red","orange","yellow","yellow green","sea green",
     "aquamarine","sky blue","blue","violet","navy"};
static XPoint points[MAXPOLY];       /* store points for polygon fill */
static char progname[62];            /* name of program */
static char *pnptr;                  /* pointer to name of program */
static long currentcolor;
struct BUTTON
{
  int x1,y1,x2,y2;
  char label[MAXBUTTONLABELSIZE];
  char active;
  int screenindex;
};
static struct BUTTON button[MAXBUTTONS];

/* internal functions */
static setuppixmap();
static colorfail();
static void getfontsize();
static void setcolor();

/* setupdisp -- this cranks up the X11 display and returns
 * the pixel depth of the display.
 */
setupdisp(pname,lnx,lny)
char *pname;                       /* name of program */
int lnx,lny;                         /* size of image */
{
  int loop;

/* initialize pixlist */
  for (loop = 0; loop < MAXIMAGES; loop++) pixlist[loop] = 0;

/* transfer program name and image size */
  if (strlen(pname) > 60) {
    fprintf(stderr,"setupdisp: program name too long\n");
    exit(1);
  }
  strcpy(progname,pname);
  pnptr = progname;
  nx = lnx;
  ny = lny;

/* initialize storage pointers */
  for (loop = 0; loop < MAXIMAGES; loop++) imagelist[loop] = 0;

/* set up the X11 connection -- take default server and screen */
  if ((mydisplay = XOpenDisplay("")) == NULL) {
    fprintf(stderr,"setupdisp: can't open default X11 display\n");
    exit(1);
  }
  myscreen = DefaultScreen(mydisplay);

/* find out things about the workstation display */
  depth = DefaultDepth(mydisplay,myscreen);
  if (depth != 1 && depth != 8) {
    fprintf(stderr,"dispimage: display of depth %d not supported\n",depth);
    exit(1);
  }
  vis = DefaultVisual(mydisplay,myscreen);
  if (depth > 1) {
    if (vis->class == GrayScale || vis->class == StaticGray) ncolors = NGRAY;
    else ncolors = NCOLORS;
  }
  else ncolors = NMONO;

/* set up hints for the window manager */
  myhint.x = 1;
  myhint.y = 1;
  myhint.width = lnx;
  myhint.height = lny;
  myhint.flags = PPosition | PSize;

/* Set background and foreground colors to white and black. */
  mybackground = WhitePixel(mydisplay,myscreen);
  myforeground = BlackPixel(mydisplay,myscreen);
  currentcolor = myforeground;

/* create a window */
  mywindow = XCreateSimpleWindow(mydisplay,DefaultRootWindow(mydisplay),
       myhint.x,myhint.y,myhint.width,myhint.height,BORDER,
       myforeground,mybackground);

/* set standard properties */
  XSetStandardProperties(mydisplay,mywindow,progname,progname,None,
       &pnptr,1,&myhint);

/* get font */
  if ((fontstruct = XLoadQueryFont(mydisplay,DEF_FONT)) == NULL) {
    fprintf(stderr,"dispimage: can't find font %s\n",DEF_FONT);
    exit(1);
  }

/* set up GCs */
  gcv.foreground = myforeground;
  gcv.background = mybackground;
  gcv.font = fontstruct->fid;
  mygc = XCreateGC(mydisplay,mywindow,(GCForeground|GCBackground|GCFont),&gcv);

/* set events of interest -- exposure, keypress, and mouse button down
 * in window
 */
  XSelectInput(mydisplay,mywindow,ExposureMask | ButtonPressMask |
       OwnerGrabButtonMask | KeyPressMask);

/* map the window */
  XMapWindow(mydisplay,mywindow);

/* wait for last exposure event and ignore button press events */
  while (1) {
    XNextEvent(mydisplay,&myevent);
    if ((myevent.type == Expose) && (myevent.xexpose.count == 0)) break;
  }

/* get default color map and colors */
  cmap = XDefaultColormap(mydisplay,myscreen);
  for (loop = 0; loop < ncolors; loop++) {
    if (XAllocNamedColor(mydisplay,cmap,cnames[loop],&exact,xcvals + loop)
         == 0) colorfail(cnames[loop]);
  }

/* return pixel depth */
  return(depth);
}

/* getcolors -- this returns a list of pixel values for the color list */
getcolors(cvals)
long cvals[];
{
  int i;
  for (i = 0; i < ncolors; i++) {
    cvals[i] = xcvals[i].pixel;
  }
  return(ncolors);
}

/* sendimage -- in the X11 version this creates initial image on
 * a pixmap.
 */
sendimage(index,ox,oy,sx,sy,image)
int index;                         /* instance of image -- position in list */
int ox,oy;                         /* offset of image from upper left corner */
int sx,sy;                         /* size of image */
unsigned char *image;              /* pointer to the image */
{

/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

/* create the image structure */
  if ((myimage = XCreateImage(mydisplay,vis,depth,ZPixmap,0,
       image,nx,ny,32,0)) == NULL) {
    fprintf(stderr,"sendimage: can't allocate ximage structure\n");
    exit(1);
  }

/* blap the image onto the pixmap */
  XSetFunction(mydisplay,mygc,GXcopy);
  XPutImage(mydisplay,pixlist[index],mygc,myimage,0,0,ox,oy,sx,sy);

/* free the image structure */
  XFree(myimage);
}

/* dispimage -- display a particular image */
dispimage(index)
int index;                       /* instance of image to be displayed */
{

/* check to see if the image exists */
  if (pixlist[index] == 0) {
    fprintf(stderr,"dispimage: trying to display nonexistent image %d\n",index);
    exit(1);
  }

/* first, check for an exposure event, and process if it is there */
  while (XCheckWindowEvent(mydisplay,mywindow,ExposureMask,&myevent) != 0);

/* display image */
  XSetFunction(mydisplay,mygc,GXcopy);
  XCopyArea(mydisplay,pixlist[index],mywindow,mygc,0,0,nx,ny,0,0);

/* synchronize */
  XSync(mydisplay,0);
}

/* drawline -- Draw a line on image in given color, or in inverting mode if 
 * the color is -1 (on color displays, this yields the normal foreground).
 * To the right is more positive in x, and down is more positive in y.
 */
drawline(index,xstart,ystart,xend,yend,linetype,color)
int index;
int xstart,ystart,xend,yend;
int linetype;
long color;
{

char pat[4];

/* linetype -- 1 = solid, 2 = dashed, 3 = dotted, 4 = dot-dashed, as in pgraf
 * 
 * color --  -1 = invert
 */

/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  setcolor(color);

/* set the line type */
  switch(linetype)
  {
    case 1 : {
      XSetLineAttributes(mydisplay,mygc,0,LineSolid,CapButt,JoinMiter);
      break; }
    case 2 : {
      pat[0] = 6;
      pat[1] = 3;
      XSetDashes(mydisplay,mygc,0,pat,2);
      XSetLineAttributes(mydisplay,mygc,0,LineOnOffDash,CapButt,JoinMiter);
      break; }
    case 3 : {
      pat[0] = 1;
      pat[1] = 3;
      XSetDashes(mydisplay,mygc,0,pat,2);
      XSetLineAttributes(mydisplay,mygc,0,LineOnOffDash,CapButt,JoinMiter);
      break; }
    case 4 : {
      pat[0] = 6;
      pat[1] = 3;
      pat[2] = 1;
      pat[3] = 3;
      XSetDashes(mydisplay,mygc,0,pat,4);
      XSetLineAttributes(mydisplay,mygc,0,LineOnOffDash,CapButt,JoinMiter);
      break; }
  }

/* do the drawing */
  XDrawLine(mydisplay,pixlist[index],mygc,xstart,ystart,xend,yend);

/* return to solid line type */
  XSetLineAttributes(mydisplay,mygc,0,LineSolid,CapButt,JoinMiter);
}


/* drawtext -- Draw text on image in the given color. */
drawtext(index,x,y,color,text)
int index;
int x,y;
long color;
char *text;
{
/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  setcolor(color);

/* do the text */
  XDrawImageString(mydisplay,pixlist[index],mygc,x,y,text,strlen(text));
}


/* mousepos -- Get position and button of mouse when clicking any button */
mousepos(index,ax,ay,abutton)
int index;
int *ax,*ay,*abutton;
{

/* display the image of interest */
  dispimage(index);

/* discard all mouse press events collected up to this point */
  while (1) {
    if (XCheckTypedEvent(mydisplay,ButtonPress,&myevent) == 0) break;
  }

/* wait for a mouse press event -- redraw on last exposure event */
  while (1) {
    XNextEvent(mydisplay,&myevent);
    if ((myevent.type == Expose) && (myevent.xexpose.count == 0)) {
      dispimage(index);
    }
    else if (myevent.type == ButtonPress) {
      *ax = myevent.xbutton.x;
      *ay = myevent.xbutton.y;
      *abutton = myevent.xbutton.button;
      break;
    }
  }
}


/* mouseclicked -- Get position and button of mouse if it has been clicked */
int mouseclicked(ax,ay,abutton)
int *ax,*ay,*abutton;
{
  if (XCheckTypedEvent(mydisplay,ButtonPress,&myevent) == 0) return(0);
  *ax = myevent.xbutton.x;
  *ay = myevent.xbutton.y;
  *abutton = myevent.xbutton.button;
  return(1);
}


/* fillpoly -- fill a polygon with the desired color; -1 gives inverting mode */
fillpoly(index,nverts,xlist,ylist,color)
int index;
int nverts;
int *xlist,*ylist;
long color;
{
  int i;

/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

/* check to see if max number of vertices is not exceeded */
  if (nverts > MAXPOLY) {
    fprintf(stderr,"fillpoly: %d is too many vertices; max = %d\n",nverts,
       MAXPOLY);
    exit(1);
  }

/* transfer points */
  for (i = 0; i < nverts; i++) {
    points[i].x = xlist[i];
    points[i].y = ylist[i];
  }

  setcolor(color);

/* fill the polygon */
  XFillPolygon(mydisplay,pixlist[index],mygc,points,nverts,
       Convex,CoordModeOrigin);
}


/* setuppixmap -- set up pixmap if needed */
static setuppixmap(index)
int index;
{
  Window troot;
  int tx,ty;
  unsigned int tw,th,tbw,td;

/* check for proper range */
  if (index < 0 || index >= MAXIMAGES) {
    fprintf(stderr,"setuppixmap: %d is too many images\n",index);
    exit(1);
  }

/* only create if needed */
  if (pixlist[index] == 0) {
    pixlist[index] = XCreatePixmap(mydisplay,mywindow,nx,ny,depth);

/* test that creation succeeded */
    tw = th = 0;
    XGetGeometry(mydisplay,pixlist[index],&troot,&tx,&ty,&tw,&th,&tbw,&td);
    if (tw != nx || th != ny) {
      fprintf(stderr,"setuppixmap: can't allocate pixmap %d\n",index);
      exit(1);
    }

/* clean it */
    XSetForeground(mydisplay,mygc,mybackground);
    XSetFunction(mydisplay,mygc,GXcopy);
    XFillRectangle(mydisplay,pixlist[index],mygc,0,0,nx,ny);

/* reset to former foreground */
    XSetForeground(mydisplay,mygc,myforeground);

/* initialize background */
    XSetBackground(mydisplay,mygc,mybackground);
  }
}

/* colorfail -- fatal error */
static colorfail(color)
char *color;
{
  fprintf(stderr,"dispimage: can't find color %s in default table\n",color);
  exit(1);
}


/* getfontsize -- get the height and width of an 'M' */
static void getfontsize(height,width)
int * height, * width;
{
  int dirhint;
  XCharStruct charstruct;
  int ascent, descent;

  XTextExtents(fontstruct,"M",1,&dirhint,&ascent,&descent,&charstruct);
  *height = ascent+descent;
  *width = XTextWidth(fontstruct,"M",1);
  return;
}


/* drawbutton -- draw a button */
void drawbutton(buttonindex)
int buttonindex;
{
  int labelx,labely;
  int cheight,cwidth;
 
  /* set up pixmap if it hasn't been done before */
  setuppixmap(button[buttonindex].screenindex);

  getfontsize(&cheight,&cwidth);

  labelx=(button[buttonindex].x1+button[buttonindex].x2-
     strlen(button[buttonindex].label)*cwidth)/2.+.5;
  labely=(button[buttonindex].y1+.5+1.5*cheight);

  setcolor(mybackground);
  XFillRectangle(mydisplay,pixlist[button[buttonindex].screenindex],mygc,
     button[buttonindex].x1,button[buttonindex].y1,button[buttonindex].x2-
     button[buttonindex].x1,button[buttonindex].y2-button[buttonindex].y1);

  setcolor(myforeground);
  XSetLineAttributes(mydisplay,mygc,0,LineSolid,CapButt,JoinMiter);
  XDrawRectangle(mydisplay,pixlist[button[buttonindex].screenindex],mygc,
     button[buttonindex].x1,button[buttonindex].y1,button[buttonindex].x2-
     button[buttonindex].x1-1,button[buttonindex].y2-button[buttonindex].y1-1);
  XDrawImageString(mydisplay,pixlist[button[buttonindex].screenindex],mygc,
    labelx,labely,button[buttonindex].label,strlen(button[buttonindex].label));
}


/* newbutton -- create a new button */
int newbutton(index,x1,y1,width,label)
int index;
int x1,y1;
int width;
char label[];
{
  int i;
  int x2,y2;
  int x[4],y[4];
  int cheight,cwidth;
  
  getfontsize(&cheight,&cwidth);
  x2=x1+width*cwidth+2*cwidth+2;
  y2=y1+2*cheight+5;

  /* find the first empty button slot */
  for(i=1;i<=MAXBUTTONS;i++)
  {
    if(!button[i].active)
      break;
  }
  if(i>MAXBUTTONS)
  {
    fprintf(stderr,"newbutton: Max buttons exceeded\n");
    exit(1);
  }

  button[i].active=1;
  button[i].x1=x1<x2?x1:x2;
  button[i].y1=y1<y2?y1:y2;
  button[i].x2=x1>x2?x1:x2;
  button[i].y2=y1>y2?y1:y2;
  strcpy(button[i].label,label);
  button[i].screenindex=index;

  drawbutton(i);

  return(i);
}


/* buttonhit -- button number at coords; zero if coords are not in a button
 */
int buttonhit(index,x,y)
int index;
int x,y;
{
  int i;
  for(i=1;i<=MAXBUTTONS;i++)
  {
    if(button[i].active && button[i].screenindex==index && button[i].x1<=x &&
       button[i].x2>x && button[i].y1<=y && button[i].y2>y) return(i);
  }
  return(0);
}


/* killbutton -- remove a button from the static list, but don't erase it from
 *               the screen
 */
void killbutton(buttonindex)
int buttonindex;
{
  button[buttonindex].active=0;
}


/* erasebutton -- erase a button from the screen and kill it.
 */
void erasebutton(buttonindex)
int buttonindex;
{
  setcolor(mybackground);
  XFillRectangle(mydisplay,pixlist[button[buttonindex].screenindex],mygc,
     button[buttonindex].x1,button[buttonindex].y1,button[buttonindex].x2-
     button[buttonindex].x1,button[buttonindex].y2-button[buttonindex].y1);
  killbutton(buttonindex);
}


/* dialoguebox -- get the results of a pop-up dialogue box
 */
char * dialoguebox(index,x,y,widthchars,prompt)
int index,x,y,widthchars;
char *prompt;
{
  int height,width;
  int cursorpos=0;
  char input[DIALOGUEBOXMAXINPUT]="";
  int i;
  char text[10];
  KeySym mykey;
  int scroll=0;
  XMappingEvent myxmapevent;
  int cheight = 0;
  int cwidth = 0;
  XRectangle cliprect[1];
  XImage * oldimage;

  /* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  getfontsize(&cheight,&cwidth);
  width=widthchars*cwidth+2*cwidth+1;
  height=4*cheight+5;

  /* save the image behind the dialogue box */
  oldimage=XGetImage(mydisplay,pixlist[index],x,y,width+1,height+1,-1,XYPixmap);
 
  /* clip further drawing to the dialogue box */
  cliprect[0].x=x;
  cliprect[0].y=y;
  cliprect[0].width=width+1;
  cliprect[0].height=height+1;
  XSetClipRectangles(mydisplay,mygc,0,0,cliprect,1,Unsorted);

  /* draw the dialogue box */
  setcolor(mybackground);
  XFillRectangle(mydisplay,pixlist[index],mygc,
     x,y,width,height);
  setcolor(myforeground);
  XDrawImageString(mydisplay,pixlist[index],mygc,x+cwidth+1,y+cheight*1.5,
     prompt,strlen(prompt));
  XDrawRectangle(mydisplay,pixlist[index],mygc,
     x,y,width,height);
  dispimage(index);

/* discard all key press events collected up to this point */
  while (1) 
    if (XCheckTypedEvent(mydisplay,KeyPress,&myevent) == 0) break;

/* wait for a key press event -- redraw on last exposure event */
  while (1) {
    XNextEvent(mydisplay,&myevent);
    if ((myevent.type == Expose) && (myevent.xexpose.count == 0))
    {	
      XSetClipMask(mydisplay,mygc,None);
      dispimage(index);
      XSetClipRectangles(mydisplay,mygc,0,0,cliprect,1,Unsorted);
    }
    else if (myevent.type == MappingNotify)
    {
      XRefreshKeyboardMapping((XMappingEvent * )&myevent);
    }
    else if (myevent.type == KeyPress) 
    {
      i = XLookupString((XKeyEvent *)&myevent,text,10,&mykey,0);

      /* exit on <Return> */
      if(i == 1 && text[0] == 13) break;

      if(i == 1)
      {
        /* Do <Delete> or <Back Space> (same thing) */
        if(text[0] == 127 || text[0] == 8)
        {
           if(cursorpos>0)
           {
             input[--cursorpos] = scroll?'\0':' ';
             if(scroll) scroll--;
             XDrawImageString(mydisplay,pixlist[index],mygc,x+cwidth+1,
                y+cheight*3.5,input+scroll,strlen(input+scroll));
             input[cursorpos]='\0';
             dispimage(index);
           }
        }
        /* Tack on and display other characters */
        else if(cursorpos<DIALOGUEBOXMAXINPUT-1)
        {
          if(cursorpos>=widthchars)
            scroll++;
          input[cursorpos++]=text[0];
          input[cursorpos]='\0';
          XDrawImageString(mydisplay,pixlist[index],mygc,x+cwidth+1,
            y+cheight*3.5,input+scroll,strlen(input+scroll));
          dispimage(index);
        }
      }
    }
  }

  /* Remove the dialogue box, redrawing what's underneath */
  XPutImage(mydisplay,pixlist[index],mygc,oldimage,0,0,x,y,width+1,height+1);
  XSetClipMask(mydisplay,mygc,None);
  dispimage(index);

  return(input);
}


/* setbackground -- set the background to black or white, and the foreground
 *                  to the opposite color. (0 = black, 1 = white).
 */
void setbackground(value)
int value;
{
  switch(value)
  {
    case 0 : 
    {
      mybackground = BlackPixel(mydisplay,myscreen);
      myforeground = WhitePixel(mydisplay,myscreen);
      break;
    }
    case 1 :
    {
      mybackground = WhitePixel(mydisplay,myscreen);
      myforeground = BlackPixel(mydisplay,myscreen);
      break;
    }
  }
}


/* bg -- returns the background pixel value */
long bg()
{
  return(mybackground);
}


/* fg -- returns the foreground pixel value */
long fg()
{
  return(myforeground);
}


/* drawcircle -- draw a circle in the given color; -1 gives inverting mode */
void drawcircle(index,x,y,radius,color)
int index;
int x,y;
int radius;
long color;
{
/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  setcolor(color);

/* do the drawing */
  XDrawArc(mydisplay,pixlist[index],mygc,x-radius,y-radius,
     radius*2,radius*2,0,23040);
}


/* fillcircle -- draw a disk in the given color; -1 gives inverting mode */
void fillcircle(index,x,y,radius,color)
int index;
int x,y;
int radius;
long color;
{
/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  setcolor(color);

/* do the drawing */
  XFillArc(mydisplay,pixlist[index],mygc,x-radius,y-radius,
     radius*2,radius*2,0,23040);
}


/* drawpoint -- draw a point in the given color; -1 gives inverting mode */
void drawpoint(index,x,y,color)
int index;
int x,y;
long color;
{
/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  setcolor(color);

/* do the drawing */
  XDrawPoint(mydisplay,pixlist[index],mygc,x,y);
}


/* fillrectangle -- fill a rectangle with the given color;
 * -1 gives inverting mode
 */
void fillrectangle(index,x1,y1,x2,y2,color)
int index;
int x1,y1,x2,y2;
long color;
{
  if(x1 > x2 || y1 > y2)
  {
    fprintf(stderr,"fillrectangle: x1 and y1 must be upper-left corner\n");
    exit(1);
  }

/* set up pixmap if it hasn't been done before */
  setuppixmap(index);

  setcolor(color);

/* do the drawing */
  XFillRectangle(mydisplay,pixlist[index],mygc,x1,y1,x2-x1+1,y2-y1+1);
}


/* clearscreen -- clear the given screen to the background color */
void clearscreen(index)
int index;
{
  setcolor(mybackground);
  XFillRectangle(mydisplay,pixlist[index],mygc,0,0,nx,ny);
}


/* setcolor -- set the foreground color and/or drawing function if necessary */
void setcolor(color)
long color;
{
  if(color == currentcolor)
    return;
  if(color >= 0 && currentcolor >= 0)
  {
    XSetForeground(mydisplay,mygc,color);
    currentcolor = color;
    return;
  }
  if(color == -1)
  {
    XSetForeground(mydisplay,mygc,myforeground);
    if(depth > 1)
      return;
    XSetFunction(mydisplay,mygc,
       (mybackground == BlackPixel(mydisplay,myscreen)) ? GXequiv : GXxor);
    currentcolor = -1;
    return;
  }
  XSetForeground(mydisplay,mygc,color);
  XSetFunction(mydisplay,mygc,GXcopy);
  currentcolor = color;
}
