/* satircan -- translates binary satellite IR data to CDF format */


#include <stdio.h>
#include "cdfhdr.h"

#define CRITTEMP (-40.0)           /* Temperatures of 47.5 C adjacent to 
                                      temps below this are considered
                                      wrap-arounds */
#define BADTEMP (-80.5)            /* Wrap-around temps are set to this */


char hbuff[HBMAX][LINE];           /* header buffer */
char c1[LINE];                     /* string buffer for a comment line */
float *sbuff,*vbuff;               /* static and variable slice buffers */
long nsbuff,nvbuff;                /* sizes of static and variable slices */
float *lon, *lat, *temp;           /* pointers to starts of data fields */
unsigned char tempindex[444*888];  /* holds the input data */
int i,j;                           /* counting variables */


main(argc,argv)
int argc;
char *argv[];
{
/* die if any arguments are passed */
  if(argc > 1)
  {
    fprintf(stderr,"%s: no arguments allowed\n",argv[0]);
    exit(1);
  }

/* make the new header */
  /* create a blank header */
  nullhdr(hbuff,HBMAX,"float");

  /* add a comment */
  sprintf(c1,"%s\n",argv[0]);
  addcline(hbuff,HBMAX,c1);

  /* add parameters describing index fields lon and lat */
  addpar(hbuff,HBMAX,"lon0","135");
  addpar(hbuff,HBMAX,"dlon","0.045045045045045");
  addpar(hbuff,HBMAX,"lat0","-10");
  addpar(hbuff,HBMAX,"dlat","0.045045045045045");

  /* add field descriptions for lon, lat, and temp */
  addfld(hbuff,HBMAX,'s',"lon",1000.,0.,'s',1,"lon",888);
  addfld(hbuff,HBMAX,'s',"lat",1000.,0.,'s',1,"lat",444);
  addfld(hbuff,HBMAX,'v',"temp",1.,0.,'s',2,"lon",888,"lat",444);

/* set up the data buffers */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbuff);

/* get pointers to lon, lat, and temp fields within the buffers */
  lon = getptr(hbuff,HBMAX,sbuff,'s','d',"lon");
  lat = getptr(hbuff,HBMAX,sbuff,'s','d',"lat");
  temp = getptr(hbuff,HBMAX,vbuff,'v','d',"temp");

/* fill the index fields with the correct values */
  for(i = 0; i < 444; i++)
    lat[i] = -10. + i * 20./444.;
  for(i = 0; i < 888; i++)
    lon[i] = 135. + i * 40./888.;

/* read in the binary input file; accept only correct file length */
  if(fread(tempindex,1,444*888,stdin) < 444*888)
  {
    fprintf(stderr,
        "%s: unexpected end of file or unable to read from standard input\n",
        argv[0]);
    exit(1);
  }
  if(fread(c1,1,1,stdin) > 0)
  {
    fprintf(stderr,"%s: input file too large\n",argv[0]);
    exit(1);
  }

/* translate byte values into temperatures in correct output buffer locations */
  for(i = 0; i < 888; i++)
    for(j = 0; j < 444; j++)
      temp[i*444+j] = (float)(tempindex[(443-j)*888+i]) / 2. - 80.;

/* Fix bad data: move along each column, then each row, making temperatures
   of 47.5 which are adjacent to temperatures below CRITTEMP into BADTEMP */
  for(i = 0; i < 888; i++)
  {
    for(j = 1; j < 444; j++)
      if(temp[i*444+j] == 47.5 && temp[i*444+(j-1)] < CRITTEMP)
        temp[i*444+j] = BADTEMP;
    for(j = 442; j >= 0; j--)
      if(temp[i*444+j] == 47.5 && temp[i*444+(j+1)] < CRITTEMP)
        temp[i*444+j] = BADTEMP;
  }
  for(j = 0; j < 444; j++)
  {
    for(i = 1; i < 888; i++)
      if(temp[i*444+j] == 47.5 && temp[(i-1)*444+j] < CRITTEMP)
        temp[i*444+j] = BADTEMP;
    for(i = 886; i >= 0; i--)
      if(temp[i*444+j] == 47.5 && temp[((i+1)*444+j)] < CRITTEMP)
        temp[i*444+j] = BADTEMP;
  }

/* write the output file */
  puthdr(stdout,hbuff,HBMAX);
  putslice(stdout,nsbuff,sbuff);
  putslice(stdout,nvbuff,vbuff);
}
