/* getblock -- Get a block of parameters from an Eldora or Dorade file.
 * Getblock is handed an open file pointer, a maximum block size, and
 * the block buffer.  It returns the block size if everything is ok,
 * -1 on error, and 0 on end of file.
 */
#include <stdio.h>
#include <math.h>

/* buffer conversion routines */
short getshort();
long getlong();
float getfloat();

long getblock(fp,max,buff)
FILE *fp;                          /* stream from which to read */
long max;                          /* maximum size of block accepted */
char *buff;                        /* buffer to hold the block --
                                      assume this is aligned on 4 byte
                                      boundary so that casting to long
                                      and float works */
{
  long asize,bsize;
  int i;

/* see if max is reasonable */
  if (max < 8) {
    fprintf(stderr,"getblock: max = %d is too small!\n",max);
    return(-1);
  }

/* get the size of the block */
  if ((bsize = fread(buff,sizeof(char),8,fp)) != 8) {

/* likely end of file */
    if (bsize == 0) return(0);
    else {
      fprintf(stderr,"getblock: initial probe yielded %d != 8 bytes\n",bsize);
      return(-1);
    }
  }
  bsize = getlong(4,buff);

/* check for reasonableness in block size */
  if (bsize < 0 || bsize > max) {
    fprintf(stderr,"getblock: block size = %d is unreasonable\n",bsize);
    return(-1);
  }

/* if expected size is 0, return a zero -- this is an error, but not fatal */
  if (bsize == 0) {
    fprintf(stderr,"getblock: warning: expected block size = 0\n");
    return(0);
  }
    
/* get the rest of the block */
  if ((asize = fread(buff + 8,sizeof(char),bsize - 8,fp) + 8) != bsize) {
    fprintf(stderr,"getblock: actual block size %d != expected size %d\n",
         asize,bsize);
    return(-1);
  }

/* everything is ok -- return size of block */
  return(bsize);
}

/* getstring -- extract a string from buffer and null terminate it */
getstring(pos,len,buff,out)
long pos,len;                  /* position and length of string */
char *buff;                    /* input buffer */
char *out;                     /* null-terminated output */
{
  long i;
  for (i = 0; i < len; i++) out[i] = buff[pos + i];
  out[len] = '\0';
}

/* extract_vold -- get useful information from VOLD block */
extract_vold(buff,proj,vol,year,month,day,hour,minute,second,source)
char *buff;                     /* the data block */
char *proj;                     /* project name -- 20 bytes */
short *vol;                     /* the volume number */
short *year;                    /* year data taken */
short *month;                   /* month data taken */
short *day;                     /* day data taken */
short *hour;                    /* hour data taken */
short *minute;                  /* minute data taken */
short *second;                  /* second data taken */
char *source;                   /* source of data -- 8 bytes */
{
  getstring(16,20,buff,proj);
  *vol = getshort(10,buff);
  *year = getshort(36,buff);
  *month = getshort(38,buff);
  *day = getshort(40,buff);
  *hour = getshort(42,buff);
  *minute = getshort(44,buff);
  *second = getshort(46,buff);
  getstring(56,8,buff,source);
}

/* extract_radd -- get useful information from RADD block */
extract_radd(buff,rad_name,rad_const,peak_power,sys_gain,rad_type,scan_mode,
     unamb_vel,unamb_range)
char *buff;                     /* the data block */
char *rad_name;                 /* name of radar -- 8 bytes */
float *rad_const;               /* the radar constant */
float *peak_power;              /* peak power in KW */
float *sys_gain;                /* system gain in db */
short *rad_type;                /* radar type -- 0 - ground; 1 - airborne fore;
                                   2 - airborne aft; 3 - airborne tail;
                                   4 - airborne lower fuselage;
                                   5 - shipborne */
short *scan_mode;               /* scan mode -- 0 - calibration;
                                   1 - PPI; 2 - co-plane; 3 - RHI;
                                   4 - vertical pointing; 5 - target;
                                   6 - manual; 7 - idle;
                                   8 - full sweep; 9 - partial sweep */
float *unamb_vel;               /* unambiguous velocity (m/s) */
float *unamb_range;             /* unambiguous range (km) */
{
  *rad_const = getfloat(16,buff);
  *peak_power = getfloat(20,buff);
  *sys_gain = getfloat(36,buff);
  *rad_type = getshort(48,buff);
  *scan_mode = getshort(50,buff);
  *unamb_vel = getfloat(92,buff);
  *unamb_range = getfloat(96,buff);
}

/* extract_cspd -- get useful information from CSPD block */
extract_cspd(buff,num_segs,dist_to_first,spacing,num_cells)
char *buff;                     /* the buffer */
short *num_segs;                /* number of segments with equal cell spacing */
short *dist_to_first;           /* distance to first segment (m) */
short spacing[6];               /* cell spacing each segment (m) */
short num_cells[6];             /* number of cells each segment */
{
  int i;
  *num_segs = getshort(8,buff);
  *dist_to_first = getshort(10,buff);
  for (i = 0; i < 6; i++) {
    spacing[i] = getshort(12 + 2*i,buff);
    num_cells[i] = getshort(24 + 2*i,buff);
  }
}

/* extract_parm -- get useful information from PARM block */
extract_parm(buff,p_name,p_desc,p_units,samples,bin_form,thresh_fld,
     thresh_val,p_scale,p_bias,bad_data)
char *buff;                     /* the buffer */
char *p_name;                   /* parameter name -- 8 bytes */
char *p_desc;                   /* parameter description -- 40 bytes */
char *p_units;                  /* parameter units -- 8 bytes */
short *samples;                 /* number of samples */
short *bin_form;                /* binary format -- 1 - char; 2 - short;
                                   3 - long; 4 - IEEE float;
                                   5 - 16 bit float (!) */
char *thresh_fld;               /* threshold field -- 8 bytes */
float *thresh_val;              /* threshold value */
float *p_scale,*p_bias;         /* parameter scale and bias -- physical
                                   value = (tape value - p_bias)/p_scale */
long *bad_data;                 /* bad data flag */
{
  getstring(8,8,buff,p_name);
  getstring(16,40,buff,p_desc);
  getstring(56,8,buff,p_units);
  *samples = getshort(76,buff);
  *bin_form = getshort(78,buff);
  getstring(80,8,buff,thresh_fld);
  *thresh_val = getfloat(88,buff);
  *p_scale = getfloat(92,buff);
  *p_bias = getfloat(96,buff);
  *bad_data = getlong(100,buff);
}

/* extract_ryib -- extract useful information from RYIB block */
extract_ryib(buff,sweep,j_day,hour,min,sec,msec,az,el,peak_power,ray_stat)
char *buff;                    /* the bloody buffer */
long *sweep;                   /* sweep number */
long *j_day;                   /* Julian day */
short *hour,*min,*sec,*msec;   /* guess! */
float *az,*el;                 /* azimuth and elevation (degrees) */
float *peak_power;             /* peak power (KW) */
long *ray_stat;                /* ray status -- 0 - normal; 1 - transition;
                                  2 - bad; */
{
  *sweep = getlong(8,buff);
  *j_day = getlong(12,buff);
  *hour = getshort(16,buff);
  *min = getshort(18,buff);
  *sec = getshort(20,buff);
  *msec = getshort(22,buff);
  *az = getfloat(24,buff);
  *el = getfloat(28,buff);
  *peak_power = getfloat(32,buff);
  *ray_stat = getlong(40,buff);
}

/* extract_asib -- extract useful information from ASIB block */
extract_asib(buff,lon,lat,z_msl,z_agl,u_aircraft,v_aircraft,w_aircraft,
     heading,roll,pitch,drift_angle,rot_angle,tilt,u,v,w,heading_rate,
     pitch_rate)
char *buff;                   /* the buffer */
float *lon,*lat;              /* longitude and latitude (degrees) */
float *z_msl,*z_agl;          /* height above sea and surface (m) */
float *u_aircraft,*v_aircraft,*w_aircraft;  /* aircraft velocity (m/s) */
float *heading,*roll,*pitch;  /* heading, roll, pitch (degrees) */
float *drift_angle;           /* angle between aircraft velocity and
                                 heading, right drift positive (degrees) */
float *rot_angle;             /* rotation angle of antenna, up is zero,
                                 positive is clockwise (degrees) */
float *tilt;                  /* forward or aft angle of radar, forward
                                 positive (degrees) */
float *u,*v,*w;               /* wind velocity at aircraft (m/s) */
float *heading_rate;          /* time rate of change of heading (degrees/s) */
float *pitch_rate;            /* time rate of change of pitch (degrees/s) */
{
  *lon = getfloat(8,buff);
  *lat = getfloat(12,buff);
  *z_msl = getfloat(16,buff);
  *z_agl = getfloat(20,buff);
  *u_aircraft = getfloat(24,buff);
  *v_aircraft = getfloat(28,buff);
  *w_aircraft = getfloat(32,buff);
  *heading = getfloat(36,buff);
  *roll = getfloat(40,buff);
  *pitch = getfloat(44,buff);
  *drift_angle = getfloat(48,buff);
  *rot_angle = getfloat(52,buff);
  *tilt = getfloat(56,buff);
  *u = getfloat(60,buff);
  *v = getfloat(64,buff);
  *w = getfloat(68,buff);
  *heading_rate = getfloat(72,buff);
  *pitch_rate = getfloat(76,buff);
}

/* extract_frad -- this one actually extracts data! returns 0 if ok */
int extract_frad(buff,bsize,word_len,total_cells,
     nparams,first_gate,last_gate,pointers,stepsize)
char *buff;                    /* da buffa */
long bsize;                    /* size of the buffer in bytes -- input */
short word_len[];              /* length of binary data word in bytes for
                                  each parameter */
short total_cells;             /* total number of cells or range gates
                                   -- input*/
short nparams;                 /* number of parameters -- input */
short *first_gate,*last_gate;  /* first and last gate recorded */
long pointers[];               /* indices in buff indicating starting point of
                                  each parameter */
long *stepsize;                /* bytes between successive values of a var */
{
  long esize,datastart;
  short i,offset;
  *first_gate = getshort(48,buff);
  *last_gate = getshort(50,buff);
  datastart = 52;

/* compute expected size of data field and pointers */
  esize = 0;
  offset = 0;
  for (i = 0; i < nparams; i++) {
    pointers[i] = datastart + offset;
    esize += word_len[i]*total_cells;
    offset += word_len[i];
  }
  *stepsize = offset;
  if (bsize - datastart != esize) {
    fprintf(stderr,"extract_frad: expected size of data area %d != actual %d\n",
         esize,bsize - datastart);
    exit(1);
  }
}

/* translate out of buff format -- should test for byte reversal, etc. */
short getshort(pos,buff)
long pos;
char *buff;
{
  return(*((short*)((buff) + (pos))));
}

long getlong(pos,buff)
long pos;
char *buff;
{
  return(*((long*)((buff) + (pos))));
}

float getfloat(pos,buff)
long pos;
char *buff;
{
  return(*((float*)((buff) + (pos))));
}
