/* getrp7 -- This subroutine reads a record (1 ray) of NCAR RP-7 data from
 * the standard input and puts the header into a structure.  A pointer to
 * a data buffer is returned.  A return of NULL indicates an
 * end of file.  If the expected number of bytes is set to 0, an rtape-type
 * record header is expected, i. e., an 8 byte ascii decimal integer.
 * If it is a positive number, that number of bytes is assumed to be
 * the size of the record, including the
 * header.  If the input parameter swap is 1, bytes
 * are swapped in the incoming header, if 0, not.
 */
#include <stdio.h>
#include <math.h>
#include "rp7.h"

unsigned char *getrp7(infile,data_flag,bytes_expected,byte_swap,header,
			   count)
FILE *infile;                      /* input file pointer, assumed open */
char data_flag;                    /* 'y' if data to be read */
int bytes_expected;                /* zero forces a look for rtape header */
int byte_swap;                     /* swap header bytes if set */
struct rp7 *header;                /* output header for record */
long count;                        /* number of requersed count */
{
  extern long lenfbeg,recordlen;
  int loop,i;                      /* looping index */
  int bytes_received;              /* bytes of data received */
  char rtheader[9];                /* buffer for rtape header */
  static unsigned char buffer[MAXDATA];  /* buffer to store data */
  static unsigned short hbuffer[100];  /* buffer to store header */
  extern int count1;                 /* number of first block */
  char ortheader[9];                 /* old block size string     */
/* check for expectation of rtape header */
  if (bytes_expected == 0 ) {
    for (loop = 0; loop < 8; loop++) 
	  if((rtheader[loop] = getc(infile)) == EOF ) {
		fprintf(stderr,"getrp7: end of file encountered after");
		fprintf(stderr," count number %d\n",count);
		exit(1);
      }
    rtheader[8] = '\0';
    bytes_expected = atoi(rtheader);
    if (bytes_expected < HSIZE || bytes_expected > 100000) {
	  if(count == count1) return((unsigned char *)NULL) ;
      /* try to locate the glitch exactly */
      /* loop downward from count to count-CNTSKIP then the previous block *
       * should contain the glitch                                         */
       for(i= count; i > count-CNTSKIP ; i--){
         /* try to go back and pick up the recordlength */
         fseek(infile,lenfbeg-recordlen,0);
         lenfbeg -= recordlen;
         /* copy current heraer size to old block position */
         strcpy(ortheader,rtheader);

         for (loop = 0; loop < 8; loop++)
            if((rtheader[loop] = getc(infile)) == EOF ) {
                  fprintf(stderr,"getrp7: end of file encountered after");
                  fprintf(stderr," count number %d\n",count);
                  exit(1);
            }

          if(atoi(rtheader)+8 == recordlen ) {
            /* found the first clean block */
            /* print out the number of the previous block */ 
            fprintf(stderr,"getrp7: glitch encountered in block %d",i);
            fprintf(stderr,"  unable to continue proscessing \n");
            fprintf(stderr,"getrp7: apparent new block size is %s\n",ortheader);
            exit(1);
          }
        }
      }
    }

/* get HSIZE word header */
  if (fread(hbuffer,sizeof(short),HSIZE,infile) != HSIZE) {
    return((unsigned char*)NULL);          /* normal return at eof */
  }

/* swap bytes if needed (not currently implemented) */
  if (byte_swap) {
    fprintf(stderr,"getrp7: byte swapping not yet implemented\n");
    exit(1);
  }

/* test for correct size of housekeeping record */
  if (hbuffer[64] != HSIZE) {
    fprintf(stderr,"getrp7: housekeeping size %d != %d\n",hbuffer[64],HSIZE);
    exit(1);
  }

/* deduct header size from bytes expected -- it is now data size */
  bytes_expected -= 2*HSIZE;

/* transfer stuff from header to header structure */
  header->log_rec_num = hbuffer[0];
  header->tape_seq_num = hbuffer[1];
  header->record_type = hbuffer[2];
  header->year = hbuffer[3];
  header->month = hbuffer[4];
  header->day = hbuffer[5];
  header->hour = hbuffer[6];
  header->minute = hbuffer[7];
  header->second = hbuffer[8];
  header->azimuth = hbuffer[9]/CF;
  header->elevation = hbuffer[10]/CF;
  header->rhozero = hbuffer[11] + hbuffer[12]/1000.;
  header->delrho = hbuffer[13];
  header->num_gates = hbuffer[14];
  header->num_samples = hbuffer[15];
  header->avgtrpow = hbuffer[17]/10;
  header->pulse_width = hbuffer[18];
  header->prf = hbuffer[19]/10.;
  header->wavelength = hbuffer[20]/100.;
  header->sweep_num = hbuffer[21];
  header->scan_mode = hbuffer[25];
  header->clock_az_lim = hbuffer[26]/CF;
  header->counter_az_lim = hbuffer[27]/CF;
  header->up_el_lim = hbuffer[28]/CF;
  header->low_el_lim = hbuffer[29]/CF;
  header->fixed_angle = hbuffer[30]/CF;
  header->prim_cop_b_angle = hbuffer[35]/CF;
  header->sec_cop_b_angle = hbuffer[36]/CF;
  header->scan_rate = hbuffer[44]/CF;
  header->volume_count = hbuffer[47];
  header->clutter_filter = hbuffer[48];
  header->polarization = hbuffer[49];
  header->mod_rec_count = hbuffer[56];
  header->altitude = hbuffer[57];
  header->longitude = hbuffer[59]/364.08889;
  header->latitude = hbuffer[58]/728.17778;
  header->transition_flag = hbuffer[60];
  header->radar_id = hbuffer[62];
  header->nfields = hbuffer[67];


/* check number of fields */
  if (header->nfields < 1 || header->nfields > NFMAX) {
    fprintf(stderr,"getrp7: %d is an unreasonable number of fields\n",
         header->nfields);
    exit(1);
  }

/* HaRP preset field names (this is temporary) */
  strcpy(header->fnames[0],"dm");
  strcpy(header->fnames[1],"ncp");
  strcpy(header->fnames[2],"vr");
  strcpy(header->fnames[3],"specwid");
  strcpy(header->fnames[4],"dz");
  strcpy(header->fnames[5],"sn");

/* parameter descriptors and scaling factors */
  for (loop = 0; loop < header->nfields; loop++) {
    header->fparams[loop] = hbuffer[68 + loop];
    header->scale[loop] = hbuffer[78 + 2*loop];
    header->bias[loop] = (short)hbuffer[79 + 2*loop];
  }

/* check consistency of data length */
  if (header->num_gates*header->nfields != bytes_expected) {
    fprintf(stderr,"getrp7: num_gates*nfields (%d) != bytes expected (%d)\n",
         header->num_gates*header->nfields,bytes_expected);
    exit(1);
  }
  if (bytes_expected > MAXDATA) {
    fprintf(stderr,"getrp7: expected data size %d exceeds buffer size\n",
         bytes_expected);
    exit(1);
  }

/* read data block */
  if (data_flag == 'y' &&
       (bytes_received = fread(buffer,sizeof(char),bytes_expected,infile))
       != bytes_expected) {
    fprintf(stderr,"getrp7: unexpected eof on data read -- %d\n",
         bytes_received);
    exit(1);
  }

/* return pointer to data block */
  return(buffer);
}
