/* Radcart2 is the second try at a ray-to-cartesian format
 * radar data converter.  This one takes the alternate approach of
 * reading rays one at a time and assigning weighted values to
 * surrounding grid points.  First order corrections for the curvature
 * of the earth are made.  */

/* 2/91, S. Barman: radcart2 will now allow the input of x,y componants
 * of the storm velocity in m/s (or km/ksec). */

/* 4/91, S. Barman: raw data calls for minute while universal format
 * data has min, hence I had to add the names for some of the time
 * variables.  The program worked fine for dx = 1, but as soon as we
 * used dx = .5 the data was limited to nx grid spaces, i.e. half the
 * range.  So now the command line uses xkm (ykm, etc.) which is the
 * range in the x (y,etc.) direction in km.  nx is then calculated from
 * nx = xkm/dx +1.  The +1 is to force a round up so as not to loose any
 * data. */

/* 2/92 S. Lewis: To average the log reflectivity is in error.  
 * Even though the reflectivity is saved as log reflectivity it 
 * should be the reflectivity that is averaged not the log 
 * reflectivity that should be averaged.  The program now expects 
 * the name of the reflectivity and will calculate 10^^(dz/10), 
 * do the averageing, then take 10log10 (of the averaged 
 * reflectivity). */

/* 11/93 D. Raymond:  Did some cleanup and changed the names of the
 * x,y, and z radar position variables to xradar, yradar, and zradar
 * for consistency with radcart and radsynth.
 */

/* 12/93 D. Raymond:  More cleanup, all incoming names now assumed.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define DEARTH 12740.
#define MAXFIELDS 25
#define DTOR 57.2958
#define TRUNC(a0,da,aloc,aa) ((int)(aa = (aloc - a0)/da + 1000.) - 1000)
#define FRACT(ia,aa,ba) (aa -= ia + 1000, ba = 1. - aa)

/* names specifying location of location of radar */
#define XRAD "xradar"
#define YRAD "yradar"
#define ZRAD "zradar"

/* storm speed variable names */
#define SVX "svx"
#define SVY "svy"

/* azimuth and elevation names */
#define AZIM "azim"
#define ELEV "elev"

/* raycount and range index field names */
#define RAYCNT "raycnt"
#define RANGE "range"

/* time names */
#define HOUR "hour"
#define MIN "min"
#define SEC "sec"

/* reflectivity names */
#define DZ "dz"
#define DBZ "dbz"

/* candis stuff */
char hbin[HBMAX][LINE];
char hbout[HBMAX][LINE];
char cline[LINE],pline[LINE];
float bad,badlim;
float *sbin,*vbin,*sbout,*vbout;
long nsbin,nvbin,nsbout,nvbout;
float *count,*range,*azim,*elev;
float *hour,*minute,*second;
float *x,*y,*z;
float *fields[MAXFIELDS];       /* input */
float *ofields[MAXFIELDS];      /* output */
float *icount[MAXFIELDS];       /* number of good instances */
int nfields;
int nrange;
int dbzflag;
float *az,*el;
float x0,yy0,z0;
float dx,dy,dz;
float xkm,ykm,zkm;
int nx,ny,nz;
long num;
float xl,yl,zl,lat0,tanlat;
float *xla,*yla,*zla;
float xloc,yloc,zloc;
float delx,dely,delz;
float weight,cx,cy,cz,bx,by,bz;
float svx,svy;
float *psvx,*psvy;
float starttime,timenow,deltatime;
float reftime;

int inane;
int cnt;

main(argc,argv)
int argc;
char **argv;
{
  int poop,loop;
  int ix,iy,iz,jx,jy,jz;
  long iloc;
  struct field *fp;
  float svalue,*sptr;
  float zfield;

  /* check arguments */
  if (argc != 14 && argc != 16 && argc != 17) {
    fprintf(stderr,"Usage:");
    fprintf(stderr," radcart2 x0 dx xkm");
    fprintf(stderr," y0 dy ykm z0 dz zkm xr yr zr lat");
    fprintf(stderr," [svx svy [reftime]]\n");
    exit(1);
  }
  x0 = atof(argv[1]);
  dx = atof(argv[2]);
  xkm = atof(argv[3]);
  yy0 = atof(argv[4]);
  dy = atof(argv[5]);
  ykm = atof(argv[6]);
  z0 = atof(argv[7]);
  dz = atof(argv[8]);
  zkm = atof(argv[9]);
  xl = atof(argv[10]);
  yl = atof(argv[11]);
  zl = atof(argv[12]);
  lat0 = atof(argv[13]);
  tanlat = tan(lat0/DTOR);

  if (argc >= 16) {
    svx = atof(argv[14]);
    svy = atof(argv[15]);
    if (argc == 17) {
      reftime = atof(argv[16]);
    }
    else {
      reftime = -101.;
    }
  }
  else {
    svx = svy = 0.;
  }

  /* the +1 is to compensate for truncation */
  nx = xkm/dx + 1.5;
  ny = ykm/dy + 1.5;
  nz = zkm/dz + 1.5;

  /* get header of input file */
  gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"radcart2: float format expected\n");
    exit(1);
  }

  /* make header of output file */
  nullhdr(hbout,HBMAX,"float");
  copycmt(hbout,HBMAX,hbin,HBMAX);

  sprintf(cline,"radcart2:\n %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbout,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s %s %s\n",argv[5],argv[6],argv[7],argv[8],
      argv[9],argv[10]);
  addcline(hbout,HBMAX,cline);
  if (argc == 14) {
    sprintf(cline,"  %s %s %s\n",argv[11],argv[12],argv[13]);
  }
  else {
    sprintf(cline,"  %s %s %s %s %s\n",argv[11],argv[12],argv[13],
        argv[14],argv[15]);
  }
  addcline(hbout,HBMAX,cline);

  /* copy parameters */
  copypar(hbout,HBMAX,hbin,HBMAX);

  /* bad data values */
  if (seekpar(hbin,HBMAX,"bad",pline) == OK) bad = atof(pline);
  else {
    bad = BAD;
    sprintf(pline,"%g",bad);
    addpar(hbout,HBMAX,"bad",pline);
  }
  if (seekpar(hbin,HBMAX,"badlim",pline) == OK) badlim = atof(pline);
  else {
    badlim = BADLIM;
    sprintf(pline,"%g",badlim);
    addpar(hbout,HBMAX,"badlim",pline);
  }

  /* index parameters */
  sprintf(pline,"%g",x0);
  addpar(hbout,HBMAX,"x0",pline);
  sprintf(pline,"%g",dx);
  addpar(hbout,HBMAX,"dx",pline);
  sprintf(pline,"%g",yy0);
  addpar(hbout,HBMAX,"y0",pline);
  sprintf(pline,"%g",dy);
  addpar(hbout,HBMAX,"dy",pline);
  sprintf(pline,"%g",z0);
  addpar(hbout,HBMAX,"z0",pline);
  sprintf(pline,"%g",dz);
  addpar(hbout,HBMAX,"dz",pline);

  /* add index fields */
  addfld(hbout,HBMAX,'s',"x",1000.,0.,'l',1,"x",nx);
  addfld(hbout,HBMAX,'s',"y",1000.,0.,'l',1,"y",ny);
  addfld(hbout,HBMAX,'s',"z",1000.,0.,'l',1,"z",nz);

  /* add variable fields */
  loop = 0;
  nfields = 0;
  dbzflag = -1;
  while ((fp = getfld(hbin,HBMAX,'v',loop++)) != NULL) {

    /* zero dimensional fields are just transferred */
    if (fp->dim == 0) {
      addfld(hbout,HBMAX,'v',fp->fname,fp->smul,fp->sadd,fp->bytes,0);
    }

    /* one dimensional fields with range dimension are cartesianized */
    else if (fp->dim == 1 && strcmp(fp->dname1,RANGE) == 0) {
      addfld(hbout,HBMAX,'v',fp->fname,fp->smul,fp->sadd,'c',3,
          "x",nx,"y",ny,"z",nz);

      /* find reflectivity field */
      if(strcmp(fp->fname,DZ) == 0 ||
          strcmp(fp->fname,DBZ) == 0 ) {
        dbzflag = nfields;
      }
      nfields++;
    }

    /* other fields are dropped */
  }

  /* add x, y, and z location of radar */
  addfld(hbout,HBMAX,'v',XRAD,100.,0.,'s',0);
  addfld(hbout,HBMAX,'v',YRAD,100.,0.,'s',0);
  addfld(hbout,HBMAX,'v',ZRAD,1000.,0.,'s',0);

  /* add storm speed */
  addfld(hbout,HBMAX,'v',SVX,100.,0.,'s',0);
  addfld(hbout,HBMAX,'v',SVY,100.,0.,'s',0);

  /* allocate buffers */
  nsbin = elemcnt(hbin,HBMAX,'s');
  nsbout = elemcnt(hbout,HBMAX,'s');
  nvbin = elemcnt(hbin,HBMAX,'v');
  nvbout = elemcnt(hbout,HBMAX,'v');
  sbin = getbuff(nsbin);
  sbout = getbuff(nsbout);
  vbin = getbuff(nvbin);
  vbout = getbuff(nvbout);

  /* get pointers for index fields */
  range = getptr2(hbin,HBMAX,sbin,'s','d',RANGE,&fp);
  if (fp->dim != 1 || strcmp(fp->fname,fp->dname1) != 0) {
    fprintf(stderr,"radcart2: %s is not an index field\n",fp->fname);
    exit(1);
  }
  nrange = fp->dsize1;
  x = getptr(hbout,HBMAX,sbout,'s','d',"x");
  y = getptr(hbout,HBMAX,sbout,'s','d',"y");
  z = getptr(hbout,HBMAX,sbout,'s','d',"z");

  /* fill in index fields */
  for (loop = 0; loop < nx; loop++) x[loop] = x0 + loop*dx;
  for (loop = 0; loop < ny; loop++) y[loop] = yy0 + loop*dy;
  for (loop = 0; loop < nz; loop++) z[loop] = z0 + loop*dz;

  /* get pointers for fields to be cartesianized */
  loop = 0;
  nfields = 0;
  while ((fp = getfld(hbin,HBMAX,'v',loop++)) != NULL) {
    if (fp->dim == 1 && strcmp(fp->dname1,RANGE) == 0) {
      fields[nfields] = getptr(hbin,HBMAX,vbin,'v','d',fp->fname);
      ofields[nfields++] = getptr2(hbout,HBMAX,vbout,'v','d',fp->fname,&fp);
    }
  }

  /* get pointers for azimuth and elevation */
  az = getptr(hbin,HBMAX,vbin,'v','d',AZIM);
  el = getptr(hbin,HBMAX,vbin,'v','d',ELEV);

  /* get pointers for time */
  hour = getptr(hbin,HBMAX,vbin,'v','d',HOUR);
  minute = getptr(hbin,HBMAX,vbin,'v','d',MIN);
  second = getptr(hbin,HBMAX,vbin,'v','d',SEC);

  /* get pointers for radar location */
  xla = getptr(hbout,HBMAX,vbout,'v','d',XRAD);
  yla = getptr(hbout,HBMAX,vbout,'v','d',YRAD);
  zla = getptr(hbout,HBMAX,vbout,'v','d',ZRAD);

  /* assign specified location values */
  *xla = xl;
  *yla = yl;
  *zla = zl;

  /* get pointers for storm speed */
  psvx = getptr(hbout,HBMAX,vbout,'v','d',SVX);
  psvy = getptr(hbout,HBMAX,vbout,'v','d',SVY);

  /* assign values */
  *psvx = svx;
  *psvy = svy;

  /* get input static slice */
  getslice(stdin,nsbin,sbin);

  /* write output header */
  puthdr(stdout,hbout,HBMAX);

  /* write output static slice */
  putslice(stdout,nsbout,sbout);

  /* initialize output fields */
  num = nx*ny*nz;
  for (poop = 0; poop < nfields; poop++) {
    icount[poop] = getbuff(num);
    for (loop = 0; loop < num; loop++) {
      ofields[poop][loop] = 0.;
      icount[poop][loop] = 0.;
    }
  }

  starttime = 0;
  deltatime = 0;
  cnt = 0;

  /* loop through input variable slices */
  while (getslice(stdin,nvbin,vbin) != EOF) {
    ++cnt;

    /* loop on range */
    for (loop = 0; loop < nrange; loop++) {

      /* get locations of range point relative to radar */
      delx = range[loop]*sin(*az/DTOR)*cos(*el/DTOR);
      dely = range[loop]*cos(*az/DTOR)*cos(*el/DTOR);
      delz = range[loop]*sin(*el/DTOR);

      /* do corrections for earth's curvature */
      dely -= delx*delx*tanlat/DEARTH;
      delz += (delx*delx + dely*dely)/DEARTH;

      /* get location relative to coordinate system origin */
      xloc = xl + delx;
      yloc = yl + dely;
      zloc = zl + delz;

      /* convert time to ksec */
      timenow = ((*hour*3600) + (*minute*60) + *second)/1000;

      /* on the first slice only is the starttime set */
      if(cnt == 1 && loop == 0)
      {

        /* if reftime is set, use as starttime -- otherwise use current time */
        if (reftime >= -100.) {
          starttime = reftime;
        }
        else {
          starttime = timenow;
        }
      }
      deltatime = timenow - starttime;

      /* translate all points to their location at the common time
           of starttime */
      xloc = xloc - (svx*deltatime);
      yloc = yloc - (svy*deltatime);


      /* find grid location */
      ix = TRUNC(x0,dx,xloc,cx);
      iy = TRUNC(yy0,dy,yloc,cy);
      iz = TRUNC(z0,dz,zloc,cz);
      FRACT(ix,cx,bx);
      FRACT(iy,cy,by);
      FRACT(iz,cz,bz);
      jx = ix + 1;
      jy = iy + 1;
      jz = iz + 1;

      /* increment fields at each corner of surrounding cube */
      if (ix >= 0 && ix < nx - 1 && iy >= 0 && iy < ny - 1 && iz >= 0 &&
          iz < nz - 1) {
        iloc = iz + nz*iy + nz*ny*ix;
        weight = bx*by*bz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = iz + nz*iy + nz*ny*jx;
        weight = cx*by*bz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = iz + nz*jy + nz*ny*ix;
        weight = bx*cy*bz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = jz + nz*iy + nz*ny*ix;
        weight = bx*by*cz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = iz + nz*jy + nz*ny*jx;
        weight = cx*cy*bz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = jz + nz*iy + nz*ny*jx;
        weight = cx*by*cz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = jz + nz*jy + nz*ny*ix;
        weight = bx*cy*cz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
        iloc = jz + nz*jy + nz*ny*jx;
        weight = cx*cy*cz;
        for (poop = 0; poop < nfields; poop++) {
          if (fabs(fields[poop][loop]) < badlim) {

            /* special treatment for reflectivity */
            if(poop == dbzflag )
            {
              zfield = pow(10.0,fields[poop][loop]/10.0);
              ofields[poop][iloc] += weight*zfield;
              icount[poop][iloc] += weight;
            }
            else
            {
              ofields[poop][iloc] += weight*fields[poop][loop];
              icount[poop][iloc] += weight;
            }
          }
        }
      }
    }
  }

  /* normalize output fields */
  for (loop = 0; loop < num; loop++) {
    for (poop = 0; poop < nfields; poop++) {
      if (icount[poop][loop] > 0.) {
        ofields[poop][loop] /= icount[poop][loop];

        /* special treatment for reflectivity */
        if (poop == dbzflag)
        {
          inane = 0;
          ofields[poop][loop] = 10*log10(ofields[poop][loop]);
        }
      }
      else {
        ofields[poop][loop] = bad;
      }
    }
  }

  /* transfer scalar fields from last input slice */
  loop = 0;
  while ((fp = getfld(hbin,HBMAX,'v',loop++)) != NULL) {
    if (fp->dim == 0) {
      svalue = vbin[fp->ptr];
      sptr = getptr(hbout,HBMAX,vbout,'v','d',fp->fname);
      *sptr = svalue;
    }
  }

  /* write variable slice */
  putslice(stdout,nvbout,vbout);
  exit(0);
}
