/* Radcart4 -- interpolate radar data to a cartesian grid. Each data point
 * is assumed to affect all eight grid points that surround it. All data
 * points that affect a particular grid point are weighted according
 * to the grid spacing minus the distance between the grid point and
 * the data point. The input file to radcart4 is produced by hrdrad
 * and the output consists of the following variables:
 * vx,vy,vz - Triple doppler particle velocities.
 * seu,sev,sewp - Standard errors in triple doppler particle velocities.
 * ux0,uy0 - Dual doppler velocities assuming zero vertical velocity.
 * gx,gy - from ux=ux0-gx*vz and uy=uy0-gy*vz, in dual analysis.
 * deu,dev - errors in dual doppler velocities assuming zero error in vz.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define DEARTH 12740.
#define MAXFIELDS 25
#define DTOR 57.2958
#define TRUNC(a0,da,aloc,aa) ((int)(aa = (aloc - a0)/da + 1000.) - 1000)
#define FRACT(ia,aa,ba) (aa -= ia + 1000, ba = 1. - aa)

/* assumed names of radial velocity and reflectivity fields */
#define VR "vr"
#define DBZ "dbz"

/* other assumed names defining positioning (meaning self evident) */
#define LON "lon"
#define LAT "lat"
#define ALT "alt"
#define TIME "time"
#define AZ "az"
#define EL "el"

/* candis stuff */
char hbin[HBMAX][LINE];
char hbout[HBMAX][LINE];
char cline[LINE],pline[LINE];
float bad,badlim;
float *sbin,*vbin,*sbout,*vbout;
long nsbin,nvbin,nsbout,nvbout;
float *range;
float *ux0,*uy0,*gx,*gy,*deu,*dev; /* dual doppler variables  */
float *seu,*sev,*sewp;          /* triple doppler standard errors  */
float *magu,*magv,*magwp;       /* magnification factors for u,v, and wp */
float *dmagu,*dmagv;            /* dual magnification factors for u and v */
float *vx,*vy,*vz;              /* triple doppler velocities */
float *time,*ctime;
float *avgtime;     /* average time to get all data at a particular grid pt. */
float *initime;                 /* initial time in grid point */
float *x,*y,*z;
float *fields[MAXFIELDS];       /* input */
float *ofields[MAXFIELDS];      /* output */
float *icount[MAXFIELDS];       /* number of good instances */
float *pcount;
float *a00,*a01,*a02,*a11,*a12,*a22;
float *w00,*w01,*w02,*w11,*w12,*w22;
float *b0,*b1,*b2;
float unx,uny,unz,rayvr,refl;
long nfields,vrtag,dbztag;
long nrange;
float *lon,*lat,*alt;
float *az,*el;
float lon0,lat0,alt0;  /* coordinates of grid origin */
float x0,yy0,z0;
float dx,dy,dz;
float xkm,ykm,zkm;
long nx,ny,nz;
long num;
float xl,yl,zl;
float xloc,yloc,zloc;
float delx,dely,delz;
float weight,cx,cy,cz,bx,by,bz;
float svx,svy,reftime;
float timenow,deltatime;
float a[3][3],ainv[3][3],c[3][3];
float det[4],b[3];
float tvaru,tvarv,tvarw;   /* Triple doppler variances  */
float dvaru,dvarv;        /* Dual doppler variances  */
long cnt;
long poop,loop,kk,nn;
long ix,iy,iz,jx,jy,jz;
long iloc;
struct field *fp;
float svalue,*sptr;
long tind;       /* if set then time goes in average for volume */

main(argc,argv)
int argc;
char **argv;
{

  /* check arguments */
  if (argc != 10 && argc != 13) {
    fprintf(stderr,"Usage: ");
    fprintf(stderr,"radcart4 lon0 dx xsize ");
    fprintf(stderr,"lat0 dy ysize alt0 dz zsize ");
    fprintf(stderr,"[reftime svx svy]\n");
    exit(1);
  }
  lon0 = atof(argv[1]);
  dx = atof(argv[2]);
  xkm = atof(argv[3]);
  lat0 = atof(argv[4]);
  dy = atof(argv[5]);
  ykm = atof(argv[6]);
  alt0 = atof(argv[7]);
  dz = atof(argv[8]);
  zkm = atof(argv[9]);

  if(argc == 13) {
    reftime = atof(argv[10]);
    svx = atof(argv[11]);
    svy = atof(argv[12]);
  }
  else {
    /* The reference time does not matter as long as svx = svy = 0. */
    reftime = svx = svy = 0.;
  }

  /* make x0,y0,z0 equal to zero, then the origin of the grid
     is at lon0,lat0,alt0. */
  x0 = yy0 = z0 = 0.;

  /* the +1 is to compensate for truncation */
  nx = xkm/dx + 1.5;
  ny = ykm/dy + 1.5;
  nz = zkm/dz + 1.5;

  /* get header of input file */
  gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"radcart4: float format expected\n");
    exit(1);
  }

  /* make header of output file */
  nullhdr(hbout,HBMAX,"float");
  copycmt(hbout,HBMAX,hbin,HBMAX);
  sprintf(cline,"radcart4:\n  %s %s %s\n",argv[1],argv[2],argv[3]);
  addcline(hbout,HBMAX,cline);
  sprintf(cline,"  %s %s %s\n",argv[4],argv[5],argv[6]);
  addcline(hbout,HBMAX,cline);
  sprintf(cline,"  %s %s %s\n",argv[7],argv[8],argv[9]);
  addcline(hbout,HBMAX,cline);
  if (argc == 13) {
    sprintf(cline,"  %s %s %s\n",argv[10],argv[11],argv[12]);
  }
  else sprintf(cline,"  no drift\n");
  addcline(hbout,HBMAX,cline);
  copypar(hbout,HBMAX,hbin,HBMAX);

  /* bad data values */
  if (seekpar(hbin,HBMAX,"bad",pline) == OK) bad = atof(pline);
  else {
    bad = BAD;
    sprintf(pline,"%g",bad);
    addpar(hbout,HBMAX,"bad",pline);
  }
  if (seekpar(hbin,HBMAX,"badlim",pline) == OK) badlim = atof(pline);
  else {
    badlim = BADLIM;
    sprintf(pline,"%g",badlim);
    addpar(hbout,HBMAX,"badlim",pline);
  }

  /* index parameters */
  sprintf(pline,"%g",lon0);
  addpar(hbout,HBMAX,"lon0",pline);
  sprintf(pline,"%g",x0);
  addpar(hbout,HBMAX,"x0",pline);
  sprintf(pline,"%g",dx);
  addpar(hbout,HBMAX,"dx",pline);
  sprintf(pline,"%g",lat0);
  addpar(hbout,HBMAX,"lat0",pline);
  sprintf(pline,"%g",yy0);
  addpar(hbout,HBMAX,"y0",pline);
  sprintf(pline,"%g",dy);
  addpar(hbout,HBMAX,"dy",pline);
  sprintf(pline,"%g",alt0);
  addpar(hbout,HBMAX,"alt0",pline);
  sprintf(pline,"%g",z0);
  addpar(hbout,HBMAX,"z0",pline);
  sprintf(pline,"%g",dz);
  addpar(hbout,HBMAX,"dz",pline);

  /* add index fields */
  addfld(hbout,HBMAX,'s',"x",1000.,0.,'l',1,"x",nx);
  addfld(hbout,HBMAX,'s',"y",1000.,0.,'l',1,"y",ny);
  addfld(hbout,HBMAX,'s',"z",1000.,0.,'l',1,"z",nz);

  /* add variable fields */
  loop = 0;
  nfields = 0;
  while ((fp = getfld(hbin,HBMAX,'v',loop++)) != NULL) {
    if (fp->dim == 1 && strcmp(fp->dname1,"range") == 0) {

      /* one dimensional fields with range dimension are cartesianized */
      addfld(hbout,HBMAX,'v',fp->fname,fp->smul,fp->sadd,'c',3,
          "x",nx,"y",ny,"z",nz);
      nfields++;
    }

    /* other fields are dropped */
  }

  /* add new fields  */
  addfld(hbout,HBMAX,'v',"vx",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"vy",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"vz",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"ux0",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"uy0",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"seu",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"sev",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"sewp",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"magu",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"magv",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"magwp",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"dmagu",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"dmagv",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"deu",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"dev",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"gx",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"gy",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"avgtime",10.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  addfld(hbout,HBMAX,'v',"time",1000.,0.,'l',0);

  /* allocate buffers */
  nsbin = elemcnt(hbin,HBMAX,'s');
  nsbout = elemcnt(hbout,HBMAX,'s');
  nvbin = elemcnt(hbin,HBMAX,'v');
  nvbout = elemcnt(hbout,HBMAX,'v');
  sbin = getbuff(nsbin);
  sbout = getbuff(nsbout);
  vbin = getbuff(nvbin);
  vbout = getbuff(nvbout);

  /* get pointers for index fields */
  range = getptr2(hbin,HBMAX,sbin,'s','d',"range",&fp);
  if (fp->dim != 1 || strcmp(fp->fname,fp->dname1) != 0) {
    fprintf(stderr,"radcart4: %s is not an index field\n",fp->fname);
    exit(1);
  }
  nrange = fp->dsize1;
  x = getptr(hbout,HBMAX,sbout,'s','d',"x");
  y = getptr(hbout,HBMAX,sbout,'s','d',"y");
  z = getptr(hbout,HBMAX,sbout,'s','d',"z");

  /* fill in index fields */
  for (loop = 0; loop < nx; loop++) x[loop] = x0 + loop*dx;
  for (loop = 0; loop < ny; loop++) y[loop] = yy0 + loop*dy;
  for (loop = 0; loop < nz; loop++) z[loop] = z0 + loop*dz;

  /* get pointers for fields to be cartesianized */
  loop = 0;
  nfields = 0;
  while ((fp = getfld(hbin,HBMAX,'v',loop++)) != NULL) {
    if (fp->dim == 1 && strcmp(fp->dname1,"range") == 0) {
      fields[nfields] = getptr(hbin,HBMAX,vbin,'v','d',fp->fname);
      ofields[nfields] = getptr2(hbout,HBMAX,vbout,'v','d',fp->fname,&fp);

/* radial velocity and reflectivity fields get special treatment */
      if(strcmp(fp->fname,VR) == 0)  {
        vrtag = nfields;
      }
      if(strcmp(fp->fname,DBZ) == 0)  {
        dbztag = nfields;
      }
      nfields++;
    }
  }

  /* get pointers for new fields   */
  vx = getptr(hbout,HBMAX,vbout,'v','d',"vx");
  vy = getptr(hbout,HBMAX,vbout,'v','d',"vy");
  vz = getptr(hbout,HBMAX,vbout,'v','d',"vz");
  ux0 = getptr(hbout,HBMAX,vbout,'v','d',"ux0");
  uy0 = getptr(hbout,HBMAX,vbout,'v','d',"uy0");
  seu = getptr(hbout,HBMAX,vbout,'v','d',"seu");
  sev = getptr(hbout,HBMAX,vbout,'v','d',"sev");
  sewp = getptr(hbout,HBMAX,vbout,'v','d',"sewp");
  magu = getptr(hbout,HBMAX,vbout,'v','d',"magu");
  magv = getptr(hbout,HBMAX,vbout,'v','d',"magv");
  magwp = getptr(hbout,HBMAX,vbout,'v','d',"magwp");
  dmagu = getptr(hbout,HBMAX,vbout,'v','d',"dmagu");
  dmagv = getptr(hbout,HBMAX,vbout,'v','d',"dmagv");
  deu = getptr(hbout,HBMAX,vbout,'v','d',"deu");
  dev = getptr(hbout,HBMAX,vbout,'v','d',"dev");
  gx = getptr(hbout,HBMAX,vbout,'v','d',"gx");
  gy = getptr(hbout,HBMAX,vbout,'v','d',"gy");
  avgtime = getptr(hbout,HBMAX,vbout,'v','d',"avgtime");

  /* get pointers for radar location */
  lon = getptr(hbin,HBMAX,vbin,'v','d',LON);
  lat = getptr(hbin,HBMAX,vbin,'v','d',LAT);
  alt = getptr(hbin,HBMAX,vbin,'v','d',ALT);

  /* get pointers for time */
  time = getptr(hbin,HBMAX,vbin,'v','d',TIME);
  ctime = getptr(hbout,HBMAX,vbout,'v','d',"time");

  /* get pointers for horizontal azimuth and elevation */
  az = getptr(hbin,HBMAX,vbin,'v','d',AZ);
  el = getptr(hbin,HBMAX,vbin,'v','d',EL);

  /* get input static slice */
  getslice(stdin,nsbin,sbin);

  /* write output header */
  puthdr(stdout,hbout,HBMAX);

  /* write output static slice */
  putslice(stdout,nsbout,sbout);

  /* initialize output fields */
  num = nx*ny*nz;
  for (poop = 0; poop < nfields; poop++) {
    icount[poop] = getbuff(num);
    for (loop = 0; loop < num; loop++) {
      ofields[poop][loop] = 0.;
      icount[poop][loop] = 0.;
    }
  }

  /* allocate buffer space for arrays A and b  */
  a00 = getbuff(num);
  a01 = getbuff(num);
  a02 = getbuff(num);
  a11 = getbuff(num);
  a12 = getbuff(num);
  a22 = getbuff(num);
  w00 = getbuff(num);
  w01 = getbuff(num);
  w02 = getbuff(num);
  w11 = getbuff(num);
  w12 = getbuff(num);
  w22 = getbuff(num);
  b0 = getbuff(num);
  b1 = getbuff(num);
  b2 = getbuff(num);
  pcount = getbuff(num);
  initime = getbuff(num);
  for (loop = 0; loop < num; loop++)  {
    b0[loop] = 0.;
    b1[loop] = 0.;
    b2[loop] = 0.;
    a00[loop] = 0.;
    a01[loop] = 0.;
    a02[loop] = 0.;
    a11[loop] = 0.;
    a12[loop] = 0.;
    a22[loop] = 0.;
    w00[loop] = 0.;
    w01[loop] = 0.;
    w02[loop] = 0.;
    w11[loop] = 0.;
    w12[loop] = 0.;
    w22[loop] = 0.;
    pcount[loop] = 0.;
    initime[loop] = bad;
    avgtime[loop] = bad;
  }

  /* loop through input variable slices */
  deltatime = 0.;
  cnt = 0;
  *ctime = 0.;
  while (getslice(stdin,nvbin,vbin) != EOF) {

    /* initialize tind for this ray (default: ray outside grid space) */
    tind = 0;

    /* get location of radar with respect to grid origin --
       these distances are approximated by the arc lengths */
    xl = (DEARTH/2.)*cos(*lat/DTOR)*((*lon - lon0)/DTOR);
    yl = (DEARTH/2.)*((*lat - lat0)/DTOR);
    zl = (*alt - alt0);


    /* loop on range */
    for (loop = 0; loop < nrange; loop++) {

      /* get locations of range point relative to radar */
      delx = range[loop]*sin(*az/DTOR)*cos(*el/DTOR);
      dely = range[loop]*cos(*az/DTOR)*cos(*el/DTOR);
      delz = range[loop]*sin(*el/DTOR);

      /* do corrections for earth's curvature */
      delz += (delx*delx + dely*dely)/DEARTH;

      /* get location relative to coordinate system origin */
      xloc = xl + delx;
      yloc = yl + dely;
      zloc = zl + delz;

      /* convert time to ksec */
      timenow = *time;

      /* Get the time difference between this data slice and the reference
         time */
      deltatime = timenow - reftime;

      /* translate all points to their location at the common time
         of reftime */
      xloc = xloc - (svx*deltatime);
      yloc = yloc - (svy*deltatime);

      /* find grid location */
      ix = TRUNC(x0,dx,xloc,cx);
      iy = TRUNC(yy0,dy,yloc,cy);
      iz = TRUNC(z0,dz,zloc,cz);
      FRACT(ix,cx,bx);
      FRACT(iy,cy,by);
      FRACT(iz,cz,bz);
      jx = ix + 1;
      jy = iy + 1;
      jz = iz + 1;

      /* check that the data point is in the grid space */
      if (ix >= 0 && ix < nx - 1 && iy >= 0 && iy < ny - 1 && iz >= 0 &&
          iz < nz - 1) {

        /* if it is, set tind to take time into account for the volume 
           time calculation and increment fields at each corner
           of surrounding cube */
        tind = 1;

        /* point 1 */
        iloc = iz + nz*iy + nz*ny*ix;
        weight = bx*by*bz;
        assign(iloc,weight);

        /* point 2 */
        iloc = iz + nz*iy + nz*ny*jx;
        weight = cx*by*bz;
        assign(iloc,weight);

        /* point 3 */
        iloc = iz + nz*jy + nz*ny*ix;
        weight = bx*cy*bz;
        assign(iloc,weight);

        /* point 4 */
        iloc = jz + nz*iy + nz*ny*ix;
        weight = bx*by*cz;
        assign(iloc,weight);

        /* point 5 */
        iloc = iz + nz*jy + nz*ny*jx;
        weight = cx*cy*bz;
        assign(iloc,weight);

        /* point 6 */
        iloc = jz + nz*iy + nz*ny*jx;
        weight = cx*by*cz;
        assign(iloc,weight);

        /* point 7 */
        iloc = jz + nz*jy + nz*ny*ix;
        weight = bx*cy*cz;
        assign(iloc,weight);

        /* point 8 */
        iloc = jz + nz*jy + nz*ny*jx;
        weight = cx*cy*cz;
        assign(iloc,weight);
      }
    }

    /* if tind is set then account for time */
    if (tind) {
      *ctime += *time;
      ++cnt;
    }
  }

  /* normalize output fields */
  *ctime /= (float)cnt;
  for (loop = 0; loop < num; loop++) {
    for (poop = 0; poop < nfields; poop++) {
      if (icount[poop][loop] > 0.) {
        ofields[poop][loop] /= icount[poop][loop];
        if (poop == dbztag) {
          ofields[poop][loop] = 10.*log10(ofields[poop][loop]);
        }
      }
      else {
        ofields[poop][loop] = bad;
      }
    }
    if (fabs(initime[loop]) > badlim) {
      avgtime[loop] = bad;
    }
    else {
      avgtime[loop] = (avgtime[loop] - initime[loop])*1000.;
    }

    /* check that there are some data in point */
    if (pcount[loop] <= 0.) {
      vx[loop] = bad;
      vy[loop] = bad;
      vz[loop] = bad;
      ux0[loop] = bad;
      uy0[loop] = bad;
      seu[loop] = bad;
      sev[loop] = bad;
      sewp[loop] = bad;
      magu[loop] = bad;
      magv[loop] = bad;
      magwp[loop] = bad;
      dmagu[loop] = bad;
      dmagv[loop] = bad;
      deu[loop] = bad;
      dev[loop] = bad;
      gx[loop] = bad;
      gy[loop] = bad;
      continue;
    }

    /* initialize the c, a, det and b arrays */
    det[3] = 0.;
    for (kk = 0; kk < 3; kk++)  {
      b[kk] = 0.;
      det[kk] = 0.;
      for (nn = 0; nn < 3; nn++)  {
        c[kk][nn] = a[kk][nn] = 0.;
      }
    }

    /* store values of matrix in arrays a and c  */
    c[0][0] = a[0][0] = a00[loop];
    c[0][1] = a[0][1] = a01[loop];
    c[0][2] = a[0][2] = a02[loop];
    c[1][0] = a[1][0] = a[0][1];
    c[1][1] = a[1][1] = a11[loop];
    c[1][2] = a[1][2] = a12[loop];
    c[2][0] = a[2][0] = a[0][2];
    c[2][1] = a[2][1] = a[1][2];
    c[2][2] = a[2][2] = a22[loop];

    /* store values of vector in b  */
    b[0] = b0[loop];
    b[1] = b1[loop];
    b[2] = b2[loop];

    /* store the value of the determinant of the system in det[3]  */
    det[3] = a[0][0]*a[1][1]*a[2][2] + a[1][0]*a[2][1]*a[0][2]
        + a[2][0]*a[1][2]*a[0][1] - a[2][0]*a[1][1]*a[0][2]
        - a[2][1]*a[1][2]*a[0][0] - a[2][2]*a[1][0]*a[0][1];

    /* if det[3] is good then find the inverse of a, ainv  */
    if(fabs(det[3]) < badlim && fabs(det[3]) > 5e-6) {
      ainv[0][0] = (a[1][1]*a[2][2] - a[2][1]*a[1][2])/det[3];
      ainv[0][1] = (a[2][1]*a[0][2] - a[0][1]*a[2][2])/det[3];
      ainv[0][2] = (a[0][1]*a[1][2] - a[1][1]*a[0][2])/det[3];
      ainv[1][0] = ainv[0][1];
      ainv[1][1] = (a[0][0]*a[2][2] - a[2][0]*a[0][2])/det[3];
      ainv[1][2] = (a[1][0]*a[0][2] - a[0][0]*a[1][2])/det[3];
      ainv[2][0] = ainv[0][2];
      ainv[2][1] = ainv[1][2];
      ainv[2][2] = (a[0][0]*a[1][1] - a[1][0]*a[0][1])/det[3];

      /* Now calculate the variances of the velocity components  */
      tvaru = ainv[0][0]*ainv[0][0]*w00[loop]
          + ainv[0][1]*ainv[0][1]*w11[loop]
          + ainv[0][2]*ainv[0][2]*w22[loop]
          + 2.*ainv[0][0]*ainv[0][1]*w01[loop]
          + 2.*ainv[0][0]*ainv[0][2]*w02[loop]
          + 2.*ainv[0][1]*ainv[0][2]*w12[loop];
      tvarv = ainv[1][0]*ainv[1][0]*w00[loop]
          + ainv[1][1]*ainv[1][1]*w11[loop]
          + ainv[1][2]*ainv[1][2]*w22[loop]
          + 2.*ainv[1][0]*ainv[1][1]*w01[loop]
          + 2.*ainv[1][0]*ainv[1][2]*w02[loop]
          + 2.*ainv[1][1]*ainv[1][2]*w12[loop];
      tvarw = ainv[2][0]*ainv[2][0]*w00[loop]
          + ainv[2][1]*ainv[2][1]*w11[loop]
          + ainv[2][2]*ainv[2][2]*w22[loop]
          + 2.*ainv[2][0]*ainv[2][1]*w01[loop]
          + 2.*ainv[2][0]*ainv[2][2]*w02[loop]
          + 2.*ainv[2][1]*ainv[2][2]*w12[loop];

      /* calculate the standard errors */
      if (tvaru < 0. || tvarv < 0. || tvarw < 0.) {
        seu[loop] = bad;
        sev[loop] = bad;
        sewp[loop] = bad;
        magu[loop] = bad;
        magv[loop] = bad;
        magwp[loop] = bad;
      }
      else {
        seu[loop] = sqrt(tvaru);
        sev[loop] = sqrt(tvarv);
        sewp[loop] = sqrt(tvarw);
        magu[loop] = sqrt(pcount[loop]*tvaru);
        magv[loop] = sqrt(pcount[loop]*tvarv);
        magwp[loop] = sqrt(pcount[loop]*tvarw);
      }
    }


    /* To calculate triple Doppler valocities --
       change each column of matrix a by vector b and calculate its 
       determinant */
    for (kk = 0; kk < 3; kk++)  {
      for (nn = 0; nn < 3; nn++)  {
        a[nn][kk] = b[nn];
      }
      det[kk] = a[0][0]*a[1][1]*a[2][2] + a[1][0]*a[2][1]*a[0][2]
          + a[2][0]*a[1][2]*a[0][1] - a[2][0]*a[1][1]*a[0][2]
          - a[2][1]*a[1][2]*a[0][0] - a[2][2]*a[1][0]*a[0][1];
      /* restore original system before changing next column */
      for (nn = 0; nn < 3; nn++) {
        a[nn][kk] = c[nn][kk];
      }
    }

    /* calculate the triple Doppler velocities */
    vx[loop] = det[0]/det[3];
    vy[loop] = det[1]/det[3];
    vz[loop] = det[2]/det[3];

    /* To calculate dual Doppler velocities -- first solve the system
       assuming vertical particle velocity is zero */
    det[1] = a[0][0]*a[1][1] - a[0][1]*a[1][0];
    ux0[loop] = (b[0]*a[1][1] - a[0][1]*b[1])/det[1];
    uy0[loop] = (a[0][0]*b[1] - b[0]*a[1][0])/det[1];

    /* now calculate the gx and gy in ux = ux0 + gx*vz, and 
       uy = uy0 + gy*vz */
    gx[loop] = (a[0][1]*a[1][2] - a[1][1]*a[0][2])/det[1];
    gy[loop] = (a[0][1]*a[0][2] - a[0][0]*a[1][2])/det[1];

    /* to calculate the standard error in the dual velocities
       first calculate the variances */
    if (fabs(det[1]) > 5e-6) {
      dvaru = a[1][1]*a[1][1]*w00[loop]/(det[1]*det[1]) -
          2.*a[1][1]*a[0][1]*w01[loop]/(det[1]*det[1]) +
          a[0][1]*a[0][1]*w11[loop]/(det[1]*det[1]);
      dvarv = a[1][0]*a[1][0]*w00[loop]/(det[1]*det[1]) -
          2.*a[1][0]*a[0][0]*w01[loop]/(det[1]*det[1]) +
          a[0][0]*a[0][0]*w11[loop]/(det[1]*det[1]);

      /* calculate the dual standard errors */
      if (dvaru < 0. || dvarv < 0.) {
        deu[loop] = bad;
        dev[loop] = bad;
        dmagu[loop] = bad;
        dmagv[loop] = bad;
      }
      else {
        deu[loop] = sqrt(dvaru);
        dev[loop] = sqrt(dvarv);
        dmagu[loop] = sqrt(pcount[loop]*dvaru);
        dmagv[loop] = sqrt(pcount[loop]*dvarv);
      }
    }
  }

  /* write variable slice */
  putslice(stdout,nvbout,vbout);
  exit(0);
}

/* assign -- do the assignment of data values to a grid point */
assign(iloc,weight)
long iloc;
float weight;
{
  unx = cos(*el/DTOR)*sin(*az/DTOR);
  uny = cos(*el/DTOR)*cos(*az/DTOR);
  unz = sin(*el/DTOR);
  a00[iloc] += weight*unx*unx;
  a01[iloc] += weight*unx*uny;
  a02[iloc] += weight*unx*unz;
  a11[iloc] += weight*uny*uny;
  a12[iloc] += weight*uny*unz;
  a22[iloc] += weight*unz*unz;
  w00[iloc] += weight*weight*unx*unx;
  w01[iloc] += weight*weight*unx*uny;
  w02[iloc] += weight*weight*unx*unz;
  w11[iloc] += weight*weight*uny*uny;
  w12[iloc] += weight*weight*uny*unz;
  w22[iloc] += weight*weight*unz*unz;
  pcount[iloc] += weight;
  for (poop = 0; poop < nfields; poop++) {
    if (fabs(fields[poop][loop]) < badlim) {
      if (poop == vrtag)  {
        rayvr = weight*fields[poop][loop];
        b0[iloc] += unx*rayvr;
        b1[iloc] += uny*rayvr;
        b2[iloc] += unz*rayvr;
      }
      if (poop == dbztag) {
        refl = pow(10.,(fields[poop][loop]/10.));
        ofields[poop][iloc] += weight*refl;
        icount[poop][iloc] += weight;
      }
      else {
        ofields[poop][iloc] += weight*fields[poop][loop];
        icount[poop][iloc] += weight;
      }
    }
  }
  if (fabs(initime[iloc]) > badlim) {
    initime[iloc] = avgtime[iloc] = timenow;
  }
  else {
    if (timenow < avgtime[iloc]) {
      if (timenow < initime[iloc]) {
        initime[iloc] = timenow;
      }
    }
    else {
      avgtime[iloc] = timenow;
    }
  }
}
