/* radcedric.c -- This program accepts a cedric file in rtape format from
 * the standard input and writes a corresponding common data format file
 * in float format to the standard output.  The cedric file specification
 * is taken from "Cedric -- Cartesian Space Data Processor (Batch Version),
 * 2/1/85" by Carl Mohr of the National Center for Atmospheric Research.
 * Radcedric assumes that data are organized in horizontal planes on input.
 * On output, one 3-D variable slice is created.  In regions with
 * non-existent data, the data points are given the value 1000.
 * A scalar variable field called time is created, with value equal
 * to the time after midnight in units of 1000 sec.
 * This program is inherently dependent on computer architecture because
 * short integer byte order (the form in which cedric stores most, but
 * not all data) differs between machines.  Set the "swap bytes" flag if
 * files from machines with the opposite byte order are being processed.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#define RECSIZE 30000     /* max allowed size of cedric records */
#define VOLHDR 510        /* length (shorts) of volume header */
#define LEVHDR 10         /* length of level header */

int nbytes;               /* actual size of cedric records */
short recbuff[RECSIZE/2]; /* input record buffer */
char *crecbuff;           /* character pointer to recbuff */
int recptr = RECSIZE;     /* input record pointer */
char hbuff[HBMAX][LINE];  /* cdf header buffer */
char temp[LINE],alpha[LINE];   /* temporary string buffers */
int loop;                 /* looping index */
int ival,junk;            /* input value and missing data value */
float bad;                /* output bad data value (=1000) */
float dval;               /* float format input value */
int nfields;              /* number of fields */
int nx,ny,nz;             /* dimension sizes */
int year,month,day,hour,minute,second,degree;   /* guess! */
float time;               /* seconds/1000 after midnight */
float sf,cf,smult;        /* scale factors for integer conversion */
float oldsmult[25];
float *sbuff,*vbuff;      /* pointers to cdf slice buffers */
long snum,vnum;           /* sizes of static and variable slices */
struct field *fp;         /* pointer to field structure */
float x0,yy0,z0,dx,dy,dz;  /* stuff for constructing index fields */
int ix,iy,iz;             /* index variables for transferring data */
int tape = 0;             /* flag to indicate tape or disk */
int bswap = 0;            /* byte swap flag */
int argl;                 /* temporary for argument checking */
short getint(),seekint();

main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if (argc < 2 || argc > 3) {
    fprintf(stderr,
         "Usage: radcedric -r or -d (rtape or direct) [-s (swap bytes)]\n");
    exit(1);
  }
  argl = 0;
  while (++argl < argc) {
    if (strcmp(argv[argl],"-r") == 0) tape = 1;
    else if (strcmp(argv[argl],"-d") == 0) tape = 0;
    else if (strcmp(argv[argl],"-s") == 0) bswap = 1;
    else {
      fprintf(stderr,"radcedric: unrecognized flag %s\n",argv[1]);
      exit(1);
    }
  }
/* set up character pointer to input record buffer */
  crecbuff = (char *) recbuff;
/* get first record */
  getrec(VOLHDR);
/* make a null header */
  nullhdr(hbuff,HBMAX,"float");
/* fill the header with all manner of stuff */
/* comment section first */
  addcline(hbuff,HBMAX,"radcedric:\n");
/* project name */
  getstr(14,4,temp);
  sprintf(alpha,"project = %s\n",temp);
  addcline(hbuff,HBMAX,alpha);
/* date */
  year = seekint(20);
  month = seekint(21);
  day = seekint(22);
  hour = seekint(23);
  minute = seekint(24);
  second = seekint(25);
  time = (3600*hour + 60*minute + second)/1000.;
  if (month == 1) strcpy(temp,"Jan");
  else if (month == 2) strcpy(temp,"Feb");
  else if (month == 3) strcpy(temp,"Mar");
  else if (month == 4) strcpy(temp,"Apr");
  else if (month == 5) strcpy(temp,"May");
  else if (month == 6) strcpy(temp,"Jun");
  else if (month == 7) strcpy(temp,"Jul");
  else if (month == 8) strcpy(temp,"Aug");
  else if (month == 9) strcpy(temp,"Sep");
  else if (month == 10) strcpy(temp,"Oct");
  else if (month == 11) strcpy(temp,"Nov");
  else if (month == 12) strcpy(temp,"Dec");
  else strcpy(temp,"Weird");
  sprintf(alpha,"date = %d%s%d\n",day,temp,year);
  addcline(hbuff,HBMAX,alpha);
/* time */
  getstr(84,4,temp);
  sprintf(alpha,"time = %d:%d:%d %s\n",hour,minute,second,temp);
  addcline(hbuff,HBMAX,alpha);
/* scaling factors */
  sf = seekint(67);
  cf = seekint(68);
/* latitude */
  degree = seekint(32);
  minute = seekint(33);
  second = seekint(34);
  sprintf(alpha,"lat. = %d:%d:%.2f\n",degree,minute,(float)(second/sf));
  addcline(hbuff,HBMAX,alpha);
/* longitude */
  degree = seekint(35);
  minute = seekint(36);
  second = seekint(37);
  sprintf(alpha,"long. = %d:%d:%.2f\n",degree,minute,(float)(second/sf));
  addcline(hbuff,HBMAX,alpha);
/* missing data value */
  junk = seekint(66);
/* parameters */
/* aximuth of x-axis */
  param("azimuth",39,cf);
/* origin */
  param("x0",159,sf);
  param("y0",164,sf);
  param("z0",169,1000.);
/* delta values */
  param("dx",162,1000.);
  param("dy",167,1000.);
  param("dz",172,1000.);
/* numbers of points */
  nx = seekint(161);
  ny = seekint(166);
  nz = seekint(171);
/* bad data values and limits */
  addpar(hbuff,HBMAX,"badlim","999.0");
  addpar(hbuff,HBMAX,"bad","1000.0");
  bad = 1000.0;
  
/* static fields */
  addfld(hbuff,HBMAX,'s',"x",100.,0.,'s',1,"x",nx);
  addfld(hbuff,HBMAX,'s',"y",100.,0.,'s',1,"y",ny);
  addfld(hbuff,HBMAX,'s',"z",100.,0.,'s',1,"z",nz);
/* variable fields */
/* time first */
  addfld(hbuff,HBMAX,'v',"time",1000.,0.,'l',0);
/* then cedric-defined fields */
  nfields = seekint(174);
  if (nfields > 25) nfields = 25;
/* set multiplicative scaling factor to 32 for all data */
  smult = 32.;
  for (loop = 0; loop < nfields; loop++) {
    getstr(350 + 10*loop,8,temp);
/* change the field name to lower case */
    lower(temp);
/* pick up and save the cedric scale factor */
    oldsmult[loop] = seekint(179 + 5*loop);
    addfld(hbuff,HBMAX,'v',temp,smult,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
/* write header to standard output */
  puthdr(stdout,hbuff,HBMAX);
/* allocate static slice */
  snum = elemcnt(hbuff,HBMAX,'s');
  if ((sbuff = (float *) malloc(sizeof(float)*snum)) == NULL) {
    fprintf(stderr,"radcedric: can't get enough memory\n");
    exit(1);
  }
/* build index fields */
  fp = seekfld(hbuff,HBMAX,'s',"x");
  seekpar(hbuff,HBMAX,"x0",temp);
  x0 = atof(temp);
  seekpar(hbuff,HBMAX,"dx",temp);
  dx = atof(temp);
  for (loop = 0; loop < nx; loop++) {
    sbuff[fp->ptr + loop] = x0 + loop*dx;
  }
  fp = seekfld(hbuff,HBMAX,'s',"y");
  seekpar(hbuff,HBMAX,"y0",temp);
  yy0 = atof(temp);
  seekpar(hbuff,HBMAX,"dy",temp);
  dy = atof(temp);
  for (loop = 0; loop < ny; loop++) {
    sbuff[fp->ptr + loop] = yy0 + loop*dy;
  }
  fp = seekfld(hbuff,HBMAX,'s',"z");
  seekpar(hbuff,HBMAX,"z0",temp);
  z0 = atof(temp);
  seekpar(hbuff,HBMAX,"dz",temp);
  dz = atof(temp);
  for (loop = 0; loop < nz; loop++) {
    sbuff[fp->ptr + loop] = z0 + loop*dz;
  }
/* write static slice */
  putslice(stdout,snum,sbuff);
/* allocate variable slice */
  vnum = elemcnt(hbuff,HBMAX,'v');
  if ((vbuff = (float *) malloc(sizeof(float)*vnum)) == NULL) {
    fprintf(stderr,"radcedric: can't get enough memory\n");
    exit(1);
  }
/* fill variable slice, time first */
  fp = seekfld(hbuff,HBMAX,'v',"time");
  vbuff[fp->ptr] = time;
/* next, stuff from cedric file */
/* loop on levels */
  for (iz = 0; iz < nz; iz++) {
/* get level header info and store z value */
    getrec(LEVHDR);
/* loop on cedric fields */
    loop = 1;
    while ((fp = getfld(hbuff,HBMAX,'v',loop++)) != NULL) {
/* set input pointer so new record is forced (cedric wastes space!) */
      recptr = RECSIZE;
/* loop in fortran order on x-y array */
      for (iy = 0; iy < ny; iy++) {
        for (ix = 0; ix < nx; ix++) {
/* get data point from input buffer */
          ival = getint();
/* let's set non-existent data to bad */
          if (ival == junk) dval = bad;
/* use saved cedric scale factor to extract physical value */
          else dval = ival/oldsmult[loop - 2];
/* put data point in output buffer */
          vbuff[fp->ptr + ix*ny*nz + iy*nz + iz] = dval;
        }
      }
    }
  }
/* all done -- put the single variable slice */
  putslice(stdout,vnum,vbuff);
  exit(0);
}

/* get cedric record from standard input */
getrec(reqshorts)
short reqshorts;
{
  int i;
  char count[9];
/* first, get mini-header and interpret if tape */
  if (tape) {
    for (i = 0; i < 8; i++) {
/* croak on eof */
      if ((count[i] = getchar()) == EOF) {
        fprintf(stderr,"radcedric: unexpected end-of-file on input\n");
        exit(1);
      }
    }
    count[8] = '\0';
    nbytes = atoi(count);
  }
/* if not tape, just use argument value */
  else nbytes = 2*reqshorts;
/* die if number of bytes is too large */
  if (nbytes >= RECSIZE) {
    fprintf(stderr,"radcedric: %d bytes in input record -- too big!\n",
      nbytes);
    exit(1);
  }
/* get record */
  if (fread(crecbuff,sizeof(char),nbytes,stdin) != nbytes) {
    fprintf(stderr,"radcedric: unexpected eof or error on input\n");
    exit(1);
  }
/* reset record pointer */
  recptr = 0;
}

/* return next short integer in record buffer */
short getint()
{
  short i,j;
  if (recptr >= nbytes) getrec(1);
  i = *((short *) (crecbuff + recptr));
  if (bswap) {
    swab(&i,&j,2);
    i = j;
  }
  recptr += 2;
  return(i);
}

/* return specified short integer from record buffer */
short seekint(pos)
int pos;
{
  short i,j;
  i = recbuff[pos];
  if (bswap) {
    swab(&i,&j,2);
    i = j;
  }
  return(i);
}

/* extract a string from record */
getstr(pos,length,buffer)
int pos,length;
char *buffer;
{
  int i;
  if (length >= LINE/2) length = LINE/2 - 1;
  for (i = 0; i < length; i++) buffer[i] = 0177 & crecbuff[i + pos];
  buffer[length] = '\0';
}

/* add a numerical parameter from cedric header */
param(name,pos,scale)
char *name;
int pos;
float scale;
{
  sprintf(temp,"%f",(float)(seekint(pos)/scale));
  addpar(hbuff,HBMAX,name,temp);
  return;
}

/* change a string to lower case */
lower(string)
char *string;
{
  int i;
  char c;
  i = 0;
  while ((c = string[i]) != '\0') {
    if ((c >= 'A') && (c <= 'Z')) string[i] += 'a' - 'A';
    if (c == ' ') {
      string[i] = '\0';
      break;
    }
    i++;
  }
}
