%{
/* radedit -- edit radar ray file for folding and bad data */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#include "pimage.h"

/* stack size */
#define SMAX 20
#define LSIZE 81

/* size of a bunch of rays */
#define BUNCH 190

/* overlap when moving */
#define NONOVERLAP 150

/* parameters for editing */
#define HALFWIDTH 5
#define TESTMIN 2

/* left and right border sizes */
#define LBORDER 100
#define RBORDER 100

/* top and bottom border sizes */
#define TBORDER 150
#define BBORDER 70

/* scan control plot parameters */
#define SCSIZE 100.
#define SCOFF 20.

/* minimum and maximum displayed values of radial velocity and reflectivity */
#define MINREF 0.
#define MAXREF 60.
#define MINVEL -25.
#define MAXVEL 25.

/* number of pixels per range gate */
#define NYPIXELS 2

/* number of pixels per ray */
#define NXPIXELS 4

/* index stuff */
#define IS(ix,iy) ((ix) + xsize*(iy))
#define II(ix,iy,ipix) ((ix) + ximage*(iy*NYPIXELS + ipix))
#define IIP(ix,iy) ((ix) + ximage*(iy))

/* name of fields (reasonable expectation to stay fixed) */
#define RANGE "range"
#define RAYCNT "raycnt"

/* getopt stuff */
int option;
extern char *optarg;
extern int optind,opterr;

/* files */
char *infile,*outfile;              /* input and output file names */
FILE *infptr,*outfptr;              /* input and output file pointers */
long nelements;                     /* eof check on input */

/* Candis stuff */
char hbuff[HBMAX][LINE];            /* header buffer for input and output */
char cline[LINE];                   /* a line character buffer */
char pline[LINE];                   /* another line character buffer */
float bad,badlim;                   /* bad data parameters */
long nsbuff,nvbuff;                 /* buffer sizes */
float *sbuff;                       /* static slice buffer */
struct field *fp;                   /* field pointer */

/* global stuff */
long loop,lx,ly;                    /* loop indices */
float *buffer[BUNCH];               /* variable slice buffers currently used */
int ateof = 0;                      /* indicates eof reached on input */
int gotbunch = 0;                   /* indicates we already have a bunch */
int xsize,ysize;                    /* screen display size */
int ximage,yimage;                  /* image display size */

/* command line names */
char *scancontrol;                  /* name of scan control variable */
char *timename;                     /* name of time variable */
char *nyquistvel;                   /* name of Nyquist velocity variable */
char *reflectivity;                 /* name of reflectivity variable */
char *radvel;                       /* name of radial velocity variable */

/* offsets in variable slice */
long raycntoffset;                  /* offset of raycnt field */
long timeoffset;                    /* offset of time field */
long scoffset;                      /* offset of scan control field */
long refloffset;                    /* offset of reflectivity */
long rveloffset;                    /* offset of radial velocity */

/* other things for input data */
float *nvptr;                       /* points to Nyquist velocity */
float nyvel;                        /* Nyquist velocity */
float *range;                       /* pointer to range index field */
int maxrange;                       /* size of range dimension */
float firstray;                     /* don't edit till we reach this ray */
int firstskip = 0;                  /* skip if this is set */
float reflthresh;                   /* threshold value for reflectivity */

/* X11 stuff */
int ncolors;                        /* number of colors available on disp */
long cvals[MAXSHADES];              /* list of pixel values */

/* local definitions */
int nstack = 0;                     /* the stack pointer */
char stack[SMAX][LSIZE];            /* the stack */
int errorflag = 0;                  /* flag errors */
int helpflag = 0;                   /* help requested */
int refreshflag = 0;                /* refresh the image */
int velflag = 0;                    /* flags indicating field displayed */
int reflflag = 0;
int getflag = 0;                    /* get a bunch of rays */
int tufflag = 0;                    /* timewise unfolding of a ray */
int quitflag = 0;                   /* quit requested */
int killflag = 0;                   /* kill requested */
int inrefl = 0;                     /* flags which image is displayed */
void docmd();                       /* do what needs to be done for cmd */
void push();                        /* push a token */
char *pop();                        /* pop a token */
void syntax();                      /* report a syntax error */
char *ftoascii();                   /* convert a float to an ascii string */
%}
%%
			fprintf(stderr,"\nradedit> ");

"\n"			{docmd(); fprintf(stderr,"radedit> ");}

help			helpflag = 1;
refresh			refreshflag = 1;
vel                     velflag = 1;
refl                    reflflag = 1;
get			getflag = 1;
edit			tufflag = 1;
kill			killflag = 1;
bye			quitflag = 1;

[-0-9.+e]+		push(yytext);
" "|"\t"		;
.			errorflag = 1;
%%

/* main program */
main(argc,argv)
int argc;
char **argv;
{

/* check command line arguments */
  opterr = 0;
  firstray = -1.;
  firstskip = 0;
  strcpy(cline,"radedit:");
  while ((option = getopt(argc,argv,"r:x:")) != -1) {
    if (option == 'r') {
      firstray = atof(optarg);
      firstskip = 0;
      strcat(cline," -r ");
      strcat(cline,optarg);
    }
    else if (option == 'x') {
      firstray = atof(optarg);
      firstskip = 1;
      strcat(cline," -x ");
      strcat(cline,optarg);
    }
  }
  strcat(cline,"\n");
  if (argc - optind != 8) {
    fprintf(stderr,
         "Usage: radedit [-r|x fray] scan time nyq refl rvel dzmin in out\n");
    exit(1);
  }

/* get positional arguments */
  scancontrol = argv[optind];
  timename = argv[optind + 1];
  nyquistvel = argv[optind + 2];
  reflectivity = argv[optind + 3];
  radvel = argv[optind + 4];
  reflthresh = atof(argv[optind + 5]);
  infile = argv[optind + 6];
  outfile = argv[optind + 7];

/* open files */
  if (strcmp(infile,outfile) == 0) {
    fprintf(stderr,"radedit: input and output files the same!\n");
    exit(1);
  }
  if ((infptr = fopen(infile,"r")) == NULL) {
    fprintf(stderr,"radedit: can't open %s for input\n",infile);
    exit(1);
  }
  if ((outfptr = fopen(outfile,"w")) == NULL) {
    fprintf(stderr,"radedit: can't open %s for output\n",outfile);
    exit(1);
  }

/* read the Candis header */
  gethdr(infptr,hbuff,HBMAX);

/* test format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"radedit: float format expected\n");
    exit(1);
  }

/* get bad data parameters */
  if (seekpar(hbuff,HBMAX,"bad",pline) == OK) bad = atof(pline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",pline) == OK) badlim = atof(pline);
  else badlim = BADLIM;

/* add a comment */
  addcline(hbuff,HBMAX,cline);
  sprintf(cline,"  %s %s %s %s %s %s %s %s\n",argv[optind],argv[optind + 1],
       argv[optind + 2],argv[optind + 3],argv[optind + 4],
       argv[optind + 5],argv[optind + 6],argv[optind + 7]);
  addcline(hbuff,HBMAX,cline);

/* get offsets of various variables */
  if ((fp = seekfld(hbuff,HBMAX,'v',RAYCNT)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s\n",RAYCNT);
    exit(1);
  }
  raycntoffset = fp->ptr;
  if ((fp = seekfld(hbuff,HBMAX,'v',timename)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s\n",timename);
    exit(1);
  }
  timeoffset = fp->ptr;
  if ((fp = seekfld(hbuff,HBMAX,'v',scancontrol)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s\n",scancontrol);
    exit(1);
  }
  scoffset = fp->ptr;
  if ((fp = seekfld(hbuff,HBMAX,'v',reflectivity)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s\n",reflectivity);
    exit(1);
  }
  refloffset = fp->ptr;
  if ((fp = seekfld(hbuff,HBMAX,'v',radvel)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s\n",radvel);
    exit(1);
  }
  rveloffset = fp->ptr;

/* allocate input buffers */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  nvbuff = elemcnt(hbuff,HBMAX,'v');

/* get range information */
  if ((range = getptr2(hbuff,HBMAX,sbuff,'s','r',RANGE,&fp)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s\n",RANGE);
    exit(1);
  }
  maxrange = fp->tsize;

/* transfer header and static slice to output */
  getslice(infptr,nsbuff,sbuff);
  puthdr(outfptr,hbuff,HBMAX);
  putslice(outfptr,nsbuff,sbuff);

/* get Nyquist velocity from static slice */
  if ((nvptr = getptr(hbuff,HBMAX,sbuff,'s','r',nyquistvel)) == NULL) {
    fprintf(stderr,"radedit: can't find variable %s in static slice\n",
         nyquistvel);
    exit(1);
  }
  nyvel = *nvptr;

/* allocate space for bunch of variable slice buffers */
  for (loop = 0; loop < BUNCH; loop++) {
    buffer[loop] = getbuff(nvbuff);
  }

/* set up display */
  xsize = LBORDER + BUNCH*NXPIXELS + RBORDER;
  ysize = BBORDER + TBORDER + maxrange*NYPIXELS;
  ximage = BUNCH*NXPIXELS;
  yimage = maxrange*NYPIXELS;
  setupdisp("radedit",xsize,ysize);
  setbackground(0);
  ncolors = getcolors(cvals);
  if (ncolors <= 2) {
    fprintf(stderr,"radedit: sorry, I don't work on monochrome displays\n");
    exit(1);
  }

/* skip or transfer a bunch of records if requested */
  if (firstray > 0.) {
    fprintf(stderr,"looking for first ray > %g ...\n",firstray);
    while (1) {
      nelements = getslice(infptr,nvbuff,buffer[0]);
      if (nelements == EOF) {
        fprintf(stderr,
             "radedit: eof encountered while searching for first ray\n");
        fclose(outfptr);
        exit(1);
      }
      rayclean(0);
      if (!firstskip) putslice(outfptr,nvbuff,buffer[0]);
      if (buffer[0][raycntoffset] >= firstray) break;
    }
    fprintf(stderr,"... found it!\n");
  }

/* set up and do get command */
  getflag = 1;
  docmd();

/* hand over control to command processor */
  yylex();
}

/* push a string onto the stack */
void push(txt)
char *txt;
{
  if (nstack >= SMAX) {
    fprintf(stderr,"radedit: too many numbers in a row\n");
    exit(1);
  }
  strcpy(stack[nstack++],txt);
}

/* pop a string from the stack */
char *pop()
{
  if (nstack <= 0) {
    fprintf(stderr,"radedit: popping empty stack\n");
    exit(1);
  }
  return(stack[--nstack]);
}

/* syntax -- report a syntax error and zero stack */
void syntax()
{
  fprintf(stderr,"syntax error\n");
  nstack = 0;
}

/* docmd -- execute the desired command */
void docmd()
{
  float *fptr;
  float velval,reflval;
  int ivel,irefl;
  int mx,my,button;
  int ray,gate;
  int ix,iy,jx,jy,nsum;
  float meanvel;

/* error */
  if (errorflag) {
    errorflag = 0;
    nstack = 0;
    fprintf(stderr,"unknown command\n");
    return;
  }

/* help */
  if (helpflag) {
    helpflag = 0;
    nstack = 0;
    fprintf(stderr,"help        Print a help statement\n");
    fprintf(stderr,"get         Get a chunk of radar data\n");
    fprintf(stderr,"edit        Enter edit mode\n");
    fprintf(stderr,"vel         Show radial velocity field\n");
    fprintf(stderr,"refl        Show reflectivity field\n");
    fprintf(stderr,"refresh     Refresh the image\n");
    fprintf(stderr,"bye         Write rest of file and depart\n");
    fprintf(stderr,"kill        Write current chunk and depart\n");
    return;
  }

/* refresh the image */
  if (refreshflag) {
    refreshflag = 0;
    nstack = 0;
    dispimage(0);
  }

/* change to velocity field */
  if (velflag) {
    velflag = 0;
    nstack = 0;

/* send the velocity image and display */
    blapimage('v');
    inrefl = 0;
  }

/* change to reflectivity field */
  if (reflflag) {
    reflflag = 0;
    nstack = 0;

/* send the reflectivity image and display */
    blapimage('r');
    inrefl = 1;
  }

/* get a bunch of rays */
  if (getflag) {
    getflag = 0;
    nstack = 0;

/* write the nonoverlapping part ... */
    for (loop = 0; loop < NONOVERLAP && loop < gotbunch; loop++) {
      putslice(outfptr,nvbuff,buffer[loop]);
    }

/* move the overlapping part down ... */
    for (loop = NONOVERLAP; loop < gotbunch; loop++) {
      fptr = buffer[loop - NONOVERLAP];
      buffer[loop - NONOVERLAP] = buffer[loop];
      buffer[loop] = fptr;
    }
    gotbunch -= NONOVERLAP;
    if (gotbunch < 0) gotbunch = 0;

/* ... and read in some more */
    while (!ateof && gotbunch < BUNCH) {
      if (getslice(infptr,nvbuff,buffer[gotbunch]) == EOF) {
        ateof = 1;
        fprintf(stderr,"end of input reached\n");
        break;
      }
      rayclean(gotbunch);
      gotbunch++;
    }

/* send the velocity image and display */
    blapimage('v');
  }

/* time-wise unfolding of a ray */
  if (tufflag) {
    tufflag = 0;
    nstack = 0;

/* send helpful message */
    fprintf(stderr,"left button to unfold, middle to advance, right to exit\n");

/* loop on mouse events -- button 3 exits, 1 or 2 does the job */
    while (1) {
      mousepos(0,&mx,&my,&button);
      if (button == 2) {
        getflag = 1;
        docmd();
      }
      if (button == 3) break;

/* find where we are in radar space */
      where(mx,my,&ray,&gate);

/* loop on gates */
      for (ly = 0; ly < maxrange; ly++) {

/* define comparison interval */
        ix = ray - HALFWIDTH;
        if (ix < 0) ix = 0;
        jx = ray + HALFWIDTH;
        if (jx >= gotbunch) jx = gotbunch - 1;

/* get average radial velocity over interval excluding problem ray */
        meanvel = 0.;
        nsum = 0;
        for (lx = ix; lx <= jx; lx++) {
          if (lx == ray) continue;
          velval = buffer[lx][rveloffset + ly];
          reflval = buffer[lx][refloffset + ly];
          if (fabs(velval) < badlim) {
            meanvel += velval;
            nsum++;
          }
        }

/* must have at least TESTMIN values */
        if (nsum < TESTMIN) continue;
        meanvel /= (float)nsum;

/* fix the problem by adding or subtracting multiples of 2*nyquist vel */
        velval = buffer[ray][rveloffset + ly];
        if (fabs(velval) > badlim) continue;
        while (velval > meanvel + nyvel) velval -= 2.*nyvel;
        while (velval < meanvel - nyvel) velval += 2.*nyvel;
        buffer[ray][rveloffset + ly] = velval;
      }

/* blap a ray */
      blapvelray(ray);
      dispimage(0);
    }
  }

/* quit */
  if (quitflag || killflag) {

/* flush out the buffer */
    for (loop = 0; loop < gotbunch; loop++) {
      putslice(outfptr,nvbuff,buffer[loop]);
    }

/* if quit, transfer rest of file */
    if (quitflag && !ateof) {
      fprintf(stderr,"transferring the rest of the file ...\n");
      while (getslice(infptr,nvbuff,buffer[0]) != EOF) {
        rayclean(0);
        putslice(outfptr,nvbuff,buffer[0]);
      }
      fprintf(stderr,"... done!\n");
    }

/* print adios */
    fclose(infptr);
    fclose(outfptr);
    fprintf(stderr,"bye!\n");
    exit(0);
  }
}

/* rayclean -- do various thresholding stuff on each ray */
rayclean(pos)
int pos;
{
  int i;
  float reflval;

/* if reflthresh exceeds -100 do thresholding */
  if (reflthresh <= -100.) return;

/* loop on range */
  for (i = 0; i < maxrange; i++) {
    reflval = buffer[pos][refloffset + i];
    if (fabs(reflval) > badlim || reflval < reflthresh) {
      buffer[pos][rveloffset + i] = bad;
      buffer[pos][refloffset + i] = bad;
    }
  }

/* kill single gate velocity values */
  for (i = 1; i < maxrange - 1; i++) {
    if (fabs(buffer[pos][rveloffset + i - 1]) > badlim &&
         fabs(buffer[pos][rveloffset + i + 1]) > badlim) {
      buffer[pos][rveloffset + i] = bad;
      buffer[pos][refloffset + i] = bad;
    }
  }
}

/* blapimage -- construct an image and display it */
blapimage(choice)
char choice;                   /* 'v' for velocity, 'r' for reflectivity */
{
  unsigned char *image;
  int ix,iy,jx,jy;
  int ivel,irefl;
  float velval,reflval;
  float svarmin,svarmax,svar;

/* clean display area first */
  rectangle(0,0,0,xsize,ysize,GRAY40);
  rectangle(0,LBORDER,TBORDER,LBORDER + ximage,TBORDER + yimage,GRAY0);

/* choose */
  if (choice == 'v') {

/* velocity ... */
    for (lx = 0; lx < gotbunch; lx++) {
      blapvelray(lx);
    }
  }

/* ... or reflectivity */
  else {
    for (lx = 0; lx < gotbunch; lx++) {
      blapreflray(lx);
    }
  }

/* range labels */
  drawtext(0,LBORDER + ximage + 5,TBORDER + YFONTOFF,
       fg(),ftoascii(range[0]));
  drawtext(0,LBORDER + ximage + 5,TBORDER + (int)(.33*yimage) + YFONTOFF,
       fg(),ftoascii(range[(int)(.33*maxrange)]));
  drawtext(0,LBORDER + ximage + 5,TBORDER + (int)(.67*yimage) + YFONTOFF,
       fg(),ftoascii(range[(int)(.67*maxrange)]));
  drawtext(0,LBORDER + ximage + 5,TBORDER + yimage + YFONTOFF,
       fg(),ftoascii(range[maxrange - 1]));
  drawtext(0,LBORDER + ximage + 5,TBORDER + yimage/2 + YFONTOFF,
       fg(),"range (km)");

  drawtext(0,LBORDER,TBORDER + yimage + 20 + YFONTOFF*2,
       fg(),ftoascii(buffer[0][raycntoffset]));
  if (gotbunch > 0) {
    drawtext(0,LBORDER + gotbunch*NXPIXELS,TBORDER + yimage + 20 + YFONTOFF*2,
         fg(),ftoascii(buffer[gotbunch - 1][raycntoffset]));
    drawtext(0,LBORDER + gotbunch*NXPIXELS/2,TBORDER + yimage + 20 + YFONTOFF*2,
         fg(),"raycount");
  }

  drawtext(0,LBORDER,TBORDER + yimage + 40 + YFONTOFF*2,
       fg(),ftoascii(buffer[0][timeoffset]));
  if (gotbunch > 0) {
    drawtext(0,LBORDER + gotbunch*NXPIXELS,TBORDER + yimage + 40 + YFONTOFF*2,
         fg(),ftoascii(buffer[gotbunch - 1][timeoffset]));
    drawtext(0,LBORDER + gotbunch*NXPIXELS/2,TBORDER + yimage + 40 + YFONTOFF*2,
         fg(),"time (ks)");
  }
  

/* find limits of scan variable */
  svarmin = 1.e10;
  svarmax = -svarmin;
  for (lx = 0; lx < gotbunch; lx++) {
    svar = buffer[lx][scoffset];
    if (svar > svarmax) svarmax = svar;
    if (svar < svarmin) svarmin = svar;
  }
  if (svarmin < -360.) svarmin = -360.;
  if (svarmax > 360.) svarmax = 360.;
  if (svarmax < svarmin + 1.) svarmax = svarmin + 1.;

/* plot scan variable */
  for (lx = 1; lx < gotbunch; lx++) {
    ix = (lx - 1)*NXPIXELS + LBORDER;
    jx = lx*NXPIXELS + LBORDER;
    svar = buffer[lx - 1][scoffset];
    if (svar < svarmin) svar = svarmin;
    if (svar > svarmax) svar = svarmax;
    iy = SCSIZE*(1. - (svar - svarmin)/(svarmax - svarmin)) + SCOFF;
    svar = buffer[lx][scoffset];
    if (svar < svarmin) svar = svarmin;
    if (svar > svarmax) svar = svarmax;
    jy = SCSIZE*(1. - (svar - svarmin)/(svarmax - svarmin)) + SCOFF;
    drawline(0,ix,iy,jx,jy,1,-1);
  }

/* draw box around it */
  ix = LBORDER - 1;
  jx = LBORDER + gotbunch*NXPIXELS;
  iy = SCSIZE + SCOFF;
  jy = SCOFF;
  box(0,ix,iy,jx,jy,1);

/* label it */
  drawtext(0,10,(int)(SCOFF + .5*SCSIZE),fg(),scancontrol);
  drawtext(0,LBORDER + ximage + 5,(int)(SCOFF + SCSIZE) + YFONTOFF,
       fg(),ftoascii(svarmin));
  drawtext(0,LBORDER + ximage + 5,(int)(SCOFF) + YFONTOFF,
       fg(),ftoascii(svarmax));

/* info line */
  sprintf(cline,"reflthresh = %.1f",reflthresh);
  drawtext(0,LBORDER,(int)(SCOFF + SCSIZE) + 2*YFONTOFF + 4,fg(),cline);

/* do the color bar */
  if (choice == 'v') {
    colorbar(10,TBORDER + 30,30,yimage - 30,FIRSTCOLOR,NCOLORS,
         MINVEL,MAXVEL,"vel (m/s)");
  }
  else {
    colorbar(10,TBORDER + 30,30,yimage - 30,FIRSTGRAY + NCOLORS,NGRAYS,
         MINREF,MAXREF,"refl (dBZ)");
  }

/* display the image */
  dispimage(0);
}

/* blapreflray -- blap a reflectivity ray onto the pixmap */
blapreflray(lx)
int lx;
{
  int irefl;
  float reflval;
  for (ly = 0; ly < maxrange; ly++) {
    reflval = buffer[lx][refloffset + ly];
    if (fabs(reflval) < badlim) {
      irefl = (int)((NGRAYS - 1)*(reflval - MINREF)/(MAXREF - MINREF) + .5);
      if (irefl < 0) irefl = 0;
      if (irefl >= NGRAYS) irefl = NGRAYS - 1;
      irefl += FIRSTGRAY + NCOLORS;
      rectangle(0,LBORDER + lx*NXPIXELS,
           TBORDER + ly*NYPIXELS,
           LBORDER + (lx + 1)*NXPIXELS,
           TBORDER + (ly + 1)*NYPIXELS,irefl);
    }
  }
}

/* blapvelray -- blap a velocity ray onto the pixmap */
blapvelray(lx)
int lx;
{
  int ivel;
  float reflval,velval;
  for (ly = 0; ly < maxrange; ly++) {
    velval = buffer[lx][rveloffset + ly];
    reflval = buffer[lx][refloffset + ly];
    if (fabs(velval) < badlim) {
      ivel = (int)((NCOLORS - 1)*(velval - MINVEL)/(MAXVEL - MINVEL) + .5);
      if (ivel < 0) ivel = 0;
      if (ivel >= NCOLORS - 1) ivel = NCOLORS - 1;
      ivel += FIRSTCOLOR;
      rectangle(0,LBORDER + lx*NXPIXELS,
           TBORDER + ly*NYPIXELS,
           LBORDER + (lx + 1)*NXPIXELS,
           TBORDER + (ly + 1)*NYPIXELS,ivel);
    }
  }
}

/* ftoascii -- convert a float to an ascii string */
char *ftoascii(val)
float val;
{
  static char aval[80];
  sprintf(aval,"%g",val);
  return(aval);
}

/* colorbar -- create a color bar on screen */
colorbar(x,y,dx,dy,startcolor,ncolors,minval,maxval,name)
int x,y,dx,dy;
int startcolor,ncolors;
float minval,maxval;
char *name;
{
  int ix,iy,jy,ddy;
  float val;

/* loop over color blocks */
  ddy = dy/ncolors;
  for (loop = 0; loop < ncolors; loop++) {

/* do a rectangle fill */
    jy = y + loop*ddy;
    rectangle(0,x,jy,x + dx,jy + ddy,startcolor + loop);

/* label it */
    val = minval + loop*(maxval - minval)/(ncolors - 1);
    drawtext(0,x + dx + 10,y + ddy*loop + ddy/2 + YFONTOFF,fg(),ftoascii(val));
  }

/* final label */
  drawtext(0,x,y - 20 + YFONTOFF,fg(),name);
}

/* rectangle -- blap a solid rectangle onto pixmap */
rectangle(index,x1,y1,x2,y2,color)
int index,x1,x2,y1,y2,color;
{
  int xlist[4],ylist[4];
  xlist[0] = x1;
  ylist[0] = y1;
  xlist[1] = x1;
  ylist[1] = y2;
  xlist[2] = x2;
  ylist[2] = y2;
  xlist[3] = x2;
  ylist[3] = y1;
  fillpoly(index,4,xlist,ylist,cvals[color]);
}

/* draw a box */
box(index,x1,y1,x2,y2,ltype)
int index,x1,y1,x2,y2,ltype;
{
  drawline(index,x1,y1,x1,y2,ltype,-1);
  drawline(index,x2,y2,x1,y2,ltype,-1);
  drawline(index,x2,y2,x2,y1,ltype,-1);
  drawline(index,x1,y1,x2,y1,ltype,-1);
}

/* where -- figure out where we are in radar space from window coords */
where(wx,wy,ray,gate)
int wx,wy;
int *ray,*gate;
{
  *ray = (wx - LBORDER)/NXPIXELS;
  if (*ray < 0) *ray = 0;
  if (*ray >= gotbunch) *ray = gotbunch - 1;
  *gate = (wy - TBORDER)/NYPIXELS;
  if (*gate < 0) *gate = 0;
  if (*gate >= maxrange) *gate = maxrange - 1;
}
