/* radeldora -- convert an Eldora file (stdin) to a series of Candis files
 *
 * This program assumes that the two Eldora antennas are putting out the
 * same parameters, as the first set of parameters is overwritten by
 * the second set.
 *
 * This program irrevokably assumes that a float is 4 bytes, a long is 4 bytes,
 * a short is 2 bytes, and a character is 1 byte.
 */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define MAX 10000            /* biggest block accepted */
#define MAXPARAMS 25         /* maximum number of parameters measured */
#define MAXVOLUMES 100       /* maximum number of files to be created */

#define USAGE "Usage: radeldora [-nfba] [-r maxgate] [-c corrs] [-f root] file ...\n"

/* common stuff for input */
FILE *infile;                /* input file pointer */
int unfold;                  /* unfold flag */
int front,back;              /* front and back ray flags */
int angrecalc;               /* recalculate azimuth and elevation */
long lbuff[MAX/4 + 1];       /* buffer to store a block */
char *buff;                  /* character cast of lbuff */
char bname[5];               /* block identifier name */
long blocksize;              /* size of the block just read */
long getblock();             /* reads an Eldora tape block */
short getshort();
long getlong();
float getfloat();

/* Candis stuff */
char hbuff[HBMAX][LINE];
char cline[LINE],pline[LINE];
float *sbuff = NULL,*vbuff = NULL;
long nsbuff,nvbuff;
struct field *fp;
float *range;
float *raycnt;
float *raydata[MAXPARAMS];
char fname[MAXPARAMS][9];
float sadd[MAXPARAMS],smul[MAXPARAMS];
char typelen[MAXPARAMS];
int floatflag[MAXPARAMS];
long cdfgates;

/* getopt stuff */
extern char *optarg;
extern int optind;
char c;

/* scalar fields */
float *c_time;
float *c_jday,*c_hour,*c_min,*c_sec,*c_msec,*c_az,*c_el;
float *c_raystat,*c_unambvel,*c_unambrange;
float *c_lon,*c_lat,*c_zmsl,*c_zagl,*c_ua,*c_va,*c_wa;
float *c_heading,*c_roll,*c_pitch;
float *c_driftangle,*c_rotangle,*c_tilt;
float *c_u,*c_v,*c_w,*c_headingrate,*c_pitchrate;

/* file control stuff */
int fileopen = 0;
int filecounter = 0;
FILE *outfile;
char filename[80];
char *fileroot;
char *defroot = "eldora";    /* default root file name */

/* vold stuff */
char proj[21] = "",source[9] = "";
short vol = 0,v_year = 0,v_mon = 0,v_day = 0,v_hour = 0,v_min = 0,v_sec = 0;

/* radd stuff */
char rad_name[9];
short rad_type,scan_mode;
float rad_const = 0.,v_peak_power = 0.,sys_gain = 0.,unamb_vel = 0.,
unamb_range = 0.;

/* cspd stuff */
short num_segs = 0,dist_to_first = 0,spacing[6],num_cells[6];
short total_cells = 0;

/* parm stuff */
char p_name[9],p_desc[41],p_units[9],thresh_fld[9];
short samples = 0,bin_form = 0;
long bad_data = 0;
float thresh_val = 0.,p_scale = 0.,p_bias = 0.;

/* ryib stuff */
short hour,min,sec,msec;
long sweep = -1,j_day = -1,ray_stat = -1;
float az = 0.,el = 0.,peak_power = 0.;

/* asib stuff */
float lon = 0.,lat = 0.,z_msl = 0.,z_agl = 0.;
float u_aircraft = 0.,v_aircraft = 0.,w_aircraft = 0.;
float heading = 0.,roll = 0.,pitch = 0.,heading_rate = 0.,pitch_rate = 0.;
float drift_angle = 0.;
float rot_angle = 0.,tilt = 0.;
float u = 0.,v = 0.,w = 0.;

/* frad stuff */
short word_len[MAXPARAMS],nparams = 0;
short first_gate = 0,last_gate = 0;
long pointers[MAXPARAMS];
long stepsize;

main(argc,argv)
int argc;
char *argv[];
{
  int i,j,k,irange,jtotal;

  /* process arguments */
  unfold = 1;
  front = back = 0;
  fileroot = defroot;
  cdfgates = 10000;
  angrecalc = 0;
  while ((c = getopt(argc,argv,"r:c:nfbao:")) != -1) {
    if (c == 'r') {
      cdfgates = atoi(optarg);
    }
    else if (c == 'c') {
      fprintf(stderr,"radeldora: corrections not implemented yet\n");
      exit(1);
    }
    else if (c == 'n') unfold = 0;
    else if (c == 'f') front = 1;
    else if (c == 'b') back = 1;
    else if (c == 'a') angrecalc = 1;
    else if (c == 'o') {
      fileroot = optarg;
    }
    else {
      fprintf(stderr,USAGE);
      exit(1);
    }
  }

  /* check to see if there is at least one input file */
  if (argc <= optind) {
    fprintf(stderr,USAGE);
    exit(1);
  }

  /* set up properly aligned pointer to character buffer */
  buff = (char*)lbuff;

  /* loop on input files */
  for (; optind < argc; optind++) {

    /* open input file */
    if ((infile = fopen(argv[optind],"r")) == NULL) {
      fprintf(stderr,"radeldora: can't open input file %s\n",argv[optind]);
      exit(1);
    }

    /* loop on blocks */
    while ((blocksize = getblock(infile,MAX,buff)) != 0) {

      /* croak if block size negative */
      if (blocksize < 0) exit(1);

      /* get block name */
      getstring(0,4,buff,bname);

      /* look at block */
      if (strcmp(bname,"VOLD") == 0) {
        extract_vold(buff,proj,&vol,
            &v_year,&v_mon,&v_day,&v_hour,&v_min,&v_sec,source);

        /* free buffers, close an open file, and increment file suffix */
        if (fileopen) {
          free(sbuff);
          free(vbuff);
          fclose(outfile);
          filecounter++;
          if (filecounter >= MAXVOLUMES) {
            fprintf(stderr,"radeldora: too many volumes!\n");
            exit(1);
          }
        }
      }
      else if (strcmp(bname,"RADD") == 0) {
        extract_radd(buff,rad_name,&rad_const,&v_peak_power,&sys_gain,&rad_type,
            &scan_mode,&unamb_vel,&unamb_range);
      }
      else if (strcmp(bname,"CSPD") == 0) {
        extract_cspd(buff,&num_segs,&dist_to_first,spacing,num_cells);
        total_cells = 0;
        for (i = 0; i < 6; i++) {
          if (i == num_segs) break;
          total_cells += num_cells[i];
        }
        if (cdfgates > total_cells) cdfgates = total_cells;

        /* zero number of parameters */
        nparams = 0;
      }
      else if (strcmp(bname,"PARM") == 0) {
        extract_parm(buff,p_name,p_desc,p_units,&samples,&bin_form,thresh_fld,
            &thresh_val,&p_scale,&p_bias,&bad_data);

        /* decypher bin_form */
        if (nparams >= MAXPARAMS) {
          fprintf(stderr,"radeldora: too many parameters!\n");
          exit(1);
        }
        if (bin_form == 1) {
          word_len[nparams] = 1;
          typelen[nparams] = 'c';
          floatflag[nparams] = 0;
        }
        else if (bin_form == 2) {
          word_len[nparams] = 2;
          typelen[nparams] = 's';
          floatflag[nparams] = 0;
        }
        else if (bin_form == 3) {
          word_len[nparams] = 4;
          typelen[nparams] = 'l';
          floatflag[nparams] = 0;
        }
        else if (bin_form == 4) {
          word_len[nparams] = 4;
          typelen[nparams] = 'l';
          floatflag[nparams] = 1;
        }
        else {
          fprintf(stderr,"radeldora: bin_form = %d is unsupported!\n",bin_form);
          exit(1);
        }

        /* fill Candis parameter tables */
        lcase(p_name,fname[nparams]);
        sadd[nparams] = p_bias;
        smul[nparams] = p_scale;

        /* increment number of parameters */
        nparams++;
      }
      else if (strcmp(bname,"RYIB") == 0) {
        extract_ryib(buff,&sweep,&j_day,&hour,&min,&sec,&msec,&az,&el,
            &peak_power,&ray_stat);
      }
      else if (strcmp(bname,"ASIB") == 0) {
        extract_asib(buff,&lon,&lat,&z_msl,&z_agl,&u_aircraft,&v_aircraft,
            &w_aircraft,&heading,&roll,&pitch,&drift_angle,&rot_angle,
            &tilt,&u,&v,&w,&heading_rate,&pitch_rate);
      }
      else if (strcmp(bname,"FRAD") == 0) {
        extract_frad(buff,blocksize,word_len,total_cells,nparams,
            &first_gate,&last_gate,pointers,&stepsize);

        /* open a Candis file if necessary, after creating a new name */
        if (!fileopen) {
          sprintf(filename,"%s.%03d",fileroot,filecounter);
          if ((outfile = fopen(filename,"w")) == NULL) {
            fprintf(stderr,"radeldora: can't open %s for output\n",filename);
            exit(1);
          }
          fileopen = 1;

          /* construct a Candis header */
          nullhdr(hbuff,HBMAX,"float");

          /* comments */
          addcline(hbuff,HBMAX,"radeldora:\n");
          sprintf(cline,"  %d-%d-%d -- %02d:%02d:%02d\n",v_year,v_mon,v_day,
              v_hour,v_min,v_sec);
          addcline(hbuff,HBMAX,cline);
          sprintf(cline,"  %s\n",proj);
          addcline(hbuff,HBMAX,cline);
          if (unfold) addcline(hbuff,HBMAX,"  unfolded\n");
          else addcline(hbuff,HBMAX,"  not unfolded\n");
          if (front && back) addcline(hbuff,HBMAX,"  front and back rays\n");
          else if (front) addcline(hbuff,HBMAX,"  front rays only\n");
          else if (back) addcline(hbuff,HBMAX,"  back rays only\n");
          else addcline(hbuff,HBMAX,"  no ray data\n");
          if (angrecalc) addcline(hbuff,HBMAX,"  recalculate az and el\n");

          /* params */
          sprintf(pline,"%d",v_year);
          addpar(hbuff,HBMAX,"vol_year",pline);
          sprintf(pline,"%d",v_mon);
          addpar(hbuff,HBMAX,"vol_month",pline);
          sprintf(pline,"%d",v_day);
          addpar(hbuff,HBMAX,"vol_day",pline);
          sprintf(pline,"%d",v_hour);
          addpar(hbuff,HBMAX,"vol_hour",pline);
          sprintf(pline,"%d",v_min);
          addpar(hbuff,HBMAX,"vol_min",pline);
          sprintf(pline,"%d",v_sec);
          addpar(hbuff,HBMAX,"vol_sec",pline);

          /* static fields */
          addfld(hbuff,HBMAX,'s',"range",1000.,0.,'l',1,"range",cdfgates);

          /* variable fields */
          addfld(hbuff,HBMAX,'v',"time",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"raycnt",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"jday",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"hour",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"min",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"sec",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"msec",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"az",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"el",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"raystat",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"unambvel",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"unambrange",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"lon",1000.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"lat",1000.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"zmsl",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"zagl",1.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"ua",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"va",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"wa",100.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"heading",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"roll",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"pitch",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"driftangle",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"rotangle",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"tilt",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"u",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"v",10.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"w",100.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"headingrate",100.,0.,'l',0);
          addfld(hbuff,HBMAX,'v',"pitchrate",100.,0.,'l',0);
          for (i = 0; i < nparams; i++) {
            addfld(hbuff,HBMAX,'v',fname[i],smul[i],sadd[i],typelen[i],
                1,"range",cdfgates);
          }

          /* allocate slice buffers */
          nsbuff = elemcnt(hbuff,HBMAX,'s');
          sbuff = getbuff(nsbuff);
          nvbuff = elemcnt(hbuff,HBMAX,'v');
          vbuff = getbuff(nvbuff);

          /* get the pointers */
          range = getptr(hbuff,HBMAX,sbuff,'s','d',"range");
          c_time = getptr(hbuff,HBMAX,vbuff,'v','d',"time");
          raycnt = getptr(hbuff,HBMAX,vbuff,'v','d',"raycnt");
          c_jday = getptr(hbuff,HBMAX,vbuff,'v','d',"jday");
          c_hour = getptr(hbuff,HBMAX,vbuff,'v','d',"hour");
          c_min = getptr(hbuff,HBMAX,vbuff,'v','d',"min");
          c_sec = getptr(hbuff,HBMAX,vbuff,'v','d',"sec");
          c_msec = getptr(hbuff,HBMAX,vbuff,'v','d',"msec");
          c_az = getptr(hbuff,HBMAX,vbuff,'v','d',"az");
          c_el = getptr(hbuff,HBMAX,vbuff,'v','d',"el");
          c_raystat = getptr(hbuff,HBMAX,vbuff,'v','d',"raystat");
          c_unambvel = getptr(hbuff,HBMAX,vbuff,'v','d',"unambvel");
          c_unambrange = getptr(hbuff,HBMAX,vbuff,'v','d',"unambrange");
          c_lon = getptr(hbuff,HBMAX,vbuff,'v','d',"lon");
          c_lat = getptr(hbuff,HBMAX,vbuff,'v','d',"lat");
          c_zmsl = getptr(hbuff,HBMAX,vbuff,'v','d',"zmsl");
          c_zagl = getptr(hbuff,HBMAX,vbuff,'v','d',"zagl");
          c_ua = getptr(hbuff,HBMAX,vbuff,'v','d',"ua");
          c_va = getptr(hbuff,HBMAX,vbuff,'v','d',"va");
          c_wa = getptr(hbuff,HBMAX,vbuff,'v','d',"wa");
          c_heading = getptr(hbuff,HBMAX,vbuff,'v','d',"heading");
          c_roll = getptr(hbuff,HBMAX,vbuff,'v','d',"roll");
          c_pitch = getptr(hbuff,HBMAX,vbuff,'v','d',"pitch");
          c_driftangle = getptr(hbuff,HBMAX,vbuff,'v','d',"driftangle");
          c_rotangle = getptr(hbuff,HBMAX,vbuff,'v','d',"rotangle");
          c_tilt = getptr(hbuff,HBMAX,vbuff,'v','d',"tilt");
          c_u = getptr(hbuff,HBMAX,vbuff,'v','d',"u");
          c_v = getptr(hbuff,HBMAX,vbuff,'v','d',"v");
          c_w = getptr(hbuff,HBMAX,vbuff,'v','d',"w");
          c_headingrate = getptr(hbuff,HBMAX,vbuff,'v','d',"headingrate");
          c_pitchrate = getptr(hbuff,HBMAX,vbuff,'v','d',"pitchrate");
          for (i = 0; i < nparams; i++) {
            raydata[i] = getptr(hbuff,HBMAX,vbuff,'v','d',fname[i]);
          }

          /* zero ray count */
          *raycnt = 0.;

          /* construct the range index field -- convert to km */
          irange = dist_to_first;
          k = 0;
          for (i = 0; i < num_segs; i++) {
            for (j = 0; j < num_cells[i]; j++) {
              if (k < cdfgates) range[k] = ((float)irange)/1000.;
              irange += spacing[i];
              if (++k > total_cells) {
                fprintf(stderr,
                     "radeldora: panic! error in range construction\n");
                exit(1);
              }
            }
          }

          /* write header and static slice */
          puthdr(outfile,hbuff,HBMAX);
          putslice(outfile,nsbuff,sbuff);
        }

        /* break out if we don't want this ray */
        if (!front && tilt <= 0.) break;
        if (!back && tilt > 0.) break;

        /* fill a variable slice */
        for (i = 0; i < nparams; i++) {
          jtotal = total_cells*nparams + pointers[i];
          if (floatflag[i]) {
            k = 0;
            for (j = pointers[i]; j < jtotal; j += stepsize) {
              if (k < cdfgates) {
                raydata[i][k++] = (getfloat(j,buff) - sadd[i])/smul[i];
              }
            }
          }
          else if (typelen[i] == 'c') {
            k = 0;
            for (j = pointers[i]; j < jtotal; j += stepsize) {
              if (k < cdfgates) {
                raydata[i][k++] = (buff[j] - sadd[i])/smul[i];
              }
            }
          }
          else if (typelen[i] == 's') {
            k = 0;
            for (j = pointers[i]; j < jtotal; j += stepsize) {
              if (k < cdfgates) {
                raydata[i][k++] = (getshort(j,buff) - sadd[i])/smul[i];
              }
            }
          }
          else if (typelen[i] == 'l') {
            k = 0;
            for (j = pointers[i]; j < jtotal; j += stepsize) {
              if (k < cdfgates) {
                raydata[i][k++] = (getlong(j,buff) - sadd[i])/smul[i];
              }
            }
          }
        }

        /* time in kiloseconds after midnight */
        *c_time = hour*3.6 + min*0.06 + sec*0.001;

        /* other scalar fields */
        *c_jday = j_day;
        *c_hour = hour;
        *c_min = min;
        *c_sec = sec;
        *c_msec = msec;
        *c_raystat = ray_stat;
        *c_unambvel = unamb_vel;
        *c_unambrange = unamb_range;
        *c_lon = lon;
        *c_lat = lat;
        *c_zmsl = z_msl;
        *c_zagl = z_agl;
        *c_ua = u_aircraft;
        *c_va = v_aircraft;
        *c_wa = w_aircraft;
        *c_heading = heading;
        *c_roll = roll;
        *c_pitch = pitch;
        *c_driftangle = drift_angle;
        *c_rotangle = rot_angle;
        *c_tilt = tilt;
        *c_u = u;
        *c_v = v;
        *c_w = w;
        *c_headingrate = heading_rate;
        *c_pitchrate = pitch_rate;

        /* recompute az and elevation if desired */
        if (angrecalc) {
          ta_angles(roll,pitch,heading,rot_angle,tilt,c_az,c_el);
        }
        else {
          *c_az = az;
          *c_el = el;
        }

        /* write a variable slice */
        putslice(outfile,nvbuff,vbuff);

        /* increment ray count */
        (*raycnt)++;
      }
    }

/* close input file */
    fclose(infile);
  }

  /* close the open output file and exit */
  fclose(outfile);
  exit(0);
}

/* convert a string to lower case */
lcase(instring,outstring)
char *instring,*outstring;
{
  int i;
  i = 0;
  while (instring[i] != '\0') {
    if (instring[i] >= 'A' && instring[i] <= 'Z') {
      outstring[i] = instring[i] - 'A' + 'a';
    }
    else outstring[i] = instring[i];
    i++;
  }
  outstring[i + 1] = '\0';
}

/*  Function to calculate the elevation and azimuth angles
    of a vector with respect to a horizontal plane over the 
    Earth from the pitch, roll, heading, elevation and azimuth
    of the same vector with respect to the WP-3 aircraft
    reference frame. The azimuth and elevation are for the
    tail radar.
    Author -- Lucio Lopez  */

/*  This function expects the angles in degrees and it also 
    calculates az and el in degrees.     */ 

ta_angles(roll,pitch,hdg,raz,rel,az,el)
float roll,pitch,hdg,raz,rel;    /* input angles */
float *az,*el;                   /* output angles */
{
  float r,p,h,a,e;
/* components of unit vector relative to horizontal reference frame */
  float vex,vey,vez;    
/* constant to change angles from degrees to radians, 2pi/360.  */
  float k;
  k = 2.*3.14159/360.;
/* transformation of angles units.  */
  r = k*roll;
  p = k*pitch;
  h = k*hdg;
  a = k*raz;
  e = k*rel;

/* It was assumed that the order in the angle rotations is heading, pitch
   and roll. So, in order to transform the aircraft coordinate system
   to a horizontal coordinate system, we first rotate back the roll angle
   then the pitch angle and finally the heading angle. It is also assumed
   that positive pitch angles are for the aircraft nose up the horizontal,
   positive roll angles are for the right wing up the horizontal and the
   positive heading is measured from the north in a clockwise sense. */

  vex = cos(e)*cos(h)*sin(a+r) - cos(e)*sin(h)*sin(p)*cos(a+r)   
        + sin(e)*sin(h)*cos(p);
  
  vey = sin(e)*cos(h)*cos(p) - cos(e)*sin(h)*sin(a+r)  
        - cos(e)*cos(h)*sin(p)*cos(a+r);

  vez = sin(p)*sin(e) + cos(p)*cos(e)*cos(a+r);  

/* now get the elevation and azimuth angles with respect to the 
   horizontal coordinate system. Then azimuth is positive from the
   north in a clockwise sense, and elevation is from the horizontal
   where positive is up the horizontal plane.    */

  *el = asin(vez)/k;
  *az = atan2(vex,vey)/k;
  if (*az < 0.) *az += 360.;
}
