/* radfudge -- This program fixes offset errors in uf/cdf/cat radar data.
 * Potential offsets fixed are in range, azimuth and reflectivity.
 * Constant values, specified on the command line, are added to
 * each of these things.  If the cluttersup argument is specified, radial
 * velocities less than the value in magnitude are set to bad.  This is
 * a primitive attempt to suppress ground clutter.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

char hbuff[HBMAX][LINE];         /* header buffer */
char hbout[HBMAX][LINE];         /* output header buffer */
char cline[LINE];                /* comment line */
char pline[LINE];                /* parameter line */
char tline[LINE];                /* want to hold one line until
                                     another is received */
float *sbuff,*vbuff;             /* buffer pointers */
long nsbuff,nvbuff;              /* buffer lengths */
float *range,*azim,*dbz,*ve;     /* interesting field pointers */
float *azim0,*dbz0,*ve0;
long azimsize,dbzsize,vesize;    /* sizes of fields */
float rangefudge,azimfudge,dbzfudge;  /* fudge values */
float mterm,pulswdth;
float cluttersup;                /* clutter suppression value */
long loop;                       /* loop temporary */
struct field *fp;                /* temporary field pointer */
float bad,badlim;                /* bad data parameters */
/* range0 is the range to the first gate as read off of the field
   format tape, urzero is the calculated correction to that value */
float range0,urzero,drange;             

main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if ((argc != 7) && (argc != 9)) {
    fprintf(stderr,"Usage: radfudge rangefudge azimfudge dbzname dbzfudge");
    fprintf(stderr," mterm pulswdth [vradname cluttersup]\n");
    exit(1);
  }
  rangefudge = atof(argv[1]);
  azimfudge = atof(argv[2]);
  dbzfudge = atof(argv[4]);
  mterm = atof(argv[5]);
  pulswdth = atof(argv[6]);
  if (argc == 9) {
    cluttersup = atof(argv[8]);
  }
  else {
    cluttersup = -1.;
  }
/* get header */
  gethdr(stdin,hbuff,HBMAX);
/* check for float format */
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"radfudge: expecting float format\n");
    exit(1);
  }
/* get bad data parameters */
  if (seekpar(hbuff,HBMAX,"badlim",cline) == FAIL) {
    badlim = BADLIM;
  }
  else badlim = atof(cline);
  if (seekpar(hbuff,HBMAX,"bad",cline) == FAIL) {
    bad = BAD;
  }
  else bad = atof(cline);
/* new output header */
  nullhdr(hbout,HBMAX,"float");
  copycmt(hbout,HBMAX,hbuff,HBMAX);
  copyfld(hbout,HBMAX,hbuff,HBMAX,'s');
  copyfld(hbout,HBMAX,hbuff,HBMAX,'v');
/* add comment line */
  addcline(hbuff,HBMAX,"radfudge:\n");
  if (argc == 7)
    sprintf(cline,"  %s %s %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4],argv[5],argv[6]);
  else
    sprintf(cline,"  %s %s %s %s %s %s %s %s\n",
      argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7],argv[8]);
  addcline(hbuff,HBMAX,cline);


/* handle parameters */
  loop = 0;
  while (getpar(hbuff,HBMAX,loop++,pline,cline) != FAIL) {
/* special treatment for range0 if it exists */
    if (strcmp(pline,"range0") == 0) {
      range0 = atof(cline);
    }
    if (strcmp(pline,"drange") == 0) {
      drange = atof(cline);
    }
 addpar(hbout,HBMAX,pline,cline);
  }
/* set up correction for range to first gate as range0 is used
   when initializing the static variable range it will not be added
   here as well */
  urzero = (1+mterm)*drange + rangefudge - 0.25*3.0e5*pulswdth;
 sprintf(cline,"%g",urzero);
 addpar(hbout,HBMAX,"urzero",cline);

/* write header */
  puthdr(stdout,hbout,HBMAX);
/* get static slice */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  getslice(stdin,nsbuff,sbuff);
/* point the range pointer (assumes that range name is "range") */
  range = getptr2(hbuff,HBMAX,sbuff,'s','d',"range",&fp);
/* fudge the range */
  for (loop = 0; loop < fp->tsize; loop++) {
    range[loop] += urzero;
  }
/* write static slice */
  putslice(stdout,nsbuff,sbuff);
/* allocate variable slice space */
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbuff);
/* point the pointers */
  azim0 = getptr2(hbuff,HBMAX,vbuff,'v','d',"azim",&fp);
  azimsize = fp->tsize;
  dbz0 = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[3],&fp);
  dbzsize = fp->tsize;
  if (argc == 9) {
    ve0 = getptr2(hbuff,HBMAX,vbuff,'v','d',argv[8],&fp);
    vesize = fp->tsize;
  }
/* loop on variable slices */
  while (getslice(stdin,nvbuff,vbuff) != EOF) {
    azim = azim0;
    dbz = dbz0;
    if (argc == 9) ve = ve0;
/* fix azimuth */
    for (loop = 0; loop < azimsize; loop++) {
      *azim += azimfudge;
      if (*azim < 0.) *azim += 360.;
      if (*azim > 360.) *azim -= 360.;
      azim++;
    }
/* fix reflectivity */
    for (loop = 0; loop < dbzsize; loop++) {
      *dbz += dbzfudge;
      dbz++;
    }
/* set small radial velocities to bad */
    if (cluttersup > 0.) {
      for (loop = 0; loop < vesize; loop++) {
        if (fabs(*ve) <= cluttersup) *ve = bad;
        ve++;
      }
    }
/* put the variable slice */
    putslice(stdout,nvbuff,vbuff);
  }
  exit(0);
}
