/* radice -- This program computes various cloud physical parameters
 * based on the assumption that the radar return is from graupel or
 * small hail.  Data used from the input file are radar reflectivity
 * in dB, terminal fall velocity in m/s (assumed negative), and air
 * density as a function of height (kg/m^3).  Default values for
 * drag coefficient and terminal velocity radius relationship are
 * assumed unless six rather than three command line arguments are
 * typed.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

/* grid to linear conversion */

#define I(ix,iy,iz) ((ix)*ny*nz + (iy)*nz + (iz))

/* argument handles */

char *reflname,*wtname,*wname,*rhoname;

/* Candis declarations */

char hbuff[HBMAX][LINE];
long nsbuff,nvbin,nvbout;
float *sbuff,*vbuff;
char cline[LINE];
struct field *fp,*fp1;
float bad,badlim;

/* input fields */

float *w,*refl,*wt,*rho;

/* new output fields */

float *dbn,*a,*aequiv,*rhoice,*dbqp,*dbprate;

/* default cloud physics constants */

/* precip particle drag coefficient */
#define CD 0.5
/* terminal velocity parameters */
#define AA 2.5
#define ALPHA 0.8
/* maximum terminal velocity (negative by definition */
#define WTMAX -2.0

/* actual values of cloud physics constants (default or command line) */

float cd,aa,alpha;

/* various stuff */

long ix,iy,iz;
long nx,ny,nz;
float prate,zice,rhoa,idens,numice,radius,qprecip,wtt,ww;

main(argc,argv)
int argc;
char *argv[];
{

/* check arguments */

  if (argc == 5) {
    cd = CD;
    aa = AA;
    alpha = ALPHA;
  }
  else if (argc == 8) {
    cd = atof(argv[5]);
    aa = atof(argv[6]);
    alpha = atof(argv[7]);
  }
  else {
    fprintf(stderr,
         "Usage: radice reflname wtname wname rhoname [Cd A alpha]\n");
    exit(1);
  }

  reflname = argv[1];
  wtname = argv[2];
  wname = argv[3];
  rhoname = argv[4];

/* get header from standard input */

  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"radice: float format expected\n");
    exit(1);
  }

/* get values of bad data parameters */

  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;

/* allocate static slice buffer */

  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);

/* find size of variable slice buffer before adding to header */

  nvbin = elemcnt(hbuff,HBMAX,'v');

/* add stuff to header */

/* comment first */

  sprintf(cline,"radice: %s %s %s %s\n",reflname,wtname,wname,rhoname);
  addcline(hbuff,HBMAX,cline);
  if (argc == 8) {
    sprintf(cline,"  %s %s %s\n",argv[5],argv[6],argv[7]);
    addcline(hbuff,HBMAX,cline);
  }

/* new fields next -- first find dim info on existing fields */

  if ((fp = seekfld(hbuff,HBMAX,'v',reflname)) == NULL) {
    fprintf(stderr,"radice: field %s doesn't exist\n",reflname);
    exit(1);
  }

/* dbn -- number of precipitation particles per m^3 of air -- decibels */

  addfld(hbuff,HBMAX,'v',"dbn",100.,0.,'s',fp->dim,
       fp->dname1,fp->dsize1,
       fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,
       fp->dname4,fp->dsize4);

/* a -- particle radius in mm */

  addfld(hbuff,HBMAX,'v',"a",100.,0.,'s',fp->dim,
       fp->dname1,fp->dsize1,
       fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,
       fp->dname4,fp->dsize4);

/* aequiv -- radius in mm of equivalent water drop */

  addfld(hbuff,HBMAX,'v',"aequiv",100.,0.,'s',fp->dim,
       fp->dname1,fp->dsize1,
       fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,
       fp->dname4,fp->dsize4);

/* rhoice -- mass density of ice particles in gm/cc */

  addfld(hbuff,HBMAX,'v',"rhoice",1000.,0.,'s',fp->dim,
       fp->dname1,fp->dsize1,
       fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,
       fp->dname4,fp->dsize4);

/* dbqp -- mixing ratio of precip in gm/kg (decibel measure) */

  addfld(hbuff,HBMAX,'v',"dbqp",1000.,0.,'s',fp->dim,
       fp->dname1,fp->dsize1,
       fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,
       fp->dname4,fp->dsize4);

/* dbprate -- precipitation rate in mm/hr (decibel measure) */

  addfld(hbuff,HBMAX,'v',"dbprate",10.,0.,'s',fp->dim,
       fp->dname1,fp->dsize1,
       fp->dname2,fp->dsize2,
       fp->dname3,fp->dsize3,
       fp->dname4,fp->dsize4);

/* allocate variable slice buffer */

  nvbout = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbout);

/* assign pointers to variables */

  refl = getptr2(hbuff,HBMAX,vbuff,'v','d',reflname,&fp);

/* assume three dimensions, x, y, and z for now */

  if (fp->dim != 3) {
    fprintf(stderr,"radice: analysis for %d-D fields not implemented\n",
       fp->dim);
    exit(1);
  }
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;

/* density is special case */

  rho = getptr2(hbuff,HBMAX,sbuff,'s','d',rhoname,&fp1);
  if ((fp1->dim != 1) || (strcmp(fp1->dname1,fp->dname3) != 0)) {
    fprintf(stderr,"radice: %s not function of z alone\n",rhoname);
    exit(1);
  }

/* the rest are simply assumed to be similar to reflectivity */

  wt = getptr(hbuff,HBMAX,vbuff,'v','d',wtname);
  w = getptr(hbuff,HBMAX,vbuff,'v','d',wname);
  dbn = getptr(hbuff,HBMAX,vbuff,'v','d',"dbn");
  a = getptr(hbuff,HBMAX,vbuff,'v','d',"a");
  aequiv = getptr(hbuff,HBMAX,vbuff,'v','d',"aequiv");
  rhoice = getptr(hbuff,HBMAX,vbuff,'v','d',"rhoice");
  dbqp = getptr(hbuff,HBMAX,vbuff,'v','d',"dbqp");
  dbprate = getptr(hbuff,HBMAX,vbuff,'v','d',"dbprate");

/* write header */

  puthdr(stdout,hbuff,HBMAX);

/* transfer static slice */

  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* loop on variable slice read */

  while (getslice(stdin,nvbin,vbuff) != EOF) {

/* cycle through grid */

    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
        for (iz = 0; iz < nz; iz++) {

/* default values are bad */

          dbn[I(ix,iy,iz)] = bad;
          a[I(ix,iy,iz)] = bad;
          aequiv[I(ix,iy,iz)] = bad;
          rhoice[I(ix,iy,iz)] = bad;
          dbqp[I(ix,iy,iz)] = bad;
          dbprate[I(ix,iy,iz)] = bad;

/* skip rest of loop if bad data or wt too large */

          if (fabs(wt[I(ix,iy,iz)]) > badlim) continue;
          if (fabs(refl[I(ix,iy,iz)]) > badlim) continue;
          if (wt[I(ix,iy,iz)] > WTMAX) continue;

/* get air density */

          rhoa = rho[iz];

/* get reflectivity in non-decibel units */

          zice = pow(10.,refl[I(ix,iy,iz)]/10.);

/* get precipitation particle radius */

          wtt = -wt[I(ix,iy,iz)];
          a[I(ix,iy,iz)] = radius = pow(pow(rhoa,0.5)*wtt/aa,1./alpha);

/* get ice particle density (constant = 3/8g) */

          idens = 0.0383*cd*aa*aa*pow(radius,2.*alpha - 1.);
          rhoice[I(ix,iy,iz)] = idens;

/* get water drop equivalent radius */

          aequiv[I(ix,iy,iz)] = radius*pow(idens,0.33333);

/* compute number of ice particles per m^3 of air -- decibel measure */

          numice = zice/(14.5*idens*idens*pow(radius,6.));
          dbn[I(ix,iy,iz)] = 10.*log10(numice);

/* compute precip mixing ratio in grams per kilogram */

          qprecip = 0.00419*pow(radius,3.)*idens*numice/rhoa;
          dbqp[I(ix,iy,iz)] = 10.*log10(qprecip);

/* compute precip rate in mm/hr */

          if (fabs(ww = w[I(ix,iy,iz)]) < badlim) {
            prate = -3.6*qprecip*rhoa*ww;
            if (prate > 0.) dbprate[I(ix,iy,iz)] = 10.*log10(prate);
            else dbprate[I(ix,iy,iz)] = bad;
          }

/* end of grid loops */

        }
      }
    }

/* write variable slice */

    putslice(stdout,nvbout,vbuff);
  }
  exit(0);
}
