/* radmflux -- This program accepts a cdf radar file processed by radvert
 * on the standard input and computes various horizontally integrated
 * quantities.  Mflux -- horizontally integrated vertical velocity times
 * density rho.  Mfluxu -- updraft part of mflux.  Mfluxd -- downdraft part
 * of mflux.  Dflux -- horizontally integrated divergence.  Dflux2 -- 
 * minus the z derivative of mflux divided by the density rho.
 * Uflux -- the u-component momentum flux.  Vflux -- the v-component
 * momentum flux.
 * The density field is assumed to occur in the static slice.  The names
 * of the divergence and vertical velocity are respectively the first
 * and second command line arguments.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#define I(ix,iy,iz) ix*ny*nz + iy*nz + iz
/* convert vertical mass fluxes to mb*km*km/s */
#define GEENORM 0.098
/* convert detrained mass fluxes to km*km/s */
#define DEENORM 0.001
/* make vertical derivative of vertical mass flux comparable to detrained */
#define DEE2NORM (DEENORM/GEENORM)

char hbin[HBMAX][LINE],hbout[HBMAX][LINE];   /* header buffers */
float *sbuff,*vbuffin,*vbuffout;             /* slice buffer pointers */
long snum,vnumin,vnumout;                    /* slice buffer lengths */
float *u,*v,*div1,*wi,*time;
float *newtime,*mflux,*mfluxu,*mfluxd,*dflux,*dflux2,*uflux,*vflux,*rho;
                                             /* field pointers */
char cline[LINE];                            /* comment line */
int nx,ny,nz;                                /* array sizes */
struct field *fp;                            /* temporary pointer */
int ix,iy,iz;                                /* looping variables */
float dx,dy,dz;                              /* array increments */
float bad,badlim;                            /* bad data value and limit */

main(argc,argv)
int argc;
char *argv[];
{
/* check usage */
  if (argc != 5) {
    fprintf(stderr,"Usage: radmflux div vert_vel x_vel y_vel\n");
    exit(1);
  }
/* get header from standard input */
  gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"radmflux: format must be float\n");
    exit(1);
  }
/* extract bad and badlim from parameter section of header */
  if (seekpar(hbin,HBMAX,"bad",cline) == FAIL) {
    fprintf(stderr,"radmflux: parameter bad missing\n");
    exit(1);
  }
  bad = atof(cline);
  if (seekpar(hbin,HBMAX,"badlim",cline) == FAIL) {
    fprintf(stderr,"radmflux: parameter badlim missing\n");
    exit(1);
  }
  badlim = atof(cline);
/* allocate space for input slices */
  snum = elemcnt(hbin,HBMAX,'s');
  vnumin = elemcnt(hbin,HBMAX,'v');
  sbuff = getbuff(snum);
  vbuffin = getbuff(vnumin);
/* define pointers to div, u, v, wi, and rho */
  div1 = getptr(hbin,HBMAX,vbuffin,'v','d',argv[1]);
  u = getptr(hbin,HBMAX,vbuffin,'v','d',argv[3]);
  v = getptr(hbin,HBMAX,vbuffin,'v','d',argv[4]);
  wi = getptr(hbin,HBMAX,vbuffin,'v','d',argv[2]);
  rho = getptr(hbin,HBMAX,sbuff,'s','d',"rho");
  time = getptr(hbin,HBMAX,vbuffin,'v','d',"time");
/* find array dimensions and grid intervals */
  fp = seekfld(hbin,HBMAX,'v',argv[1]);
  if (fp->dim != 3) {
    fprintf(stderr,"radmflux: input fields don't have 3 dimensions\n");
    exit(1);
  }
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;
  if (seekpar(hbin,HBMAX,"dx",cline) == FAIL) croak();
  dx = atof(cline);
  if (seekpar(hbin,HBMAX,"dy",cline) == FAIL) croak();
  dy = atof(cline);
  if (seekpar(hbin,HBMAX,"dz",cline) == FAIL) croak();
  dz = atof(cline);
/* make new header */
  nullhdr(hbout,HBMAX,"float");
  copycmt(hbout,HBMAX,hbin,HBMAX);
  copypar(hbout,HBMAX,hbin,HBMAX);
  copyfld(hbout,HBMAX,hbin,HBMAX,'s');
  sprintf(cline,"radmflux: %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbout,HBMAX,cline);
  addfld(hbout,HBMAX,'v',"mflux",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"mfluxu",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"mfluxd",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"dflux",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"dflux2",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"uflux",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"vflux",32.,0.,'s',1,"z",nz);
  addfld(hbout,HBMAX,'v',"time",1000.,0.,'l',0);
/* allocate output variable slice buffer */
  vnumout = elemcnt(hbout,HBMAX,'v');
  vbuffout = getbuff(vnumout);
/* make pointers to mass flux fields */
  mflux = getptr(hbout,HBMAX,vbuffout,'v','d',"mflux");
  mfluxu = getptr(hbout,HBMAX,vbuffout,'v','d',"mfluxu");
  mfluxd = getptr(hbout,HBMAX,vbuffout,'v','d',"mfluxd");
  dflux = getptr(hbout,HBMAX,vbuffout,'v','d',"dflux");
  dflux2 = getptr(hbout,HBMAX,vbuffout,'v','d',"dflux2");
  uflux = getptr(hbout,HBMAX,vbuffout,'v','d',"uflux");
  vflux = getptr(hbout,HBMAX,vbuffout,'v','d',"vflux");
  newtime = getptr(hbout,HBMAX,vbuffout,'v','d',"time");
/* write output header */
  puthdr(stdout,hbout,HBMAX);
/* transfer static slice */
  getslice(stdin,snum,sbuff);
  putslice(stdout,snum,sbuff);
/* loop on variable slices */
  while (getslice(stdin,vnumin,vbuffin) != EOF) {
/* transfer time value */
    *newtime = *time;
/* calculate mass fluxes */
    mfluxcal();
/* write variable slice */
    putslice(stdout,vnumout,vbuffout);
  }
  exit(0);
}

/* failure to find desired parameters */
croak()
{
  fprintf(stderr,"radmflux: can't find parameter named dx, dy, or dz\n");
  exit(1);
}

/* compute mass fluxes */
mfluxcal()
{
  int valid,dvalid,validx,validy,jz,kz;
  float sumu,sumd,dsum,sumx,sumy,utemp,vtemp,wtemp,dtemp,delz;
  float umean,vmean;
/* loop on levels */
  for (iz = 0; iz < nz; iz++) {
/* find means of horizontal winds */
    validx = 0;
    validy = 0;
    umean = vmean = 0.;
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
        if (fabs(utemp = u[I(ix,iy,iz)]) < badlim) {
          umean += utemp;
          validx++;
        }
        if (fabs(vtemp = v[I(ix,iy,iz)]) < badlim) {
          vmean += vtemp;
          validy++;
        }
      }
    }
    if (validx) umean = umean/validx;
    if (validy) vmean = vmean/validy;
/* integrate at given level, omitting bad data */
    sumu = 0.;
    sumd = 0.;
    dsum = 0.;
    sumx = 0.;
    sumy = 0.;
    valid = 0;
    dvalid = 0;
    validx = 0;
    validy = 0;
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
        if (fabs(wtemp = wi[I(ix,iy,iz)]) < badlim) {
          if (wtemp >= 0.) sumu += wtemp;
          else sumd += wtemp;
          valid++;
          if (fabs(utemp = u[I(ix,iy,iz)]) < badlim) {
            sumx += (utemp - umean)*wtemp;
            validx++;
          }
          if (fabs(vtemp = v[I(ix,iy,iz)]) < badlim) {
            sumy += (vtemp - vmean)*wtemp;
            validy++;
          }
        }
        if (fabs(dtemp = div1[I(ix,iy,iz)]) < badlim) {
          dsum += dtemp;
          dvalid++;
        }
      }
    }
/* normalize */
    if (valid) {
      mfluxu[iz] = GEENORM*rho[iz]*dx*dy*sumu;
      mfluxd[iz] = GEENORM*rho[iz]*dx*dy*sumd;
      mflux[iz] = mfluxu[iz] + mfluxd[iz];
    }
    else mfluxu[iz] = mfluxd[iz] = mflux[iz] = bad;
    if (dvalid) {
      dflux[iz] = DEENORM*dx*dy*dsum;
    }
    else dflux[iz] = bad;
    if (validx) {
      uflux[iz] = rho[iz]*dx*dy*sumx;
    }
    else uflux[iz] = bad;
    if (validy) {
      vflux[iz] = rho[iz]*dx*dy*sumy;
    }
    else vflux[iz] = bad;
  }
/* compute dflux2 from mflux as consistency check */
  for (iz = 0; iz < nz; iz++) {
/* default is no data */
    dflux2[iz] = bad;
/* try to find centered indices for differentiation with good data */
/* if necessary, try off-centered differences */
    jz = iz + 1;
    kz = iz - 1;
    if (jz >= nz) jz = iz;
    if (kz < 0) kz = iz;
    if (fabs(mflux[jz]) > badlim) jz = iz;
    if (fabs(mflux[jz]) > badlim) continue;
    if (fabs(mflux[kz]) > badlim) kz = iz;
    if (fabs(mflux[kz]) > badlim) continue;
    if (jz <= kz) continue;
/* found usable mflux data -- differentiate */
    delz = (jz - kz)*dz;
    dflux2[iz] = -DEE2NORM*(mflux[jz] - mflux[kz])/(delz*rho[iz]);
  }
}
