/* radsynth -- This program synthesizes cartesian velocity components
 * from radial velocity components associated with two or more radars.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define SCALEHT 8.3                         /* scale height of atmosphere */
#define MAXRADARS 10
#define XTEST 0.001
#define TINY 1.e-10
#define I(ix,iy,iz) (iz + nz*(iy + ny*ix))

/* names assumed in input file */
#define VE "vr"                             /* radial velocity */
#define DZ "dz"                             /* reflectivity */
#define XRAD "xradar"                       /* x-location of radar */
#define YRAD "yradar"                       /* y-location of radar */
#define ZRAD "zradar"                       /* z-location of radar */
#define XRAD2 "xloc"                        /* alternate x-location of radar */
#define YRAD2 "yloc"                        /* alternate y-location of radar */
#define ZRAD2 "zloc"                        /* alternate z-location of radar */

char hbin[HBMAX][LINE],hbout[HBMAX][LINE];  /* header buffers */
char cline[LINE],pline[LINE];               /* general purpose line buffers */
float *sbuff,*vbuff;                        /* slice buffer pointers */
long nsbuffin,nvbuffin,nsbuff,nvbuff;       /* slice buffer sizes */
float *eu,*ev,*ew,*dbz,*u,*v,*w,*wp,*time;  /* potentially new fields */
float *year,*mon,*day,*hour,*min,*sec;      /* time and date output */
float *ve[MAXRADARS];                       /* radial velocity fields */
float *dz[MAXRADARS];                       /* reflectivity fields */
float *radloc[MAXRADARS][3];                /* locations of radars */
float n[MAXRADARS][3];                      /* radial unit vectors */
float wt[MAXRADARS];                        /* radar weights */
float *yearin,*monin,*dayin;                /* date input */
float *hourin,*minin,*secin;                /* time input */
float *xin[MAXRADARS];                      /* input index fields */
float *yin[MAXRADARS];
float *zin[MAXRADARS];
float *x,*y,*z;                             /* output index fields */
int ix,iy,iz;                               /* x, y, and z loop variables */
long index;                                 /* index for accessing 3-d */
int nx,ny,nz;                               /* size of cartesian domain */
int minradars,numradars;                    /* minimum number for synth
                                               and actual number */
struct field *fp;                           /* field structure pointer */
int goop,poop,loop;                         /* general looping variables */
long tcount;                                /* number of rtime elements */
float sum,term,denom;                       /* float temporaries */
float uerror,verror,werror;
float a[3][3],b[3],c[3][3];                 /* linear equation coeffs */
float bad,badlim;                           /* bad data variables */
int goodradars;                             /* number of radars with data */
int nrefl;                                  /* number of good reflectivities */
int surfextr,totalvel;                      /* flags controlling w calc */
float det(),termcalc(),totalcalc(),divcalc();

/* getopt flag stuff */
extern char *optarg;
extern int optind;
int ccc;

main(argc,argv)
int argc;
char *argv[];
{

/* process flags */
  surfextr = totalvel = 0;
  while ((ccc = getopt(argc,argv,"xw")) != -1) {
    if (ccc == 'x') {
      surfextr = 1;
    }
    else if (ccc == 'w') {
      totalvel = 1;
    }
    else {
      fprintf(stderr,"radsynth: unknown flag %c\n",ccc);
      exit(1);
    }
  }

/* check arguments */
  if ((argc - optind < 3) || (argc - optind > MAXRADARS + 1)) {
    fprintf(stderr,"Usage: radsynth [-xw] minradars suffix1 suffix2 ...\n");
    exit(1);
  }
  minradars = atoi(argv[optind]);
  numradars = argc - optind - 1;
  if (minradars < 2) {
    fprintf(stderr,"radsynth: minradars must be two or greater\n");
    exit(1);
  }
  if (minradars > numradars) {
    fprintf(stderr,"radsynth: minradars exceeds actual number specified\n");
    exit(1);
  }

/* get header */
  gethdr(stdin,hbin,HBMAX);

/* check for float format */
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"radsynth: expecting float format\n");
    exit(1);
  }

/* get bad data parameters */
  if (seekpar(hbin,HBMAX,"bad",cline) == FAIL) bad = BAD;
  else bad = atof(cline);
  if (seekpar(hbin,HBMAX,"badlim",cline) == FAIL) badlim = BADLIM;
  else badlim = atof(cline);

/* get dimensions of cartesian space */
  sprintf(cline,"%s%s",VE,argv[1 + optind]);
  if ((fp = seekfld(hbin,HBMAX,'v',cline)) == NULL) {
    fprintf(stderr,"radsynth: vr%s field not found\n",argv[1 + optind]);
    exit(1);
  }
  if (fp->dim != 3) {
    fprintf(stderr,"radsynth: unexpected dimension to %s field\n",cline);
    exit(1);
  }
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;

/* create output header */
  nullhdr(hbout,HBMAX,"float");

/* copy over stuff from old header */
  copycmt(hbout,HBMAX,hbin,HBMAX);
  copypar(hbout,HBMAX,hbin,HBMAX);
  copyfld(hbout,HBMAX,hbin,HBMAX,'s');
  copyfld(hbout,HBMAX,hbin,HBMAX,'v');

/* add comment */
  sprintf(cline,"radsynth:");
  if (surfextr) {
    sprintf(pline," -x");
    strcat(cline,pline);
  }
  if (totalvel) {
    sprintf(pline," -w");
    strcat(cline,pline);
  }
  sprintf(pline," %s\n",argv[optind]);
  strcat(cline,pline);
  addcline(hbout,HBMAX,cline);
  strcpy(cline," ");
  for (loop = 1 + optind; loop < argc; loop++) {
    strcat(cline," ");
    strcat(cline,argv[loop]);
  }
  strcat(cline,"\n");
  addcline(hbout,HBMAX,cline);

/* add x, y, z index parameters */
  newipars("x",argv[1 + optind]);
  newipars("y",argv[1 + optind]);
  newipars("z",argv[1 + optind]);

/* add other parameters */
  sprintf(pline,"%g",(float)SCALEHT);
  addpar(hbout,HBMAX,"SCALEHT",pline);

/* add x, y, z index fields if they don't exist */
  if (seekfld(hbin,HBMAX,'s',"x") == NULL) {
    addfld(hbout,HBMAX,'s',"x",100.,0.,'s',1,"x",nx);
  }
  if (seekfld(hbin,HBMAX,'s',"y") == NULL) {
    addfld(hbout,HBMAX,'s',"y",100.,0.,'s',1,"y",ny);
  }
  if (seekfld(hbin,HBMAX,'s',"z") == NULL) {
    addfld(hbout,HBMAX,'s',"z",100.,0.,'s',1,"z",nz);
  }

/* add eu, ev, ew, u, v, w, dbz, and time fields */
  if (seekfld(hbin,HBMAX,'v',"u") == NULL) {
    addfld(hbout,HBMAX,'v',"u",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"v") == NULL) {
    addfld(hbout,HBMAX,'v',"v",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"w") == NULL) {
    addfld(hbout,HBMAX,'v',"w",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"wp") == NULL) {
    addfld(hbout,HBMAX,'v',"wp",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"eu") == NULL) {
    addfld(hbout,HBMAX,'v',"eu",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"ev") == NULL) {
    addfld(hbout,HBMAX,'v',"ev",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"ew") == NULL) {
    addfld(hbout,HBMAX,'v',"ew",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"dbz") == NULL) {
    addfld(hbout,HBMAX,'v',"dbz",100.,0.,'s',3,"x",nx,"y",ny,"z",nz);
  }
  if (seekfld(hbin,HBMAX,'v',"time") == NULL) {
    addfld(hbout,HBMAX,'v',"time",1000.,0.,'l',0);
  }
  if (seekfld(hbin,HBMAX,'v',"year") == NULL) {
    addfld(hbout,HBMAX,'v',"year",1000.,0.,'l',0);
  }
  if (seekfld(hbin,HBMAX,'v',"mon") == NULL) {
    addfld(hbout,HBMAX,'v',"mon",1000.,0.,'l',0);
  }
  if (seekfld(hbin,HBMAX,'v',"day") == NULL) {
    addfld(hbout,HBMAX,'v',"day",1000.,0.,'l',0);
  }
  if (seekfld(hbin,HBMAX,'v',"hour") == NULL) {
    addfld(hbout,HBMAX,'v',"hour",1000.,0.,'l',0);
  }
  if (seekfld(hbin,HBMAX,'v',"min") == NULL) {
    addfld(hbout,HBMAX,'v',"min",1000.,0.,'l',0);
  }
  if (seekfld(hbin,HBMAX,'v',"sec") == NULL) {
    addfld(hbout,HBMAX,'v',"sec",1000.,0.,'l',0);
  }

/* write output header */
  puthdr(stdout,hbout,HBMAX);

/* allocate slice buffers */
  nsbuffin = elemcnt(hbin,HBMAX,'s');
  nsbuff = elemcnt(hbout,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  nvbuffin = elemcnt(hbin,HBMAX,'v');
  nvbuff = elemcnt(hbout,HBMAX,'v');
  vbuff = getbuff(nvbuff);

/* get static slice */
  getslice(stdin,nsbuffin,sbuff);

/* allocate all input field pointers */
  for (loop = 0; loop < numradars; loop++) {
    sprintf(cline,"%s%s",VE,argv[loop + optind + 1]);
    ve[loop] = getptr2(hbin,HBMAX,vbuff,'v','d',cline,&fp);

    sprintf(cline,"%s%s",DZ,argv[loop + optind + 1]);
    dz[loop] = getptr2(hbin,HBMAX,vbuff,'v','r',cline,&fp);

    xin[loop] = getptr(hbin,HBMAX,sbuff,'s','d',fp->dname1);
    yin[loop] = getptr(hbin,HBMAX,sbuff,'s','d',fp->dname2);
    zin[loop] = getptr(hbin,HBMAX,sbuff,'s','d',fp->dname3);

    if ((fp->dsize1 != nx) || (fp->dsize2 != ny) || (fp->dsize3 != nz)) {
      fprintf(stderr,"radsynth: inconsistent cartesian space dimensions\n");
      exit(1);
    }

/* check for alternate radar location names and locations */
    sprintf(cline,"%s%s",XRAD,argv[loop + optind + 1]);
    if ((radloc[loop][0] = getptr(hbin,HBMAX,sbuff,'s','r',cline)) == NULL) {
      if ((radloc[loop][0] = getptr(hbin,HBMAX,vbuff,'v','r',cline)) == NULL) {
        sprintf(cline,"%s%s",XRAD2,argv[loop + optind + 1]);
        if ((radloc[loop][0] = getptr(hbin,HBMAX,sbuff,'s','r',cline)) ==
             NULL) {
          radloc[loop][0] = getptr(hbin,HBMAX,vbuff,'v','d',cline);
        }
      }
    }

    sprintf(cline,"%s%s",YRAD,argv[loop + optind + 1]);
    if ((radloc[loop][1] = getptr(hbin,HBMAX,sbuff,'s','r',cline)) == NULL) {
      if ((radloc[loop][1] = getptr(hbin,HBMAX,vbuff,'v','r',cline)) == NULL) {
        sprintf(cline,"%s%s",YRAD2,argv[loop + optind + 1]);
        if ((radloc[loop][1] = getptr(hbin,HBMAX,sbuff,'s','r',cline)) ==
             NULL) {
          radloc[loop][1] = getptr(hbin,HBMAX,vbuff,'v','d',cline);
        }
      }
    }

    sprintf(cline,"%s%s",ZRAD,argv[loop + optind + 1]);
    if ((radloc[loop][2] = getptr(hbin,HBMAX,sbuff,'s','r',cline)) == NULL) {
      if ((radloc[loop][2] = getptr(hbin,HBMAX,vbuff,'v','r',cline)) == NULL) {
        sprintf(cline,"%s%s",ZRAD2,argv[loop + optind + 1]);
        if ((radloc[loop][2] = getptr(hbin,HBMAX,sbuff,'s','r',cline)) ==
             NULL) {
          radloc[loop][2] = getptr(hbin,HBMAX,vbuff,'v','d',cline);
        }
      }
    }
  }

/* get date and time from first radar on list */
  sprintf(cline,"year%s",argv[optind + 1]);
  yearin = getptr(hbin,HBMAX,vbuff,'v','d',cline);
  sprintf(cline,"mon%s",argv[optind + 1]);
  monin = getptr(hbin,HBMAX,vbuff,'v','d',cline);
  sprintf(cline,"day%s",argv[optind + 1]);
  dayin = getptr(hbin,HBMAX,vbuff,'v','d',cline);
  sprintf(cline,"hour%s",argv[optind + 1]);
  hourin = getptr(hbin,HBMAX,vbuff,'v','d',cline);
  sprintf(cline,"min%s",argv[optind + 1]);
  minin = getptr(hbin,HBMAX,vbuff,'v','d',cline);
  sprintf(cline,"sec%s",argv[optind + 1]);
  secin = getptr(hbin,HBMAX,vbuff,'v','d',cline);

/* allocate output index field pointers */
  x = getptr(hbout,HBMAX,sbuff,'s','d',"x");
  y = getptr(hbout,HBMAX,sbuff,'s','d',"y");
  z = getptr(hbout,HBMAX,sbuff,'s','d',"z");

/* check consistency of input index fields */
  for (loop = 1; loop < numradars; loop++) {
    for (ix = 0; ix < nx; ix++) {
      if (fabs(xin[loop][ix] - xin[0][ix]) > XTEST) {
        fprintf(stderr,"radsynth: inconsistent index fields\n");
        exit(1);
      }
    }
    for (iy = 0; iy < ny; iy++) {
      if (fabs(yin[loop][iy] - yin[0][iy]) > XTEST) {
        fprintf(stderr,"radsynth: inconsistent index fields\n");
        exit(1);
      }
    }
    for (iz = 0; iz < nz; iz++) {
      if (fabs(zin[loop][iz] - zin[0][iz]) > XTEST) {
        fprintf(stderr,"radsynth: inconsistent index fields\n");
        exit(1);
      }
    }
  }

/* set values of new index fields */
  for (ix = 0; ix < nx; ix++) x[ix] = xin[0][ix];
  for (iy = 0; iy < ny; iy++) y[iy] = yin[0][iy];
  for (iz = 0; iz < nz; iz++) z[iz] = zin[0][iz];

/* write static slice */
  putslice(stdout,nsbuff,sbuff);

/* allocate the output field pointers */
  u = getptr(hbout,HBMAX,vbuff,'v','d',"u");
  v = getptr(hbout,HBMAX,vbuff,'v','d',"v");
  w = getptr(hbout,HBMAX,vbuff,'v','r',"w");
  wp = getptr(hbout,HBMAX,vbuff,'v','r',"wp");
  eu = getptr(hbout,HBMAX,vbuff,'v','d',"eu");
  ev = getptr(hbout,HBMAX,vbuff,'v','d',"ev");
  ew = getptr(hbout,HBMAX,vbuff,'v','r',"ew");
  dbz = getptr(hbout,HBMAX,vbuff,'v','d',"dbz");
  time = getptr(hbout,HBMAX,vbuff,'v','d',"time");
  year = getptr(hbout,HBMAX,vbuff,'v','d',"year");
  mon = getptr(hbout,HBMAX,vbuff,'v','d',"mon");
  day = getptr(hbout,HBMAX,vbuff,'v','d',"day");
  hour = getptr(hbout,HBMAX,vbuff,'v','d',"hour");
  min = getptr(hbout,HBMAX,vbuff,'v','d',"min");
  sec = getptr(hbout,HBMAX,vbuff,'v','d',"sec");

/* read the variable slice (only one assumed) */
  getslice(stdin,nvbuffin,vbuff);

/* loop on cartesian points */
  for (iz = 0; iz < nz; iz++) {
    for (iy = 0; iy < ny; iy++) {
      for (ix = 0; ix < nx; ix++) {
        index = I(ix,iy,iz);

/* extrapolate vr and dz down to surface if desired */
        if (iz == 0 && surfextr) {
          for (loop = 0; loop < numradars; loop++) {
            if (fabs(ve[loop][I(ix,iy,iz)]) > badlim &&
                 fabs(ve[loop][I(ix,iy,iz + 1)]) < badlim) {
              ve[loop][I(ix,iy,iz)] = ve[loop][I(ix,iy,iz + 1)];
            }

            if (dz[loop] && fabs(dz[loop][I(ix,iy,iz)]) > badlim &&
                 fabs(dz[loop][I(ix,iy,iz + 1)]) < badlim) {
              dz[loop][I(ix,iy,iz)] = dz[loop][I(ix,iy,iz + 1)];
            }
          }
        }

/* compute average reflectivity from all radars -- convert from log to linear */
        nrefl = 0;
        sum = 0.;
        for (loop = 0; loop < numradars; loop++) {
          if (dz[loop] && fabs(dz[loop][index]) < badlim) {
            sum += pow(10.,dz[loop][index]/10.);
            nrefl++;
          }
        }
        if (nrefl == 0) dbz[index] = bad;
        else dbz[index] = sum/nrefl;
      }
    }
  }

/* loop on cartesian points yet again (separate loops to reduce swapping) */
  for (iz = 0; iz < nz; iz++) {
    for (iy = 0; iy < ny; iy++) {
      for (ix = 0; ix < nx; ix++) {
        index = I(ix,iy,iz);

/* see if there are minimum good data */
        goodradars = 0;
        for (loop = 0; loop < numradars; loop++) {
          if (fabs(ve[loop][index]) < badlim) goodradars++;
        }

/* continue only if sufficient good data */
        u[index] = v[index] = w[index] = eu[index] = ev[index] =
             ew[index] = bad;
        if (goodradars >= minradars) {

/* compute unit vectors pointing in ray directions and radar weights*/
          for (loop = 0; loop < numradars; loop++) {
            sum = 0.;
            term = n[loop][0] = x[ix] - *radloc[loop][0];
            sum += term*term;
            term = n[loop][1] = y[iy] - *radloc[loop][1];
            sum += term*term;
            term = n[loop][2] = z[iz] - *radloc[loop][2];
            sum += term*term;

/* triple doppler analysis assumes unit weights */
            if (minradars > 2) wt[loop] = 1.;

/* double doppler analysis weights by square of distance to radar */
            else wt[loop] = sum;
            sum = sqrt(sum);
            for (poop = 0; poop < 3; poop++) n[loop][poop] /= sum;
          }

/* compute the correlations */
          for (goop = 0; goop < 3; goop++) {
            b[goop] = 0.;
            for (loop = 0; loop < numradars; loop++) {
              if (fabs(ve[loop][index]) < badlim)
                   b[goop] += n[loop][goop]*ve[loop][index]*wt[loop];
            }
            for (poop = 0; poop < 3; poop++) {
              a[goop][poop] = 0.;
              for (loop = 0; loop < numradars; loop++) {
                if (fabs(ve[loop][index]) < badlim)
                     a[goop][poop] += n[loop][goop]*n[loop][poop]*wt[loop];
              }
            }
          }

/* dual doppler analysis */
          if (minradars == 2) {

/* compute w and wp at this level */
            if (iz == 0) {

/* zero w at lowest level and compute wp there */
              w[index] = 0.;
              wp[index] = w[index] + termcalc(ix,iy,iz);
            }
            else {

/* do computation at higher level -- only calc w here if totalvel selected */
              if (totalvel) w[index] = totalcalc(ix,iy,iz);
              else w[index] = 0.;

/* compute wp assuming a bad w is a zero w */
              if (fabs(w[index]) < badlim) {
                wp[index] = w[index] + termcalc(ix,iy,iz);
              }
              else {
                wp[index] = termcalc(ix,iy,iz);
              }
            }

/* if wp bad, kill everything */
            if (fabs(wp[index]) > badlim) {
              u[index] = v[index] = bad;
            }
            else {

/* otherwise, add wp compensation to b */
              b[0] -= wp[index]*a[0][2];
              b[1] -= wp[index]*a[1][2];

/* compute the denominator */
              denom = a[0][0]*a[1][1] - a[0][1]*a[1][0];

/* check on linear independence */
              if (fabs(denom) < TINY) {
                u[index] = v[index] = bad;
              }

/* if ok, compute the velocities */
              else {
                u[index] = (a[1][1]*b[0] - a[0][1]*b[1])/denom;
                v[index] = -(a[1][0]*b[0] - a[0][0]*b[1])/denom;
              }

/* compute the error amplification factors */
              uerror = verror = 0.;
              for (loop = 0; loop < numradars; loop++) {
                if (fabs(ve[loop][index]) < badlim) {
                  term = (n[loop][0]*a[1][1] - n[loop][1]*a[0][1])*wt[loop];
                  uerror += term*term;
                  term = (n[loop][0]*a[1][0] - n[loop][1]*a[0][0])*wt[loop];
                  verror += term*term;
                }
              }
              eu[index] = sqrt(uerror)/denom;
              ev[index] = sqrt(verror)/denom;
            }
          }

/* multiple (>2) doppler analysis */
          else {

/* compute the denominator */
            denom = det(a);

/* set things to bad if denom is tiny */
            if (fabs(denom) < TINY) {
              u[index] = v[index] = wp[index] = bad;
            }
            else {

/* compute the velocities */
              setarray(c,a);
              for (goop = 0; goop < 3; goop++) c[0][goop] = b[goop];
              u[index] = det(c)/denom;
              setarray(c,a);
              for (goop = 0; goop < 3; goop++) c[1][goop] = b[goop];
              v[index] = det(c)/denom;
              setarray(c,a);
              for (goop = 0; goop < 3; goop++) c[2][goop] = b[goop];
              wp[index] = det(c)/denom;

/* compute the error amplification factors */
              uerror = verror = werror = 0.;
              for (loop = 0; loop < numradars; loop++) {
                if (fabs(ve[loop][index]) < badlim) {
                  setarray(c,a);
                  for (goop = 0; goop < 3; goop++)
                       c[0][goop] = n[loop][goop]*wt[loop];
                  term = det(c);
                  uerror += term*term;
                  setarray(c,a);
                  for (goop = 0; goop < 3; goop++)
                       c[1][goop] = n[loop][goop]*wt[loop];
                  term = det(c);
                  verror += term*term;
                  setarray(c,a);
                  for (goop = 0; goop < 3; goop++)
                       c[2][goop] = n[loop][goop]*wt[loop];
                  term = det(c);
                  werror += term*term;
                }
              }
              eu[index] = sqrt(uerror)/denom;
              ev[index] = sqrt(verror)/denom;
              ew[index] = sqrt(werror)/denom;
            }
          }
        }
      }
    }
  }

/* final loop */
  for (iz = 0; iz < nz; iz++) {
    for (iy = 0; iy < ny; iy++) {
      for (ix = 0; ix < nx; ix++) {
        index = I(ix,iy,iz);

/* recompute the vertical air velocity */
        if (iz == 0) {
          w[index] = 0.;
        }
        else {
          w[index] = totalcalc(ix,iy,iz);
        }

/* recompute particle velocity aloft if dual doppler */
        if (minradars == 2) {
          if (fabs(w[index]) < badlim) {
            wp[index] = w[index] + termcalc(ix,iy,iz);
          }
          else wp[index] = bad;
        }

/* decibelize the reflectivity */
        if (fabs(dbz[index]) < badlim) dbz[index] = 10.*log10(dbz[index]);
      }
    }
  }

/* transfer date info */
  *year = *yearin;
  *mon = *monin;
  *day = *dayin;
  *hour = *hourin;
  *min = *minin;
  *sec = *secin;

/* compute time in ks */
  *time = (*sec + *min*60. + *hour*3600.)/1000.;

/* done -- write variable slice and quit */
  putslice(stdout,nvbuff,vbuff);
  exit(0);
}

/* compute 3-d determinant */
float det(a)
float a[3][3];
{
  return(a[0][0]*a[1][1]*a[2][2] + a[0][1]*a[1][2]*a[2][0] +
       a[1][0]*a[2][1]*a[0][2] - a[2][0]*a[1][1]*a[0][2] -
       a[0][1]*a[1][0]*a[2][2] - a[2][1]*a[1][2]*a[0][0]);
}

/* copy 3 by 3 matrix */
setarray(c,a)
float c[9],a[9];
{
  int i;
  for (i = 0; i < 9; i++) c[i] = a[i];
}

/* add new index parameters, with values matching old ones */
newipars(basename,suffix)
char *basename,*suffix;
{
  char new[LINE],old[LINE];

/* compute new basename0 parameter */
  sprintf(new,"%s0",basename);

/* add it if it doesn't exist */
  if (seekpar(hbout,HBMAX,new,cline) == FAIL) {
    sprintf(old,"%s%s0",basename,suffix);
    if (seekpar(hbout,HBMAX,old,cline) == FAIL) {
      fprintf(stderr,"radsynth: can't find parameter %s\n",old);
      exit(1);
    }
    addpar(hbout,HBMAX,new,cline);
  }

/* compute new dbasename parameter */
  sprintf(new,"d%s",basename);

/* add it if it doesn't exist */
  if (seekpar(hbout,HBMAX,new,cline) == FAIL) {
    sprintf(old,"d%s%s",basename,suffix);
    if (seekpar(hbout,HBMAX,old,cline) == FAIL) {
      fprintf(stderr,"radsynth: can't find parameter %s\n",old);
      exit(1);
    }
    addpar(hbout,HBMAX,new,cline);
  }
}

/* termcalc -- Compute (downward) terminal velocity from reflectivity
 * using Joss-Waldvogel (1970) relation.  Always returns a non-bad value.
 */
float termcalc(ix,iy,iz)
int ix,iy,iz;                      /* location of desired calc */
{
  float zzz;
  int jz;

/* if we don't find a good reflectivity at this level, look above */
  for (jz = iz; jz < nz; jz++) {
    zzz = dbz[I(ix,iy,jz)];
    if (fabs(zzz) < badlim) return(-2.6*pow(zzz,0.107));
  }
  return(-2.);
}

/* totalcalc -- Compute the air velocity at a level using mass continuity
 * and the vertical velocity at the next lowest level.
 */
float totalcalc(ix,iy,iz)
int ix,iy,iz;
{
  float div1,div2,div,wstar,wmid;
  int nlevels,kz;

/* at lowest level just return zero */
  if (iz == 0) return(0.);

/* if lower level w is bad, return bad */
  kz = iz - 1;
  if (fabs(wstar = w[I(ix,iy,kz)]) > badlim) return(bad);

/* otherwise use divergence at two levels to calculate w at this level */
  else {
    div1 = divcalc(ix,iy,kz);
    div2 = divcalc(ix,iy,iz);
    div = 0.;
    nlevels = 0;
    if (fabs(div1) < badlim) {
      div += div1;
      nlevels++;
    }
    if (fabs(div2) < badlim) {
      div += div2;
      nlevels++;
    }
    if (nlevels) {
      div /= (float)nlevels;
      wmid = wstar + .5*(wstar/SCALEHT - div)*(z[iz] - z[kz]);
      return(wstar + (wmid/SCALEHT - div)*(z[iz] - z[kz]));
    }
    else {
      return(bad);
    }
  }
}

/* divcalc -- calculate divergence at a specified point */
float divcalc(ix,iy,iz)
int ix,iy,iz;
{
  float divx,divy;
  int jx,jy,kx,ky;
  if (ix == 0 || fabs(u[I(ix - 1,iy,iz)]) > badlim) kx = ix;
  else kx = ix - 1;
  if (ix == nx - 1 || fabs(u[I(ix + 1,iy,iz)]) > badlim) jx = ix;
  else jx = ix + 1;
  if (jx > kx && fabs(u[I(jx,iy,iz)]) < badlim &&
       fabs(u[I(kx,iy,iz)]) < badlim) {
  divx = (u[I(jx,iy,iz)] - u[I(kx,iy,iz)])/(x[jx] - x[kx]);
  }
  else divx = bad;
  if (iy == 0 || fabs(v[I(ix,iy - 1,iz)]) > badlim) ky = iy;
  else ky = iy - 1;
  if (iy == ny - 1 || fabs(v[I(ix,iy + 1,iz)]) > badlim) jy = iy;
  else jy = iy + 1;
  if (jy > ky && fabs(v[I(ix,jy,iz)]) < badlim &&
       fabs(v[I(ix,ky,iz)]) < badlim) {
    divy = (v[I(ix,jy,iz)] - v[I(ix,ky,iz)])/(y[jy] - y[ky]);
  }
  else divy = bad;
  if (fabs(divx) < badlim && fabs(divy) < badlim) return(divx + divy);
  else return(bad);
}
