/* radtop -- This program creates 6 2-D fields from radar data files:
 * 1) Top level of reflectivity, topz; 2) Top level of vertical velocity,
 * topw; 3) Vertical velocity at that top level, topwval; 4) Reflectivity
 * at that top level, topzval; 5) Uppermost level at which divergence
 * is defined, topd; 6) Value of divergence at that level, topdval.
 * Input arguments are the field names of the reflectivity, the
 * divergence, and the vertical velocity.  The
 * new fields are output into an new common data format file.  The index fields
 * x, y, and z are assumed to exist.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#define I(ix,iy,iz) ix*ny*nz + iy*nz + iz
#define J(ix,iy) ix*ny + iy

char hbuff[HBMAX][LINE],hbout[HBMAX][LINE];
char cline[LINE];
long nsb,nvb0,nvb;
float *sb,*vb0,*vb,*wval,*zval,*dval;
float *topz,*topw,*topwval,*topzval,*topd,*topdval,*z;
float bad,badlim;           /* bad data value and limit */
int ix,iy,iz,nx,ny,nz;
int loop,nscalar;
long oldsptr[HBMAX],newsptr[HBMAX];
struct field *fp;

main(argc,argv)
int argc;
char *argv[];
{
/* check usage */
  if (argc != 4) {
    fprintf(stderr,"Usage: radtop reflectivity vert_vel divergence\n");
    exit(1);
  }
/* get header */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"radtop: must have float format\n");
    exit(1);
  }
  nsb = elemcnt(hbuff,HBMAX,'s');
  nvb0 = elemcnt(hbuff,HBMAX,'v');
/* extract bad and badlim from parameter section of header */
  if (seekpar(hbuff,HBMAX,"bad",cline) == FAIL) {
    fprintf(stderr,"radtop: parameter bad missing\n");
    exit(1);
  }
  bad = atof(cline);
  if (seekpar(hbuff,HBMAX,"badlim",cline) == FAIL) {
    fprintf(stderr,"radtop: parameter badlim missing\n");
    exit(1);
  }
  badlim = atof(cline);
/* get sizes of x, y, and z dimensions */
  fp = seekfld(hbuff,HBMAX,'s',"x");
  if (fp == NULL) {
    fprintf(stderr,"radtop: no x field\n");
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,"x") != 0)) {
    fprintf(stderr,"radtop: x not index field\n");
    exit(1);
  }
  nx = fp->dsize1;
  fp = seekfld(hbuff,HBMAX,'s',"y");
  if (fp == NULL) {
    fprintf(stderr,"radtop: no y field\n");
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,"y") != 0)) {
    fprintf(stderr,"radtop: y not index field\n");
    exit(1);
  }
  ny = fp->dsize1;
  fp = seekfld(hbuff,HBMAX,'s',"z");
  if (fp == NULL) {
    fprintf(stderr,"radtop: no z field\n");
    exit(1);
  }
  if ((fp->dim != 1) || (strcmp(fp->dname1,"z") != 0)) {
    fprintf(stderr,"radtop: z not index field\n");
    exit(1);
  }
  nz = fp->dsize1;
/* create new header */
  nullhdr(hbout,HBMAX,"float");
  copycmt(hbout,HBMAX,hbuff,HBMAX);
  copypar(hbout,HBMAX,hbuff,HBMAX);
  copyfld(hbout,HBMAX,hbuff,HBMAX,'s');
/* copy all scalar fields */
  nscalar = loop = 0;
  while ((fp = getfld(hbuff,HBMAX,'v',loop++)) != NULL) {
    if (fp->dim == 0) {
      oldsptr[nscalar] = fp->ptr;
      addfld(hbout,HBMAX,'v',fp->fname,fp->smul,fp->sadd,fp->bytes,0);
      fp = seekfld(hbout,HBMAX,'v',fp->fname);
      newsptr[nscalar++] = fp->ptr;
    }
  }
/* add new fields to header */
  addfld(hbout,HBMAX,'v',"topz",32.,0.,'s',2,"x",nx,"y",ny);
  addfld(hbout,HBMAX,'v',"topw",32.,0.,'s',2,"x",nx,"y",ny);
  addfld(hbout,HBMAX,'v',"topd",32.,0.,'s',2,"x",nx,"y",ny);
  addfld(hbout,HBMAX,'v',"topwval",32.,0.,'s',2,"x",nx,"y",ny);
  addfld(hbout,HBMAX,'v',"topzval",32.,0.,'s',2,"x",nx,"y",ny);
  addfld(hbout,HBMAX,'v',"topdval",32.,0.,'s',2,"x",nx,"y",ny);
/* add comment line to header */
  sprintf(cline,"radtop: %s %s %s\n",argv[1],argv[2],argv[3]);
  addcline(hbout,HBMAX,cline);
/* find size of new variable slice */
  nvb = elemcnt(hbout,HBMAX,'v');
/* allocate space */
  if ((sb = (float*)malloc(sizeof(float)*nsb)) == NULL) {
    fprintf(stderr,"radtop: can't get enough memory\n");
    exit(1);
  }
  if ((vb0 = (float*)malloc(sizeof(float)*nvb0)) == NULL) {
    fprintf(stderr,"radtop: can't get enough memory\n");
    exit(1);
  }
  if ((vb = (float*)malloc(sizeof(float)*nvb)) == NULL) {
    fprintf(stderr,"radtop: can't get enough memory\n");
    exit(1);
  }
/* get static slice */
  getslice(stdin,nsb,sb);
/* set up field pointers */
/* reflectivity */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[1])) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n",argv[1]);
    exit(1);
  }
  zval = vb0 + fp->ptr;
/* vertical velocity */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[2])) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n",argv[2]);
    exit(1);
  }
  wval = vb0 + fp->ptr;
/* divergence */
  if ((fp = seekfld(hbuff,HBMAX,'v',argv[3])) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n",argv[3]);
    exit(1);
  }
  dval = vb0 + fp->ptr;
/* height of highest defined reflectivity */
  if ((fp = seekfld(hbout,HBMAX,'v',"topz")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","topz");
    exit(1);
  }
  topz = vb + fp->ptr;
/* height of highest defined vertical velocity */
  if ((fp = seekfld(hbout,HBMAX,'v',"topw")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","topw");
    exit(1);
  }
  topw = vb + fp->ptr;
/* height of highest defined divergence */
  if ((fp = seekfld(hbout,HBMAX,'v',"topd")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","topd");
    exit(1);
  }
  topd = vb + fp->ptr;
/* vertical velocity at highest point defined */
  if ((fp = seekfld(hbout,HBMAX,'v',"topwval")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","topwval");
    exit(1);
  }
  topwval = vb + fp->ptr;
/* reflectivity at highest point defined */
  if ((fp = seekfld(hbout,HBMAX,'v',"topzval")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","topzval");
    exit(1);
  }
  topzval = vb + fp->ptr;
/* divergence at highest point defined */
  if ((fp = seekfld(hbout,HBMAX,'v',"topdval")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","topdval");
    exit(1);
  }
  topdval = vb + fp->ptr;
/* z index field */
  if ((fp = seekfld(hbuff,HBMAX,'s',"z")) == NULL) {
    fprintf(stderr,"radtop: %s field doesn't exist\n","z");
    exit(1);
  }
  z = sb + fp->ptr;
/* put header */
  puthdr(stdout,hbout,HBMAX);
/* put static slice */
  putslice(stdout,nsb,sb);
/* loop on variable slices */
  while(getslice(stdin,nvb0,vb0) != EOF) {
/* transfer scalars */
    for (loop = 0; loop < nscalar; loop++) {
      vb[newsptr[loop]] = vb0[oldsptr[loop]];
    }
/* loop on horizontal position */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
/* default values -- bad */
        topz[J(ix,iy)] = topw[J(ix,iy)] = topd[J(ix,iy)] =
        topzval[J(ix,iy)] = topwval[J(ix,iy)] = topdval[J(ix,iy)] = bad;
/* find topz and topzval by starting down from top */
        for (iz = nz - 1; iz >= 0; iz--) {
          if (fabs(zval[I(ix,iy,iz)]) < badlim) {
            topz[J(ix,iy)] = z[iz];
            topzval[J(ix,iy)] = zval[I(ix,iy,iz)];
            break;
          }
        }
/* find topw and topwval the same way */
        for (iz = nz - 1; iz >= 0; iz--) {
          if (fabs(wval[I(ix,iy,iz)]) < badlim) {
            topw[J(ix,iy)] = z[iz];
            topwval[J(ix,iy)] = wval[I(ix,iy,iz)];
            break;
          }
        }
/* ditto for divergence */
        for (iz = nz - 1; iz >= 0; iz--) {
          if (fabs(dval[I(ix,iy,iz)]) < badlim) {
            topd[J(ix,iy)] = z[iz];
            topdval[J(ix,iy)] = dval[I(ix,iy,iz)];
            break;
          }
        }
      }
    }
/* write variable slice */
    putslice(stdout,nvb,vb);
  }
  exit(0);
}
