/* radvert.c -- This filter computes various quantities having to do
 * with vertical motion on radar data, namely, horizontal divergence,
 * continuity-generated vertical velocity, and particle terminal
 * velocity.  Also computed are vertical profiles of
 * density (rho), as well as the pressure (pres)
 * and the scale height (zs).
 * These are computed on a constant lapse rate hypothesis.
 * Vertical wind is only defined where the divergence is not missing.
 * Expected on the standard input is a cdfcedric file
 * containing horizontal velocity components, and optionally, vertical
 * particle velocity.  These fields are passed to the standard
 * output, along with the new fields, as a common data format file.
 * If the vertical particle velocity doesn't exist, the vertical
 * terminal velocity isn't computed.  Likewise, if the divergence,
 * integrated vertical velocity, or terminal velocity
 * fields already exist, they aren't recomputed.
 * The command arguments are the names of the various fields:
 * 1 = x velocity; 2 = y velocity; 3 = z particle velocity;
 * 4 = horizontal divergence; 5 = continuity-generated vertical
 * velocity; 6 = particle terminal velocity.
 * Optional 7th and 8th arguments are the name of the reflectivity
 * field and a cloud physics generated vertical velocity.
 * Normally, only one variable slice is expected.  However, this
 * program will loop until all variable slices are exhausted.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#define PMSL 100.0      /* sea level pressure (kpa) */
#define TMSL 300.0      /* sea level temperature (K) */
#define GAMMA 7.0       /* temperature lapse rate (K/km) */
#define GEE 9.8         /* acceleration of gravity (m/s^2) */
#define RGAS 287.6      /* gas constant for air (J/kg/K) */
#define SCALE 32.0      /* integer packing scale factor */
#define WTCLDTOP 2.0    /* default terminal velocity at cloud top */
#define WCPA 31.8       /* constants for reflectivity-terminal velocity */
#define WCPB 0.918      /* relationship                                 */
#define WTMIN 2.0       /* assumed terminal velocity for undefined refl */
#define WTVAR1 1.0      /* top terminal velocity for no divergence */
#define WTVAR2 5.0      /* max terminal velocity for max divergence */
#define I(ix,iy,iz) (ix)*ny*nz + (iy)*nz + (iz)  /* array index macro */

char hbin[HBMAX][LINE],hbout[HBMAX][LINE];  /* header buffers */
int nhbin;              /* number of lines in input header buffer */
float *sbuff,*vbuff;    /* slice buffers */
long nsbuffin,nsbuffot,nvbuffin,nvbuffot;   /* sizes of slice buffers */
float *u,*v,*w,*div1,*wi,*wterm,*dbz,*wcp;  /* field pointers */
float *rho,*pres,*z,*zs;
float z1,mu,rhomsl;     /* intermediates for pressure and density calcs */
float wtcldtop;         /* cloud top terminal velocity */
int wtmod = 0;          /* flag indicating cloud top wt modified */
int varflag = 0;        /* flag indicating div-dependent top wt to be used */
float dvalue;           /* temporary for variable wt-top calculation */
float divmax;           /* maximum divergence value accepted */
int i,ix,iy,iz,nx,ny,nz;   /* looping variables and array limits */
float dx,dy,dz;         /* grid intervals */
int widone =0;          /* field status flags */
int divdone = 0;
int wtermdon = 0;
int wno = 0;
char alpha[LINE];       /* line buffer */
char d1[LINE],d2[LINE],d3[LINE];  /* dimension names */
struct field *fp;       /* field structure pointer */
float bad,badlim;       /* bad data stuff */
float *getptr();

main(argc,argv)
int argc;
char *argv[];
{
/* check arguments */
  if (argc > 1 && strcmp(argv[1],"-v") == 0) {
    for (i = 2; i < argc; i++) argv[i - 1] = argv[i];
    argc--;
    varflag = 1;
  }
  if (argc > 2 && strcmp(argv[1],"-w") == 0) {
    wtcldtop = atof(argv[2]);
    for (i = 3; i < argc; i++) argv[i - 2] = argv[i];
    argc -= 2;
    wtmod = 1;
  }
  else wtcldtop = WTCLDTOP;
  if ((argc != 7) && (argc != 9)) {
    fprintf(stderr,
         "Usage: radvert [-w wtval]|[-v] u v w div wi wterm [dbz wcp]\n");
    exit(1);
  }
/* get header from standard input */
  nhbin = gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"radvert: input file must be float format\n");
    exit(1);
  }
/* extract bad and badlim from parameter section of header */
  if (seekpar(hbin,HBMAX,"bad",alpha) == FAIL) {
    fprintf(stderr,"radvert: parameter bad missing\n");
    exit(1);
  }
  bad = atof(alpha);
  if (seekpar(hbin,HBMAX,"badlim",alpha) == FAIL) {
    fprintf(stderr,"radvert: parameter badlim missing\n");
    exit(1);
  }
  badlim = atof(alpha);
/* check for existing and non-existent fields */
/* x-velocity */
  if ((fp = seekfld(hbin,HBMAX,'v',argv[1])) == NULL) {
    fprintf(stderr,"radvert: no %s field\n",argv[1]);
    exit(1);
  }
/* get dimension names and sizes */
  if (fp->dim != 3) {
    fprintf(stderr,"radvert: incorrect field dimensions = %d\n",fp->dim);
    exit(1);
  }
/* we don't do extensive error checking to see if its really a cedric file */
  strcpy(d1,fp->dname1);
  strcpy(d2,fp->dname2);
  strcpy(d3,fp->dname3);
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;
/* y-velocity */
  if (seekfld(hbin,HBMAX,'v',argv[2]) == NULL) {
    fprintf(stderr,"radvert: no %s field\n",argv[2]);
    exit(1);
  }
/* z-particle velocity */
  if (seekfld(hbin,HBMAX,'v',argv[3]) == NULL) {
    fprintf(stderr,"radvert warning: no %s field\n",argv[3]);
    wno = 1;
  }
/* optional reflectivity */
  if ((argc == 9) && (seekfld(hbin,HBMAX,'v',argv[7]) == NULL)) {
    fprintf(stderr,"radvert: no %s field\n",argv[7]);
    exit(1);
  }
/* div */
  if (seekfld(hbin,HBMAX,'v',argv[4]) != NULL) {
    fprintf(stderr,"radvert warning: %s already exists\n",argv[4]);
    divdone = 1;
  }
/* wi */
  if (seekfld(hbin,HBMAX,'v',argv[5]) != NULL) {
    fprintf(stderr,"radvert warning: %s already exists\n",argv[5]);
    widone = 1;
  }
/* wterm */
  if (seekfld(hbin,HBMAX,'v',argv[6]) != NULL) {
    fprintf(stderr,"radvert warning: %s already exists\n",argv[6]);
    wtermdon = 1;
  }
/* wcp */
  if ((argc == 9) && (seekfld(hbin,HBMAX,'v',argv[8]) != NULL)) {
    fprintf(stderr,"radvert: %s already exists\n",argv[8]);
    exit(1);
  }
/* get deltas */
  if (seekpar(hbin,HBMAX,"dx",alpha) == FAIL) {
    fprintf(stderr,"radvert: can't find a needed parameter\n");
    exit(1);
  }
  dx = atof(alpha);
  if (seekpar(hbin,HBMAX,"dy",alpha) == FAIL) {
    fprintf(stderr,"radvert: can't find a needed parameter\n");
    exit(1);
  }
  dy = atof(alpha);
  if (seekpar(hbin,HBMAX,"dz",alpha) == FAIL) {
    fprintf(stderr,"radvert: can't find a needed parameter\n");
    exit(1);
  }
  dz = atof(alpha);
/* make new header */
/*start by copying over old header */
  for (i = 0; i < nhbin; i++) strcpy(hbout[i],hbin[i]);
/* add divergence */
  if (divdone == 0)
    addfld(hbout,HBMAX,'v',argv[4],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
/* add continuity-generated vertical velocity */
  if (widone == 0)
    addfld(hbout,HBMAX,'v',argv[5],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
/* add terminal velocity of hydrometeors if it can be computed */
  if ((wtermdon == 0) && (wno == 0))
    addfld(hbout,HBMAX,'v',argv[6],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
/* optionally add wcp */
  if (argc == 9)
    addfld(hbout,HBMAX,'v',argv[8],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
/* add density to static slice */
  if (seekfld(hbout,HBMAX,'s',"rho") == NULL)
    addfld(hbout,HBMAX,'s',"rho",10000.,0.,'s',1,d3,nz);
/* add pressure to static slice */
  if (seekfld(hbout,HBMAX,'s',"pres") == NULL)
    addfld(hbout,HBMAX,'s',"pres",10.,0.,'s',1,d3,nz);
/* add density scale height to static slice */
  if (seekfld(hbout,HBMAX,'s',"zs") == NULL)
    addfld(hbout,HBMAX,'s',"zs",100.,0.,'s',1,d3,nz);
/* add comments */
  if (wtmod) {
    sprintf(alpha,"radvert: -h %.1f %s %s %s %s %s %s\n",wtcldtop,
         argv[1],argv[2],argv[3],argv[4],argv[5],argv[6]);
  }
  else {
    sprintf(alpha,"radvert: %s %s %s %s %s %s\n",argv[1],argv[2],argv[3],
         argv[4],argv[5],argv[6]);
  }
  addcline(hbout,HBMAX,alpha);
  if (argc == 9) {
    sprintf(alpha,"     %s %s\n",argv[7],argv[8]);
    addcline(hbout,HBMAX,alpha);
  }
  if (wno) addcline(hbout,HBMAX,"zero ubc\n");
  else if (varflag) addcline(hbout,HBMAX,"variable particle ubc\n");
  else addcline(hbout,HBMAX,"particle ubc\n");
/* allocate space for slices */
  nsbuffin = elemcnt(hbin,HBMAX,'s');
  nsbuffot = elemcnt(hbout,HBMAX,'s');
  if ((sbuff = (float *) malloc(sizeof(float)*nsbuffot)) == NULL) {
    fprintf(stderr,"radvert: can't get enough memory\n");
    exit(1);
  }
  nvbuffot = elemcnt(hbout,HBMAX,'v');
  nvbuffin = elemcnt(hbin,HBMAX,'v');
  if ((vbuff = (float *) malloc(sizeof(float)*nvbuffot)) == NULL) {
    fprintf(stderr,"radvert: can't get enough memory\n");
    exit(1);
  }
/* set pointers for fields -- NULL if it doesn't exist */
  u = getptr(argv[1]);
  v = getptr(argv[2]);
  w = getptr(argv[3]);
  div1 = getptr(argv[4]);
  wi = getptr(argv[5]);
  wterm = getptr(argv[6]);
  if (argc == 9) {
    dbz = getptr(argv[7]);
    wcp = getptr(argv[8]);
  }
  else dbz = wcp = NULL;
  rho = getptr("rho");
  pres = getptr("pres");
  z = getptr("z");
  zs = getptr("zs");
/* put new header */
  puthdr(stdout,hbout,HBMAX);
/* get static slice */
  getslice(stdin,nsbuffin,sbuff);
/* fill density, pressure, and scale height fields if they exist */
  if ((rho != NULL) && (pres != NULL) && (zs != NULL)) {
    z1 = TMSL/GAMMA;
    mu = GEE/(RGAS*GAMMA*0.001);
    rhomsl = 1000.0*PMSL/(RGAS*TMSL);
    for (iz = 0; iz < nz; iz++) {
      pres[iz] = PMSL*pow(1. - z[iz]/z1,mu);
      rho[iz] = rhomsl*pow(1. - z[iz]/z1,mu - 1.);
      zs[iz] = (z1 - z[iz])/(mu - 1.);
    }
  }
  else {
    fprintf(stderr,"radvert: panic -- rho, pres, and zs don't exist!\n");
    exit(1);
  }
/* put static slice */
  putslice(stdout,nsbuffot,sbuff);
/* get variable slice */
  while (getslice(stdin,nvbuffin,vbuff) != EOF) {
/* do computations */
/* compute raw divergence field */
    if (divdone == 0) {
      divcalc();
    }
/* compute integrated vertical velocity */
    if (widone == 0) {
      wicalc();
    }
/* compute particle terminal velocity */
    if ((wno == 0) && (wtermdon == 0)) {
      wtermcal();
    }
/* optional cloud physics vertical velocity -- overwrites wterm with wtermcp */
    if (argc == 9) wcpcalc();
/* put variable slice */
    putslice(stdout,nvbuffot,vbuff);
  }
}

/* set float pointer in vbuff or sbuff */
float *getptr(name)
char *name;
{
  fp = seekfld(hbout,HBMAX,'v',name);
  if (fp == NULL) {
    fp = seekfld(hbout,HBMAX,'s',name);
    if (fp == NULL) return(NULL);
    else return(sbuff + fp->ptr);
  }
  else return(vbuff + fp->ptr);
}

/* compute divergence field */
divcalc()
{
  int jx,kx,jy,ky,jz;
  float delx,dely,tempdiv,u1,u2,v1,v2;
/* loop on z */
  for (iz = 0; iz < nz; iz++) {
/* loop on x and y */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
/* set up for centered differences except at boundaries */
        jx = ix + 1;
        if (jx >= nx) jx = nx - 1;
        kx = ix - 1;
        if (kx < 0) kx = 0;
        jy = iy + 1;
        if (jy >= ny) jy = ny - 1;
        ky = iy - 1;
        if (ky < 0) ky = 0;
/* be a pessimist -- default is missing data */
        div1[I(ix,iy,iz)] = bad;
/* check to see if data defining divergence at this point are ok */
/* go to off-centered differences if necessary */
        if (fabs((u2 = u[I(jx,iy,iz)])) > badlim) {
          jx = ix;
          if (fabs((u2 = u[I(jx,iy,iz)])) > badlim) continue;
        }
        if (fabs((u1 = u[I(kx,iy,iz)])) > badlim) {
          kx = ix;
          if (fabs((u1 = u[I(kx,iy,iz)])) > badlim) continue;
        }
        if (fabs((v2 = v[I(ix,jy,iz)])) > badlim) {
          jy = iy;
          if (fabs((v2 = v[I(ix,jy,iz)])) > badlim) continue;
        }
        if (fabs((v1 = v[I(ix,ky,iz)])) > badlim) {
          ky = iy;
          if (fabs((v1 = v[I(ix,ky,iz)])) > badlim) continue;
        }
/* check to see whether we are dividing by zero */
        if ((jx <= kx) || (jy <= ky)) continue;
        delx = dx*(jx - kx);
        dely = dy*(jy - ky);
/* hurray! everything is defined */
        tempdiv = (u2 - u1)/delx + (v2 - v1)/dely;
/* everything appears ok -- define the divergence at this point */
        div1[I(ix,iy,iz)] = tempdiv;
      }
    }
  }
  exit(0);
}

/* compute continuity-generated vertical velocity */
wicalc()
{
  int jz,kz,lz,valid,start,good;
  float tempdiv,sum;
/* loop on x and y */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
/* integrate down using previously computed divergence */
/* start indicates whether integration has been initialized */
      start = 0;
      for (iz = nz - 1; iz >= 0; iz--) {
        jz = iz + 1;
/* default -- missing data */
        wi[I(ix,iy,iz)] = bad;
/* have we initialized integration yet? */
        if (start) {
/* average divergence at iz and jz, or just take one if other is bad */
          valid = 0;
          sum = 0.;
          if (fabs((tempdiv = div1[I(ix,iy,jz)])) < badlim) {
            valid++;
            sum += tempdiv;
          }
          if (fabs((tempdiv = div1[I(ix,iy,iz)])) < badlim) {
            valid++;
            sum += tempdiv;
          }
/* if no valid divergence data, stop and jump to next level */
          if (valid) tempdiv = sum/valid;
          else {
            start = 0;
            continue;
          }
/* do integration */
          wi[I(ix,iy,iz)] = wi[I(ix,iy,jz)] -
            (wi[I(ix,iy,jz)]/zs[iz] - tempdiv)*dz;
        }
/* haven't started yet -- check to see if particle velocity defined */
        else if (w != NULL) {
/* it is -- check to see if particle velocity exists at that level,
 * and if divergence exists at that level or just below */
          kz = iz > 0 ? iz - 1 : 0;
          good = (fabs(w[I(ix,iy,iz)]) < badlim) &&
               ((fabs(div1[I(ix,iy,iz)]) < badlim) ||
               (fabs(div1[I(ix,iy,kz)]) < badlim));
          if (good) {
/* modify wtcldtop dependent upon divergence there if varflag set */
            if (varflag) {
              if (fabs(div1[I(ix,iy,iz)]) < badlim)
                   dvalue = div1[I(ix,iy,iz)];
              else if (fabs(div1[I(ix,iy,kz)]) < badlim)
                   dvalue = div1[I(ix,iy,kz)];
              else dvalue = 0.;
              divmax = WTVAR2/dz;
              if (dvalue < 0) dvalue = 0.;
              if (dvalue > divmax) dvalue = divmax;
              wtcldtop = dvalue*(WTVAR2 - WTVAR1)/divmax + WTVAR1;
            }
/* datum is there -- initialize wi using particle velocity */
            wi[I(ix,iy,iz)] = w[I(ix,iy,iz)] + wtcldtop;
            start = 1;
          }
        }
/* haven't started and vertical particle velocity not defined */
        else {
/* start by setting wi = 0 just above where div first defined */
          lz = iz - 1;
          if (lz < 0) lz = 0;
          if (fabs(div1[I(ix,iy,lz)]) < badlim) {
            wi[I(ix,iy,iz)] = 0.;
            start = 1;
          }
        }
      }
    }
  }
}

/* compute particle terminal velocity */
wtermcal()
{
  float tempw,tempwi;
/* loop on x, y, and z */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {
/* compute terminal fall velocity (convention is negative) */
        wterm[I(ix,iy,iz)] = bad;
        if (fabs((tempw = w[I(ix,iy,iz)])) > badlim) continue;
        if (fabs((tempwi = wi[I(ix,iy,iz)])) > badlim) continue;
        wterm[I(ix,iy,iz)] = tempw - tempwi;
      }
    }
  }
}

/* optionally compute cloud physics vertical air velocity
 * also computes a revised terminal velocity and converts the
 * 2-D divergence into a residual of the continuity equation */
wcpcalc()
{
  int jz,kz;
  float tempw,tempz,wtermcp,delz;
/* die if no w field */
  if (w == NULL) {
    fprintf(stderr,"radvert: w doesn't exist, so wcp can't be created\n");
    exit(1);
  }
/* die if no dbz field */
  if (dbz == NULL) {
    fprintf(stderr,"radvert: dbz doesn't exist, so wcp can't be created\n");
    exit(1);
  }
/* loop on z, y, z */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {
/* default is bad */
        wcp[I(ix,iy,iz)] = bad;
/* good only if particle vertical velocity defined */
        if (fabs((tempw = w[I(ix,iy,iz)])) > badlim) continue;
/* if reflectivity is bad, assume low terminal velocity */
        if (fabs((tempz = dbz[I(ix,iy,iz)])) > badlim) wtermcp = WTMIN;
/* cloud physics formula from Raymond theory book, p 4424 */
        else wtermcp = exp(dbz[I(ix,iy,iz)]/WCPA + WCPB);
        wcp[I(ix,iy,iz)] = w[I(ix,iy,iz)] + wtermcp;
/* if wterm field exists, set wterm = wtermcp */
        if (wterm != NULL) wterm[I(ix,iy,iz)] = -wtermcp;
      }
/* add vertical divergence to horizontal divergence to get cont eqn resid */
      if (div1 == NULL) continue;
      for (iz = 0; iz < nz; iz++) {
        jz = iz + 1;
        kz = iz - 1;
        if (jz == nz) jz = nz - 1;
        if (kz < 0) kz = 0;
        if ((fabs(wcp[I(ix,iy,iz)]) > badlim) ||
            (fabs(wcp[I(ix,iy,jz)]) > badlim) ||
            (fabs(wcp[I(ix,iy,kz)]) > badlim) ||
            (fabs(div1[I(ix,iy,iz)]) > badlim)) div1[I(ix,iy,iz)] = bad;
        else {
          delz = (jz - kz)*dz;
          div1[I(ix,iy,iz)] += (wcp[I(ix,iy,jz)] - wcp[I(ix,iy,kz)])/delz
               - wcp[I(ix,iy,iz)]/zs[iz];
        }
      }
    }
  }
}
