/* radvert3.c -- This program computes various quantities having to do
 * with vertical motion for dual doppler radar data, namely, horizontal
 * divergence, continuity-generated vertical velocity, particle
 * terminal velocity (wtZ, from Jorgensen's approach),
 * and vertical particle velocity (wpart).
 * Also computed are vertical profiles of density (rho), as well as
 * pressure (pres), and scale height (zs). These are computed on a
 * constant lapse rate hypothesis.
 * Expected in the standard input is file produced by radcart5 
 * containing horizontal dual doppler velocities assuming vertical
 * particle velocity is zero, gx and gy factors from dual doppler 
 * analysis, standard error fieldseur, evr, eup, and evp, and optionally,
 * vertical particle velocity.  These 
 * variables are assumed to have the names given by this program.
 * These fields are passed to the output
 * along with the new fields, as a common data format file. If the
 * vertical particle velocity doesn't exist downward integration
 * is initialized from zero. Upward integration is always initialized
 * from zero.
 * The command arguments are the names of the various fields:
 * 1 = name for horizontal divergence;
 * 2 = name for continuity-generated vertical wind speed;
 * 3 = name for particle terminal velocity;
 * Downward integration is optionally initialized in one of three
 * different ways according to chosen flag, -w, -v or -Z. 
 * The optional flag -c sets the divergence correction in columns option.
 * The optional flag -t indicates that thresholding in horizontal velocities
 * errors and divergence of g are desired. If flag -t is chosen then the
 * next two arguments are the thresholding values for gdiv and u,v wind errors
 * respectively. 
 */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define PMSL 100.0      /* sea level pressure (kpa) */
#define TMSL 298.0      /* sea level temperature (K) */
#define GAMMA 5.0       /* temperature lapse rate (K/km) */
#define GEE 9.8         /* acceleration of gravity (m/s^2) */
#define RGAS 287.6      /* gas constant for air (J/kg/K) */
#define SCALE 32.0      /* integer packing scale factor */
#define WTCLDTOP 2.0    /* default terminal velocity at cloud top */
#define WCPA 31.8       /* constants for reflectivity-terminal velocity */
#define WCPB 0.918      /* relationship                                 */
#define WTMIN 2.0       /* assumed terminal velocity for undefined refl */
#define WTVAR1 1.0      /* top terminal velocity for no divergence */
#define WTVAR2 5.0      /* max terminal velocity for max divergence */
#define BOTT_SNOW 7.    /* height of bottom of snow layer */
#define TOP_RAIN 4.25   /* height of top of rain layer */
#define I(ix,iy,iz) (ix)*ny*nz + (iy)*nz + (iz)  /* array index macro */

char hbin[HBMAX][LINE],hbout[HBMAX][LINE];  /* header buffers */
int nhbin;              /* number of lines in input header buffer */
float *sbuff,*vbuff;    /* slice buffers */
long nsbuffin,nsbuffot,nvbuffin,nvbuffot;   /* sizes of slice buffers */
float *w,*wterm,*wpart,*dbz;  /* field pointers */
float *gx,*gy,*ux0,*uy0,*wi2,*div2;
float *gdiv,*eur,*evr,*eup,*evp;
float *wtZ,*u,*v;
float *rho,*pres,*z,*zs;
float z1,mu,rhomsl;     /* intermediates for pressure and density calcs */
float wtcldtop;         /* cloud top terminal velocity */
int divco;              /* divergence correction flag */
int wtmod = 0;          /* flag indicating cloud top wt modified */
int varflag = 0;        /* flag indicating div-dependent top wt to be used */
int wtZflag = 0;      /* flag indicating wtZ relationship top wt to be used */
float dvalue;           /* temporary for variable wt-top calculation */
float divmax;           /* maximum divergence value accepted */
int i,ix,iy,iz,nx,ny,nz;   /* looping variables and array limits */
float dx,dy,dz;         /* grid intervals */
int wno = 0;            /* flag indicates vert. part. vel. exists */
char alpha[LINE];       /* line buffer */
char d1[LINE],d2[LINE],d3[LINE];  /* dimension names */
struct field *fp;       /* field structure pointer */
float bad,badlim;       /* bad data stuff */
float *getptr();
float lowpass();        /* low pass filter */
float ggradwi();
float wpthres,vrthres;  /* thresholding values */
int threshmod;          /* threshold flag  */

main(argc,argv)
int argc;
char *argv[];
{
  int loop;

  /* check arguments */
  /* various options for cloud top terminal velocity for determination
     of air velocity there for initialization of downward integration --
     none of these selected means wair = 0 at cloud top */

  /* cloud top terminal velocity depends on divergence there */
  if (argc > 1 && strcmp(argv[1],"-v") == 0) {
    for (i = 2; i < argc; i++) argv[i - 1] = argv[i];
    argc--;
    varflag = 1;
  }

  /* take cloud top terminal velocity as specified by Jorgensen relation */
  if (argc > 1 && strcmp(argv[1],"-Z") == 0) {
    for (i = 2; i < argc; i++) argv[i - 1] = argv[i];
    argc--;
    wtZflag = 1;
  }

  /* specify the terminal velocity at cloud top */
  if (argc > 2 && strcmp(argv[1],"-w") == 0) {
    wtcldtop = atof(argv[2]);
    for (i = 3; i < argc; i++) argv[i - 2] = argv[i];
    argc -= 2;
    wtmod = 1;
  }

  /* use default value for cloud top terminal velocity */
  else wtcldtop = WTCLDTOP;

  /* -c means divergence correction */
  if (argc > 1 && strcmp(argv[1],"-c") == 0) {
    for (i = 2; i < argc; i++) argv[i - 1] = argv[i];
    argc --;
    divco = 1;
  }
  else divco = 0;

  /* -t wperr vrerr means do thresholding on radcart5-generated error values */
  if (argc > 3 && strcmp(argv[1],"-t") == 0) {
    wpthres = atof(argv[2]);
    vrthres = atof(argv[3]);  
    for (i = 4; i < argc; i++) argv[i - 3] = argv[i];
    argc -= 3;
    threshmod = 1;
  }
  else threshmod = 0;

  if (argc != 5) {
    fprintf(stderr,
     "Usage: radvert3 [-w wtval|-v|-Z] [-c] [-t wperr vrerr] wp div wi wterm\n");
    exit(1);
  }

  /* get header from standard input */
  nhbin = gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"radvert3: input file must be float format\n");
    exit(1);
  }

  /* extract bad and badlim from parameter section of header */
  if (seekpar(hbin,HBMAX,"bad",alpha) == FAIL) {
    fprintf(stderr,"radvert3: parameter bad missing\n");
    exit(1);
  }
  bad = atof(alpha);
  if (seekpar(hbin,HBMAX,"badlim",alpha) == FAIL) {
    fprintf(stderr,"radvert3: parameter badlim missing\n");
    exit(1);
  }
  badlim = atof(alpha);

  /* check for existing and non-existent fields */
  /* x-velocity */
  if ((fp = seekfld(hbin,HBMAX,'v',"ux0")) == NULL) {
    fprintf(stderr,"radvert3: no ux0 field\n");
    exit(1);
  }

  /* get dimension names and sizes */
  if (fp->dim != 3) {
    fprintf(stderr,"radvert3: incorrect field dimensions = %d\n",fp->dim);
    exit(1);
  }

  /* we don't do extensive error checking */
  strcpy(d1,fp->dname1);
  strcpy(d2,fp->dname2);
  strcpy(d3,fp->dname3);
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;

  /* y-velocity */
  if (seekfld(hbin,HBMAX,'v',"uy0") == NULL) {
    fprintf(stderr,"radvert3: no uy0 field\n");
    exit(1);
  }

  /* z-particle velocity */
  if (seekfld(hbin,HBMAX,'v',argv[1]) == NULL) {
    fprintf(stderr,"radvert3: warning: no %s field\n",argv[1]);
    wno = 1;
  }

  /* reflectivity */
  if (seekfld(hbin,HBMAX,'v',"dbz") == NULL) {
    fprintf(stderr,"radvert3: no dbz field\n");
    exit(1);
  }

  /* div */
  if (seekfld(hbin,HBMAX,'v',argv[2]) != NULL) {
    fprintf(stderr,"radvert3: %s already exists\n",argv[2]);
    exit(1);
  }

  /* wi */
  if (seekfld(hbin,HBMAX,'v',argv[3]) != NULL) {
    fprintf(stderr,"radvert3: %s already exists\n",argv[3]);
    exit(1);
  }

  /* wterm */
  if (seekfld(hbin,HBMAX,'v',argv[4]) != NULL) {
    fprintf(stderr,"radvert3: %s already exists\n",argv[4]);
    exit(1);
  }

  /* get deltas */
  if (seekpar(hbin,HBMAX,"dx",alpha) == FAIL) {
    fprintf(stderr,"radvert3: can't find a needed parameter\n");
    exit(1);
  }
  dx = atof(alpha);
  if (seekpar(hbin,HBMAX,"dy",alpha) == FAIL) {
    fprintf(stderr,"radvert3: can't find a needed parameter\n");
    exit(1);
  }
  dy = atof(alpha);
  if (seekpar(hbin,HBMAX,"dz",alpha) == FAIL) {
    fprintf(stderr,"radvert3: can't find a needed parameter\n");
    exit(1);
  }
  dz = atof(alpha);

  /* make new header */
  /* start by copying over old header */
  for (i = 0; i < nhbin; i++) strcpy(hbout[i],hbin[i]);

  /* add terminal velocity from wt-Z relationship   */
  addfld(hbout,HBMAX,'v',"wtZ",SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add horizontal wind components */
  addfld(hbout,HBMAX,'v',"u",SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);
  addfld(hbout,HBMAX,'v',"v",SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add g field divergence */
  addfld(hbout,HBMAX,'v',"gdiv",SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add dual doppler divergence */
  addfld(hbout,HBMAX,'v',argv[2],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add dual doppler continuity-generated vertical velocity */
  addfld(hbout,HBMAX,'v',argv[3],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add terminal velocity of hydrometeors if it can be computed */
  if (!wno)
    addfld(hbout,HBMAX,'v',argv[4],SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add vertical particle velocity from wt-Z and wi computations */
  addfld(hbout,HBMAX,'v',"wpart",SCALE,0.,'s',3,d1,nx,d2,ny,d3,nz);

  /* add density to static slice */
  if (seekfld(hbout,HBMAX,'s',"rho") == NULL)
    addfld(hbout,HBMAX,'s',"rho",10000.,0.,'s',1,d3,nz);

  /* add pressure to static slice */
  if (seekfld(hbout,HBMAX,'s',"pres") == NULL)
    addfld(hbout,HBMAX,'s',"pres",10.,0.,'s',1,d3,nz);

  /* add density scale height to static slice */
  if (seekfld(hbout,HBMAX,'s',"zs") == NULL)
    addfld(hbout,HBMAX,'s',"zs",100.,0.,'s',1,d3,nz);

  /* add comments */
  if (wtmod) {
    sprintf(alpha,"radvert3: -w %.1f %s %s %s %s\n",wtcldtop,
	    argv[1],argv[2],argv[3],argv[4]);
  }
  else {
    sprintf(alpha,"radvert3: %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  }
  addcline(hbout,HBMAX,alpha);
  if (threshmod) 
    sprintf(alpha,"  thresh: wperr %.2f vrerr %.2f\n",wpthres,vrthres); 
  addcline(hbout,HBMAX,alpha);
  if (divco) addcline(hbout,HBMAX,"  column div correction\n"); 
  if (wno) addcline(hbout,HBMAX,"  zero ubc\n");
  else if (varflag) addcline(hbout,HBMAX,"  variable particle ubc\n");
  else if (wtZflag) addcline(hbout,HBMAX,"  wt-Z ubc\n");
  else addcline(hbout,HBMAX,"  particle ubc\n");

  /* allocate space for slices */
  nsbuffin = elemcnt(hbin,HBMAX,'s');
  nsbuffot = elemcnt(hbout,HBMAX,'s');
  if ((sbuff = (float *) malloc(sizeof(float)*nsbuffot)) == NULL) {
    fprintf(stderr,"radvert3: can't get enough memory\n");
    exit(1);
  }
  nvbuffot = elemcnt(hbout,HBMAX,'v');
  nvbuffin = elemcnt(hbin,HBMAX,'v');
  if ((vbuff = (float *) malloc(sizeof(float)*nvbuffot)) == NULL) {
    fprintf(stderr,"radvert3: can't get enough memory\n");
    exit(1);
  }

  /* set pointers for fields -- NULL if it doesn't exist */
  ux0 = getptr("ux0");
  uy0 = getptr("uy0");
  w = getptr(argv[1]);
  div2 = getptr(argv[2]);
  wtZ = getptr("wtZ");
  u = getptr("u");
  v = getptr("v");
  wi2 = getptr(argv[3]);
  wterm = getptr(argv[4]);
  wpart = getptr("wpart");
  dbz = getptr("dbz");
  gx = getptr("gx");
  gy = getptr("gy");
  gdiv = getptr("gdiv");
  eur = getptr("eur");
  evr = getptr("evr");
  eup = getptr("eup");
  evp = getptr("evp");
  rho = getptr("rho");
  pres = getptr("pres");
  z = getptr("z");
  zs = getptr("zs");

  /* put new header */
  puthdr(stdout,hbout,HBMAX);

  /* get static slice */
  getslice(stdin,nsbuffin,sbuff);
  
  /* fill density, pressure, and scale height fields if they exist */
  if ((rho != NULL) && (pres != NULL) && (zs != NULL)) {
    z1 = TMSL/GAMMA;
    mu = GEE/(RGAS*GAMMA*0.001);
    rhomsl = 1000.0*PMSL/(RGAS*TMSL);
    for (iz = 0; iz < nz; iz++) {
      pres[iz] = PMSL*pow(1. - z[iz]/z1,mu);
      rho[iz] = rhomsl*pow(1. - z[iz]/z1,mu - 1.);
      zs[iz] = (z1 - z[iz])/(mu - 1.);
    }
  }
  else {
    fprintf(stderr,"radvert3: panic -- rho, pres, and zs don't exist!\n");
    exit(1);
  }

  /* put static slice */
  putslice(stdout,nsbuffot,sbuff);

  /* get variable slice */
  while (getslice(stdin,nvbuffin,vbuff) != EOF) {
    /* do computations */
    /* set all fields to bad in region with dbz bad */
    threshdbz();  
    
    /* compute particle terminal velocity from wt-Z relationship  */
    wtZcalc();
    
    /* compute g field divergence */
    gdivcalc();
    
    /* initialize wi2 field to bad */
    wi2init();
    
    /* compute dual divergence field */
    divcalc2();
    
    /* compute the dual doppler integrated vertical wind speed */
    /* integration going down */
    wi2down();
    
    /* compute the correction to the divergence field */
    if (divco) divcorr(); 
    
    /* compute particle terminal velocity */
    if (!wno) {
      wtermcal();
    }
    
    /* compute vertical particle velocities from wt-Z and wi computations */
    wpartcal();
    
    /* compute the horizontal wind components from ux0,uy0 and wpart */
    uvcalc();  
    
    /* put variable slice */
    putslice(stdout,nvbuffot,vbuff);
  }

  /* done! */
  exit(0);
}

/* set float pointer in vbuff or sbuff */
float *getptr(name)
char *name;
{
  fp = seekfld(hbout,HBMAX,'v',name);
  if (fp == NULL) {
    fp = seekfld(hbout,HBMAX,'s',name);
    if (fp == NULL) return(NULL);
    else return(sbuff + fp->ptr);
  }
  else return(vbuff + fp->ptr);
}

/* threshold in dbz and in dual errors eur, evr, eup, evp */
threshdbz()
{

  /* loop in all indices and threshold */
  for (iz = 0; iz < nz; iz++) {
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {

	/* threshold in bad values of dbz  */
        if (fabs(dbz[I(ix,iy,iz)]) > badlim) {
          ux0[I(ix,iy,iz)] = bad;
          uy0[I(ix,iy,iz)] = bad;
          gx[I(ix,iy,iz)] = bad;
          gy[I(ix,iy,iz)] = bad;
          if (!wno) w[I(ix,iy,iz)] = bad;
        }

	/* threshold in standard errors if threshmod set */
        if (threshmod && (fabs(eur[I(ix,iy,iz)]) > vrthres ||
			  fabs(evr[I(ix,iy,iz)]) > vrthres ||
			  fabs(eup[I(ix,iy,iz)]) > wpthres ||
			  fabs(evp[I(ix,iy,iz)]) > wpthres)) {
          ux0[I(ix,iy,iz)] = bad;
          uy0[I(ix,iy,iz)] = bad;
          gx[I(ix,iy,iz)] = bad;
          gy[I(ix,iy,iz)] = bad;
          if (!wno) w[I(ix,iy,iz)] = bad;
        }
      }
    }
  }
}

/* initialize wi2 field to bad */
wi2init()
{

  /* loop on iz,iy,ix and initialize */
  for (iz = 0; iz < nz; iz++) {
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
        wi2[I(ix,iy,iz)] = bad; 
      }
    }
  }
}

/* compute g field divergence */
gdivcalc()
{
  int jx,kx,jy,ky;
  float delx,dely,tempdiv,u1,u2,v1,v2;

  /* loop on z */
  for (iz = 0; iz < nz; iz++) {

    /* loop on x and y */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {

	/* set up for centered differences except at boundaries */
        jx = ix + 1;
        if (jx >= nx) jx = nx - 1;
        kx = ix - 1;
        if (kx < 0) kx = 0;
        jy = iy + 1;
        if (jy >= ny) jy = ny - 1;
        ky = iy - 1;
        if (ky < 0) ky = 0;

	/* be a pessimist -- default is missing data */
        gdiv[I(ix,iy,iz)] = bad;

	/* if data at this point are bad then continue */
        if (fabs(gx[I(ix,iy,iz)]) > badlim || fabs(gy[I(ix,iy,iz)])
             > badlim) continue;

	/* check to see if data defining divergence at this point are ok */
	/* go to off-centered differences if necessary */
        if (fabs((u2 = gx[I(jx,iy,iz)])) > badlim) {
          jx = ix;
          if (fabs((u2 = gx[I(jx,iy,iz)])) > badlim) continue;
        }
        if (fabs((u1 = gx[I(kx,iy,iz)])) > badlim) {
          kx = ix;
          if (fabs((u1 = gx[I(kx,iy,iz)])) > badlim) continue;
        }
        if (fabs((v2 = gy[I(ix,jy,iz)])) > badlim) {
          jy = iy;
          if (fabs((v2 = gy[I(ix,jy,iz)])) > badlim) continue;
        }
        if (fabs((v1 = gy[I(ix,ky,iz)])) > badlim) {
          ky = iy;
          if (fabs((v1 = gy[I(ix,ky,iz)])) > badlim) continue;
        }

	/* check to see whether we are dividing by zero */
        if ((jx <= kx) || (jy <= ky)) continue;
        delx = dx*(jx - kx);
        dely = dy*(jy - ky);

	/* hurray! everything is defined */
        tempdiv = (u2 - u1)/delx + (v2 - v1)/dely;

	/* everything appears ok -- define the divergence at this point */
        gdiv[I(ix,iy,iz)] = tempdiv;
      }
    }
  }
}

/* compute divergence field  for dual case */
divcalc2()
{
  int jx,kx,jy,ky;
  float delx,dely,tempdiv,u1,u2,v1,v2;

  /* loop on z */
  for (iz = 0; iz < nz; iz++) {

    /* loop on x and y */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {

	/* set up for centered differences except at boundaries */
        jx = ix + 1;
        if (jx >= nx) jx = nx - 1;
        kx = ix - 1;
        if (kx < 0) kx = 0;
        jy = iy + 1;
        if (jy >= ny) jy = ny - 1;
        ky = iy - 1;
        if (ky < 0) ky = 0;

	/* be a pessimist -- default is missing data */
        div2[I(ix,iy,iz)] = bad;

	/* if data at this point is bad then continue */
        if (fabs(ux0[I(ix,iy,iz)]) > badlim || fabs(uy0[I(ix,iy,iz)])
            > badlim) continue;

	/* check to see if data defining divergence at this point are ok */
	/* go to off-centered differences if necessary */
        if (fabs(ux0[I(jx,iy,iz)]) > badlim) {
          jx = ix;
          if (fabs(ux0[I(jx,iy,iz)]) > badlim) continue;
          else {
            if (fabs(gx[I(jx,iy,iz)]) > badlim || fabs(wtZ[I(jx,iy,iz)])
                > badlim) u2 = ux0[I(jx,iy,iz)];
            else u2 = ux0[I(jx,iy,iz)] + gx[I(jx,iy,iz)]*wtZ[I(jx,iy,iz)]; 
          }
        }
        else {
          if (fabs(gx[I(jx,iy,iz)]) > badlim || fabs(wtZ[I(jx,iy,iz)])
              > badlim) u2 = ux0[I(jx,iy,iz)];
          else u2 = ux0[I(jx,iy,iz)] + gx[I(jx,iy,iz)]*wtZ[I(jx,iy,iz)]; 
        }
        if (fabs(ux0[I(kx,iy,iz)]) > badlim) {
          kx = ix;
          if (fabs(ux0[I(kx,iy,iz)]) > badlim) continue;
          else {
            if (fabs(gx[I(kx,iy,iz)]) > badlim || fabs(wtZ[I(kx,iy,iz)])
                > badlim) u1 = ux0[I(kx,iy,iz)];
            else u1 = ux0[I(kx,iy,iz)] + gx[I(kx,iy,iz)]*wtZ[I(kx,iy,iz)]; 
          }
        }
        else {
          if (fabs(gx[I(kx,iy,iz)]) > badlim || fabs(wtZ[I(kx,iy,iz)])
              > badlim) u1 = ux0[I(kx,iy,iz)];
          else u1 = ux0[I(kx,iy,iz)] + gx[I(kx,iy,iz)]*wtZ[I(kx,iy,iz)]; 
        }
        if (fabs(uy0[I(ix,jy,iz)]) > badlim) {
          jy = iy;
          if (fabs(uy0[I(ix,jy,iz)]) > badlim) continue;
          else {
            if (fabs(gy[I(ix,jy,iz)]) > badlim || fabs(wtZ[I(ix,jy,iz)])
                > badlim) v2 = uy0[I(ix,jy,iz)];
            else v2 = uy0[I(ix,jy,iz)] + gy[I(ix,jy,iz)]*wtZ[I(ix,jy,iz)]; 
          }
        }
        else {
          if (fabs(gy[I(ix,jy,iz)]) > badlim || fabs(wtZ[I(ix,jy,iz)])
              > badlim) v2 = uy0[I(ix,jy,iz)];
          else v2 = uy0[I(ix,jy,iz)] + gy[I(ix,jy,iz)]*wtZ[I(ix,jy,iz)]; 
        }
        if (fabs(uy0[I(ix,ky,iz)]) > badlim) {
          ky = iy;
          if (fabs(uy0[I(ix,ky,iz)]) > badlim) continue;
          else {
            if (fabs(gy[I(ix,ky,iz)]) > badlim || fabs(wtZ[I(ix,ky,iz)])
                > badlim) v1 = uy0[I(ix,ky,iz)];
            else v1 = uy0[I(ix,ky,iz)] + gy[I(ix,ky,iz)]*wtZ[I(ix,ky,iz)]; 
          }
        }
        else {
          if (fabs(gy[I(ix,ky,iz)]) > badlim || fabs(wtZ[I(ix,ky,iz)])
              > badlim) v1 = uy0[I(ix,ky,iz)];
          else v1 = uy0[I(ix,ky,iz)] + gy[I(ix,ky,iz)]*wtZ[I(ix,ky,iz)]; 
        }

	/* check to see whether we are dividing by zero */
        if ((jx <= kx) || (jy <= ky)) continue;
        delx = dx*(jx - kx);
        dely = dy*(jy - ky);

	/* hurray! everything is defined */
        tempdiv = (u2 - u1)/delx + (v2 - v1)/dely;

	/* everything appears ok -- define the divergence at this point */
        div2[I(ix,iy,iz)] = tempdiv;
      }
    }
  }
}

/* compute the correction to the divergence field */
divcorr()
{
  float zr,wib,rhob;
  int zt,flag;
  float temp;
  flag = 0;
  
  /* value of density at z = 0 (boundary) */
  rhob = rho[0];
  
  /* loop on x and y */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      
      /* if bad data at surface do not do anything */
      if (fabs((wib = wi2[I(ix,iy,0)])) > badlim) continue;
      
      /* if vert. wind is not a number at surface, set it to bad  */
      temp = wib;
      if (temp != wib) {
        wi2[I(ix,iy,0)] = bad;
        continue;
      }

      /* loop in column to find the last defined wind	 */
      for (iz = 0; iz < nz; iz++) {
        if (fabs(wi2[I(ix,iy,iz)]) > badlim) {
          flag = 1;
          break;
        }
      }
      zt = iz;
      if (flag) zt -= 1;

      /* if zt is zero then do not do anything */
      if (zt == 0) {
	wi2[I(ix,iy,0)] = 0.;
	continue;
      }
      
      /* correct the wind field in this column */
      for (iz = 0; iz <= zt; iz++) { 
        if (fabs(wi2[I(ix,iy,iz)]) < badlim) {
          zr = (float) iz;
          zr /= (float) zt;
          wi2[I(ix,iy,iz)] += (rhob/rho[iz])*wib*(zr - 1.0);
        }
        else {
          fprintf(stderr,"bad vertical wind in this column\n");
          fprintf(stderr,"iz = %d  wi = %f\n",iz,wi2[I(ix,iy,iz)]);
          fprintf(stderr,"bad = %f  badlim = %f\n",bad,badlim);
          fprintf(stderr,"zt = %d  x = %d  y = %d\n",zt,ix,iy);
        }
      }
    }
  }
}

/* compute wt-Z particle terminal velocity.  */
wtZcalc()
{
  float vts,vtw,Zw,Zs,rhos;
  float temp,term1,gamma;
  for (iz = 0; iz < nz; iz++)  {
    for ( ix = 0; ix < nx; ix++)  {
      for ( iy = 0; iy < ny; iy++)  {
	wtZ[I(ix,iy,iz)] = 0.;
	if (fabs(dbz[I(ix,iy,iz)]) > badlim) continue;
	Zw = pow(10.,(dbz[I(ix,iy,iz)]/10.));
	Zs = 4.72*pow(10.,(dbz[I(ix,iy,iz)]/10.));
	
	/* calculate the density factor */
	temp = TMSL - GAMMA*z[iz];
	if (rho[iz] != 0.0) {
	  term1 = pow(10.0,(0.43*log10(rhomsl/rho[iz]) -
			    0.4*pow(log10(rhomsl/rho[iz]),2.5)));
	  rhos = (1.+0.0023*(1.1-(rho[iz]/rhomsl))*(TMSL-temp))*term1;
	}
	else continue;
	vts = -rhos*0.817*pow(Zs,0.063);
	
	/* calculate the terminal velocity for rain */
	gamma = pow((720.*0.08/(1.0e-12*Zw)),(1./7.));
	vtw = -rhos*(9.65-10.3*pow(gamma/(gamma+6.),7));
	
	/*        vtw = -rhos*2.6*pow(Zw,0.107);      */
	if (z[iz] >= BOTT_SNOW) wtZ[I(ix,iy,iz)] = vts;
	else {
	  if (z[iz] <= TOP_RAIN) wtZ[I(ix,iy,iz)] = vtw;
	  else {
	    wtZ[I(ix,iy,iz)] = (BOTT_SNOW - z[iz])*vtw/(BOTT_SNOW - TOP_RAIN)
	      + (z[iz] - TOP_RAIN)*vts/(BOTT_SNOW - TOP_RAIN);
	  }
	}
      }
    }
  }
}

/* compute continuity-generated vertical velocity using downward integration
   for dual Doppler analysis  */
wi2down()
{
  int jz,kz,lz,valid,start,good,gridpt;
  float tempdiv,sum,oldwi2;
  float ggrad,ggdiv;
  start = 0;
  gridpt = 0;
  for (iz = nz - 1; iz >= 0; iz--) {
    jz = iz + 1;

    /* loop on x and y */
    for (ix = 0; ix < nx; ix++) {
      for (iy = 0; iy < ny; iy++) {
   
	/* check if integration has started on this column --
	   we can't start at very top, i. e., if jz = iz + 1 = nz */
        if (jz < nz && fabs(wi2[I(ix,iy,jz)]) < badlim) start = 1;
        else start = 0;

	/* have we initialized integration yet? */
        if (start) {

	  /* get term involving g*grad(wi2)  */
          ggrad = ggradwi(ix,iy,iz);

	  /* average divergence at iz and jz, or just take one
	     if other is bad */
          valid = 0;
          sum = 0.;
          if (fabs((tempdiv = div2[I(ix,iy,jz)])) < badlim) {
            valid++;
            sum += tempdiv;
          }
          if (fabs((tempdiv = div2[I(ix,iy,iz)])) < badlim) {
            valid++;
            sum += tempdiv;
          }

	  /* if no valid divergence data, stop and jump to next level */
          if (valid) {
            tempdiv = sum/valid;
          }
          else {
            start = 0;
            continue;
          }

	  /* check the value of gdiv */
          if (fabs((ggdiv = gdiv[I(ix,iy,jz)])) > badlim) {
            if (fabs((ggdiv = gdiv[I(ix,iy,iz)])) > badlim) ggdiv = 0.0;
          }

	  /* do integration */
          wi2[I(ix,iy,iz)] = wi2[I(ix,iy,jz)] -
            (wi2[I(ix,iy,jz)]*(1./zs[iz] - ggdiv) - tempdiv
             - ggrad)*dz;
        }
	
	/* haven't started yet -- check to see if particle velocity defined */
        else if (w != NULL) {
	  
	  /* it is -- check to see if particle velocity exists at that level,
	     and if divergence exists at that level or just below */
          kz = iz > 0 ? iz - 1 : 0;
          good = (fabs(w[I(ix,iy,iz)]) < badlim) &&
               ((fabs(div2[I(ix,iy,iz)]) < badlim) ||
               (fabs(div2[I(ix,iy,kz)]) < badlim));
          if (good) {

	    /* modify wtcldtop dependent upon divergence there if
	       varflag set */
            if (varflag) {
              if (fabs(div2[I(ix,iy,iz)]) < badlim)
		dvalue = div2[I(ix,iy,iz)];
              else if (fabs(div2[I(ix,iy,kz)]) < badlim)
		dvalue = div2[I(ix,iy,kz)];
              else dvalue = 0.;
              divmax = WTVAR2/dz;
              if (dvalue < 0) dvalue = 0.;
              if (dvalue > divmax) dvalue = divmax;
              wtcldtop = dvalue*(WTVAR2 - WTVAR1)/divmax + WTVAR1;
            }

	    /* modify wtcldtop with wtZ relationship value if wtZflag set */
            else {
              if (wtZflag)   wtcldtop = -1.*wtZ[I(ix,iy,iz)];
            }

	    /* datum is there -- initialize wi using particle velocity  */
            wi2[I(ix,iy,iz)] = w[I(ix,iy,iz)] + wtcldtop;  
          }
        }

	/* haven't started and vertical particle velocity not defined */
        else {

	  /* start by setting wi = 0 just above where div first defined */
          lz = iz - 1;
          if (lz < 0) lz = 0;
          if (fabs(div2[I(ix,iy,lz)]) < badlim) {
            wi2[I(ix,iy,iz)] = 0.;
          }
        }
      }
    }
  }
}

/* ggradwi computes gx*dw/dx + gy*dw/dy in previous level */
float ggradwi(lx,ly,lz)
int lx,ly,lz;
{
  float dwdx,dwdy,ggx,ggy;
  float wx2,wxc,wy2,wyc,term1,term2;
  int kx,ky,kz,jx,jy,jz;

  /* set up for upstream differences  except at the boundaries */
  jz = lz + 1;
  jx = lx + 1;
  if (jx >= nx) return(0.0);
  kx = lx - 1;
  if (kx < 0) return(0.0);
  jy = ly + 1;
  if (jy >= ny) return(0.0);
  ky = ly - 1;
  if (ky < 0) return(0.0);

  /* check that gx and gy are good */
  if (fabs((ggx = gx[I(lx,ly,lz)])) > badlim) {
    if (fabs((ggx = gx[I(lx,ly,jz)])) > badlim) return(0.0);
  }
  if (fabs((ggy = gy[I(lx,ly,lz)])) > badlim) {
    if (fabs((ggy = gy[I(lx,ly,jz)])) > badlim) return(0.0);
  }

  /* Courant condition */
   if (fabs(ggx) > .5*dx/dz || 
        fabs(ggy) > .5*dy/dz) return(0.0);

  /* check to see if data at this point are ok */
  if (fabs((wxc = wi2[I(lx,ly,jz)])) > badlim) return(0.0);
  wyc = wxc;
  if (ggx > 0.0) {
    if (fabs((wx2 = wi2[I(jx,ly,jz)])) > badlim) return(0.0);
    dwdx = (wx2 - wxc)/dx;
  }
  else {
    if (fabs((wx2 = wi2[I(kx,ly,jz)])) > badlim) return(0.0);
    dwdx = (wxc - wx2)/dx;
  }
  if (ggy > 0.0) {
    if (fabs((wy2 = wi2[I(lx,jy,jz)])) > badlim) return(0.0);
    dwdy = (wy2 -wyc)/dy;
  }
  else {
    if (fabs((wy2 = wi2[I(lx,ky,jz)])) > badlim) return(0.0);
    dwdy = (wyc - wy2)/dy;
  }

  /* now return the value of ggradwi */
  if (fabs(ggx*dwdx + ggy*dwdy) > badlim) return(0.0);
  else return((ggx*dwdx + ggy*dwdy));
}

/* compute particle terminal velocity */
wtermcal()
{
  float tempw,tempwi;

  /* loop on x, y, and z */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {

	/* compute terminal fall velocity (convention is negative) */
        wterm[I(ix,iy,iz)] = bad;
        if (fabs((tempw = w[I(ix,iy,iz)])) > badlim) continue;
        if (fabs((tempwi = wi2[I(ix,iy,iz)])) > badlim) continue;
        wterm[I(ix,iy,iz)] = tempw - tempwi;
      }
    }
  }
}

/* compute vertical particle velocities from wt-Z and wi */
wpartcal()
{
  float tempwi2,tempwtZ;

  /* loop on x, y, and z */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {

	/* compute vertical velocity */
        wpart[I(ix,iy,iz)] = bad;
        if (fabs((tempwtZ = wtZ[I(ix,iy,iz)])) > badlim) continue;
        if (fabs((tempwi2 = wi2[I(ix,iy,iz)])) > badlim) continue;
        wpart[I(ix,iy,iz)] = tempwi2 + tempwtZ;
      }
    }
  }
}

/* compute uv winds */
uvcalc()
{
  float term1,term2,term3;

  /* loop on x, y, and z  */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {

	/* compute the horizontal wind */
       u[I(ix,iy,iz)] = bad;
       v[I(ix,iy,iz)] = bad;
       if (fabs((term1 = ux0[I(ix,iy,iz)])) < badlim && 
           fabs((term2 = gx[I(ix,iy,iz)])) < badlim &&
           fabs((term3 = wpart[I(ix,iy,iz)])) < badlim)
       u[I(ix,iy,iz)] = term1 + term2*term3;
       if (fabs((term1 = uy0[I(ix,iy,iz)])) < badlim && 
           fabs((term2 = gy[I(ix,iy,iz)])) < badlim &&
           fabs((term3 = wpart[I(ix,iy,iz)])) < badlim)
       v[I(ix,iy,iz)] = term1 + term2*term3;
      }
    }
  }
}
