/* radvort -- compute absolute vorticity and vort flux from radar data */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

/* angles */
#define PI 3.14159
#define DPR (180./PI)

/* names of output fields */
#define RHO "rho"
#define OX "ox"
#define OY "oy"
#define OZ "oz"
#define FX "fx"
#define FY "fy"
#define AVX "avx"
#define AVY "avy"
#define SOZ "soz"

/* density constants */
#define RHO0 1.2
#define SCALEHT 8.3

/* Coriolis parameter stuff */
#define F0 (4.*PI/86400.)

/* index calculator */
#define I(ix,iy,iz) ((iz) + nz*((iy) + ny*(ix)))

/* Candis stuff */
char hbuf[HBMAX][LINE];                 /* header buffer */
char cline[LINE];                       /* character line buffer */
float *sbuff,*vbuff;                    /* slice buffer pointers */
long nsbuffin,nvbuffin,nsbuffout,nvbuffout;   /* slice buffer sizes */
float *vx,*vy,*vz;                      /* pointers to air velocity */
float *sx,*sy;                          /* pointers to storm speed */
float *rho;                             /* pointer to density */
float *ox,*oy,*oz;                      /* pointers to abs vorticity */
float *fx,*fy;                          /* pointers to force */
float *avx,*avy;                        /* absolute vorticity flux */
float *soz;                             /* absolute vorticity source */
long nx,ny,nz;                          /* grid size */
char xname[LINE],yname[LINE],zname[LINE];  /* names of index fields */
float *x,*y,*z;                         /* index fields */
struct field *fp;                       /* field pointer */
float bad,badlim;                       /* bad data values */
long ix,iy,iz,ipos;                     /* indexing variables */
float f;                                /* Coriolis parameter */

/* input stuff */
char *vxname,*vyname,*vzname;           /* names of air motion components */
char *sxname,*syname;                   /* names of storm motion components */
float latitude,gam0,gam1,bldepth,cdrag; /* constants */

/* functions */
float xd(),yd(),zd();                   /* derivative takers */

main(argc,argv)
int argc;
char *argv[];
{

/* check arguments */
  if (argc != 9) {
    fprintf(stderr,"Usage: radvort un vn wn sxn syn lat bld cd\n");
    exit(1);
  }
  vxname = argv[1];
  vyname = argv[2];
  vzname = argv[3];
  sxname = argv[4];
  syname = argv[5];
  latitude = atof(argv[6]);
  bldepth = atof(argv[7]);
  cdrag = atof(argv[8]);

/* get header and check for float format */
  gethdr(stdin,hbuf,HBMAX);
  if (getfmt(hbuf,HBMAX) != 'f') {
    fprintf(stderr,"radvort: float format expected\n");
    exit(1);
  }

/* get bad data parameters */
  if (seekpar(hbuf,HBMAX,"bad",cline) == FAIL) bad = BAD;
  else bad = atof(cline);
  if (seekpar(hbuf,HBMAX,"badlim",cline) == FAIL) badlim = BADLIM;
  else badlim = atof(cline);

/* get the dimensions of input fields */
  if ((fp = seekfld(hbuf,HBMAX,'v',vxname)) == NULL) {
    fprintf(stderr,"radvort: can't find field %s in input\n",vxname);
    exit(1);
  }
  if (fp->dim != 3) {
    fprintf(stderr,"radvort: field %s is not 3-D\n",vxname);
    exit(1);
  }
  nx = fp->dsize1;
  ny = fp->dsize2;
  nz = fp->dsize3;

/* get the names of index fields */
  strcpy(xname,fp->dname1);
  strcpy(yname,fp->dname2);
  strcpy(zname,fp->dname3);

/* get sizes of input slices */
  nsbuffin = elemcnt(hbuf,HBMAX,'s');
  nvbuffin = elemcnt(hbuf,HBMAX,'v');

/* add the output fields to the buffer */
  addfld(hbuf,HBMAX,'s',RHO,10000.,0.,'s',1,zname,nz);
  addfld(hbuf,HBMAX,'v',OX,100.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',OY,100.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',OZ,100.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',FX,100000.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',FY,100000.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',AVX,100000.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',AVY,100000.,0.,'s',3,xname,nx,yname,ny,zname,nz);
  addfld(hbuf,HBMAX,'v',SOZ,100000.,0.,'s',3,xname,nx,yname,ny,zname,nz);

/* add comment and parameters */
  addcline(hbuf,HBMAX,"radvort:\n");
  sprintf(cline,"  %s %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4],argv[5]);
  addcline(hbuf,HBMAX,cline);
  sprintf(cline,"  %s %s %s\n",argv[6],argv[7],argv[8]);
  addcline(hbuf,HBMAX,cline);
  sprintf(cline,"%f",(float)cdrag);
  addpar(hbuf,HBMAX,"cdrag",cline);
  sprintf(cline,"%f",(float)latitude);
  addpar(hbuf,HBMAX,"latitude",cline);
  sprintf(cline,"%f",(float)bldepth);
  addpar(hbuf,HBMAX,"bldepth",cline);

/* write the header */
  puthdr(stdout,hbuf,HBMAX);

/* get sizes of output slices */
  nsbuffout = elemcnt(hbuf,HBMAX,'s');
  nvbuffout = elemcnt(hbuf,HBMAX,'v');

/* allocate space for slices */
  sbuff = getbuff(nsbuffout);
  vbuff = getbuff(nvbuffout);

/* get the static and variable slices */
  if (getslice(stdin,nsbuffin,sbuff) == EOF) {
    fprintf(stderr,"radvort: eof on static slice read\n");
    exit(1);
  }
  if (getslice(stdin,nvbuffin,vbuff) == EOF) {
    fprintf(stderr,"radvort: eof on variable slice read\n");
    exit(1);
  }

/* set up the pointers to the index fields */
  x = getptr(hbuf,HBMAX,sbuff,'s','d',xname);
  y = getptr(hbuf,HBMAX,sbuff,'s','d',yname);
  z = getptr(hbuf,HBMAX,sbuff,'s','d',zname);

/* set up the pointers to the input variables */
  vx = getptr(hbuf,HBMAX,vbuff,'v','d',vxname);
  vy = getptr(hbuf,HBMAX,vbuff,'v','d',vyname);
  vz = getptr(hbuf,HBMAX,vbuff,'v','d',vzname);
  sx = getptr(hbuf,HBMAX,vbuff,'v','d',sxname);
  sy = getptr(hbuf,HBMAX,vbuff,'v','d',syname);

/* set up the pointers to the output variables */
  rho = getptr(hbuf,HBMAX,sbuff,'s','d',RHO);
  ox = getptr(hbuf,HBMAX,vbuff,'v','d',OX);
  oy = getptr(hbuf,HBMAX,vbuff,'v','d',OY);
  oz = getptr(hbuf,HBMAX,vbuff,'v','d',OZ);
  fx = getptr(hbuf,HBMAX,vbuff,'v','d',FX);
  fy = getptr(hbuf,HBMAX,vbuff,'v','d',FY);
  avx = getptr(hbuf,HBMAX,vbuff,'v','d',AVX);
  avy = getptr(hbuf,HBMAX,vbuff,'v','d',AVY);
  soz = getptr(hbuf,HBMAX,vbuff,'v','d',SOZ);

/* do the calculations */
  docalcs();

/* write the output */
  putslice(stdout,nsbuffout,sbuff);
  putslice(stdout,nvbuffout,vbuff);

/* done */
  exit(0);
}

/* docalcs -- do the calculations */
docalcs()
{
  long nbl;
  float dbl;
  float u,v,w,urel,vrel,usurf,vsurf,uabs;
  float xdoz,ydoz,xdavx,ydavy;

/* compute rho */
  for (iz = 0; iz < nz; iz++) {
    rho[iz] = RHO0*exp(-z[iz]/SCALEHT);
  }

/* compute Coriolis parameter */
  f = F0*sin(latitude/DPR);

/* compute top of boundary layer */
  nbl = bldepth/(z[1] - z[0]) + 0.5;
  if (nbl < 0) nbl = 0;
  if (nbl > nz) nbl = nz;
  dbl = 1./((z[1] - z[0])*(nbl + 0.5));

/* fix things at lowest level if possible */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {
        ipos = I(ix,iy,iz);
        if (iz == 0 && fabs(vx[I(ix,iy,0)]) > badlim &&
             fabs(vx[I(ix,iy,1)]) < badlim) {
          vx[I(ix,iy,0)] = vx[I(ix,iy,1)];
        }
        if (iz == 0 && fabs(vy[I(ix,iy,0)]) > badlim &&
             fabs(vy[I(ix,iy,1)]) < badlim) {
          vy[I(ix,iy,0)] = vy[I(ix,iy,1)];
        }
        if (iz == 0) vz[I(ix,iy,iz)] = 0.;
      }
    }
  }

/* loop on volume */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {
        ipos = I(ix,iy,iz);

/* velocity components */
        usurf = vx[I(ix,iy,0)];
        vsurf = vy[I(ix,iy,0)];
        u = vx[I(ix,iy,iz)];
        v = vy[I(ix,iy,iz)];
        w = vz[I(ix,iy,iz)];

/* compute storm-relative velocity components */
        if (fabs(u) < badlim) {
          urel = u - *sx;
        }
        else urel = bad;
        if (fabs(v) < badlim) {
          vrel = v - *sy;
        }
        else vrel = bad;

/* compute absolute vorticity components */
        ox[ipos] = - zd(ix,iy,iz,vy);
        oy[ipos] = zd(ix,iy,iz,vx);
        xdoz = xd(ix,iy,iz,vy);
        ydoz = yd(ix,iy,iz,vx);
        if (fabs(xdoz) < badlim && fabs(ydoz) < badlim) {
          oz[ipos] = xdoz - ydoz + f;
        }
        else oz[ipos] = bad;

/* assume horizontal vorticity zero at surface */
        if (iz == 0) ox[ipos] = oy[ipos] = 0.;

/* compute friction */
        if (iz <= nbl) {
          if (fabs(usurf) < badlim && fabs(vsurf) < badlim) {
            uabs = pow(usurf*usurf + vsurf*vsurf,.5);
            fx[ipos] = - cdrag*uabs*usurf*dbl*rho[0]/rho[iz];
            fy[ipos] = - cdrag*uabs*vsurf*dbl*rho[0]/rho[iz];
          }
          else {
            fx[ipos] = fy[ipos] = bad;
          }
        }
        else {
          fx[ipos] = fy[ipos] = 0.;
        }

/* compute absolute vorticity flux */
        if (fabs(urel) < badlim && fabs(ox[ipos]) < badlim && fabs(oz[ipos])
             < badlim && fabs(vz[ipos]) < badlim && fabs(fy[ipos]) < badlim) {
          avx[ipos] = urel*oz[ipos] - vz[ipos]*ox[ipos] - fy[ipos];
        }
        else avx[ipos] = bad;
        if (fabs(vrel) < badlim && fabs(oy[ipos]) < badlim && fabs(oz[ipos])
             < badlim && fabs(vz[ipos]) < badlim && fabs(fx[ipos]) < badlim) {
          avy[ipos] = vrel*oz[ipos] - vz[ipos]*oy[ipos] + fx[ipos];
        }
        else avy[ipos] = bad;
      }
    }
  }

/* compute absolute vorticity flux convergence */
  for (ix = 0; ix < nx; ix++) {
    for (iy = 0; iy < ny; iy++) {
      for (iz = 0; iz < nz; iz++) {
        ipos = I(ix,iy,iz);
        xdavx = xd(ix,iy,iz,avx);
        ydavy = yd(ix,iy,iz,avy);
        if (fabs(xdavx) < badlim && fabs(ydavy) < badlim) {
          soz[ipos] = - xdavx - ydavy;
        }
        else soz[ipos] = bad;
      }
    }
  }
}

/* derivatives */
float xd(ix,iy,iz,var)
long ix,iy,iz;
float *var;
{
  float v1,v2;
  if (ix > 0 && ix < nx - 1 &&
       fabs(v1 = var[I(ix - 1,iy,iz)]) < badlim &&
       fabs(v2 = var[I(ix + 1,iy,iz)]) < badlim) {
    return((v2 - v1)/(x[ix + 1] - x[ix - 1]));
  }
  else {
    return(bad);
  }
}

float yd(ix,iy,iz,var)
long ix,iy,iz;
float *var;
{
  float v1,v2;
  if (iy > 0 && iy < ny - 1 &&
       fabs(v1 = var[I(ix,iy - 1,iz)]) < badlim &&
       fabs(v2 = var[I(ix,iy + 1,iz)]) < badlim) {
    return((v2 - v1)/(y[iy + 1] - y[iy - 1]));
  }
  else {
    return(bad);
  }
}

float zd(ix,iy,iz,var)
long ix,iy,iz;
float *var;
{
  float v1,v2;
  if (iz > 0 && iz < nz - 1 &&
       fabs(v1 = var[I(ix,iy,iz - 1)]) < badlim &&
       fabs(v2 = var[I(ix,iy,iz + 1)]) < badlim) {
    return((v2 - v1)/(z[iz + 1] - z[iz - 1]));
  }
  else {
    return(bad);
  }
}
