/*
  Rafalpha.c
  
  usage: rafalpha [n]

  Lists ptime and all variable field elements of a Candis file
  in ASCII scientific notation with n significant digits,
  seperated by spaces, one line per second of data.
*/




#include <stdio.h>
#include <math.h>
#include <string.h>
#include "cdfhdr.h"

#define DEFSIGDIGS 6      /*  Default number of significant digits  */

char hbuff[HBMAX][LINE];  /*  Header buffer  */
float *sbuff,*vbuff;      /*  Static and variable slice buffers  */
long nsbuff,nvbuff;       /*  Number of elements in slice buffers  */
float *var[200];          /*  Pointers to beginning of each field in 
                                 variable slice  */
struct field *fp;         /*  Pointer to a field description structure  */
long nt;                  /*  Number of elements in each field  */
int i=0,j=0;              /*  Loop counters  */
char format[40];          /*  format string for output  */
int numfields=0;          /*  Number of fields in variable slice  */
int sigdigs=0;            /*  Number of significant digits to be written  */
char indexname[80];       /*  Name of index field and var field dimensions  */

int main(argc,argv)
int argc;
char *argv[];
{

  /*  Output usage if more than one argument is present or arg1 is bad */

  if(argc>2 || (argc==2 && (atoi(argv[1])<1 || atoi(argv[1])>10)))
  {
    if(argc>2)
      fprintf(stderr,"%s: Invalid number of arguments!\n",argv[0]);

    if(argc>=2 && (atoi(argv[1])<1 || atoi(argv[1])>10))
      fprintf(stderr,"%s: Invalid significant digits argument!\n",argv[0]);

    fprintf(stderr,"\n%s: Usage:\n",argv[0]);
    fprintf(stderr,"%s:    %s [n]\n",argv[0],argv[0]);
    fprintf(stderr,"%s:    n = number of significant digits to output, 1-10\n"
        ,argv[0]);

    exit(1);
  }


  /*  Set number of significant digits written from argument 1  */

  if(argc==1)
    sigdigs=DEFSIGDIGS;
  else
    sigdigs=atoi(argv[1]);

  switch(sigdigs)
  {
  case 1:
    strcpy(format,"%2.0e ");
    break;
  case 2:
    strcpy(format,"%3.1e ");
    break;
  case 3:
    strcpy(format,"%4.2e ");
    break;
  case 4:
    strcpy(format,"%5.3e ");
    break;
  case 5:
    strcpy(format,"%6.4e ");
    break;
  case 6:
    strcpy(format,"%7.5e ");
    break;
  case 7:
    strcpy(format,"%8.6e ");
    break;
  case 8:
    strcpy(format,"%9.7e ");
    break;
  case 9:
    strcpy(format,"%10.8e ");
    break;
  case 10:
    strcpy(format,"%11.9e ");
    break;
  }


  /*  Read header, make sure of float format  */

  gethdr(stdin,hbuff,HBMAX);
  if(getfmt(hbuff,HBMAX)!='f')
  {
    fprintf(stderr,"%s: Input file must be float format!\n",argv[0]);
    exit(1);
  }


  /*  Allocate and fill buffers for static and variable slice  */

  nvbuff=elemcnt(hbuff,HBMAX,'v');
  vbuff=getbuff(nvbuff);
  nsbuff=elemcnt(hbuff,HBMAX,'s');
  sbuff=getbuff(nsbuff);

  getslice(stdin,nsbuff,sbuff);

  getslice(stdin,nvbuff,vbuff);


  /*  Get pointer to start of index field  */

  if((fp = getfld(hbuff,HBMAX,'v',0)) == NULL)
  {
    fprintf(stderr,"%s: Input file must contain at least one variable field!"
        ,argv[0]);
    exit(0);
  }

  strcpy(indexname,fp->dname1);
  if((fp=seekfld(hbuff,HBMAX,'s',indexname))==NULL)
  {
    fprintf(stderr,"%s: Index field (%s) must be present!\n",argv[0],indexname);
    exit(1);
  }
  if(fp->dim!=1)
  {
    fprintf(stderr,"%s: Index field (%s) must be one-dimensional!\n",argv[0]
       ,indexname);
    exit(1);
  }
  nt=fp->dsize1;
  var[1]=getptr(hbuff,HBMAX,sbuff,'s','d',fp->fname);
  numfields=1;


  /*  Get pointers to starts of variable fields  */

  while((fp=getfld(hbuff,HBMAX,'v',numfields-1))!=NULL)
  {
    var[++numfields]=getptr(hbuff,HBMAX,vbuff,'v','d',fp->fname);
    if(fp->dim!=1)
    {
      fprintf(stderr,"%s: Error in field \"%s\":\n",argv[0],fp->fname);
      fprintf(stderr,"%s: All variable fields must be one-dimensional!\n"
          ,argv[0]);
      exit(1);
    }
    if(fp->dsize1!=nt)
    {
      fprintf(stderr,"%s: Error in field \"%s\":\n",argv[0],fp->fname);
      fprintf(stderr,"%s: All variable fields must have same lengths!\n"
          ,argv[0]);
      exit(1);
    }
    if(strcmp(indexname,fp->dname1))
    {
      fprintf(stderr,"%s: Error in field \"%s\":\n",argv[0],fp->fname);
      fprintf(stderr,"%s: All variable fields must have same dimension name!\n"
         ,argv[0]);
      exit(1);
    }
  }


  /*  Write fields to standard output  */

  for(i=1;i<=nt;i++)
  {
    for(j=1;j<=numfields;j++)
    {
      printf(format,var[j][i-1]);
    }
    printf("\n");
  }

  exit(0);
}


