/* rafclouds -- Do corrections on various in-cloud thermodynamic
 * parameters in particular:
 * 1. Despike the Ophir radiometer, remove the overshoot, and combine
 *    the scaled high pass part with the low pass part of the Rosemount
 *    temperature probe.
 * 2. Despike the UV hygrometer and combine the high pass part with
 *    the low pass part of the dewpointer mixing ratio.
 * 3. Delag the total water probe mixing ratio and combine the high
 *    pass part with the low pass part of the dewpointer mixing ratio.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

/* RAF variable names */
#define TIME "ptime"
#define OPHIR "xato"
#define TEMP "atb"
#define MIXRAT "mr"
#define UVMIXRAT "xmruv"
#define LYMANALPHA "vla"
#define TOWMIXRAT "xmrtow"
#define LIQUID "plwcf"

/* new variables */
#define OPHIR2 "oat2"               /* despiked and overshoot removed */
#define UVMIXRAT2 "mruv2"           /* despiked mruv */
#define CTEMP "ctemp"               /* corrected temperature */
#define CMR "cmr"                   /* corrected mixing ratio */
#define CMRT "cmrt"                 /* corrected total water mixing ratio */
#define CLEANATB "cleanatb"         /* liquid water response removed */

/* threshold and cutoff for temperature despiking */
#define TEMPSPIKE 0.5
#define MRSPIKE 0.5
#define DSCUTOFF 0.2

/* dry out time and dry threshold for atb in seconds */
#define DRYTIME 20
#define DRYTHRESH 0.03
#define DRYMAX 10.

/* bad data value for cleanatb */
#define BARF 10000.

/* scale factors for high freq response of various instruments */
#define OATSCALE 1.0
#define MRUVSCALE 1.0
#define MRTOWSCALE 1.0
#define VLASCALE 3.1

/* command line argument values */
float tau_temp,tau_mr,tau_mrt;
long mrtlag;

/* Candis stuff */
char hbuf[HBMAX][LINE];
char cline[LINE];
float *sbuff,*vbuff;
long nsbuffin,nvbuffin;
long nsbuff,nvbuff;
long nvalues;
float delt;
float *tptime,*oat,*atb,*liquid,*mr,*mruv,*vla,*xmrtow;
float *cleanatb,*oat2,*mruv2,*ctemp,*cmr,*cmrt;
struct field *fp;

/* other stuff */
float fastmrscale;

main(argc,argv)
int argc;
char *argv[];
{
  long poop,loop;
  float cuton,cutoff;
  float datbdt,lsdatbdt;

/* get args */
  if (argc != 5) {
    fprintf(stderr,"Usage: rafclouds tau_temp tau_mr tau_mrt mrtlag\n");
    exit(1);
  }
  tau_temp = atof(argv[1]);
  tau_mr = atof(argv[2]);
  tau_mrt = atof(argv[3]);
  mrtlag = atoi(argv[4]);
  if (mrtlag < 0) {
    fprintf(stderr,"rafclouds: mrtlag cannot be negative\n");
    exit(1);
  }

/* read the header */
  gethdr(stdin,hbuf,HBMAX);
  if (getfmt(hbuf,HBMAX) != 'f') {
    fprintf(stderr,"rafclouds: float format expected\n");
    exit(1);
  }

/* get input slice sizes */
  nsbuffin = elemcnt(hbuf,HBMAX,'s');
  nvbuffin = elemcnt(hbuf,HBMAX,'v');

/* add comment */
  if (gotfld(UVMIXRAT)) addcline(hbuf,HBMAX,"rafclouds(mruv):\n");
  else addcline(hbuf,HBMAX,"rafclouds(vla):\n");
  sprintf(cline,"  %s %s %s %s\n",argv[1],argv[2],argv[3],argv[4]);
  addcline(hbuf,HBMAX,cline);

/* add parameters */
  sprintf(cline,"%g",TEMPSPIKE);
  addpar(hbuf,HBMAX,"tempspike",cline);
  sprintf(cline,"%g",DSCUTOFF);
  addpar(hbuf,HBMAX,"dscutoff",cline);
  sprintf(cline,"%g",OATSCALE);
  addpar(hbuf,HBMAX,"oatscale",cline);
  sprintf(cline,"%g",MRUVSCALE);
  addpar(hbuf,HBMAX,"mruvtscale",cline);
  sprintf(cline,"%g",VLASCALE);
  addpar(hbuf,HBMAX,"vlascale",cline);
  sprintf(cline,"%g",MRTOWSCALE);
  addpar(hbuf,HBMAX,"mrtowtscale",cline);
  sprintf(cline,"%g",DRYTIME);
  addpar(hbuf,HBMAX,"drytime",cline);
  sprintf(cline,"%g",DRYTHRESH);
  addpar(hbuf,HBMAX,"drythresh",cline);
  sprintf(cline,"%g",DRYMAX);
  addpar(hbuf,HBMAX,"drymax",cline);

/* see if we have a time index field */
  if ((fp = seekfld(hbuf,HBMAX,'s',TIME)) == NULL || fp->dim != 1 ||
       strcmp(fp->dname1,TIME) != 0) {
    fprintf(stderr,"rafclouds: index field %s not found\n",TIME);
    exit(1);
  }
  nvalues = fp->dsize1;

/* add the new fields */
  if (gotfld(OPHIR) && gotfld(TEMP) && gotfld(LIQUID)) {
    addfld(hbuf,HBMAX,'v',OPHIR2,10.,0.,'s',1,TIME,nvalues);
    addfld(hbuf,HBMAX,'v',CTEMP,10.,0.,'s',1,TIME,nvalues);
    addfld(hbuf,HBMAX,'v',CLEANATB,10.,0.,'s',1,TIME,nvalues);
  }
  if (gotfld(MIXRAT) && (gotfld(UVMIXRAT) || gotfld(LYMANALPHA))) {
    addfld(hbuf,HBMAX,'v',UVMIXRAT2,10.,0.,'s',1,TIME,nvalues);
    addfld(hbuf,HBMAX,'v',CMR,10.,0.,'s',1,TIME,nvalues);
  }
  if (gotfld(MIXRAT) && gotfld(TOWMIXRAT)) {
    addfld(hbuf,HBMAX,'v',CMRT,10.,0.,'s',1,TIME,nvalues);
  }

/* get the output slice sizes and allocate space */
  nsbuff = elemcnt(hbuf,HBMAX,'s');
  nvbuff = elemcnt(hbuf,HBMAX,'v');
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbuff);

/* define the pointers */
  tptime = getptr(hbuf,HBMAX,sbuff,'s','d',TIME);
  oat = getptr(hbuf,HBMAX,vbuff,'v','r',OPHIR);
  atb = getptr(hbuf,HBMAX,vbuff,'v','r',TEMP);
  liquid = getptr(hbuf,HBMAX,vbuff,'v','r',LIQUID);
  mr = getptr(hbuf,HBMAX,vbuff,'v','r',MIXRAT);
  mruv = getptr(hbuf,HBMAX,vbuff,'v','r',UVMIXRAT);
  vla = getptr(hbuf,HBMAX,vbuff,'v','r',LYMANALPHA);
  xmrtow = getptr(hbuf,HBMAX,vbuff,'v','r',TOWMIXRAT);
  oat2 = getptr(hbuf,HBMAX,vbuff,'v','r',OPHIR2);
  cleanatb = getptr(hbuf,HBMAX,vbuff,'v','r',CLEANATB);
  mruv2 = getptr(hbuf,HBMAX,vbuff,'v','r',UVMIXRAT2);
  ctemp = getptr(hbuf,HBMAX,vbuff,'v','r',CTEMP);
  cmr = getptr(hbuf,HBMAX,vbuff,'v','r',CMR);
  cmrt = getptr(hbuf,HBMAX,vbuff,'v','r',CMRT);

/* read the slices */
  getslice(stdin,nsbuffin,sbuff);
  getslice(stdin,nvbuffin,vbuff);

/* get the time step */
  delt = tptime[1] - tptime[0];
  if (delt != 1.) {
    fprintf(stderr,"rafclouds: warning: delt = %g != 1 second\n",delt);
  }

/* do the temperature computations */
  if (oat && atb && liquid) {

/* linearly interpolate atb across regions with liquid water */
    dryclean(nvalues,atb,liquid,cleanatb,DRYTHRESH,DRYMAX,DRYTIME);

/* despike the Ophir temperature */
    despike(nvalues,oat,oat2,DSCUTOFF,TEMPSPIKE);

/* mix scaled high freq part of oat2 and low freq part of cleaned atb */
    if (tau_temp < delt) {
      fprintf(stderr,
           "rafclouds: tau_temp is too small -- make greater than delt\n");
      exit(1);
    }
    cutoff = delt/tau_temp;
    combine(nvalues,cutoff,OATSCALE,cleanatb,oat2,ctemp);
  }

/* high freq part of mruv plus low freq part of mr */
  if (mr && (mruv || vla)) {

/* despike mruv if we have it */
    if (mruv) {
      despike(nvalues,mruv,mruv2,DSCUTOFF,MRSPIKE);
      fastmrscale = MRUVSCALE;
    }

/* else get vla */
    else {
      for (loop = 0; loop < nvalues; loop++) {
        mruv2[loop] = vla[loop];
      }
      fastmrscale = VLASCALE;
    }

/* combine */
    if (tau_mr < delt) {
      fprintf(stderr,
           "rafclouds: tau_mr is too small -- make greater than delt\n");
      exit(1);
    }
    cutoff = delt/tau_mr;
    combine(nvalues,cutoff,fastmrscale,mr,mruv2,cmr);
  }

/* do the total water computations */
  if (mr && xmrtow) {

/* remove lag in total water mixing ratio */
    for (loop = 0; loop < nvalues; loop++) {
      poop = loop + mrtlag;
      if (poop >= nvalues) poop = nvalues - 1;
      xmrtow[loop] = xmrtow[poop];
    }

/* combine high frequency part of total water mixing ratio with low freq mr */
    if (tau_mrt < delt) {
      fprintf(stderr,
           "rafclouds: tau_mrt is too small -- make greater than delt\n");
      exit(1);
    }
    cutoff = delt/tau_mrt;
    combine(nvalues,cutoff,MRTOWSCALE,mr,xmrtow,cmrt);
  }

/* write the header and slices */
  puthdr(stdout,hbuf,HBMAX);
  putslice(stdout,nsbuff,sbuff);
  putslice(stdout,nvbuff,vbuff);
  exit(0);
}

/* gotfld -- checks to see if field exists and has proper characteristics */
gotfld(fname)
char *fname;
{
  if ((fp = seekfld(hbuf,HBMAX,'v',fname)) != NULL &&
       fp->dim == 1 && strcmp(fp->dname1,TIME) == 0) return(1);
  else return(0);
}

/* combine -- This program combines the low frequency part of one
 * field with the high frequency part of a second field and returns
 * the results in a third field.  The third field may be either of
 * the input fields or a separate field.
 */
combine(nv,cutoff,hiscale,lofld,hifld,newfld)
long nv;                        /* number of points */
float cutoff;                   /* filter parameter */
float hiscale;                  /* scaling factor for hi freq field */
float *lofld,*hifld,*newfld;    /* fields */
{
  long loop;
  float *lotmp,*hitmp;

/* allocate temporary space */
  lotmp = getbuff(nv);
  hitmp = getbuff(nv);

/* do filtering */
  lopass(nv,cutoff,lofld,lotmp);
  hipass(nv,cutoff,hifld,hitmp);

/* combine results */
  for (loop = 0; loop < nv; loop++) {
    newfld[loop] = lotmp[loop] + hiscale*hitmp[loop];
  }

/* free temporary space */
  free(lotmp);
  free(hitmp);
}

/* despike -- This routine takes out one time step spikes greater than
 * a specified magnitude */
despike(nv,fld,newfld,cutoff,thresh)
long nv;                        /* number of points */
float *fld;                     /* field to be despiked */
float *newfld;                  /* despiked field */
float cutoff;                   /* low pass cutoff for despiking */
float thresh;                   /* spiking threshold */
{
  long poop,loop,pos1,pos2;
  float *tmpfld;
  float startval,endval,left,right,denom;

/* allocate temporary space */
  tmpfld = getbuff(nv);

/* copy fld to newfld */
  for (loop = 0; loop < nv;loop++) newfld[loop] = fld[loop];

/* low pass filter the field of interest */
  lopass(nv,cutoff,fld,tmpfld);

/* declare as bad (0) where abs(tmpfld - fld) exceeds thresh */
  for (loop = 0; loop < nv; loop++) {
    if (fabs(tmpfld[loop] - fld[loop]) < thresh) tmpfld[loop] = 1.;
    else tmpfld[loop] = 0.;
  }

/* fill in regions that have had spike values declared by linear interp */
  pos1 = 0;
  for (loop = 0; loop < nv - 1; loop++) {

/* bad to good */
    if (tmpfld[loop] == 0. && tmpfld[loop + 1] > 0.) {
      pos2 = loop;
      endval = fld[loop + 1];
      if (pos1 == 0) startval = endval;

/* fill in gap */
      for (poop = pos1; poop <= pos2; poop++) {
        denom = pos2 - pos1 + 2;
        left = (pos2 + 1 - poop)/denom;
        right = 1. - left;
        newfld[poop] = left*startval + right*endval;
      }
    }

/* good to bad */
    if (tmpfld[loop] > 0. && tmpfld[loop + 1] == 0.) {
      pos1 = loop + 1;
      startval = fld[loop];
    }
  }

/* free temporary space */
  free(tmpfld);
}

/* lopass -- do a low pass filtering job */
lopass(nv,cutoff,fld,newfld)
long nv;                        /* number of points */
float cutoff;                   /* filter parameter */
float *fld;                     /* field to be filtered */
float *newfld;                  /* result of filtering */
{
  long loop;
  float cuton;

/* check parameter range */
  if (cutoff < 0. || cutoff >= 1.) {
    fprintf(stderr,"lopass: cutoff parameter %g out of range\n");
    exit(1);
  }
  cuton = 1. - cutoff;

/* forward filtering */
  newfld[0] = fld[0];
  for (loop = 1; loop < nv; loop++) {
    newfld[loop] = cutoff*fld[loop] + cuton*newfld[loop - 1];
  }

/* backward filtering */
  for (loop = nv - 2; loop >= 0; loop--) {
    newfld[loop] = cutoff*newfld[loop] + cuton*newfld[loop + 1];
  }
}

/* hipass -- do a high pass filtering job */
hipass(nv,cutoff,fld,newfld)
long nv;                        /* number of points */
float cutoff;                   /* filter parameter */
float *fld;                     /* field to be filtered */
float *newfld;                  /* result of filtering */
{
  long loop;
  lopass(nv,cutoff,fld,newfld);
  for (loop = 0; loop < nv; loop++) {
    newfld[loop] = fld[loop] - newfld[loop];
  }
}

/* dryclean -- ignore atb values inside clouds -- linearly interp across */
dryclean(nv,atb,liquid,cleanatb,drythresh,drymax,drytime)
long nv;                        /* number of points */
float *atb;                     /* Rosemount temp */
float *liquid;                  /* liquid water (g/kg) */
float *cleanatb;                /* cleaned Rosemount temp */
float drythresh;                /* threshold below which air considered dry */
float drymax;                   /* threshold above which measurement bad */
long drytime;                   /* steps it takes for Rosemount to dry out */
{
  float denom,a,b;
  long loop,poop,bflag;

/* copy atb to cleanatb */
  for (loop = 0; loop < nv; loop++) cleanatb[loop] = atb[loop];

/* loop thru making bad regions immediately before and after liquid is found */
  for (loop = drytime + 1; loop < nv - drytime - 1; loop++) {
    bflag = 0;
    for (poop = loop - drytime; poop <= loop + drytime; poop++) {
      if (liquid[poop] > drythresh && liquid[poop] < drymax) bflag++;
    }
    if (bflag) cleanatb[loop] = BARF;
  }

/* loop again to interpolate */
  bflag = 0;
  for (loop = 1; loop < nv; loop++) {
    if (cleanatb[loop] == BARF && cleanatb[loop - 1] != BARF) {
      bflag = loop - 1;
    }
    if (cleanatb[loop] != BARF && cleanatb[loop - 1] == BARF) {
      for (poop = bflag + 1; poop < loop; poop++) {
        denom = loop - bflag;
        if (denom < 2.) denom = 2.;
        b = (poop - bflag)/denom;
        a = 1. - b;
        cleanatb[poop] = a*cleanatb[bflag] + b*cleanatb[loop];
      }
    }
  }
}
