/* rafhirate -- This routine provides a means of plotting high data
 * rate (> 1 Hz) data translated from NCAR RAF tapes by rtape and
 * rafread.  This is needed because of the peculiar way RAF writes
 * its data tapes.  Data collected over one second intervals are
 * all included in one logical record.  Data from particular sensors
 * are stored sequentially within this logical record, with one
 * element for 1 Hz data, 5 for 5 Hz data, etc.  Thus, data from
 * a single sensor are not stored sequentially.  The solution to this
 * problem is to use cdfcat to rearrange data into two dimensional
 * arrays for each sensor, with the high rate on one axis and the
 * low rate on the other axis, and then to use this program to
 * convert these 2-D arrays into 1-D arrays.  The way cdfcat arranges
 * the arrays, all data will be in the correct order.  In addition,
 * rafihrate rearranges the time index fields so that there will be
 * something to plot the data against.
 */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

char hbin[HBMAX][LINE],hbout[HBMAX][LINE]; /* header buffers */
char cline[LINE];                          /* comment line buffer */
char tptime[LINE];                         /* tapetimename */
float *sbuffin,*sbuffout,*vbuff;           /* slice buffers */
long nsbuffin,nsbuffout,nvbuff;            /* buffer lengths */
struct field *fp;                          /* temp pointer */
long tpsize,tpoffset;                      /* tptime field info */
int goop,loop,poop;                        /* looping vars */
int rate;                                  /* data sample rate */
float *tpin,*tpout,*et,*hrate;             /* field pointers */

main(argc,argv)
int argc;
char *argv[];
{
/* check args */
  if (argc != 2) {
    fprintf(stderr,"Usage: rafhirate tape_time_name\n");
    exit(1);
  }
  strcpy(tptime,argv[1]);
/* get header from standard input */
  gethdr(stdin,hbin,HBMAX);
  if (getfmt(hbin,HBMAX) != 'f') {
    fprintf(stderr,"rafhirate: need float format on input\n");
    exit(1);
  }
/* make new header */
  nullhdr(hbout,HBMAX,"float");
  copycmt(hbout,HBMAX,hbin,HBMAX);
  copypar(hbout,HBMAX,hbin,HBMAX);
  sprintf(cline,"rafhirate: %s\n",tptime);
  addcline(hbout,HBMAX,cline);
/* find tape time static field */
  if ((fp = seekfld(hbin,HBMAX,'s',tptime)) == NULL) {
    fprintf(stderr,"rafhirate: can't find %s in static slice\n",tptime);
    exit(1);
  }
/* extract size and offset */
  tpsize = fp->tsize;
  tpoffset = fp->ptr;
/* add to new header */
  addfld(hbout,HBMAX,'s',tptime,1.,0.,'l',1,tptime,tpsize);
/* add a companion starting from zero */
  addfld(hbout,HBMAX,'s',"etime",1.,0.,'l',1,"etime",tpsize);
/* look for high rate index fields and store rate data */
  loop = poop = 0;
  while ((fp = getfld(hbin,HBMAX,'s',loop++)) != NULL) {
    if ((rate = getrate(fp->fname)) > 0) {
/* add modified index field */
      addfld(hbout,HBMAX,'s',fp->fname,1000.,0.,'l',1,fp->fname,rate*tpsize);
    }
  }
/* fix up variable slice section */
  loop = 0;
  while ((fp = getfld(hbin,HBMAX,'v',loop++)) != NULL) {
    if ((fp->dim == 2) && (strcmp(fp->dname1,tptime) == 0)) {
      addfld(hbout,HBMAX,'v',fp->fname,fp->smul,fp->sadd,'l',1,fp->dname2,
           fp->dsize1*fp->dsize2);
    }
    else if ((fp->dim == 1) && (strcmp(fp->dname1,tptime) == 0)) {
      addfld(hbout,HBMAX,'v',fp->fname,fp->smul,fp->sadd,'l',fp->dim,
           "etime",fp->dsize1);
    }
    else {
      fprintf(stderr,"rafhirate: %s dimension unexpectedly = %d\n",
           fp->fname,fp->dim);
    }
  }
/* write new header to standard output */
  puthdr(stdout,hbout,HBMAX);
/* allocate buffers */
  nsbuffin = elemcnt(hbin,HBMAX,'s');
  nsbuffout = elemcnt(hbout,HBMAX,'s');
  nvbuff = elemcnt(hbin,HBMAX,'v');
  if (nvbuff != elemcnt(hbout,HBMAX,'v')) {
    fprintf(stderr,"rafhirate: variable slice section of output wrong\n");
    exit(1);
  }
  sbuffin = getbuff(nsbuffin);
  sbuffout = getbuff(nsbuffout);
  vbuff = getbuff(nvbuff);
/* get static slice */
  getslice(stdin,nsbuffin,sbuffin);
/* generate new static slice */
  tpin = getptr(hbin,HBMAX,sbuffin,'s','d',tptime);
  tpout = getptr(hbout,HBMAX,sbuffout,'s','d',tptime);
  et = getptr(hbout,HBMAX,sbuffout,'s','d',"etime");
  for (loop = 0; loop < tpsize; loop++) {
    tpout[loop] = tpin[loop];
    et[loop] = tpin[loop] - tpin[0];
  }
  goop = 0;
  while ((fp = getfld(hbout,HBMAX,'s',goop++)) != NULL) {
    if ((rate = getrate(fp->fname)) > 0) {
      hrate = sbuffout + fp->ptr;
      for (loop = 0; loop < tpsize; loop++) {
        for (poop = 0; poop < rate; poop++) {
          hrate[loop*rate + poop] = et[loop] + ((float)poop)/((float)rate);
        }
      }
    }
  }
/* write static slice */
  putslice(stdout,nsbuffout,sbuffout);
/* transfer variable slices */
  while (getslice(stdin,nvbuff,vbuff) != EOF) {
    putslice(stdout,nvbuff,vbuff);
  }
  exit(0);
}

/* see if name is of form "time"number and return number */
int getrate(text)
char *text;
{
  if (strlen(text) < 5) return(0);
  if (strncmp(text,"time",4) == 0) return(atoi(text + 4));
  else return(0);
}
