/* rafwinds -- This program computes winds from raw RAF data. \
 * We assume certain RAF variable names, and insert certain offsets \
 * and scale factors, as well as a delay between arinc and sync data. \
 * If gpsflag is set, GPS updating is used. \
 */
#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"

#define TCUT_VZ 30.               /* low pass/high pass cutoff (sec) for vz */
#define TCUT_GPS 250.             /* low pass/high pass cutoff (sec) for GPS */

#define ROLL "roll"               /* aircraft roll */
#define PITCH "pitch"             /* aircraft pitch */
#define HEADING "thi"             /* arinc aircraft heading */
#define HEADING2 "thdg"           /* heading on laser INS */
#define ARINC_HEADING 1           /* set to 1 if using arinc heading */
#define UP "vew"                  /* arinc aircraft EW ground speed */
#define VP "vns"                  /* arinc aircraft NS ground speed */
#define WPOLD "vzi"               /* arinc aircraft vertical speed */
#define WPOLD2 "ivspd"            /* new vertical aircraft speed (was wp3) */
#define TAS "tasr"                /* true airspeed */
#define ATTACK "akrd"             /* angle of attack */
#define SIDESLIP "ssrd"           /* sideslip angle */
#define GUP "gvew"                /* GPS EW ground speed */
#define GVP "gvns"                /* GPS NS ground speed */
#define ALON "alon"               /* arinc INS longitude */
#define ALAT "alat"               /* arinc INS latitude */
#define GLON "glon"               /* GPS longitude */
#define GLAT "glat"               /* GPS latitude */
#define TIME "tptime"             /* seconds after midnight (index field) */
#define TIME2 "ptime"             /* seconds after midnight (index field) */
#define UNEW "unew"               /* my new EW wind */
#define VNEW "vnew"               /* my new NS wind */
#define WNEW "wnew"               /* my new vertical wind */
#define PVEL "pvel"               /* aircraft vertical vel from pres rate */
#define WPCOR "vzic"              /* WPOLD with constant and trend removed */
#define WP "vud"                  /* pres corr vertical aircraft velocity */
#define PRES "psfdc"              /* pressure in mb */
#define TEMP "atrf"               /* reverse flow temp in C */
#define TMLAG "tmlag"             /* time lag between arinc and sync data */
#define MR "mr"                   /* mixing ratio, g/kg */

#define DTOR 57.2958              /* degrees per radian */

#define MRFACT 2.99e-4            /* correction factor for mixing ratio on
                                   * true airspeed -- derived from unpublished
                                   * work of DLR */

/* DLR TAS correction:
 *
 *     V = (T*(A + B*r)*((1 + pd/ps)^(2/7 - C*r) - 1))^0.5
 * 
 *     T = temperature (K)
 *     pd = dynamic pressure (mb)
 *     ps = static pressure (mb)
 *     r = water vapor mixing ratio (g/g)
 *     A = 2009.35
 *     B = 1660
 *     C = 0.065
 *     
 *     This approximates to
 *     
 *     V = V_dry*(1 + (B/(2*A) - 7*C/4)*r) = V_dry*(1 + 0.299*r)
 *     
 *     since pd << ps in regions where it matters.
 */

/* bounds on mixing ratio for airspeed correction */
#define MRMIN 0.
#define MRMAX 20.

/* candis stuff */
char hbuff[HBMAX][LINE];
char cline[LINE],pline[LINE];
long nsbuff,nvbin,nvbout;
float *sbuff,*vbuff;
float *roll,*pitch,*heading,*up,*vp,*wpold,*tas,*attack,*sideslip;
float *gup,*gvp,*alon,*alat,*glon,*glat,*tmlag;
float *tptime;
float *unew,*vnew,*wnew;
float *pres,*temp;
float *pvel;
float *wpcor;
float *wp;
float *mr;
float *temporary;
struct field *fp;
float at0,fat,si0,fsi,ftas,fwp,lag;
int gpsflag;
int tasfix;
long nvalues;
float delt,ratdt;
int dtsamp;
int lstep;
float lbot,ltop;
float cutoff;
float rho;
float wpslope;
double wpzero;
float rheading,rpitch,rroll,tanalpha,tanbeta,uaod;
float spsi,cpsi,stheta,ctheta,sphi,cphi;
float derived_lag;
long intlag;
char timename[20];
int optind;

main(argc,argv)
int argc;
char **argv;
{
  float boundmr;
  long loop;
  int syslag;
  int c;

/* check args */
  gpsflag = tasfix = 0;
  while ((c = getopt(argc,argv,"gt")) != -1) {
    if (c == 'g') gpsflag = 1;
    if (c == 't') tasfix = 1;
  }
  if (argc != optind + 7) {
    fprintf(stderr,
         "Usage: rafwinds [-g] [-t] at0 fat si0 fsi ftas fwp lag\n");
    exit(1);
  }
  at0 = atof(argv[optind]);
  fat = atof(argv[optind + 1]);
  si0 = atof(argv[optind + 2]);
  fsi = atof(argv[optind + 3]);
  ftas = atof(argv[optind + 4]);
  fwp = atof(argv[optind + 5]);
  lag = atof(argv[optind + 6]);

/* get header from standard input */
  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"rafwinds: float format expected\n");
    exit(1);
  }

/* get input slice buffer sizes */
  nsbuff = elemcnt(hbuff,HBMAX,'s');
  nvbin = elemcnt(hbuff,HBMAX,'v');

/* add comment to header */
  sprintf(cline,"rafwinds:");
  if (gpsflag) strcat(cline," -g");
  if (tasfix) strcat(cline," -t");
  strcat(cline,"\n");
  addcline(hbuff,HBMAX,cline); 
  sprintf(cline,"  %s %s %s %s\n",argv[optind],argv[optind + 1],
       argv[optind + 2],argv[optind + 3]);
  addcline(hbuff,HBMAX,cline); 
  sprintf(cline,"  %s %s %s\n",argv[optind + 4],argv[optind + 5],
       argv[optind + 6]);
  addcline(hbuff,HBMAX,cline); 

/* add parameters */
  sprintf(cline,"%g",(float)TCUT_VZ);
  addpar(hbuff,HBMAX,"tcut_vz",cline);
  sprintf(cline,"%g",(float)TCUT_GPS);
  addpar(hbuff,HBMAX,"tcut_gps",cline);

/* look for index field */
  strcpy(timename,"noname");
  if ((fp = seekfld(hbuff,HBMAX,'s',TIME)) == NULL) {
    fprintf(stderr,"rafwinds: can't find index field %s, trying %s\n",
         TIME,TIME2);
    if ((fp = seekfld(hbuff,HBMAX,'s',TIME2)) == NULL) {
      fprintf(stderr,"rafwinds: out of ideas on name of time field!\n");
      exit(1);
    }
    else strcpy(timename,TIME2);
  }
  else strcpy(timename,TIME);
  if (fp->dim != 1 || strcmp(fp->fname,fp->dname1) != 0) {
    fprintf(stderr,"rafwinds: %s not an index field\n",timename);
    exit(1);
  }
  nvalues = fp->dsize1;

/* allocate temporary space */
  temporary = getbuff(nvalues);

/* add new fields to header */
  addfld(hbuff,HBMAX,'v',UNEW,100.,0.,'s',1,fp->dname1,nvalues);
  addfld(hbuff,HBMAX,'v',VNEW,100.,0.,'s',1,fp->dname1,nvalues);
  addfld(hbuff,HBMAX,'v',WNEW,100.,0.,'s',1,fp->dname1,nvalues);
  addfld(hbuff,HBMAX,'v',WPCOR,100.,0.,'s',1,fp->dname1,nvalues);
  addfld(hbuff,HBMAX,'v',PVEL,100.,0.,'s',1,fp->dname1,nvalues);
  addfld(hbuff,HBMAX,'v',WP,100.,0.,'s',1,fp->dname1,nvalues);

/* get new variable slice buffer size */
  nvbout = elemcnt(hbuff,HBMAX,'v');

/* allocate buffers */
  sbuff = getbuff(nsbuff);
  vbuff = getbuff(nvbout);

/* allocate pointers */
  tptime = getptr(hbuff,HBMAX,sbuff,'s','d',timename);
  roll = getptr(hbuff,HBMAX,vbuff,'v','d',ROLL);
  pitch = getptr(hbuff,HBMAX,vbuff,'v','d',PITCH);
  if ((heading = getptr(hbuff,HBMAX,vbuff,'v','r',HEADING)) == NULL) {
    heading = getptr(hbuff,HBMAX,vbuff,'v','d',HEADING2);
    fprintf(stderr,"rafwinds: %s not available, using %s\n",HEADING,HEADING2);
  }
  up = getptr(hbuff,HBMAX,vbuff,'v','d',UP);
  vp = getptr(hbuff,HBMAX,vbuff,'v','d',VP);
  if ((wpold = getptr(hbuff,HBMAX,vbuff,'v','r',WPOLD)) == NULL) {
    wpold = getptr(hbuff,HBMAX,vbuff,'v','d',WPOLD2);
    fprintf(stderr,"rafwinds: %s not available, using %s\n",WPOLD,WPOLD2);
  }
  tas = getptr(hbuff,HBMAX,vbuff,'v','d',TAS);
  attack = getptr(hbuff,HBMAX,vbuff,'v','d',ATTACK);
  sideslip = getptr(hbuff,HBMAX,vbuff,'v','d',SIDESLIP);
  gup = getptr(hbuff,HBMAX,vbuff,'v','d',GUP);
  gvp = getptr(hbuff,HBMAX,vbuff,'v','d',GVP);
  alon = getptr(hbuff,HBMAX,vbuff,'v','d',ALON);
  alat = getptr(hbuff,HBMAX,vbuff,'v','d',ALAT);
  glon = getptr(hbuff,HBMAX,vbuff,'v','d',GLON);
  glat = getptr(hbuff,HBMAX,vbuff,'v','d',GLAT);
  mr = getptr(hbuff,HBMAX,vbuff,'v','d',MR);

/* if system time lag doesn't exist, make tmlag a temporary field */
  if ((tmlag = getptr(hbuff,HBMAX,vbuff,'v','r',TMLAG)) == NULL) {
    fprintf(stderr,"rafwinds: no time lag variable -- make system lag zero\n");
    tmlag = getbuff(nvalues);
    syslag = 0;
  }
  else syslag = 1;
  unew = getptr(hbuff,HBMAX,vbuff,'v','d',UNEW);
  vnew = getptr(hbuff,HBMAX,vbuff,'v','d',VNEW);
  wnew = getptr(hbuff,HBMAX,vbuff,'v','d',WNEW);
  pvel = getptr(hbuff,HBMAX,vbuff,'v','d',PVEL);
  wpcor = getptr(hbuff,HBMAX,vbuff,'v','d',WPCOR);
  wp = getptr(hbuff,HBMAX,vbuff,'v','d',WP);
  pres = getptr(hbuff,HBMAX,vbuff,'v','d',PRES);
  temp = getptr(hbuff,HBMAX,vbuff,'v','d',TEMP);

/* write header */
  puthdr(stdout,hbuff,HBMAX);

/* read and write static slice */
  getslice(stdin,nsbuff,sbuff);
  putslice(stdout,nsbuff,sbuff);

/* read variable slice (only one assumed) */
  getslice(stdin,nvbin,vbuff);

/***************** do computations *****************/

/* get time step */
  if(nvalues > 10) delt = (tptime[dtsamp = 10] - tptime[0])/10.;
  else delt = (tptime[dtsamp = nvalues - 1] - tptime[0])/(nvalues - 1);

/* check for gaps that would mess up delt calculation */
  for (loop = 0; loop < dtsamp; loop++) {
    ratdt = (tptime[loop + 1] - tptime[loop])/delt;
    if (ratdt < .999 || ratdt > 1.001) {
      fprintf(stderr,"rafwinds: missing record in delt calc, ratdt = %g\n",
           ratdt);
      exit(1);
    }
  }

/* compute derived time lag and put in tmlag field if system lag exists */
  for (loop = 0; loop < nvalues; loop++) {
    if (syslag) {
      intlag = tptime[loop]/60.;
      derived_lag = tptime[loop] - 60*intlag;
      derived_lag = derived_lag - tmlag[loop];
      while (derived_lag < -30.) derived_lag += 60.;
      while (derived_lag > 30.) derived_lag -= 60.;
      tmlag[loop] = derived_lag;
    }

/* no system lag -- just set lag to zero before correcting */
    else tmlag[loop] = 0.;

/* add in correction */
    tmlag[loop] += lag;
  }

/* correct lags on all arinc fields */
  if (ARINC_HEADING) lagfix(tmlag,delt,nvalues,heading,temporary);
  lagfix(tmlag,delt,nvalues,up,temporary);
  lagfix(tmlag,delt,nvalues,vp,temporary);
  lagfix(tmlag,delt,nvalues,alon,temporary);
  lagfix(tmlag,delt,nvalues,alat,temporary);

/* do DLR's moisture correction to true airspeed */
  if (tasfix) {
    fprintf(stderr,"rafwinds: doing TAS correction for moisture\n");
    for (loop = 0; loop < nvalues; loop++) {
      boundmr = mr[loop];
      if (boundmr < 0.) boundmr = MRMIN;
      if (boundmr > 20.) boundmr = MRMAX;
      tas[loop] *= (1. + MRFACT*boundmr);
    }
  }

/* scale all suspect fields */
  for (loop = 0; loop < nvalues; loop++) {
    attack[loop] = at0 + attack[loop]*fat;
    sideslip[loop] = si0 + sideslip[loop]*fsi;
    tas[loop] = tas[loop]*ftas;
    wpold[loop] = wpold[loop]*fwp;
  }

/* do gps updating if requested */
  if (gpsflag) {
    cutoff = delt/TCUT_GPS;
    combine(nvalues,cutoff,1.,glon,alon,alon);
    combine(nvalues,cutoff,1.,glat,alat,alat);
    combine(nvalues,cutoff,1.,gup,up,up);
    combine(nvalues,cutoff,1.,gvp,vp,vp);
  }

/* compute vertical velocity from pressure rate */
  for (loop = 1; loop < nvalues - 1; loop++) {
    pvel[loop] = (pres[loop + 1] - pres[loop - 1])/(2.*delt);
  }
  pvel[0] = pvel[1];
  pvel[nvalues - 1] = pvel[nvalues - 2];
  for (loop = 0; loop < nvalues; loop++) {
    rho = pres[loop]/(2.8705*(temp[loop] + 273.15));
    pvel[loop] = -pvel[loop]/(.098*rho);
  }

/* take out trend and constant in INS-derived vertical velocity */
  wpslope = (wpold[nvalues - 1] - wpold[0])/(tptime[nvalues - 1] - tptime[0]);
  wpzero = 0.;
  for (loop = 0; loop < nvalues; loop++) {
    wpcor[loop] = wpold[loop] - (tptime[loop] - tptime[0])*wpslope;
    wpzero += wpcor[loop];
  }
  wpzero /= (float)nvalues;
  for (loop = 0; loop < nvalues; loop++) wpcor[loop] -= wpzero;

/* merge INS and pressure vertical velocities */
  cutoff = delt/TCUT_VZ;
  combine(nvalues,cutoff,1.,pvel,wpcor,wp);

/* compute new winds */
  for (loop = 0; loop < nvalues; loop++) {
    rheading = heading[loop]/DTOR;
    rpitch = pitch[loop]/DTOR;
    rroll = roll[loop]/DTOR;
    tanalpha = tan(attack[loop]/DTOR);
    tanbeta = tan(sideslip[loop]/DTOR);
    uaod = tas[loop]/pow(1. + tanalpha*tanalpha + tanbeta*tanbeta,.5);
    spsi = sin(rheading);
    cpsi = cos(rheading);
    stheta = sin(rpitch);
    ctheta = cos(rpitch);
    sphi = sin(rroll);
    cphi = cos(rroll);
    unew[loop] = up[loop] - uaod*(spsi*ctheta + tanbeta*(cpsi*cphi +
         spsi*stheta*sphi) + tanalpha*(spsi*stheta*cphi - cpsi*sphi));
    vnew[loop] = vp[loop] - uaod*(cpsi*ctheta - tanbeta*(spsi*cphi -
         cpsi*stheta*sphi) + tanalpha*(cpsi*stheta*cphi + spsi*sphi));
    wnew[loop] = wp[loop] - uaod*(stheta - tanbeta*ctheta*sphi -
         tanalpha*ctheta*cphi);
  }

/******************* done **************************/

/* write variable slice */
  putslice(stdout,nvbout,vbuff);
  exit(0);
}

/* lagfix -- This routine fixes lags by linear interpolation. */
lagfix(lag,delt,nv,field,temp)
float *lag;                      /* creeping time lag */
float delt;                      /* time step */
long nv;                         /* number of points */
float *field;                    /* the field to be lag corrected */
float *temp;                     /* temporary space */
{
  long loop,pb,pt,ls;
  float lb,lt;

/* loop over field */
  for (loop = 0; loop < nv; loop++) {
    ls = lag[loop]/delt + 100.;
    ls -= 100;
    lb = lag[loop]/delt - ls;
    lt = 1. - lb;
    pt = loop + ls;
    pb = pt + 1;
    if (pt < 0) pt = 0;
    if (pt >= nv) pt = nv - 1;
    if (pb < 0) pb = 0;
    if (pb >= nv) pb = nv - 1;
    temp[loop] = lb*field[pb] + lt*field[pt];
  }

/* return data to field */
  for (loop = 0; loop < nv; loop++) {
    field[loop] = temp[loop];
  }
}

/* combine -- This program combines the low frequency part of one
 * field with the high frequency part of a second field and returns
 * the results in a third field.  The third field may be either of
 * the input fields or a separate field.
 */
combine(nv,cutoff,hiscale,lofld,hifld,newfld)
long nv;                        /* number of points */
float cutoff;                   /* filter parameter */
float hiscale;                  /* scaling factor for hi freq field */
float *lofld,*hifld,*newfld;    /* fields */
{
  long loop;
  float *lotmp,*hitmp;

/* allocate temporary space */
  lotmp = getbuff(nv);
  hitmp = getbuff(nv);

/* do filtering */
  lopass(nv,cutoff,lofld,lotmp);
  hipass(nv,cutoff,hifld,hitmp);

/* combine results */
  for (loop = 0; loop < nv; loop++) {
    newfld[loop] = lotmp[loop] + hiscale*hitmp[loop];
  }

/* free temporary space */
  free(lotmp);
  free(hitmp);
}

/* lopass -- do a low pass filtering job */
lopass(nv,cutoff,fld,newfld)
long nv;                        /* number of points */
float cutoff;                   /* filter parameter */
float *fld;                     /* field to be filtered */
float *newfld;                  /* result of filtering */
{
  long loop;
  float cuton;

/* check parameter range */
  if (cutoff < 0. || cutoff >= 1.) {
    fprintf(stderr,"lopass: cutoff parameter %g out of range\n");
    exit(1);
  }
  cuton = 1. - cutoff;

/* forward filtering */
  newfld[0] = fld[0];
  for (loop = 1; loop < nv; loop++) {
    newfld[loop] = cutoff*fld[loop] + cuton*newfld[loop - 1];
  }

/* backward filtering */
  for (loop = nv - 2; loop >= 0; loop--) {
    newfld[loop] = cutoff*newfld[loop] + cuton*newfld[loop + 1];
  }
}

/* hipass -- do a high pass filtering job */
hipass(nv,cutoff,fld,newfld)
long nv;                        /* number of points */
float cutoff;                   /* filter parameter */
float *fld;                     /* field to be filtered */
float *newfld;                  /* result of filtering */
{
  long loop;
  lopass(nv,cutoff,fld,newfld);
  for (loop = 0; loop < nv; loop++) {
    newfld[loop] = fld[loop] - newfld[loop];
  }
}
