/* uniget -- Translate a netCDF file to Candis format */
#include <stdio.h>
#include <math.h>
#include "netcdf.h"
#include "cdfhdr.h"

/* netCDF stuff */

static long cdfid;                        /* handle for the netcdf file */
static long ndims;                        /* number of dimensions */
static long nvars;                        /* number of variables */
static long natts;                        /* number of attributes */
static long xtendim;                      /* handle for UNLIMITED dimension */
static long exflag[MAX_NC_VARS];          /* flags to indicate UNLIMITED
					    dimension */
static long dimsize[MAX_NC_DIMS];         /* size of dimensions */
static char dimname[MAX_NC_DIMS][MAX_NC_NAME];  /* names of dimensions */
static long dummy[MAX_NC_DIMS];           /* flag for non-existent
					    index fields */
static long varatts;                      /* number of variable-specific
					    attributes */
static char varname[MAX_NC_VARS][MAX_NC_NAME];  /* variable names */
static nc_type datatype[MAX_NC_VARS];    /* type of variable data */
static nc_type atttype;                  /* type of attribute data */
static long vardims[MAX_NC_VARS];         /* number of dimensions for
					    variable */
static long vdlist[MAX_NC_VARS][MAX_VAR_DIMS]; /* dimension list */
static long rvdlist[MAX_NC_VARS][MAX_VAR_DIMS]; /* reduced (no UNLIM)
						  dimension list */
static char history[HBMAX*LINE];         /* history attribute string
					    lists of variable-specific
					    dimensions */
static char attname[MAX_NC_NAME];        /* attribute name */
static long attlen;                       /* number of elements in attribute */
static double attbuff[10];               /* attribute value buffer */
static float validrange[MAX_NC_VARS][2]; /* values of range attributes */
static char *ncbuffer = (char*)NULL;     /* buffer for storing data from
					    ncvarget */
static long buffersize = 0;              /* current size of above buffer */
static long start[5],count[5];            /* arrays for controling ncvarget */

/* Candis stuff */

static char hbuff[HBMAX][LINE];          /* header buffer */
static char cline[LINE];                 /* line buffer */
static char pval[LINE];                  /* parameter value in ascii */
static float bad,badlim;                 /* bad data stuff */
static char ftype[MAX_NC_VARS];          /* field type (s or v) */
static float *sbuff,*vbuff;              /* slice buffer pointers */
static long nsbuff,nvbuff;               /* slice buffer lengths */
static float *fpointer;                  /* general field pointer */
static struct field *fp;                 /* pointer to field structure */
static struct field ptable[MAX_NC_VARS]; /* table for storing field data*/
static float idval,ival0;                /* index parameter values */
static char *infile;                     /* name of input file */

main(argc,argv)
int argc;
char *argv[];
{
  int i,j,k,record,maxrecords;

/* check arguments */

  if (argc != 2) {
    fprintf(stderr,"Usage: uniget netcdf_file\n");
    exit(1);
  }
  infile = argv[argc - 1];

/* open the netcdf */

  if ((cdfid = ncopen(infile,NC_NOWRITE)) == -1) {
    fprintf(stderr,"uniget: can't open netcdf file %s\n",infile);
    exit(1);
  }

/* inquire about its health and get dimension information */

  ncinquire(cdfid,&ndims,&nvars,&natts,&xtendim);
  for (i = 0; i < ndims; i++) {
    ncdiminq(cdfid,i,dimname[i],dimsize + i);

/* transform to lower case */

    lowercase(dimname[i]);
  }

/* get the global attributes */

  strcpy(history,"");
  for (i = 0; i < natts; i++) {
    ncattname(cdfid,NC_GLOBAL,i,attname);
    ncattinq(cdfid,NC_GLOBAL,attname,&atttype,&attlen);

/* check for history stuff */

    if (strcmp(attname,"history") == 0) {
      if (attlen > (HBMAX - 10)*LINE) {
        fprintf(stderr,"uniget: input history attribute too long\n");
        exit(1);
      }
      ncattget(cdfid,NC_GLOBAL,"history",history);
    }
  }

/* get the variables */

  badlim = 0.;
  for (i = 0; i < nvars; i++) {
    ncvarinq(cdfid,i,varname[i],datatype + i,vardims + i,vdlist[i],&varatts);

/* transform to lower case */
    lowercase(varname[i]);

/* eliminate the unlimited dimension from reduced list */

    exflag[i] = k = 0;
    for (j = 0; j < vardims[i]; j++) {
      rvdlist[i][k] = vdlist[i][j];
      if (xtendim == vdlist[i][j]) exflag[i] = 1;
      else k++;
    }

/* look for valid range attribute for each variable */

    validrange[i][0] = -BADLIM;
    validrange[i][1] = BADLIM;
    for (j = 0; j < varatts; j++) {
      ncattname(cdfid,i,j,attname);
      ncattinq(cdfid,i,attname,&atttype,&attlen);
      if (strcmp(attname,"validmin") == 0) {
        if (attlen != 1) {
          fprintf(stderr,"uniget: validmin attribute length not 1\n");
          exit(1);
        }
        ncattget(cdfid,i,"validmin",attbuff);
        floatcon(attbuff,&validrange[i][0],atttype,1);
      }
      if (strcmp(attname,"validmax") == 0) {
        if (attlen != 1) {
          fprintf(stderr,"uniget: validmax attribute length not 1\n");
          exit(1);
        }
        ncattget(cdfid,i,"validmax",attbuff);
        floatcon(attbuff,&validrange[i][1],atttype,1);
      }
      if (strcmp(attname,"valid_range") == 0) {
        if (attlen != 2) {
          fprintf(stderr,"uniget: valid_range attribute length not 2\n");
          exit(1);
        }
        ncattget(cdfid,i,"valid_range",attbuff);
        floatcon(attbuff,validrange[i],atttype,2);
      }
      if (fabs(validrange[i][0]) > badlim) badlim = fabs(validrange[i][0]);
      if (fabs(validrange[i][1]) > badlim) badlim = fabs(validrange[i][1]);
    }
  }

/* need to define bad */

  bad = 1.001*badlim;

/* make a Candis header */

  nullhdr(hbuff,HBMAX,"float");

/* comment */

  if (strcmp(history,"") == 0) {
    sprintf(cline,"uniget: %s\n",infile);
    addcline(hbuff,HBMAX,cline);
  }
  else extractcom(history);

/* bad data parameters */

  sprintf(cline,"%g",bad);
  addpar(hbuff,HBMAX,"bad",cline);
  sprintf(cline,"%g",badlim);
  addpar(hbuff,HBMAX,"badlim",cline);

/* fields */

  for (i = 0; i < nvars; i++) {

/* check maximum dimensions of fields */

    if ((vardims[i] - exflag[i]) > 4) {
      fprintf(stderr,"uniget: can't handle variable with > 4 dimensions\n");
      exit(1);
    }

/* no NC_UNLIMITED dimension -- just index fields go into static slice */

    if (xtendim == -1) {
      if ((vardims[i] == 1)&&(strcmp(varname[i],dimname[vdlist[i][0]]) == 0)) {
        ftype[i] = 's';
      }
      else ftype[i] = 'v';
    }

/* otherwise, fields with NC_UNLIMITED dimension go into variable slices */

    else {
      if (exflag[i]) ftype[i] = 'v';
      else ftype[i] = 's';
    }

/* add field description to header */

    addfld(hbuff,HBMAX,ftype[i],varname[i],1.,0.,'l',vardims[i] - exflag[i],
         dimname[rvdlist[i][0]],dimsize[rvdlist[i][0]],
         dimname[rvdlist[i][1]],dimsize[rvdlist[i][1]],
         dimname[rvdlist[i][2]],dimsize[rvdlist[i][2]],
         dimname[rvdlist[i][3]],dimsize[rvdlist[i][3]]);
    fp = seekfld(hbuff,HBMAX,ftype[i],varname[i]);
    ptable[i] = *fp;
  }

/* add dummy index fields if none exist in input file */

  for (i = 0; i < ndims; i++) {
    if ((i != xtendim) && (seekfld(hbuff,HBMAX,'s',dimname[i]) == NULL)) {
      dummy[i] = 1;
      addfld(hbuff,HBMAX,'s',dimname[i],1.,0.,'l',1,dimname[i],dimsize[i]);
    }
    else dummy[i] = 0;
  }

/* allocate slice buffers */

  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbuff);

/* transfer data from netcdf to static slice */

  for (i = 0; i < nvars; i++) {
    datatrans(i,'s',sbuff,0);
  }

/* fill dummy index fields */

  for (i = 0; i < ndims; i++) {
    if (dummy[i]) {
      if ((fp = seekfld(hbuff,HBMAX,'s',dimname[i])) == NULL) {
        fprintf(stderr,"uniget: internal error -- lost dummy index field %s\n",
             dimname[i]);
        exit(1);
      }
      for (j = 0; j < dimsize[i]; j++) {
        sbuff[fp->ptr + j] = j;
      }
    }
  }

/* make index parameters from index field info and fill static slice tables */

  i = 0;
  while ((fp = getfld(hbuff,HBMAX,'s',i)) != NULL) {
    if ((fp->dim == 1) && (strcmp(fp->fname,fp->dname1) == 0)) {
      fpointer = sbuff + fp->ptr;
      ival0 = fpointer[0];
      if (fp->tsize <= 1) idval = 0.;
      else idval = (fpointer[fp->tsize - 1] - fpointer[0])/(fp->tsize - 1);
      sprintf(pval,"%g",ival0);
      sprintf(cline,"%s0",fp->fname);
      lowercase(cline);
      addpar(hbuff,HBMAX,cline,pval);
      sprintf(pval,"%g",idval);
      sprintf(cline,"d%s",fp->fname);
      lowercase(cline);
      addpar(hbuff,HBMAX,cline,pval);
    }
    i++;
  }

/* write the Candis header and the static slice to standard output */

  puthdr(stdout,hbuff,HBMAX);
  putslice(stdout,nsbuff,sbuff);

/* transfer the variable slice or slices */

  if (xtendim < 0) maxrecords = 1;
  else maxrecords = dimsize[xtendim];
  for (record = 0; record < maxrecords; record++) {
    for (i = 0; i < nvars; i++) {
      if (datatrans(i,'v',vbuff,record) == -1) {
        ncclose(cdfid);
        fprintf(stderr,"uniget: data read failure\n");
        exit(1);
      }
    }
    putslice(stdout,nvbuff,vbuff);
  }
  ncclose(cdfid);
  exit(0);
}


/* floatcon -- convert everything to float type */

floatcon(inpointer,outpointer,type,num)
char *inpointer;
float *outpointer;
nc_type type;
long num;
{
  long i;
  switch (type) {
    case NC_BYTE :
      for (i = 0; i < num; i++) outpointer[i] = *((char*)inpointer + i);
      break;
    case NC_CHAR :
      for (i = 0; i < num; i++) outpointer[i] = *((char*)inpointer + i);
      break;
    case NC_SHORT :
      for (i = 0; i < num; i++) outpointer[i] = *((short*)inpointer + i);
      break;
    case NC_LONG :
      for (i = 0; i < num; i++) outpointer[i] = *((long*)inpointer + i);
      break;
    case NC_FLOAT :
      for (i = 0; i < num; i++) outpointer[i] = *((float*)inpointer + i);
      break;
    case NC_DOUBLE :
      for (i = 0; i < num; i++) outpointer[i] = *((double*)inpointer + i);
      break;
    default :
      fprintf(stderr,"uniget: unknown netCDF data type %d\n",type);
      exit(1);
  }
}


/* datatrans -- transfer data from variable to field */

int datatrans(var,stype,sbuff,record)
int var;
char stype;
float *sbuff;
long record;
{
  int j;

/* exit if wrong field type */

  if (stype != ftype[var]) return(0);

/* exit if no NC_UNLIMITED dimension and record > 0 */

  if ((xtendim < 0) && (record > 0)) return(-1);

/* get pointers and buffers set up -- once buffer gets big enough,
   leave it alone */

  fp = &ptable[var];
  fpointer = fp->ptr + sbuff;
  if (buffersize < fp->tsize) {
    if (ncbuffer != (char*)NULL) free(ncbuffer);
    if ((ncbuffer = (char*)malloc(sizeof(double)*fp->tsize)) == (char*)NULL) {
      fprintf(stderr,"uniget: can't get enough memory\n");
      exit(1);
    }
    buffersize = fp->tsize;
  }

/* loop on dimensions, setting up start and count stuff for ncvarget */

  for (j = 0; j < vardims[var]; j++) {
    if (xtendim == vdlist[var][j]) {
      start[j] = record;
      count[j] = 1;
    }
    else {
      start[j] = 0;
      count[j] = dimsize[vdlist[var][j]];
    }
  }

/* get the data from the netcdf file */

  if (ncvarget(cdfid,var,start,count,ncbuffer) == -1) {
    return(-1);
  }

/* convert to float form, fix bad data stuff, and store in slice buffer */

  floatcon(ncbuffer,fpointer,datatype[var],fp->tsize);
  for (j = 0; j < fp->tsize; j++) {
    if (fpointer[j] < validrange[var][0]) fpointer[j] = bad;
    if (fpointer[j] > validrange[var][1]) fpointer[j] = bad;
  }
  return(0);
}

/* extract comment from history string */

extractcom(history)
char *history;
{
  int i,n;
  char c;
  n = i = 0;
  while ((c = history[i++]) != '\0') {
    if (n < LINE - 1) cline[n++] = c;
    else {
      fprintf(stderr,"uniget: history line too long\n");
      exit(1);
    }
    if (c == '\n') {
      cline[n] = '\0';
      addcline(hbuff,HBMAX,cline);
      n = 0;
    }
  }
}

/* transform a string to lower case */
lowercase(text)
char *text;
{
  int i;
  i = 0;
  while (text[i] != '\0') {
    if (text[i] >= 'A' && text[i] <= 'Z') text[i] += ('a' - 'A');
    i++;
  }
}
