/* uniput -- This program converts a Candis file to a NETCDF file
 * using the Unidata Project C binding for the NETCDF system calls.
 * All fields are converted to NETCDF data type float.  Index fields in 
 * the Candis file define the dimensionality of the NETCDF file, so if
 * Candis file has fields without an index field for each dimension
 * name, an error exit occurs.  The value of badlim defines the NETCDF
 * valid data limits for all fields.  The name of the output NETCDF file
 * is the only command line argument.
 */

#include <stdio.h>
#include <math.h>
#include "cdfhdr.h"
#include "netcdf.h"

/* general stuff */

static long goop,poop,loop;                 /* general looping variables */

/* Candis stuff */

static char hbuff[HBMAX][LINE];             /* header buffer */
static char comment[HBMAX][LINE];           /* comment buffer */
static char cline[LINE];                    /* character line buffer */
static char stype;                          /* slice type ('s' or 'v') */
static long nsbuff,nvbuff;                  /* slice buffer lengths */
static float *sbuff,*vbuff;                 /* slice buffer pointers */
static float bad,badlim;                    /* bad data stuff */
static int nfields;                         /* number of candis fields */
static int fieldsizes[HBMAX][5];            /* dimension sizes of candis
					       fields */
static int start[5],count[5];               /* cube control arrays for
					       ncvarput */
static float *fpoint[HBMAX];                /* field pointers */
static struct field *fp;                    /* temporary field structure
					       pointer */

/* CDF stuff */

static int cdfid;                           /* CDF handle */
static int varid[HBMAX];                    /* variable handles */
static char ftype[HBMAX];                   /* variable types */
static int vdims[HBMAX];                    /* number of dimensions in a
					       variable */
static int vdimlist[5];                     /* list of dimension id's for
					       variable */
static int vid;                             /* temporary variable handle */
static int actualdims;                      /* number of index fields */
static char ixnames[MAX_NC_DIMS][LINE];     /* names of index fields */
static int dimsizes[MAX_NC_DIMS];           /* sizes of dimensions */
static int dimlist[MAX_NC_DIMS];            /* list of dimension id's
					       for NETCDF */
static float validrange[2];                 /* valid data limits */
static char history[HBMAX*LINE];            /* history list */
static long comht;                          /* comment height */


/* main program */

main(argc,argv)
int argc;
char *argv[];
{

/* check command line arguments */

  if (argc != 2) {
    fprintf(stderr,"Usage: uniput netCDF_file\n");
    exit(1);
  }

/* get header */

  gethdr(stdin,hbuff,HBMAX);
  if (getfmt(hbuff,HBMAX) != 'f') {
    fprintf(stderr,"uniput: float format expected\n");
    exit(1);
  }

/* bad data stuff */

  if (seekpar(hbuff,HBMAX,"bad",cline) == OK) bad = atof(cline);
  else bad = BAD;
  if (seekpar(hbuff,HBMAX,"badlim",cline) == OK) badlim = atof(cline);
  else badlim = BADLIM;
  validrange[0] = -badlim;
  validrange[1] = badlim;

/* allocate slice buffers */

  nsbuff = elemcnt(hbuff,HBMAX,'s');
  sbuff = getbuff(nsbuff);
  nvbuff = elemcnt(hbuff,HBMAX,'v');
  vbuff = getbuff(nvbuff);

/* scan index fields */

  loop = 0;
  actualdims = 1;
  strcpy(ixnames[0],"record");  /* dimension zero is always record dim */
  dimsizes[0] = NC_UNLIMITED;
  while ((fp = getfld(hbuff,HBMAX,'s',loop++)) != NULL) {
    if ((fp->dim == 1) && (strcmp(fp->fname,fp->dname1) == 0)) {
      if (actualdims >= MAX_NC_DIMS) {
        fprintf(stderr,"uniput: dimensionality %d of input file too big\n",
             actualdims);
        exit(1);
      }
      dimsizes[actualdims] = fp->dsize1;
      strcpy(ixnames[actualdims++],fp->fname);
    }
  }

/* create the NETCDF, clobbering any pre-existing file */

  if ((cdfid = nccreate(argv[1],NC_CLOBBER)) == -1) {
    fprintf(stderr,"uniput: can't create a NETCDF named %s\n",argv[1]);
    exit(1);
  }

/* create dimensions in NETCDF, one per index field */

  for (loop = 0; loop < actualdims; loop++) {
    if ((dimlist[loop] = ncdimdef(cdfid,ixnames[loop],dimsizes[loop])) == -1) {
      fprintf(stderr,"uniput: error in creating dimension %s\n",ixnames[loop]);
      exit(1);
    }
  }

/* create the variables, first initializing things */

  nfields = 0;
  stype = 's';

/* loop on slice type */

  for (loop = 0; loop < 2; loop++) {

/* loop on field */

    poop = 0;
    while ((fp = getfld(hbuff,HBMAX,stype,poop++)) != NULL) {

/* set internal info tables */

      if (stype == 's') {
        fpoint[nfields] = sbuff + fp->ptr;
        ftype[nfields] = 's';
        vdims[nfields] = 0;
      }
      else {
        fpoint[nfields] = vbuff + fp->ptr;
        ftype[nfields] = 'v';
        vdimlist[0] = dimlist[0];
        vdims[nfields] = 1;
      }
      if (fp->dim > 0) scanit(vdims[nfields]++,fp->dname1,fp->dsize1);
      if (fp->dim > 1) scanit(vdims[nfields]++,fp->dname2,fp->dsize2);
      if (fp->dim > 2) scanit(vdims[nfields]++,fp->dname3,fp->dsize3);
      if (fp->dim > 3) scanit(vdims[nfields]++,fp->dname4,fp->dsize4);

/* create the field in the NETCDF file */

      if ((vid = varid[nfields] =
           ncvardef(cdfid,fp->fname,NC_FLOAT,vdims[nfields],vdimlist)) == -1) {
        fprintf(stderr,"uniput: can't create variable %s\n",fp->fname);
        exit(1);
      }

/* create the local attributes */

      if (ncattput(cdfid,vid,"valid_range",NC_FLOAT,2,validrange) == -1) {
        fprintf(stderr,"uniput: can't create valid range attribute\n");
        exit(1);
      }

/* increment field number */

      nfields++;
    }

/* increment slice type */

    stype = 'v';
  }

/* create the global history attribute */

  comht = getcmt(hbuff,HBMAX,comment);
  strcpy(history,"");
  for (loop = 0; loop < comht; loop++) {
    strcat(history,comment[loop]);
  }
  if (ncattput(cdfid,NC_GLOBAL,"history",
       NC_CHAR,strlen(history),history) == -1) {
    fprintf(stderr,"uniput: can't create history attribute\n");
    exit(1);
  }

/* take NETCDF out of define mode */

  ncendef(cdfid);

/* transfer static slice */

  getslice(stdin,nsbuff,sbuff);
  putcdf(0,'s');

/* transfer variable slices */

  loop = 0;
  while (getslice(stdin,nvbuff,vbuff) != EOF) {
    putcdf(loop++,'v');
  }

/* close the NETCDF and exit */

  ncclose(cdfid);
  exit(0);
}

/* do variable table assignments */

scanit(position,name,size)
int position;
char *name;
int size;
{
  int i,cdfpos;

/* fill candis tables */

  fieldsizes[nfields][position] = size;

/* scan index field names to match dimension names */

  cdfpos = -1;
  for (i = 1; i < actualdims; i++) {
    if (strcmp(name,ixnames[i]) == 0) {
      cdfpos = i;
      break;
    }
  }
  if (cdfpos < 0) {
    fprintf(stderr,"uniput: dimension name %s lacks index field\n",name);
    exit(1);
  }
  vdimlist[position] = dimlist[i];
}

/* write a static or variable candis slice to a NETCDF */

putcdf(record,type)
int record;
char type;
{

/* loop on fields */

  for (poop = 0; poop < nfields; poop++) {

/* do only type requested */

    if (ftype[poop] != type) continue;

/* put a data field to netCDF */

    for (goop = 0; goop < vdims[poop]; goop++) {
      start[goop] = 0;
      count[goop] = fieldsizes[poop][goop];
    }
    if (type == 'v') {
      start[0] = record;
      count[0] = 1;
    }
    if (ncvarput(cdfid,varid[poop],start,count,fpoint[poop]) == -1) {
      fprintf(stderr,"uniput: ncvarput failure\n");
      exit(1);
    }
  }
}
