/* auxio.c -- Contains all non-float type Candis I/O routines */

#include <stdio.h>
#include "cdfhdr.h"

/* getilist -- This routine gets information from a header buffer about
 * the positions and sizes of fields and elements in a slice in integer
 * format.
 */

getilist(buffer,max,stype,size,length,offset,add,mul)
char buffer[][LINE];             /* header buffer */
int max;                         /* max lines in header buffer */
char stype;                      /* 's' or 'v', static or variable slice */
int *size;                       /* size of integers in each field */
long *length;                    /* number of elements in each field */
long *offset;                    /* offset of each field in elements */
float *add,*mul;                 /* scaling parameters */
{
  int i;
  struct field *fp;
/* loop on fields */
  i = 0;
  while ((fp = getfld(buffer,max,stype,i)) != NULL) {
    if (fp->bytes == 'c') size[i] = sizeof(char);
    else if (fp->bytes == 's') size[i] = sizeof(short);
    else if (fp->bytes == 'l') size[i] = sizeof(long);
    else if (fp->bytes == 'p') size[i] = -sizeof(float);
    else {
      fprintf(stderr,"getilist: unknown integer type '%c'\n",fp->bytes);
      exit(1);
    }
    length[i] = fp->tsize;
    offset[i] = fp->ptr;
    add[i] = fp->sadd;
    mul[i] = fp->smul;
    i++;
  }
/* set end with zero size */
  size[i] = 0;
}

/* gislice -- get a slice in integer format and return stuff in float fmt */

long gislice(stream,size,length,offset,add,mul,num,sbuff,swapflag)
FILE *stream;                   /* file pointer to input file */
int *size;                      /* size of integers in each field */
long *length;                   /* number of elements in each field */
long *offset;                   /* offset of each field in elements */
float *add,*mul;                /* scaling factors */
long num;                       /* number of expected elements */
float *sbuff;                   /* pointer to slice buffer */
char swapflag;                  /* swap bytes on integers */
{
  void *ipointer;               /* buffer for read */
  int i;                        /* loop counter */
  long j;                       /* long loop counter */
  long eltcount;                /* element count from data file */
/* get element count */
  if(num < 0 || num > MAXSLICE) {
    fprintf(stderr,
        "gislice: bizarre number of elements requested = %ld\n",num);
    exit(1);
  }
  if((eltcount = getelcnt(stream)) == EOF) return(EOF);
  if(num != eltcount) {
    fprintf(stderr,
        "gislice: requested element count = %ld, actual = %ld\n",num,eltcount);
    exit(1);
  }
  if (num == 0) return(0);
/* loop on fields, converting one by one */
  i = 0;
  while (size[i] != 0) {
/* first, allocate a temporary buffer for the read */
    if ((ipointer = (void*)malloc(abs(size[i])*length[i])) == NULL) {
      fprintf(stderr,"gislice: can't get enough memory\n");
      exit(1);
    }
/* read the stuff */
    if (fread(ipointer,abs(size[i]),length[i],stream) != length[i]) {
      fprintf(stderr,"gislice: error in read\n");
      exit(1);
    }
/* swap bytes if necessary */
    if (swapflag && size[i] == 2) swap2((short*)ipointer,length[i]);
    else if (swapflag && size[i] == 4) swap4((long*)ipointer,length[i]);
/* convert to float format */
    for (j = 0; j < length[i]; j++) {
      if (size[i] == sizeof(char)) sbuff[offset[i] + j] =
           ((float)(*((char*)ipointer + j)) - add[i])/mul[i];
      else if (size[i] == sizeof(short)) sbuff[offset[i] + j] =
           ((float)(*((short*)ipointer + j)) - add[i])/mul[i];
      else if (size[i] == sizeof(long)) sbuff[offset[i] + j] =
           ((float)(*((long*)ipointer + j)) - add[i])/mul[i];
/* special case for pixel data -- no conversion */
      else if (size[i] < 0) sbuff[offset[i] + j] = *((float*)ipointer + j);
    }
/* free the buffer and increment field index */
    free(ipointer);
    i++;
  }
/* return the number of elements */
  return(eltcount);
}

/* swap2 -- swap bytes on 16 bit words */
swap2(ptr,len)
short *ptr;
long len;
{
  long i;
  char a;
/* loop over buffer */
  for (i = 0; i < len; i++) {
    a = *((char*)(ptr + i));
    *((char*)(ptr + i)) = *((char*)(ptr + i) + 1);
    *((char*)(ptr + i) + 1) = a;
  }
}

/* swap4 -- swap bytes on 32 bit words */
swap4(ptr,len)
long *ptr;
long len;
{
  long i;
  char a;
/* loop over buffer */
  for (i = 0; i < len; i++) {
    a = *((char*)(ptr + i));
    *((char*)(ptr + i)) = *((char*)(ptr + i) + 3);
    *((char*)(ptr + i) + 3) = a;
    a = *((char*)(ptr + i) + 1);
    *((char*)(ptr + i) + 1) = *((char*)(ptr + i) + 2);
    *((char*)(ptr + i) + 2) = a;
  }
}


/* pislice -- put a slice in integer format from a slice buffer */

long pislice(stream,size,length,offset,add,mul,num,sbuff)
FILE *stream;                   /* file pointer to input file */
int *size;                      /* size of integers in each field */
long *length;                   /* number of elements in each field */
long *offset;                   /* offset of each field in elements */
float *add,*mul;                /* scaling factors */
long num;                       /* number of expected elements */
float *sbuff;                   /* pointer to slice buffer */
{
  void *ipointer;               /* buffer for read */
  int i;                        /* loop counter */
  long j;                       /* long loop counter */
  long eltcount;                /* element count from data file */
  char elct[9];                 /* ascii buffer for eltcount */
  float hold;                   /* temporary for intermediate result */
/* do some plausibility checking */
  if(num < 0 || num > MAXSLICE) {
    fprintf(stderr,"putslice: weird element count = %ld\n",num);
    exit(1);
  }
/* write element count */
  sprintf(elct,"%8ld",num);
  fputs(elct,stream);
  if (num == 0) return(0);
/* loop on fields */
  i = 0;
  eltcount = 0;
  while (size[i] != 0) {
    eltcount += length[i];
/* allocate temporary buffer */
    if ((ipointer = (void*)malloc(size[i]*length[i])) == NULL) {
      fprintf(stderr,"pislice: can't get enough memory\n");
      exit(1);
    }
/* do conversion */
    for (j = 0; j < length[i]; j++) {
      if (size[i] > 0) {
        hold = sbuff[offset[i] + j]*mul[i] + add[i];
#if ROUNDTOZERO
        hold = (hold < 0.) ? hold - .5 : hold + .5;
#else
        hold = hold + .5;
#endif
        if (size[i] == sizeof(char)) *((char*)ipointer + j) = (char)hold;
        if (size[i] == sizeof(short)) *((short*)ipointer + j) = (short)hold;
        if (size[i] == sizeof(long)) *((long*)ipointer + j) = (long)hold;
      }
/* special case for pixel data -- no conversion */
      else {
        *((float*)ipointer + j) = sbuff[offset[i] + j];
      }
    }
/* write buffer to stream */
    if (fwrite(ipointer,size[i],length[i],stream) != length[i]) {
      fprintf(stderr,"pislice: write error\n");
      exit(1);
    }
/* free buffer and increment field index */
    free(ipointer);
    i++;
  }
/* return element count */
  return(eltcount);
}

/* gaslice -- get a slice in ascii form - works like getslice except
 * that input slice is in ascii
 */

long gaslice(stream,size,length,offset,nelements,sbuff)
FILE *stream;                    /* input file, assumed open */
int *size;                       /* size of integers in each field */
long *length;                    /* number of elements in each field */
long *offset;                    /* offset of each field in elements */
long nelements;                  /* number of expected elements from header */
float *sbuff;                    /* pointer to slice buffer */
{
  long i,j,elcount;
  int intvalue,k;
  char *caddr;
/* get element count */
  if (fscanf(stream,"%ld",&elcount) <= 0) return(EOF);
  if (elcount != nelements) {
    fprintf(stderr,"gaslice: %ld elements expected from header, %ld found\n",
         nelements,elcount);
    exit(1);
  }
/* loop on fields */
  i = 0;
  while (size[i] != 0) {
/* check whether float or pixel field -- then get data */
    if (size[i] < 0) {        /* pixel */
      for (j = 0; j < length[i]; j++) {
        caddr = (char*)(sbuff + offset[i] + j);
        for (k = 0; k < sizeof(float); k++) {
          if (fscanf(stream,"%d",&intvalue) <= 0) {
            fprintf(stderr,"gaslice: unexpected eof or data error\n");
            exit(1);
          }
          *(caddr + k) = (char)(intvalue & 0xff);
        }
      }
    }
    else {                    /* float */
      for (j = 0; j < length[i]; j++) {
        if (fscanf(stream,"%f",sbuff + offset[i] + j) <= 0) {
          fprintf(stderr,"gaslice: unexpected eof or data error\n");
          exit(1);
        }
      }
    }
    i++;
  }
/* return number of elements found */
  return(elcount);
}

/* paslice -- put a slice in ascii form - works like putslice except
 * for format of output
 */

long paslice(stream,size,length,offset,nelements,sbuff)
FILE *stream;                    /* input file, assumed open */
int *size;                       /* size of integers in each field */
long *length;                    /* number of elements in each field */
long *offset;                    /* offset of each field in elements */
long nelements;                  /* number of expected elements from header */
float *sbuff;                    /* pointer to slice buffer */
{
  long i,j;
  int k;
  char *caddr;
/* write element count */
  fprintf(stream,"\n%8ld\n",nelements);
/* loop on fields */
  i = 0;
  while (size[i] != 0) {
/* check whether float or pixel field -- then put data */
    if (size[i] < 0) {        /* pixel */
      fprintf(stream,"\n");
      for (j = 0; j < length[i]; j++) {
        caddr = (char*)(sbuff + offset[i] + j);
        for (k = 0; k < sizeof(float); k++) {
          fprintf(stream," %d",(int)caddr[k]);
        }
        if((j - (j/5)*5) == 0) fprintf(stream,"\n");
      }
    }
    else {                    /* float */
      fprintf(stream,"\n");
      for (j = 0; j < length[i]; j++) {
        fprintf(stream," %g",sbuff[offset[i] + j]);
        if((j - (j/5)*5) == 0) fprintf(stream,"\n");
      }
    }
    i++;
  }
/* return number written */
  return(nelements);
}
