/* cdfio.c -- contains all float type Candis I/O routines */

#include <stdio.h>
#include "cdfhdr.h"

/* gethdr.c -- This routine reads a common data format header line-wise
 * into buffer from the named stream.  The asterisk at the end is included.
 * The number of lines read is returned.  Only max lines are stored,
 * to provide a way of protecting the buffer from overages.
 */

int gethdr(stream,buffer,max)
FILE *stream;
char buffer[][LINE];
int max;
{
  int count,ptr;
  count = ptr = 0;
  for (;;) {
    if (fgets(buffer[ptr],LINE,stream) == NULL) {
      fprintf(stderr,"gethdr: unexpected end of file\n");
      exit(1);
    }
    count++;
    if (ETEST(ptr)) return(count);
    ptr++;
    if (ptr >= max) ptr = max - 1;
  }
}

/* puthdr.c -- This routine writes a common data format header from
 * buffer to the named stream.  It depends on finding a last line
 * that consists of a single asterisk as its end of buffer condition.
 * However, in no case will it write more than max lines.
 */

puthdr(stream,buffer,max)
FILE *stream;
char buffer[][LINE];
int max;
{
  int count;
  for (count = 0; count < max; count++) {
    if (fputs(buffer[count],stream) == EOF) {
      fprintf(stderr,"puthdr: error on write\n");
      exit(1);
    }
    if (ETEST(count)) break;
  }
}

/* Getslice accepts a pointer to an open input stream, the number of elements
 * expected in the slice, and the buffer to fill and reads the data file,
 * comparing the element count from the file with that expected, and
 * filling the buffer with that many data elements if succssful. Dies with
 * an error message if the counts do not match and assumes enough room in the
 * buffer for that amount. Everything is in float except the element count
 * from the data file which is a decimal integer "%8ld". Returns number read
 * or EOF if end of file encountered.
 */

long getslice(stream,num,sbuff)
FILE *stream;   /* pointer into data file of the start of the slice */
long num;   /* number of elements expected in slice */
float sbuff[];    /* return the data in this buffer */
{
  long eltcount;  /* element count from data file */
  long nread; /* number of floats read */
  if(num < 0 || num > MAXSLICE) {
    fprintf(stderr,
        "getslice: bizarre number of elements requested = %ld\n",num);
    exit(1);
  }
  if((eltcount = getelcnt(stream)) == EOF) return(EOF);
  if(num != eltcount) {
    fprintf(stderr,
        "getslice: requested element count = %ld, actual = %ld\n",
        num,eltcount);
    exit(1);
  }
  if (num == 0) return(0);
  nread = fread(sbuff,sizeof(float),num,stream);
  if(nread != num) {
    fprintf(stderr,"getslice: not enough data read\n");
    exit(1);
  }
  if (FLOATSWAP) floatswap(num,sbuff);
  return(eltcount);
}

/* Putslice accepts a buffer of floats, an open file pointer, and the element
 * count, writes the number in 8 byte decimal ascii, then writes the
 * floats using fwrite.  Returns the number of floats written.
 */

long putslice(stream,num,sbuff)
FILE *stream; /* the data file pointer */
long num; /* the number of elements to write */
float sbuff[];  /* the buffer of floats */
{
  long nwritten;  /* number of elements written */
  char elct[9];
  if(num < 0 || num > MAXSLICE) {
    fprintf(stderr,"putslice: weird element count = %ld\n",num);
    exit(1);
  }
  if (FLOATSWAP) floatswap(num,sbuff);
  sprintf(elct,"%8ld",num);
  fputs(elct,stream);
  if (num == 0) return(0);
  nwritten = fwrite(sbuff,sizeof(float),num,stream);
  if(nwritten != num) {
    fprintf(stderr,"putslice: botched fwrite\n");
    exit(1);
  }
  if (FLOATSWAP) floatswap(num,sbuff);
  return(nwritten);
}

/* getelcnt accepts a stream, reads eight characters from it converting
 * non-blank characters to a number, and returns the number as a long. It
 * is need for reading the element count from files of integers because
 * scanf will ignore leading blanks and read anything that looks like a legal
 * integer character or until it exhausts the field. This may include
 * characters beyond the eight which would have been read had blank characters
 * been counted.
 */

long getelcnt(stream)
FILE *stream;
{
  long eltcount;  /* the value calculated and returned */
  char str[9];  /* the element count in a string */
  char *i;        /* return string pointer from string getter */
  if ((i = fgets(str,9,stream)) == NULL) return((long) EOF);
  if (strlen(i) != 8) {
    fprintf(stderr,"getelcnt: error in element count format\n");
    exit(1);
  }
  eltcount = (long) atoi(str);
  return(eltcount);
}
